/* ===================================================================== */
/*     DRAWxtl                                                           */
/*                                                                       */
/* Routine to read crystallographic information such as unit cell, space */
/* group, atom positions, and thermal ellipsoid parameters to generate   */
/* POLYHEDRON, SPHERE, ELLIPSOID and BOND geometry POV and VRML scenes.  */
/*                                                                       */
/* Part 2                                                                */
/*                                                                       */
/* Coded by:   L.W. Finger                                               */
/*            Geophysical Laboratory                                     */
/*            Carnegie Institution of Washington                         */
/*            Washington DC 20015                                        */
/*            Larry.Finger@lwfinger.net                                  */
/*                                                                       */
/* Original Version: April 92                                            */
/*            Revised January 1993 to add bonds.                         */
/*            Converted to AVS Module February 1993.                     */
/*            Thermal Ellipsoids added April 1996                        */
/*            Thermal ellipsoids for POV and VRML finished Nov. 1996     */
/*                                                                       */
/*                                                                       */
/* Converted to standalone POV scene file generator Oct. 1994/March 1995 */
/* and VRML generator added September 1995/March 1996                    */
/* by M. Kroeker, Organic Chemistry, Technische Hochschule Darmstadt,    */
/* Petersenstr. 22, D-64287 Darmstadt, Germany                           */
/*                                                                       */
/* Two versions merged  July 1995                                        */
/* ===================================================================== */

/*************************************************************************
 * This file contains routines:
 *
 *  determinant - return the determinant of a 3x3 matrix
 *  generate_ellipsoids - generate all the ellipsoids for this structure
 *  generate_planes - generate planes
 *  generate_poly - generate polyhedra
 *  generate_texts - generate any text in the drawing
 *  import_gsas - read structure information from a GSAS exp file
 *  import_shelx - read structure information from a SHELX ins file
 *  import_fdat - read structure information from an fdat file
 *  Unique_Atom - determine if this position is unique in the unit cell
 *  show_slab - generate the commands to define a slab of structure
 *  push_cylinder - generate an openGL cylinder corresponding to a unit-cell edge
 *  generate_gl_texts - plot text in an openGL window
 *  findsys - determine crystal system from symmetry operations
 *  getsym - parse SHELX symmetry operator information
 *
 *************************************************************************/

#include "drawxtl.h"
#include "draw_ext.h"
void findsys(void);
void getsym(char*, int, int);

/* ************************************************************** */
/* ************************************************************** */

float determinant (double rot[3][3])
{
  double temp;

  temp = rot[0][0] * (rot[1][1] * rot[2][2] - rot[1][2] * rot[2][1])
	- rot[1][0] * (rot[0][1] * rot[2][2] - rot[0][2] * rot[2][1])
	+ rot[2][0] * (rot[0][1] * rot[1][2] - rot[0][2] * rot[1][1]);
  return (float) temp;
}

/* ************************************************************** */
/* ************************************************************** */

void generate_ellipsoids (void)

/* routine to generate ellipsoid objects and add then to edit list */
{
  float ev[3];					 /* scaled eigenvalue */
  double rot[3][3];				 /* rotation matrix for ellipsoid */
  int i, j, k, l, m, n, o, oo;	 /* loop counters */
  int Vo[3];
  int n_ellipsoids;
  float EV[3][3];				 /* array for eigenvectors */
  float alpha1 = 0.0f, alpha2 = 0.0f;
  float alpha3 = 0.0f;          /* terms needed to calculate rotations */
  float d1 = 0.0f, d2 = 0.0f, d3 = 0.0f, alpha = 0.0f, sinalp, radius, temp;
  float sinal1, sinal2, sinal3, cosal1, cosal2, cosal3;
#ifdef OPENGL
  char string[40];
  float glr, glg, glb;
  GLdouble clipx[4] = { 0., 0., 0., 0. };
  GLdouble clipy[4] = { 0., 0., 0., 0. };
  GLdouble clipz[4] = { 0., 0., 0., 0. };
#endif
  
  int outside = 0;
  int ntest[8][3] =
	{ {1, 1, 1}, {-1, 1, 1}, {1, -1, 1}, {1, 1, -1}, {-1, -1, 1}, 
	{1, -1, -1}, {-1, 1, -1}, {-1, -1, -1}
  };						 /* list of octant coordinates to check */
  double eye[3], vmin, dist;	 /* used to find octant for cutout */
  int best;					 /* octant list pointer */
  int l1;
  int omit;
  extern int nomits,omit1[],omit2[];  
  
  n_ellipsoids = 0;
  radius = 0.00015f * Scale;

  if (Ellipaxis_width != 0.)
    radius = Ellipaxis_width;
  fprintf (drvui->flout, "\nGenerated %6.1f percent probability ellipsoids for:\n\n",
	   100. * Ellipsoid_Prob);
  for (i = 1; i < n_ellips; ++i) {		 /* loop through ellipsoids */
    for (j = 0; j < 3; j++) {
      ev[j] = drvui->ellips_RMS[i][j] * Ellipsoid_Scale;
      for (k = 0; k < 3; k++)
        EV[j][k] = drvui->ellips_EV[i][j][k];
    }
    for (j = 0; j < natom; ++j) {      /* loop through atoms */
      nvert = 0;                       /* clear the vertex list */
      if (((drvui->atom_n[j] >> 24) & 255) == i)/* this ellipsoid */
        find_all_in_box (j);           /* generate all locations for this atom */
      if (nvert > 0) {           /* if some present */
        if (do_ellipsoids == 0) {      /* Biso sphere */
          float *radii;                /* allocate storage for radii of spheres */

	      radii = (float *)
		  malloc ((unsigned) (nvert * sizeof (float)));
          for (k = 0; k < nvert; ++k)		 /* copy radii info to storage */
            radii[k] = drvui->ellips_RMS[i][0] * Ellipsoid_Scale;
          Output_Spheres (radii, i);
          free (radii);
        } else {
          for (k = 0; k < nvert; ++k) {		 /* loop through the atom positions */
            if (clipflag != 0) {

              outside = 0;
              for (l1 = 0; l1 < 3; l1++)
                if (o_vert[3 * k + l1] < drvui->clip_lim[drvui->frame_no][l1] ||
                  o_vert[3 * k + l1] > drvui->clip_lim[drvui->frame_no][l1 + 3])
                  outside = 1;
            }
            omit = 0;
            for (l1=0;l1<nomits;l1++) {
              if (omit1[l1]==i*100000 && omit2[l1] ==k) omit = 1;
            }

            if ((clipflag == 0 || outside == 0)&& omit==0) {
              o = vert_sym_no[k];
              fprintf (drvui->flout,
		         "  %c%c%c%c %1d at %8.5f %8.5f %8.5f with sym op %d\n",
		         drvui->atom_l[j][0], drvui->atom_l[j][1], drvui->atom_l[j][2],
		         drvui->atom_l[j][3], drvui->sv_atom_n[j]
		         , o_vert[3 * k], o_vert[3 * k + 1],
		         o_vert[3 * k + 2], o);
	          if (no_comment == 0)
		        fprintf (drvui->fpoutv,
			        "# Ellipsoid around %c%c%c%c %1d at %8.5f %8.5f %8.5f \n",
			        drvui->atom_l[j][0], drvui->atom_l[j][1], drvui->atom_l[j][2],
                    drvui->atom_l[j][3], drvui->sv_atom_n[j], o_vert[3 * k]
			        , o_vert[3 * k + 1], o_vert[3 * k + 2]);
	          fprintf (drvui->fpoutp,
		           "/* Ellipsoid around %c%c%c%c %1d at %8.5f %8.5f %8.5f  */\n",
		           drvui->atom_l[j][0], drvui->atom_l[j][1], drvui->atom_l[j][2],
		           drvui->atom_l[j][3], drvui->sv_atom_n[j], o_vert[3 * k]
		           , o_vert[3 * k + 1], o_vert[3 * k + 2]);
	          for (l = 0; l < 3; ++l) {
		        for (m = 0; m < 3; ++m) {
		          rot[l][m] = 0.0;
		          for (n = 0; n < 3; ++n) {
		            rot[l][m] += EV[l][n] * drvui->rss[o][n][m];
                  }				 /* now have combined rotation matrix for ellipsoid */
                }				 /* for on m */
              }					 /* for on l */

/* make rotation right-handed (det > 0) */

              if (determinant (rot) < 0.0) {
                for (l = 0; l < 3; l++) {   /* make rot right handed */
                  for (m = 0; m < 3; m++) {
                    rot[l][m] *= -1.0;
                  }
                }
              }
              if (fabs ((temp = determinant (rot)) - (float) 1.0) > 1.0e-3)
                fprintf (drvui->flout, "Determinant of rot = %8.5f\n", temp);

/* calculate ellipsoid rotation angles for POV and openGL */

              alpha2 = (float) asin (rot[0][2]);
              cosal2 = (float) cos (alpha2);
              if (fabs (cosal2) > 0.00001) {
                alpha1 = (float) atan2 (-rot[1][2] / cosal2, rot[2][2] / cosal2);
                alpha3 = (float) atan2 (-rot[0][1] / cosal2, rot[0][0] / cosal2);
              } else {
                alpha3 = (float) atan2 (rot[1][0], -rot[2][0]);
                alpha1 = 0.0f;
              }

/* recalculate rot using alpha1,2,3 from above */
              
              sinal2 = (float) sin(alpha2);
              sinal1 = (float) sin(alpha1);
              cosal1 = (float) cos(alpha1);
              sinal3 = (float) sin(alpha3);
              cosal3 = (float) cos(alpha3);
              rot[0][0] =  cosal2 * cosal3;
              rot[1][0] =  -cosal2 * sinal3;
              rot[2][0] = sinal2;
              rot[0][1] = cosal1 * sinal3 + sinal1 * sinal2 * cosal3;
              rot[1][1] = cosal1 * cosal3 - sinal1 * sinal2 * sinal3;
              rot[2][1] = -sinal1 * cosal2;
              rot[0][2] = sinal1 * sinal3 - cosal1 * sinal2 * cosal3;
              rot[1][2] = sinal1 * cosal3 + cosal1 * sinal2 * sinal3;
              rot[2][2] = cosal1 * cosal2;

              if (El_Cutout == 1) {		 /* select octant for cutout */
                double fullrot[3][3];
                vmin =  -1000000.;            /* initialize to impossible value */
                best = 0;
                for (m = 0; m < 3; m++) {     /* calculate combined rotation matrix */
                    for (n = 0; n < 3; n++) {
                        fullrot[m][n] = 0.0;
                        for (oo = 0; oo < 3; oo++) {
                            fullrot[m][n] += rot[oo][m] * G_Rot[oo][n];
                        }
                    }
                }
                if (fabs ((temp = determinant (fullrot)) - 1.0f) > 1.0e-3)
                    fprintf (drvui->flout, "Determinant of fullrot = %8.5f\n", temp);

                eye[0] = fullrot[0][2];       /* get the rotated position of the viewer (z axis) */
                eye[1] = fullrot[1][2];
                eye[2] = fullrot[2][2];

                for (n = 0; n < 8; n++) {
                  for (m = 0; m < 3; m++)     /* take next triple from list */
                    Vo[m] = ntest[n][m];

                  dist = eye[0] * Vo[0] + eye[1] * Vo[1] + eye[2] * Vo[2];
                  if (dist > vmin) {          /* maximum is the one we want */
                    vmin = dist;
                    best = n;
                  }
                }

                for (m = 0; m < 3; m++)       /* set the octant to display */
                  Vo[m] = ntest[best][m];

              }           /* ... if cutout */

/* output this ellipsoid to the POV file */

              fprintf (drvui->fpoutp," object{\n");
              fprintf (drvui->fpoutp,"   union{\n");
              fprintf (drvui->fpoutp,"     object{\n");
              fprintf (drvui->fpoutp,"      cylinder{ < 0,0,-0.5 > , < 0,0,0.5 >, 1.\n");  /* first principal ellipse */
              fprintf (drvui->fpoutp,"       texture{pigment{color %s}}\n",Ellipaxis_color);
              fprintf (drvui->fpoutp,"       scale<%8.5f,%8.5f,%8.5f>\n      }\n     }\n",
                     ev[0], ev[1], radius);
              fprintf (drvui->fpoutp,"     object{\n      cylinder{ < 0,-0.5,0 > , < 0,0.5,0 >, 1.\n");  /* second principal ellipse */
              fprintf (drvui->fpoutp,"       texture{pigment{color %s}}\n",Ellipaxis_color);
              fprintf (drvui->fpoutp,"       scale<%8.5f,%8.5f,%8.5f>\n      }\n     }\n",
                     ev[0], radius, ev[2]);
              fprintf (drvui->fpoutp,"     object{\n      cylinder{ < -0.5,0,0 > , < 0.5,0,0 >, 1.\n");  /* third principal ellipse */
              fprintf (drvui->fpoutp,"       texture{pigment{color %s}}\n",Ellipaxis_color);
              fprintf (drvui->fpoutp,"       scale<%8.5f,%8.5f,%8.5f>\n      }\n     }\n",
                     radius, ev[1], ev[2]);
              fprintf (drvui->fpoutp,"     object{\n");
              if (El_Cutout)
                fprintf (drvui->fpoutp,"      difference {\n");
              fprintf (drvui->fpoutp,"       sphere{< 0,0,0 > 1}\n");
              if (El_Cutout) {
                fprintf (drvui->fpoutp,"       box{<0,0,0>,<%d,%d,%d>}\n",
                    Vo[0], Vo[1], Vo[2]);
              }
              fprintf (drvui->fpoutp,"       texture{pigment{color %s  }}\n",
                       drvui->ellips_col_p[i]);
              fprintf (drvui->fpoutp,"       finish{phong %5.2f phong_size %5.2f reflection 0.1}\n",
                       drvui->Phong_Value, drvui->Phong_Size);
              fprintf (drvui->fpoutp,"       scale<%8.5f,%8.5f,%8.5f>\n",
                       ev[0], ev[1], ev[2]);
              if (El_Cutout)
                fprintf (drvui->fpoutp,"      }\n");
              fprintf (drvui->fpoutp,"     }\n   }\n");
              fprintf (drvui->fpoutp,"   rotate <%8.5f,0,0>\n", -alpha1 * RAD);
              fprintf (drvui->fpoutp,"   rotate <0,%8.5f,0>\n", -alpha2 * RAD);
              fprintf (drvui->fpoutp,"   rotate <0,0,%8.5f>\n", -alpha3 * RAD);
              fprintf (drvui->fpoutp,"   translate <%8.5f, %8.5f, %8.5f>\n }\n",
                 s_vert[3 * k], s_vert[3 * k + 1], s_vert[3 * k + 2]);

/* Output the ellipsoid to the VRML file */

              if (!Vrml2)
                fprintf (drvui->fpoutv, " Separator {\n");
              fprintf (drvui->fpoutv, "  Transform {\n");
              d1 = (float) (0.5 * (rot[2][1] - rot[1][2]));
              d2 = (float) (0.5 * (rot[0][2] - rot[2][0]));
              d3 = (float) (0.5 * (rot[1][0] - rot[0][1]));

/* d's contain direction cosines * sin(alpha) */

              alpha = (float) (0.5 * (rot[0][0] + rot[1][1] +
			      rot[2][2] - (float) 1.0));
              alpha = (float) acos (alpha);
              sinalp = (float) sin (alpha);
              if (fabs (sinalp) > 0.0001) {

                d1 = d1 / sinalp;
                d2 = d2 / sinalp;
                d3 = d3 / sinalp;

/* Make sure alpha is in correct quadrant */

                if (fabs (d1) > 0.0001) {
                  temp = (float) (1.0 -  cos (alpha)) * d2 * d3 -
                       d1 * (float) sin (alpha);
                  if (fabs (rot[1][2] - temp) > 0.0001)
                    alpha = -alpha;
                } else {
                  if (fabs (d2) > 0.0001) {
                    temp = (float) (1.0 - cos (alpha)) * d1 *
                          d3 - d2 * (float) sin (alpha);
                    if (fabs (rot[2][0] - temp) > 0.0001)
                      alpha = -alpha;
                  } else {
                    if (fabs (d3) > 0.0001) {
                      temp = (float) (1.0 - cos (alpha)) *
                             d1 * d2 - d3 * (float) sin (alpha);
                      if (fabs (rot[0][1] - temp) > 0.0001)
                        alpha = -alpha;
                    }
                  }
                }
              } else {
                d1 = (float) sqrt (fabs(rot[0][0]));
                d2 = (float) sqrt (fabs(rot[1][1]));
                d3 = (float) sqrt (fabs(rot[2][2]));
              }
              temp = (float) sqrt (d1 * d1 + d2 * d2 + d3 * d3);
              d1 = d1 / temp;
              d2 = d2 / temp;
              d3 = d3 / temp;

	        fprintf (drvui->fpoutv,
		       "   rotation %6.3f %6.3f %6.3f %7.3f\n",
		       d1, d2, d3, -alpha);
	        fprintf (drvui->fpoutv,
		       "   translation %8.5f %8.5f %8.5f\n",
		       s_vert[3 * k], s_vert[3 * k + 1], s_vert[3 * k + 2]);
#ifdef OPENGL
/* output ellipsoid to screen with openGL */
            if (gl_view) {
              glPushMatrix ();
              glLoadName (i*100000);
              glPushName (k);
              (void)sscanf (drvui->ellips_col_v[i], "%f %f %f", &glr, &glg, &glb);
              glColor3f (glr, glg, glb);
              glTranslatef (s_vert[3 * k],s_vert[3 * k + 1], s_vert[3 * k + 2]);
              glRotatef (-alpha3 * (float) RAD, 0.0f, 0.0f, 1.0f);
              glRotatef (-alpha2 * (float) RAD, 0.0f, 1.0f, 0.0f);
              glRotatef (-alpha1 * (float) RAD, 1.0f, 0.0f, 0.0f);
              glScalef (ev[0], ev[1], ev[2]);
              if (El_Cutout) {
                clipx[0] = -(GLdouble)Vo[0];
                clipy[1] = -(GLdouble)Vo[1];
                clipz[2] = -(GLdouble)Vo[2];
                glClipPlane (GL_CLIP_PLANE0, clipx);
                glClipPlane (GL_CLIP_PLANE1, clipy);
                glClipPlane (GL_CLIP_PLANE2, clipz);
                glEnable (GL_CLIP_PLANE0);
                glutSolidSphere (1., 10, 10);
                glDisable (GL_CLIP_PLANE0);
                glEnable (GL_CLIP_PLANE1);
                glutSolidSphere (1., 10, 10);
                glDisable (GL_CLIP_PLANE1);
                glEnable (GL_CLIP_PLANE2);
                glutSolidSphere (1., 10, 10);
                glDisable (GL_CLIP_PLANE2);
                strncpy(string,Ellipaxis_color,39);
                trim_string(string,40);
                check_color(string);
                (void)sscanf(string,"%f %f %f",&glr,&glg,&glb);
                glColor3f (glr, glg, glb);
                glDisable (GL_LIGHTING);
                glBegin (GL_LINES);
                glVertex3f (0.0f, 0.0f, 0.0f);
                glVertex3f ((float) Vo[0], 0.0f, 0.0f);
                glVertex3f (0.0f, 0.0f, 0.0f);
                glVertex3f (0.0f, (float) Vo[1], 0.0f);
                glVertex3f (0.0f, 0.0f, 0.0f);
                glVertex3f (0.0f, 0.0f, (float) Vo[2]);
                glEnd ();
                glEnable (GL_LIGHTING);
              } else {
                glutSolidSphere (1., 10, 10);
              }                /* if El_Cutout */
              glPopName();
              glPopMatrix ();
            }
#endif
            if (Vrml2) {
              if (strlen (drvui->ellips_col_v[i]) == 0)
                strcpy (drvui->ellips_col_v[i], "1 0 0");
              fprintf (drvui->fpoutv,
                  "   scale %8.5f %8.5f %8.5f\n", ev[0], ev[1], ev[2]);
              fprintf (drvui->fpoutv, "  children [ Shape {");
              fprintf (drvui->fpoutv, " appearance Appearance {\n");
              fprintf (drvui->fpoutv,
                "  material Material {diffuseColor  %s}\n", drvui->ellips_col_v[i]);
              fprintf (drvui->fpoutv, "  }\n");
              fprintf (drvui->fpoutv, "  geometry Sphere {");
              fprintf (drvui->fpoutv, "   radius     1.}}]}\n");
            } else {
              if (strlen (drvui->ellips_col_v[i]) == 0)
                strcpy (drvui->ellips_col_v[i], "1 0 0");
              fprintf (drvui->fpoutv,"   scaleFactor %8.5f %8.5f %8.5f\n",
                  ev[0], ev[1], ev[2]);
              fprintf (drvui->fpoutv, "  }\n");
              fprintf (drvui->fpoutv, "  Material {\n");
              fprintf (drvui->fpoutv, "   shininess 0.3\n");
              fprintf (drvui->fpoutv, "   diffuseColor  %s\n", drvui->ellips_col_v[i]);
              fprintf (drvui->fpoutv, "  }\n");
              fprintf (drvui->fpoutv, "  Sphere {\n");
              fprintf (drvui->fpoutv, "   radius     1\n  }\n");
              fprintf (drvui->fpoutv, " }\n");
            }
            n_ellipsoids++;
            }					 /* if clipflag == 0 */
          }					 /* for on k */
        }					 /* if ellipsoid */
      }					         /* if nvert > 0 */
    }                                            /* for all atoms j */
  }						 /* for ellipsoids i */
  printf ("%4d ellipsoids\n", n_ellipsoids);
  fprintf (drvui->flout, "%4d ellipsoids generated\n", n_ellipsoids);
}						 /* end of generate_ellipsoids */

/* ************************************************************** */
/* ************************************************************** */

void generate_planes (void)

/* procedure to generate corners of planes */
{
  int i, j, k;             /* loop counters, etc */
  int nvert_start;         /* starting number for cations */
  int planeno;             /* number of current polyhedron */
  int Plane_Count;         /* number of planes output */
  int ii, jj, ll, l;
  int *no, *ns;
  float xm, ym, zm, vx, vy, vz, bm, vxs, vys, vzs, bms, phi, phi0, cosphi,
  radius;
  int outside;
#ifdef OPENGL
  float glr, glg, glb;
#endif
  
  radius = rad_edge * 0.01f * Scale;
  if (drvui->nplane == 1)
    return;           /* exit NOW if no work */
  Plane_Count = 0;
  for (i = 0; i < 3; ++i)
    boxlim[i] += 3.0f;          /* expand box for cation generation */
  nvert = 0;                    /* make list empty */
  for (i = 0; i < natom; ++i)   /* loop through atoms */
      find_all_in_box (i);      /* generate all potential cations */
  for (i = 0; i < 3; ++i)
    boxlim[i] -= 3.0f;          /* shrink box to original size */
  if (nvert == 0)
    return;                     /* if no anions, no polyhedra */
  nvert_start = nvert;    /* save start of cation list */
  for (i = 0; i < natom; ++i) { /* start loop for cations */
    nvert = nvert_start;  /* get rid of previous cations */
    planeno = (drvui->atom_n[i] >> 16) & 255;     /* save plane number */
    if (planeno > 0 && drvui->plane_fn[planeno] == drvui->frame_no) { /* greater than zero for polyhedral centers */
      find_all_in_box (i);      /* add all polyhedral centers in box */
      if (nvert_start < nvert) {/* check for some found */
        draw_list = 0;          /* initialize the draw list */
        for (j = nvert_start; j < nvert; ++j) {   /* loop through cations */
          outside = 0;
	  if (clipflag) {
	    for (ii = 0; ii < 3; ii++) {
	      if (o_vert[3 * j + ii] < drvui->clip_lim[drvui->frame_no][ii] -0.01
		  || o_vert[3 * j + ii] > drvui->clip_lim[drvui->frame_no][ii + 3] +0.01)
		outside = 1;
	    }
	  }
          if (domolcomp && packflag) {
            for (ii = 0; ii < 3; ii++) {
	      if (o_vert[3 * j + ii] < drvui->cryst_lim[drvui->frame_no][ii] -0.01
		  || o_vert[3 * j + ii] > drvui->cryst_lim[drvui->frame_no][ii + 3] +0.01)
                outside = 1;
            }
          }
          numb_list = 0;            /* list clear */
          for (k = 0; k < nvert_start; ++k) {   /* loop through anions */
            if (dist (j, k) <= 0.) continue;
            if (dist (j, k) <= drvui->plane_size[planeno] + 0.0001 && !outside)  /* add vertex k to poly list */
              vertex_list[numb_list++] = k;
          }

          if (numb_list > 2) {       /* only do if at least a triangle */
            fprintf (drvui->flout,"Polygon with %d vertices around %c%c%c%c at %8.5f %8.5f %8.5f\n",
                   numb_list, drvui->atom_l[i][0], drvui->atom_l[i][1],drvui->atom_l[i][2],
                   drvui->atom_l[i][3], s_vert[3 * j], s_vert[3 * j + 1],
                   s_vert[3 * j + 2]);
            if (no_comment == 0)
              fprintf (drvui->fpoutv,"# Polygon (%d) around %c%c%c%c at %8.5f %8.5f %8.5f \n",
                   numb_list, drvui->atom_l[i][0], drvui->atom_l[i][1], drvui->atom_l[i][2],
                   drvui->atom_l[i][3], o_vert[3 * j], o_vert[3 * j + 1],
                   o_vert[3 * j + 2]);
            if (Vrml2) {
              fprintf (drvui->fpoutv, " Shape {");
              fprintf (drvui->fpoutv, "appearance Appearance {\n");
              fprintf (drvui->fpoutv,"   material Material {diffuseColor %s} \n \n",
                  drvui->plane_col_v[planeno]);
            } else {
              fprintf (drvui->fpoutv, " Separator {\n        ");
              fprintf (drvui->fpoutv, "  Material {\n");
              fprintf (drvui->fpoutv, "   diffuseColor %s \n  }\n",drvui->plane_col_v[planeno]);
            }
            Plane_Count++;       /* increment number output */
            poly_list[0] = numb_list;
/* Make vertices consecutive - find center of polygon */
            xm = ym = zm = 0.0f;
            no = (int *)malloc ((unsigned)((4 * numb_list) * sizeof (int)));
            ns = (int *)malloc ((unsigned)((4 * numb_list) * sizeof (int)));
            for (ii = 0; ii < numb_list; ii++) {
              no[ii] = vertex_list[ii];
              xm = xm + s_vert[3 * no[ii]];
              ym = ym + s_vert[3 * no[ii] + 1];
              zm = zm + s_vert[3 * no[ii] + 2];
            }
            xm = xm / (float) numb_list;
            ym = ym / (float) numb_list;
            zm = zm / (float) numb_list;

/* sort vertices in consecutive order based on minimal angles between vectors */

            for (ii = 0; ii < numb_list; ii++)
              ns[ii] = 0;
            ns[0] = no[0];
            no[numb_list] = 0;
            for (ii = 0; ii < numb_list - 1; ii++) {
              phi0 = 1000.0f;
              vxs = s_vert[3 * ns[ii]] - xm;
              vys = s_vert[3 * ns[ii] + 1] - ym;
              vzs = s_vert[3 * ns[ii] + 2] - zm;
              bms = (float) sqrt (vxs * vxs + vys * vys + vzs * vzs);
              for (jj = 0; jj < numb_list; jj++) {
                for (ll = 0; ll < ii; ll++)
                  if (ns[ll] == no[jj])
                    jj++;
                if (ns[ii] != no[jj] && jj < numb_list) {
                  vx = s_vert[3 * no[jj]] - xm;
                  vy = s_vert[3 * no[jj] + 1] - ym;
                  vz = s_vert[3 * no[jj] + 2] - zm;
                  bm = (float) sqrt (vx * vx + vy * vy + vz * vz);
                  cosphi = (vxs * vx + vys * vy + vzs * vz) / (bms * bm);
                  if (cosphi < -1.)
                    cosphi = -1.0f;
                  if (cosphi > 1.)
                    cosphi = 1.0f;
                  phi = (float) acos (cosphi) * (float) RAD;
                  if (phi < phi0) {
                    phi0 = phi;
                    ns[ii + 1] = no[jj];
                  }
                }
              }
            }
            draw_list = 1;
            poly_list[0] = numb_list;
            for (l = 0; l < numb_list; ++l)
              poly_list[draw_list++] = ns[l] + 1;
            poly_list[draw_list++] = 0;     /* terminate polygon list */
            plot_vrml_poly (planeno);
            fprintf (drvui->fpoutp," /* Polygon (%d) around %c%c%c%c at %8.5f %8.5f %8.5f */ \n",
                  numb_list, drvui->atom_l[i][0], drvui->atom_l[i][1], drvui->atom_l[i][2],
                  drvui->atom_l[i][3], o_vert[3 * j], o_vert[3 * j + 1],
                  o_vert[3 * j + 2]);
#ifdef OPENGL
            if (gl_view){
              glPushMatrix ();
              (void)sscanf (drvui->plane_col_v[planeno], "%f %f %f", &glr, &glb, &glg);
              glColor3f (glr, glb, glg);
              glBegin (GL_TRIANGLES);
            }
#endif
            for (l = 0; l < numb_list - 2; ++l) {
              fprintf (drvui->fpoutp," triangle {< %8.5f, %8.5f, %8.5f>,\n",
                   s_vert[3 * ns[0]], s_vert[3 * ns[0] + 1],s_vert[3 * ns[0] + 2]);
              fprintf (drvui->fpoutp,"<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>\n",
                   s_vert[3 * ns[l + 1]],s_vert[3 * ns[l + 1] + 1],
                   s_vert[3 * ns[l + 1] + 2], s_vert[3 * ns[l + 2]],
                   s_vert[3 * ns[l + 2] + 1], s_vert[3 * ns[l + 2] + 2]);
              fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n}\n",drvui->plane_col_p[planeno]);
#ifdef OPENGL
              if (gl_view) {
                glVertex3f (s_vert[3 * ns[0]],s_vert[3 * ns[0] + 1], s_vert[3 * ns[0] + 2]);
                glVertex3f (s_vert[3 * ns[l + 1]],s_vert[3 * ns[l + 1] + 1],s_vert[3 * ns[l + 1] + 2]);
                glVertex3f (s_vert[3 * ns[l + 2]],s_vert[3 * ns[l + 2] + 1],s_vert[3 * ns[l + 2] + 2]);
              }
#endif
            }
#ifdef OPENGL
            if (gl_view) {
              glEnd ();
              glPopMatrix ();
            }
#endif
            if (edges == 1) {
              for (l = 0; l < numb_list - 1; ++l) {
                fprintf (drvui->fpoutp,
                     " cylinder { <%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f> , %8.5f \n",
                     s_vert[3 * ns[l]], s_vert[3 * ns[l] + 1], s_vert[3 * ns[l] + 2],
                     s_vert[3 * ns[l + 1]], s_vert[3 * ns[l + 1] + 1],
                     s_vert[3 * ns[l + 1] + 2], radius);
                fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n}\n", col_edge_p);
              }
              fprintf (drvui->fpoutp,
                   " cylinder { <%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f> , %8.5f \n",
                   s_vert[3 * ns[0]], s_vert[3 * ns[0] + 1],
                   s_vert[3 * ns[0] + 2], s_vert[3 * ns[numb_list - 1]],
                   s_vert[3 * ns[numb_list - 1] + 1],
                   s_vert[3 * ns[numb_list - 1] + 2], radius);
              fprintf (drvui->fpoutp," texture{pigment{color %s  }}\n}\n", col_edge_p);
            }
            draw_list = 0;
            free (no);
            free (ns);
          }           /* numb_list > 2 */
        }           /* loop on j through cations */
      }             /* nvert_start < nvert */
    }
  }             /* end of loop through cations */
  printf ("%4d planes.\n", Plane_Count);
  fprintf (drvui->flout, "Generated %4d planes.\n", Plane_Count);
}

/* ************************************************************** */
/* ************************************************************** */

void generate_poly (void)

/* procedure to generate vertices of polyhedra */
{
  int i, j, k;             /* loop counters, etc */
  int nvert_start;         /* starting number for cations */
  int polyno;              /* number of current polyhedron */
  int Poly_Count;          /* Number of polyhedra output */
  int Max_Vertices;        /* Maximum number of vertices */
  float radius;
  int ii, outside;
  int type_check;          /* if atom type checking needs to be done */

  radius = rad_edge * 0.01f *Scale;

  if (drvui->npoly == 1)
    return;                /* exit NOW if no work */
  Poly_Count = 0;
  for (i = 0; i < 3; ++i)
    boxlim[i] += 3.0f;     /* expand box for anion generation */
  nvert = 0;               /* make list empty */
  for (i = 0; i < natom; ++i) {       /* loop through atoms */
      find_all_in_box (i); /* generate all potential anions */
  }
  for (i = 0; i < 3; ++i)
    boxlim[i] -= 3.0f;     /*shrink box to original size */
  if (nvert == 0)
    return;                /* if no anions, no polyhedra */
  nvert_start = nvert;/* save start of cation list */
  for (i = 0; i < natom; ++i) {       /* start loop for cations */
    nvert = nvert_start;   /* get rid of previous cations */
    type_check = 0;
    polyno = (drvui->atom_n[i] >> 8) & 255;  /* save polyhedron number */
    if (polyno > 0 && drvui->poly_fn[polyno] == drvui->frame_no) {  /* greater than zero for polyhedral centers */
      if (drvui->poly_t[polyno][0] != '\0')
        type_check = 1;
      find_all_in_box (i);       /* add all polyhedral centers in box */
      if (nvert_start < nvert) {     /* check for some found */
        draw_list = 0;         /* initialize the draw list */
        Max_Vertices = 0;
        for (j = nvert_start; j < nvert; ++j) {   /* first loop through cations */
          numb_list = 0;       /* list clear */
          outside = 0;
          if (clipflag != 0) {
            for (ii = 0; ii < 3; ii++) {
              if (o_vert[3 * j + ii] < drvui->clip_lim[drvui->frame_no][ii] - 0.01
                  || o_vert[3 * j + ii] > drvui->clip_lim[drvui->frame_no][ii + 3] + 0.01)
                outside = 1;
            }
          }
          if (outside == 1)
            continue;
          for (k = 0; k < nvert_start; ++k) {   /* loop through anions */
            float d;
            int kk;

            kk = drvui->orig_atom_no[k];
            d = dist (j, k);
            if (d > 0.0 && d <= drvui->poly_size[polyno]) {  /* add vertex k to poly list */
              if ((!type_check) || ((drvui->atom_l[kk][0] == drvui->poly_t[polyno][0])
                   && check_atom_name(drvui->atom_l[kk], drvui->poly_t[polyno])))
                if (d >= drvui->poly_min[polyno])
                  numb_list++;
            }
          }
          if (numb_list > Max_Vertices)
            Max_Vertices = numb_list;
        }                      /* end of first pass through list */
        draw_list = 0;         /* initialize the draw list */
        for (j = nvert_start; j < nvert; ++j) {   /* second loop through cations */
          numb_list = 0;       /* list clear */
          for (k = 0; k < nvert_start; ++k) {   /* loop through anions */
            float d;
            int kk;

            kk = drvui->orig_atom_no[k];
            outside = 0;
            if (domolcomp && packflag) {
              for (ii = 0; ii < 3; ii++) {
                if (o_vert[3 * k + ii] < drvui->cryst_lim[drvui->frame_no][ii] - 0.01 
                 || o_vert[3 * k + ii] > drvui->cryst_lim[drvui->frame_no][ii + 3] + 0.01)
                  outside = 1;
              }
            }
            if (clipflag != 0) {
              for (ii = 0; ii < 3; ii++) {
                if (o_vert[3 * k + ii] < drvui->clip_lim[drvui->frame_no][ii] - 0.01
                 || o_vert[3 * k + ii] > drvui->clip_lim[drvui->frame_no][ii + 3] + 0.01)
                  outside = 1;
              }
            }
            d = dist (j, k);
            if (d > 0.0 && (d <= drvui->poly_size[polyno] + 0.000001) && !outside) {  /* add vertex k to poly list */
              if (d >= drvui->poly_min[polyno]) {
                if ((type_check == 0) || 
                         (check_atom_name(drvui->atom_l[kk], drvui->poly_t[polyno])))
                  vertex_list[numb_list++] = k;
              }
            }
          }
          if (numb_list > 3 && numb_list == Max_Vertices) {  /* only do if polyhedron complete */
            fprintf (drvui->flout,"Polyhedron with %d vertices around %c%c%c%c at %8.5f %8.5f %8.5f \n",
                  numb_list, drvui->atom_l[i][0], drvui->atom_l[i][1], drvui->atom_l[i][2],
                  drvui->atom_l[i][3], s_vert[3 * j], s_vert[3 * j + 1],
                  s_vert[3 * j + 2]);
            fprintf (drvui->fpoutp," /* Polyhedron (%d) around %c%c%c%c at %8.5f %8.5f %8.5f */ \n",
                  numb_list, drvui->atom_l[i][0], drvui->atom_l[i][1], drvui->atom_l[i][2],
                  drvui->atom_l[i][3], o_vert[3 * j], o_vert[3 * j + 1],
                  o_vert[3 * j + 2]);
            if (no_comment == 0)
              fprintf (drvui->fpoutv,"# Polyhedron (%d) around %c%c%c%c at %8.5f %8.5f %8.5f \n",
                  numb_list, drvui->atom_l[i][0], drvui->atom_l[i][1],drvui->atom_l[i][2],
                  drvui->atom_l[i][3], o_vert[3 * j], o_vert[3 * j + 1],
                  o_vert[3 * j + 2]);
            if (Vrml2) {
              fprintf (drvui->fpoutv, " Shape {");
              fprintf (drvui->fpoutv,"appearance Appearance {\n  material Material { diffuseColor  %s} \n \n",
                   drvui->poly_col_v[polyno]);
            } else {
              fprintf (drvui->fpoutv, " Separator {\n    ");
              fprintf (drvui->fpoutv,"  Material {\n   diffuseColor  %s \n  }\n",drvui->poly_col_v[polyno]);
            }
            if (drvui->p_rad_edge[polyno] != 0.)
              radius = drvui->p_rad_edge[polyno] * 0.01f *Scale;

            build_poly_list (numb_list, polyno); /* select polygons */
            poly_list[draw_list++] = 0;     /* terminate polygon list */
            plot_vrml_poly (polyno);     /* pov and openGL part already done in build_poly_list */
            draw_list = 0;
            Poly_Count++;
          }
        }
      }             /* nvert_start < nvert */
    }
  }             /* end of loop through cations */
  printf ("%4d polyhedra.\n", Poly_Count);
  fprintf (drvui->flout, "Generated %4d polyhedra.\n", Poly_Count);
}

/* ************************************************************** */
/* ************************************************************** */

void generate_texts (void)

/* procedure to create POV and VRML command sequences for rendering of
   user-defined labels (labeltext command) - the GL rendering is done
   in generate_gl_texts */
{
  int i, n;
  int Label_Count = 0;
  float vert[3];
  float size;

  if (drvui->nlabel == 1)
    return;

  for (n = 1; n < drvui->nlabel; n++) {
    if (drvui->triple[0] == n) continue;         /* skip "vector_trip" */
    if (drvui->frame_no != drvui->label_fn[n]) continue;

    nvert = 0;           /* initialize vertex list */
    for (i = 0; i < 3; i++) {
      vert[i] = drvui->label_x[n][i];
      if (Display_axes && ((n==drvui->triple[0]) || (n==drvui->triple[1])
                 || (n == drvui->triple[2]) || (n==drvui->triple[3])))
          vert[i] += origin[i];
    }
    add_vert_nc (vert);

    if (Display_axes && ((n==drvui->triple[0]) || (n==drvui->triple[1])
                 || (n == drvui->triple[2]) || (n==drvui->triple[3])))
        for (i=0; i<3; i++)
            s_vert[i] += offset[i];
    Label_Count++;
    fprintf (drvui->flout, "Label %.3f %.3f %.3f %s\n", vert[0], vert[1], vert[2]
        ,drvui->label_label[n]);
    fprintf (drvui->fpoutp, "/* Labels */\n");
    fprintf (drvui->fpoutv, "# Labels\n");
    size = drvui->label_scale * 8.0f * Text_Size;
    fprintf (drvui->fpoutp, "  text { ttf \"crystal.ttf\",\"%s\" 0.15,0.\n",
	     drvui->label_label[n]);
    fprintf (drvui->fpoutp, "    scale <%4.2f,%4.2f,%4.2f>\n", size, size, size);
    fprintf (drvui->fpoutp, "    rotate <0, 0, -zrot>\n");
    fprintf (drvui->fpoutp, "    rotate <0, -yrot, 0>\n");
    fprintf (drvui->fpoutp, "    rotate <-xrot,0, 0>\n");
    fprintf (drvui->fpoutp,
	     "    translate <%8.5f,%8.5f,%8.5f> pigment{color Black}\n",
	     s_vert[0], s_vert[1], s_vert[2]);
    fprintf (drvui->fpoutp, "  }\n");
    if (Vrml2) {
      fprintf (drvui->fpoutv, "   Transform {");
      fprintf (drvui->fpoutv, " translation %8.5f %8.5f %8.5f \n", s_vert[0],
	       s_vert[1], s_vert[2]);
      fprintf (drvui->fpoutv, " rotation 1 0 0 %f\n",-xrot/RAD);
      fprintf (drvui->fpoutv, "children Transform { rotation 0 1 0 %f\n",-yrot/RAD);
      fprintf (drvui->fpoutv, "children Transform { rotation 0 0 1 %f\n",-zrot/RAD); 
      fprintf (drvui->fpoutv, "  children [ \n");
      fprintf (drvui->fpoutv, "   Billboard {\n    axisOfRotation 0 1 0\n");
      fprintf (drvui->fpoutv, "    children [ \n");
      fprintf (drvui->fpoutv, "     Shape {\n");
      fprintf (drvui->fpoutv, "      appearance Appearance{\n");
      fprintf (drvui->fpoutv, "       material Material {diffuseColor 0 0 0}\n");
      fprintf (drvui->fpoutv, "      }\n");
      fprintf (drvui->fpoutv, "      geometry Text{string [\"%s\"]}\n", drvui->label_label[n]);
      fprintf (drvui->fpoutv, "     }\n    ]\n");
      fprintf (drvui->fpoutv, "   }\n  ]\n }}}\n");
    } else {
      fprintf (drvui->fpoutv, " Separator {\n  Translation {");
      fprintf (drvui->fpoutv, " translation %8.5f %8.5f %8.5f }\n", s_vert[0],
	       s_vert[1], s_vert[2]);
      fprintf (drvui->fpoutv, "  Rotation{ rotation 1 0 0 %8.5f}\n", -xrot / RAD);
      fprintf (drvui->fpoutv, "  Rotation{ rotation 0 1 0 %8.5f}\n", -yrot / RAD);
      fprintf (drvui->fpoutv, "  Rotation{ rotation 0 0 1 %8.5f}\n", -zrot / RAD);
      size = drvui->label_scale * Text_Size;
      fprintf (drvui->fpoutv, "  Scale{ scaleFactor %4.2f %4.2f %4.2f }\n",
	       size, size, size);
      fprintf (drvui->fpoutv, "  Material{ diffuseColor 0. 0. 0. }\n");
      fprintf (drvui->fpoutv, "  AsciiText { string  \"%s\" }\n", drvui->label_label[n]);
      fprintf (drvui->fpoutv, " }\n");
    }
  }						 /* end of loop on l0   */
  printf ("%4d labels.\n", Label_Count);
  fprintf (drvui->flout, "Generated %4d labels.\n", Label_Count);
}


/* ************************************************************** */
/* ************************************************************** */

void import_gsas (char input[100], int in_line, int Quick)

/* routine to extract structural information from a GSAS exp file */
{
  char string[100], filename[100], temp[5], tstring[20];
  int phase_no, i;
  FILE *impin;

  if (in_line) {
    fprintf (stderr,
	     "Cannot read inline GSAS data. Please use the import instruction\n");
    delay (5);
    exit (0);
  }
  (void) sscanf (input, "%s %s %d", filename, filename, &phase_no);	/* get file name and phase number */
  if ((phase_no < 1) || (phase_no > 9)) {
    fprintf (stderr,
	     "Illegal phase number in GSAS import instruction, Run aborted.\n");
    fprintf (drvui->flout,
	     "Illegal phase number in GSAS import instruction, Run aborted.\n");
    delay (5);
    exit (0);
  }
  if ((impin = fopen (filename, "r")) == NULL) {
    fprintf (stderr, "Cannot Open GSAS Import File, Run aborted.\n");
    fprintf (drvui->flout, "Cannot Open GSAS Import File, Run aborted.\n");
    delay (5);
    exit (0);
  }
  sprintf (temp, "CRS%d", phase_no);		 /* make search object */
  for (;;) {
    if (!fgets (string, 100, impin)) {;		 /* search for phase info */
      fprintf (stderr,
	       "Error reading phase info from GSAS Import File, Run aborted.\n");
      if (!Quick)
	fprintf (drvui->flout,
		 "Error reading phase info from GSAS Import File, Run aborted.\n");
      delay (5);
      exit (0);
    }
    if (strncmp (string, temp, 4) == 0)
      break;
  }
  for (;;) {					 /* search for lattice constants */
    if (!fgets (string, 100, impin)) {
      fprintf (stderr,
	       "Error reading lattice constants GSAS Import File, Run aborted.\n");
      if (!Quick)
	fprintf (drvui->flout,
		 "Error reading lattice constants GSAS Import File, Run aborted.\n");
      delay (5);
      exit (0);
    }
    (void) sscanf (string, "%s %s", tstring, tstring);
    if (strncmp (tstring, "ABC", 3) == 0) {
      (void) sscanf (string, "%s %s %f %f %f", tstring, tstring, &drvui->lat_con[0],
		     &drvui->lat_con[1]
		     , &drvui->lat_con[2]);		 /* read lattice constants */
      break;
    }
  }						 /* search for lattice constants */
  for (;;) {					 /* search for angles */
    if (!fgets (string, 100, impin)) {
      fprintf (stderr,
	       "Error reading lattice angles GSAS Import File, Run aborted.\n");
      if (!Quick)
	fprintf (drvui->flout,
		 "Error reading lattice angles GSAS Import File, Run aborted.\n");
      delay (5);
      exit (0);
    }
    (void) sscanf (string, "%s %s", tstring, tstring);
    if (strncmp (tstring, "ANGLES", 6) == 0) {
      (void) sscanf (string, "%s %s %f %f %f", tstring, tstring, &drvui->lat_con[3],
		     &drvui->lat_con[4], &drvui->lat_con[5]);		 /* read lattice constants */
      break;
    }
  }						 /* search for angles */
  if (!Quick)
    fprintf (drvui->flout,
	     "******      Cell     %8.5f %8.5f %8.5f %8.3f %8.3f %8.3f\n",
	     drvui->lat_con[0], drvui->lat_con[1], drvui->lat_con[2], drvui->lat_con[3], drvui->lat_con[4],
	     drvui->lat_con[5]);
  for (;;) {					 /* position to atomic coordinates */
    if (!fgets (string, 100, impin)) {
      fprintf (stderr,
	       "Error reading atomic coordinates from GSAS Import File, Run aborted.\n");
      if (!Quick)
	fprintf (drvui->flout,
		 "Error reading atomic coordinates from GSAS Import File, Run aborted.\n");
      delay (5);
      exit (0);
    }
    (void) sscanf (string, "%s %s", tstring, tstring);
    if (strncmp (tstring, "AT", 2) == 0)
      break;
  }						 /* position to atomic coordinates */

/* at this point, positioned at first atom line */
  for (;;) {					 /* process atom parameters */
    (void) sscanf (string, "%s %s", tstring, tstring);
    if (strncmp (tstring, "AT", 2) != 0)
      break;					 /* out when atom stuff done */
    (void) sscanf (string, "%s %s %s %s %f %f %f", tstring, tstring, tstring,
		   tstring, &drvui->atom_xyz[natom][0], &drvui->atom_xyz[natom][1],
		   &drvui->atom_xyz[natom][2]);
    drvui->atom_n[natom] = 1;
    for (i = 0; i < 4; i++)
      drvui->atom_l[natom][i] = ' ';			 /* initialize atom name */
    if (string[62] >= '0' && string[62] <= '9')
      drvui->atom_n[natom] = (int) string[62] - 48;
    else
      drvui->atom_l[natom][0] = string[62];
    if (string[63] >= '0' && string[63] <= '9')
      drvui->atom_n[natom] = 10 * drvui->atom_n[natom] + (int) string[63] - 48;
    else
      drvui->atom_l[natom][1] = string[63];
    if (string[64] >= '0' && string[64] <= '9')
      drvui->atom_n[natom] = 10 * drvui->atom_n[natom] + (int) string[64] - 48;
    else
      drvui->atom_l[natom][2] = string[64];
    if (string[65] >= '0' && string[65] <= '9')
      drvui->atom_n[natom] = 10 * drvui->atom_n[natom] + (int) string[65] - 48;
    else
      drvui->atom_l[natom][3] = string[65];
    if (!Quick)
      fprintf (drvui->flout,
	       "******      Atom %c%c%c%c%3d %8.5f %8.5f %8.5f\n",
	       drvui->atom_l[natom][0], drvui->atom_l[natom][1],
           drvui->atom_l[natom][2], drvui->atom_l[natom][3],
	       drvui->atom_n[natom], drvui->atom_xyz[natom][0],
           drvui->atom_xyz[natom][1], drvui->atom_xyz[natom][2]);
    (void) fgets (string, 100, impin);
    if (string[74] == 'A') {
      drvui->ell_type[n_ellips] = 1;			 /* Uij */
      (void) sscanf (string, "%s %s %s %f %f %f %f %f %f", tstring, tstring, tstring,
          &drvui->ellips[n_ellips][0], &drvui->ellips[n_ellips][1], &drvui->ellips[n_ellips][2],
          &drvui->ellips[n_ellips][3], &drvui->ellips[n_ellips][4], &drvui->ellips[n_ellips][5]);	/* read unique coefficients */
      for (i = 0; i < 4; i++)
        drvui->ellips_l[n_ellips][i] = drvui->atom_l[natom][i];
      drvui->ellips_n[n_ellips] = drvui->atom_n[natom];
      if (!Quick)
        fprintf (drvui->flout,
		 "******      Uij  %c%c%c%c%3d %8.5f %8.5f %8.5f %8.5f %8.5f %8.5f\n",
		 drvui->atom_l[natom][0], drvui->atom_l[natom][1],
         drvui->atom_l[natom][2], drvui->atom_l[natom][3],
         drvui->atom_n[natom], drvui->ellips[n_ellips][0],
		 drvui->ellips[n_ellips][1], drvui->ellips[n_ellips][2], drvui->ellips[n_ellips][3],
         drvui->ellips[n_ellips][4], drvui->ellips[n_ellips][5]);
      if (++n_ellips >= MAX_EL) {
        fprintf (stderr,"Too many ellipsoid types specified - increase MAX_EL.\n");
        delay (5);
        exit (0);
      }
    }						 /* string[74] == 'A' */
    if (++natom >= MAX_ATOM) {
      fprintf (stderr,
	       "Too many atoms in asymmetric unit - increase MAX_ATOM.\n");
      delay (5);
      exit (0);
    }
    if (!fgets (string, 100, impin)) {
      fprintf (stderr, "Error reading GSAS Import File, Run aborted.\n");
      if (!Quick)
	fprintf (drvui->flout, "Error reading GSAS Import File, Run aborted.\n");
      delay (5);
      exit (0);
    }
  }						 /* process atom parameters */
  for (;;) {					 /* start loop for Space Group information */
    if (!fgets (string, 100, impin)) {
      fprintf (stderr, "Error reading GSAS Import File, Run aborted.\n");
      if (!Quick)
	fprintf (drvui->flout, "Error reading GSAS Import File, Run aborted.\n");
      delay (5);
      exit (0);
    }
    (void) sscanf (string, "%s %s", tstring, tstring);
    if (strncmp (tstring, "SG", 2) == 0)
      break;					 /* out when space group found */
  }						 /* start loop for Space Group information */
  string[0] = 's';
  string[1] = 'p';
  string[2] = 'g';
  string[3] = 'r';
  for (i = 3; i < 60; i++)
    string[i + 1] = string[i + 10];
  string[60] = '\0';
  fprintf (drvui->flout, "******      %s\n", string);
  symop (string);
  (void) fclose (impin);
}

/* ************************************************************** */
/* ************************************************************** */

void import_shelx (char string[100], int in_line, int Quick)

/* routine to extract structural information from a SHELX data file */
{
  FILE *impin;
  char input[200];				 /* temporary for continuation cards */
  char filename[100];				 /* string for filename */
  int intype;					 /* type of datum read */
  int i, j, k;					 /* temporary */
  float xray;					 /* xray wavelength (read, but not used) */
  int read_start;				 /* flag for start of atom list */
  int read_uij;					 /* flag for uij card */
  float u11 = 0.0f, u22 = 0.0f;  /* temporary storage for uij */
  float site_occup = 0.0f;       /* occupancy (read, but not used) */
  static char shcmds[][68] =
	{ "ZERR", "LATT", "SYMM", "SFAC", "DISP", "UNIT", "LAUE",
    "REM ", "MORE", "TIME", "OMIT", "SHEL", "BASF", "TWIN",
    "EXTI", "SWAT", "HOPE", "MERG", "SPEC", "RESI", "MOVE",
    "ANIS", "AFIX", "HFIX", "FRAG", "FEND", "EXYZ", "EADP",
    "EQIV", "CONN", "PART", "BIND", "FREE", "DFIX", "DANG",
    "BUMP", "SAME", "SADI", "CHIV", "FLAT", "DELU", "SIMU",
    "DEFS", "ISOR", "NCSY", "SUMP", "L.S.", "CGLS", "BLOC",
    "DAMP", "STIR", "WGHT", "FVAR", "BOND", "CONF", "MPLA",
    "RTAB", "HTAB", "LIST", "ACTA", "SIZE", "TEMP", "WPDB",
    "FMAP", "GRID", "PLAN", "MOLE", "REM\0"
  };
  static char cmd[] = "    ";			 /*temporary storage for start of line */
  float fvarpar[30];
  int nfvar=0;
  int fv;
  int latt;
  char *p,*pp;
    
  memset (input, ' ', (size_t) 199);

  if (in_line) {
    impin = drvui->fpin;
  } else {
    (void) sscanf (string, "%*s %s", filename);	 /* get file name */
    if ((impin = fopen (filename, "r")) == NULL) {
      fprintf (stderr, "Cannot Open SHELX Import File, Run aborted.\n");
      exit (0);
    }
  }
  read_start = 0;
  read_uij = 0;
  latt = 0;
  if (drvui->sys == 0) { /* only do this if no spgr command given */
    /* initialize symmetry arrays */
    for (i = 0; i <= 23; ++i) {
      for (j = 0; j <= 2; ++j) {
        drvui->ts[i][j] = 0.0;
        for (k = 0; k <= 2; ++k)
          drvui->ss[i][j][k] = 0;
      }
    }
    for (i = 0; i <= 2; ++i) {
      drvui->lat_pos[0][i] = 0.0;
      drvui->ss[0][i][i] = 1;
      for (j = 0; j <= 3; ++j)
        drvui->spg[i][j] = ' ';
    }

  /* add trivial x,y,z symmetry */
    drvui->ts[0][0]=0;
    drvui->ts[0][1]=0;
    drvui->ts[0][2]=0;
    drvui->ss[0][0][0]=1;
    drvui->ss[0][1][1]=1;
    drvui->ss[0][2][2]=1;
    drvui->ng=1;
  }

  if (!Quick)
    fprintf (drvui->flout, "****** Now interpreting SHELX-style data\n");
  while (!feof (impin)) {
    if (!fgets (string, 100, impin)) {		 /* read a line */
      if (!Quick)
        fprintf (drvui->flout, "Error reading file.");
      exit (1);
    }
    trim_string(string,100);
    strcpy (input, string);
    if (strchr (input, '=') != NULL && 
    (strncmp(input,"TITL",4)&& strncmp(input,"REM",3))) {
      (*strchr (input, '=')) = '\0';
      if (!fgets (string, 100, impin)) {
        fprintf (stderr, "Error reading file.");
        exit (1);
      }
      strcat (input, string);
    }
    intype = 0;
    for (i = 0; i < 4; i++) {
      cmd[i] = (char) toupper (input[i]);
      if (input[i] == '\0')
	intype = 3;
    }

    if (strcmp (cmd, "TITL") == 0)
      intype = 1;				 /* echo and ignore */
    if (strcmp (cmd, "CELL") == 0)
      intype = 2;

    if (strcmp (cmd, "HKLF") == 0)
      intype = 4;				 /* trigger exit */
    if (strncmp (cmd, "END", 3) == 0)
      intype = 4;				 /* trigger exit */

    if (strncmp (cmd, "FVAR", 4) == 0)
      intype = 5;				 /* store fixed parameters */

    if (drvui->sys == 0) { /* ignore LATT/SYMM if spgr already seen */
      if (strncmp (cmd, "LATT", 4) == 0)
        intype = 6;

      if (strncmp (cmd, "SYMM", 4) == 0)
        intype = 7;
    }

    if (intype == 0) {				 /* check against list of 'other' commands */
      for (i = 0; i < 68; i++) {
	if (strlen (cmd) < 4 || strcmp (cmd, shcmds[i]) == 0) {
	  intype = 3;
	  break;
	}
      }
    }
    if ((strncmp (cmd, "REM", 3) == 0) || (strncmp(cmd,"    ",4) == 0))
      intype = 3;
    switch (intype) {
    case 1:
      trim_string(input,51);
      if (!Quick)
        fprintf (drvui->flout, "******      %s\n", input);
      break;					 /* ignore title except for listing */
    case 2:
      for (i = 3; i <= 5; ++i)
        drvui->lat_con[i] = 0.0f;
      (void) sscanf (input, "%*s %f %f %f %f %f %f %f", &xray, &drvui->lat_con[0]
		     , &drvui->lat_con[1], &drvui->lat_con[2], &drvui->lat_con[3], &drvui->lat_con[4]
		     , &drvui->lat_con[5]);
      if (!Quick)
          fprintf (drvui->flout,"******      Cell     %8.5f %8.5f %8.5f %8.3f %8.3f %8.3f\n",
              drvui->lat_con[0], drvui->lat_con[1], drvui->lat_con[2], drvui->lat_con[3],
              drvui->lat_con[4], drvui->lat_con[5]);
      break;
    case 3:					 /* just ignore most SHELX commands */
      break;
    case 4:
      read_start = 0;
      if (!Quick)
          fprintf (drvui->flout, "****** HKLF or END - end of SHELX dataset\n");
      if (!in_line)
          (void) fclose (impin);
      if (latt == 0 && drvui->sys == 0) {
          drvui->acentric = 0;    /* No LATT line given, spgp = P -1 */
          latt = 1;
          drvui->nlat = 1;
          drvui->nbr = 1;
      }
      if (latt == 3) drvui->sys = 5;
      if (latt == 4) drvui->sys = 6;
      if (drvui->sys == 0) findsys();
      if (drvui->autolabel==2) {
        for (i=0;i<natom;i++) {
          int nn=0;
          drvui->label_fn[drvui->nlabel]=drvui->frame_no;
          drvui->label_x[drvui->nlabel][0]=drvui->atom_xyz[i][0];
          drvui->label_x[drvui->nlabel][1]=drvui->atom_xyz[i][1];
          drvui->label_x[drvui->nlabel][2]=drvui->atom_xyz[i][2];
          for (j=0;j<4;j++) {
            if (drvui->atom_l[i][j] != ' ')
              drvui->label_label[drvui->nlabel][nn++]=drvui->atom_l[i][j];
          }
          drvui->label_label[drvui->nlabel][nn]='\0';
          sprintf(cmd,"%d",drvui->atom_n[i]);
          strcat(drvui->label_label[drvui->nlabel],cmd);

          drvui->nlabel++;
        }
        drvui->autolabel=1;
      }
      return;
    case 5: 
      i = sscanf(input,"%*s %f %f %f %f %f %f %f",
    		&fvarpar[nfvar],&fvarpar[nfvar+1],&fvarpar[nfvar+2],
    		&fvarpar[nfvar+3],&fvarpar[nfvar+4],&fvarpar[nfvar+5],
    		&fvarpar[nfvar+6]);
    		nfvar +=7;
      if (i<7) nfvar -= (7-i);
      break;
    case 6:
      (void) sscanf(input,"%*s %d",&latt);
      if (latt <0) {
        drvui->acentric = 1;
        latt *= -1;
      } else
        drvui->acentric = 0;
      switch (latt) {
      case 1:
        drvui->nlat = 1;
        drvui->nbr = 1;
        break;
      case 2:
        drvui->nlat = 2;
        drvui->nbr = 6;
        drvui->lat_pos[1][0] = 0.5f;
        drvui->lat_pos[1][1] = 0.5f;
        drvui->lat_pos[1][2] = 0.5f;
        break;
      case 3:
        drvui->nlat = 3;
        drvui->nbr = 7;
        drvui->lat_pos[1][0] = (1.f / 3.f);
        drvui->lat_pos[1][1] = (2.f / 3.f);
        drvui->lat_pos[1][2] = (2.f / 3.f);
        drvui->lat_pos[2][0] = (2.f / 3.f);
        drvui->lat_pos[2][1] = (1.f / 3.f);
        drvui->lat_pos[2][2] = (1.f / 3.f);
        break;
      case 4:
        drvui->nlat = 4;
        drvui->nbr = 5;
        drvui->lat_pos[1][0] = 0.0;
        drvui->lat_pos[1][1] = 0.5f;
        drvui->lat_pos[1][2] = 0.5f;
        drvui->lat_pos[2][0] = 0.5f;
        drvui->lat_pos[2][1] = 0.0;
        drvui->lat_pos[2][2] = 0.5f;
        drvui->lat_pos[3][0] = 0.5f;
        drvui->lat_pos[3][1] = 0.5f;
        drvui->lat_pos[3][2] = 0.0;
        break;
      case 5:
        drvui->nlat = 2;
        drvui->nbr = 2;
        drvui->lat_pos[1][0] = 0.0;
        drvui->lat_pos[1][1] = 0.5f;
        drvui->lat_pos[1][2] = 0.5f;
        break;
      case 6:
        drvui->nlat = 2;
        drvui->nbr = 3;
        drvui->lat_pos[1][0] = 0.5f;
        drvui->lat_pos[1][1] = 0.0;
        drvui->lat_pos[1][2] = 0.5f;
        break;
      case 7:
        drvui->nlat = 2;
        drvui->nbr = 4;
        drvui->lat_pos[1][0] = 0.5f;
        drvui->lat_pos[1][1] = 0.5f;
        drvui->lat_pos[1][2] = 0.0;
      }
      break;

    case 7: /* SYMM */
      drvui->ng++;
      input[0]=' ';
      input[1]=' ';
      input[2]=' ';
      input[3]=' ';
      p = strstr(input,",");
      *p='\0';
      getsym(input,drvui->ng-1,0);
      p++;
      pp = strstr(p,",");
      *pp='\0';
      getsym(p,drvui->ng-1,1);
      pp++;
      getsym(pp,drvui->ng-1,2);
      break;

    default:
      site_occup=1.;
      u11=u22=0.05f;
      i = sscanf (input, "%c%c%c%c %d %f %f %f %f %f %f %f %f %f %f",
		  &drvui->atom_l[natom][0], &drvui->atom_l[natom][1],
		  &drvui->atom_l[natom][2], &drvui->atom_l[natom][3],
		  &drvui->atom_n[natom], &drvui->atom_xyz[natom][0],
		  &drvui->atom_xyz[natom][1], &drvui->atom_xyz[natom][2],
		  &site_occup, &u11, &u22, &drvui->ellips[n_ellips][2],
		  &drvui->ellips[n_ellips][5], &drvui->ellips[n_ellips][4],
		  &drvui->ellips[n_ellips][3]);
      for (j = 0; j < 3; j++) {
        if (fabs(drvui->atom_xyz[natom][j]) >15.) {
          fv=((int)fabs(drvui->atom_xyz[natom][j]))/10;
          if (drvui->atom_xyz[natom][j] <0.)
            drvui->atom_xyz[natom][j] = 1.0f- fvarpar[fv -1]* (drvui->atom_xyz[natom][j]-fv*10);
          else
            drvui->atom_xyz[natom][j] = fvarpar[fv -1]* (drvui->atom_xyz[natom][j]-fv*10);
        } else { 
          if (drvui->atom_xyz[natom][j] > 9.)
            drvui->atom_xyz[natom][j] -= 10.0f;
        }
      }
      if (fabs(site_occup) >11.) { 
         fv= ((int)fabs(site_occup))/10;
         if (site_occup <0)
           site_occup = 1.0f-fvarpar[fv-1];
         else
           site_occup = fvarpar[fv-1];
      }
      if (fabs(u11) >10.) { 
        fv = ((int)fabs(u11))/10;
        if (u11 <0.)
          u11 = 1.0f- fvarpar[fv-1] *((float)fabs(u11)-fv);
        else
          u11 = fvarpar[fv-1] * (u11-fv);
      }	

      drvui->atom_n[natom] = 0;
      k = 0;
      for (j = 0; j < 4; j++)			 /* extract atom number if part of SHELX name */
        if (drvui->atom_l[natom][j] >= '0' && drvui->atom_l[natom][j] <= '9') {
          drvui->atom_n[natom] = 10 * drvui->atom_n[natom]
              + (int) drvui->atom_l[natom][j] - 48;
        } else {
            drvui->atom_l[natom][k++] = drvui->atom_l[natom][j];
        }
      for (j=k;j<4;j++)
        drvui->atom_l[natom][j] = ' ';
      if (drvui->atom_n[natom] == 0)
        drvui->atom_n[natom] = 1;

      if (i > 11) {
        drvui->ell_type[n_ellips] = 1;			 /* Uij */
        strcpy(drvui->ellips_l[n_ellips],drvui->atom_l[natom]);
        drvui->ellips_n[n_ellips] = drvui->atom_n[natom];
        if (fabs(u22) >10.) { 
          fv = ((int)fabs(u22))/10;
          if (u22 <0.)
            u22 = 1.0f - fvarpar[fv-1] * ((float)fabs(u22)-fv*10);
          else
            u22 = fvarpar[fv-1] * (u22-fv*10);
        }
        if (fabs(drvui->ellips[n_ellips][2]) >10.) { 
          fv= ((int)fabs(drvui->ellips[n_ellips][2]))/10;
          if (drvui->ellips[n_ellips][2] <0. )
            drvui->ellips[n_ellips][2] = 1.0f -fvarpar[fv-1] * ((float)fabs(drvui->ellips[n_ellips][2])-fv*10);
          else
            drvui->ellips[n_ellips][2] =  fvarpar[fv-1] * (drvui->ellips[n_ellips][2]-fv*10);
        }
        if (fabs(drvui->ellips[n_ellips][5]) >10.) { 
          fv= ((int)fabs(drvui->ellips[n_ellips][5]))/10;
          if (drvui->ellips[n_ellips][5] <0. )
            drvui->ellips[n_ellips][5] = 1.0f -fvarpar[fv-1] * ((float)fabs(drvui->ellips[n_ellips][5])-fv*10);
          else
            drvui->ellips[n_ellips][5] =  fvarpar[fv-1] * (drvui->ellips[n_ellips][5]-fv*10);
        }
        if (drvui->ellips[n_ellips][4] >10.) { 
          fv= ((int)fabs(drvui->ellips[n_ellips][4]))/10;
          if (drvui->ellips[n_ellips][4] <0. )
            drvui->ellips[n_ellips][4] = 1.0f -fvarpar[fv-1] * ((float)fabs(drvui->ellips[n_ellips][4])-fv*10);
          else
            drvui->ellips[n_ellips][4] =  fvarpar[fv-1] * (drvui->ellips[n_ellips][4]-fv*10);
        }
        if (drvui->ellips[n_ellips][3] >10.) { 
          fv= ((int)fabs(drvui->ellips[n_ellips][3]))/10;
          if (drvui->ellips[n_ellips][3] <0. )
            drvui->ellips[n_ellips][3] = 1.0f -fvarpar[fv-1] * ((float)fabs(drvui->ellips[n_ellips][3])-fv*10);
          else
            drvui->ellips[n_ellips][3] =  fvarpar[fv-1] * (drvui->ellips[n_ellips][3]-fv*10);
        }

        drvui->ellips[n_ellips][0] = u11;
        drvui->ellips[n_ellips][1] = u22;
        if (!Quick)
         fprintf (drvui->flout,
		   "******      Atom %c%c%c%c%3d %8.5f %8.5f %8.5f \n",
		   drvui->atom_l[natom][0], drvui->atom_l[natom][1],
		   drvui->atom_l[natom][2], drvui->atom_l[natom][3]
		   , drvui->atom_n[natom], drvui->atom_xyz[natom][0],
		   drvui->atom_xyz[natom][1], drvui->atom_xyz[natom][2]);
        if (!Quick)
          fprintf (drvui->flout,
		   "******      Uij  %c%c%c%c%3d %8.5f %8.5f %8.5f %8.5f %8.5f %8.5f\n",
		   drvui->atom_l[natom][0], drvui->atom_l[natom][1],
		   drvui->atom_l[natom][2], drvui->atom_l[natom][3]
		   , drvui->atom_n[natom], drvui->ellips[n_ellips][0],
		   drvui->ellips[n_ellips][1], drvui->ellips[n_ellips][2]
		   , drvui->ellips[n_ellips][3], drvui->ellips[n_ellips][4],
		   drvui->ellips[n_ellips][5]);
        do_ellipsoids = 1;
        drvui->save_el_number[n_ellips] = drvui->ellips_n[n_ellips];
        memset(drvui->ellips_col_p[n_ellips],0,40);
        memset(drvui->ellips_col_v[n_ellips],0,40);
        strcpy(drvui->ellips_col_p[n_ellips],"Gray30");
        strcpy(drvui->ellips_col_v[n_ellips],"0.3 0.3 0.3");
        n_ellips++;
        if (n_ellips >= MAX_EL) {
          fprintf (stderr,
		   "Too many ellipsoid types specified - increase MAX_EL.\n");
          exit (0);
        }
      } else {
        if (i < 10) {
          if (!Quick)
            fprintf (drvui->flout,
		     "******WARNING - too few parameters given for following atom:\n");
          if (site_occup < 0)
            site_occup = 0.;
          if (u11 < 0)
            u11 = 0.;
        } 
        if (!Quick)
          fprintf (drvui->flout,
		   "******      Atom %c%c%c%c%3d %8.5f %8.5f %8.5f %8.5f %8.5f\n",
		   drvui->atom_l[natom][0], drvui->atom_l[natom][1],
		   drvui->atom_l[natom][2], drvui->atom_l[natom][3]
		   , drvui->atom_n[natom], drvui->atom_xyz[natom][0],
		   drvui->atom_xyz[natom][1], drvui->atom_xyz[natom][2], site_occup, u11);
      }
      natom++;
      if (natom >= MAX_ATOM) {
        fprintf (stderr,
		 "Too many atoms in asymmetric unit - increase MAX_ATOM.\n");
        exit (0);
      }
      break;                     /* end of default */
    }                            /* end of switch */
  }
  if (!in_line)
    (void) fclose (impin);
  if (drvui->autolabel==2) {
    for (i=0;i<natom;i++) {
      int nn=0;
      drvui->label_fn[drvui->nlabel]=drvui->frame_no;
      drvui->label_x[drvui->nlabel][0]=drvui->atom_xyz[i][0];
      drvui->label_x[drvui->nlabel][1]=drvui->atom_xyz[i][1];
      drvui->label_x[drvui->nlabel][2]=drvui->atom_xyz[i][2];
      for (j=0;j<4;j++) {
        if (drvui->atom_l[i][j] != ' ')
          drvui->label_label[drvui->nlabel][nn++]=drvui->atom_l[i][j];
      }
      drvui->label_label[drvui->nlabel][nn]='\0';
      sprintf(cmd,"%d",drvui->atom_n[i]);
      strcat(drvui->label_label[drvui->nlabel],cmd);

      drvui->nlabel++;
    }
    drvui->autolabel=1;
  }
}						 /* end of import_shelx */

/* ************************************************************** */
/* ************************************************************** */

void import_fdat (char input[100], int in_line, int Quick)

/* routine to extract structural information from a CSD 'FDAT' data file */
{
  FILE *impin;
  char string[100];				 /* temporary string */
  char filename[100];				 /* string for filename */
  int i, j;					 /* temporary */
  int k, l, k3, k4;				 /*flags for space group symbol parsing */
  char spstring[100];				 /* space group symbol */
  char lat_s[6][7];				 /*temporary storage for lattice constants */
  int ldec[6];					 /*temporary storage for decimal places of lattice constants */
  char number[4];
  char ncons[4], nrfacs[4], nrems[4], ndiss[4], nerrs[4], noprs[4],
	nrads[4], nats[4];
  char sats[4], nbnds[4];
  int ncards, nrfac = 0, nrem = 0, ndis = 0, nerr = 0, nopr = 0, nrad =
	0, nat = 0, sat = 0, ncon = 0, cellflag;
  int intflag, coordflag, centflag, errflag = 0;
  int nskip;					 /* lines to skip between lattice constants and atoms */
  int natomin;

  memset (spstring, 0, 100);
  memset (string, 0, 100);
  for (i = 0; i < 6; i++)
    memset (lat_s[i], 0, 7);
  if (in_line) {
    impin = drvui->fpin;
  }
  else {
    (void) sscanf (input, "%s %s", filename, filename);	/* get file name */
    if ((impin = fopen (filename, "r")) == NULL) {
      fprintf (stderr, "Cannot Open CSD FDAT Import File, Run aborted.\n");
      delay (5);
      exit (0);
    }
  }

  fprintf (drvui->flout, "****** Now interpreting CSD-style data\n");

  if (!fgets (input, 100, impin)) {		 /* read a line */
    fprintf (drvui->flout, "Error reading file.");
    delay (5);
    exit (1);
  }

/* the first line should contain the REFCODE of the database entry */
/* and information about the quality of the data and the length of */
/* the following sections                                          */

  if (strncmp (input, "#", 1) == 0) {

    (void) sscanf (input,
		   "%23c%3d%3c%3c%3c%3c%3c%3c%3c%3c%3c%3c%1d%1d%1d%1d%1d",
		   string, &ncards, nrfacs, nrems, ndiss, nerrs, noprs, nrads,
		   nats, sats, nbnds, ncons, &cellflag, &intflag, &coordflag,
		   &centflag, &errflag);

    nrfac = atoi (nrfacs);
    ncon = atoi (ncons);
    nrem = atoi (nrems);
    ndis = atoi (ndiss);
    nerr = atoi (nerrs);
    nopr = atoi (noprs);
    nrad = atoi (nrads);
    nat = atoi (nats);
    sat = atoi (sats);
    input[0] = ' ';
    input[9] = '\0';				 /* truncate title */
    fprintf (drvui->flout, "******      %s\n", input);
    if (!cellflag) {
      fprintf (drvui->flout,
	       "Error - this CSD entry has neither CELL nor COORDINATES\n");
      delay (5);
      exit (1);
    }

    if (!coordflag) {
      fprintf (drvui->flout, "Error - this CSD entry has no COORDINATES\n");
      delay (5);
      exit (1);
    }

    if (errflag) {
      fprintf (drvui->flout,
	       "Warning - this CSD entry has the ERROR flag set -\n at least some coordinates may be wrong");
    }

    if (!fgets (input, 100, impin)) {		 /* read a line */
      fprintf (drvui->flout, "Error reading file.");
      delay (5);
      exit (1);
    }
  }

/* lattice constants come in a fixed FORTRAN format followed by integers */
/* encoding precision, esd's, etc and finally by the spacegroup symbol */

  for (i = 0; i <= 5; ++i)
    drvui->lat_con[i] = 0.0f;

  (void) sscanf (input, "%6c%6c%6c%6c%6c%6c%1d%1d%1d%1d%1d%1d%21c%s",
		 lat_s[0], lat_s[1], lat_s[2], lat_s[3], lat_s[4], lat_s[5],
		 &ldec[0], &ldec[1], &ldec[2], &ldec[3], &ldec[4], &ldec[5],
		 string, string);

  for (i = 0; i < 6; ++i) {
    drvui->lat_con[i] = (float) (atof (lat_s[i]) / pow (10, (double) ldec[i]));
  }

/* separate class symbol from other symmetry elements */
  spstring[0] = 's';
  spstring[1] = 'p';
  spstring[2] = 'g';
  spstring[3] = 'r';
  spstring[4] = ' ';
  spstring[5] = string[0];
  spstring[6] = ' ';
  j = 6;
  k = 0;
  k3 = 0;
  k4 = 0;
  l = (int) strlen (string);
  for (i = 1; i <= l; i++) {
    switch (string[i]) {
    case ' ':
      break;
    case '/':
      spstring[++j] = string[i];
      break;
    default:
      j++;
      /* do we have to collate this symbol (for screw axes or mirror planes) ? */
      if (string[i] > '0' && string[i] <= '9') { /* it is a number */
	if (string[0] == 'R' ||			 /* no screw axes in R */
	    l == 4 ||				 /* no screw axes in four letter names */
	    k != 0 ||				 /* have just appended to a number */
	    k3 != 0 ||				 /* cannot have further screw axes after 3_x */
	    (k4 != 0 && spstring[j - 1] == '3')) {	/* cannot have 3_x after a 4_x */
	  spstring[j] = ' ';			 /* so separate it */
	  k = 0;				 /* and reset the collation flag */
	  j++;
	}
	else if (spstring[j - 1] != '-' && (spstring[j - 1] <= string[i]
					    || spstring[j - 1] >= 'a')) {
	  /* neither a minus sign nor a higher number precedes it */
	  spstring[j] = ' ';			 /* so separate it */
	  k = 0;				 /* and reset the collation flag */
	  j++;
	}
	else {
	  k++;					 /* set flag for collation */
	  if (spstring[j - 1] == '3')
	    k3 = 1;				 /* if 3_x screw, prohibit further screw axes */
	  if (spstring[j - 1] == '4')
	    k4 = 1;				 /* if 4_x screw, prohibit 3_x screw axes */
	}
      }
      else {					 /* not a number */
	if (spstring[j - 1] != '/') {		 /*but not a mirror plane either */
	  spstring[j] = ' ';			 /* separate it */
	  j++;
	}
      }
      spstring[j] = string[i];			 /* now add the symbol */
      break;
    }						 /*switch */
  }						 /*for i */
  spstring[j + 1] = '\0';

/* Space group 90, P 4 21 2, is not correctly identified by this code. 
   It accounts for only 14 of the 175000 entries in the CSD (and not all 
   of them have coordinates), so this quick and dirty fix is all it can
   expect from us. Does this spacegroup really exist ? :-) */
  if (!strcmp (string, "P4212"))
    strcpy (spstring, "P 4 21 2");

  fprintf (drvui->flout, "******      Space group *%s*\n", spstring);
  symop (spstring);

  fprintf (drvui->flout,
	   "******      Cell     %8.5f %8.5f %8.5f %8.3f %8.3f %8.3f\n",
	   drvui->lat_con[0], drvui->lat_con[1], drvui->lat_con[2], drvui->lat_con[3], drvui->lat_con[4],
	   drvui->lat_con[5]);

/*skip R value, any textual information, symmetry codes and radii data */

  nskip = (nrfac + nrem + ndis + nerr) / 80 + 1; /* skip comments (80chars per line) */
  nskip += nopr / 5 + 1;			 /* skip symmetry operators (five per line) */
  nskip += nrad / 16 + 1;			 /* skip atomic radii (sixteen per line) */


  for (i = 0; i < nskip; i++) {
    if (!fgets (input, 100, impin)) {		 /* read a line */
      fprintf (drvui->flout, "Error reading file while skipping records.");
      fprintf (drvui->flout,
	       "nrfac:%d, nrem:%d, ndis:%d,nerr:%d, nopr:%d, nrad:%d, nskip:%d\n",
	       nrfac, nrem, ndis, nerr, nopr, nrad, nskip);
      delay (5);
      exit (1);
    }
  }

  natomin = natom;				 /* number of atoms already in list */

/* read coordinates until empty atom symbol encountered */

  for (i = 0; i < (nat + sat) / 3 + 1; i++) {	 /*for all expected atom records */

    if (!fgets (input, 100, impin)) {		 /* read a line */
      fprintf (drvui->flout, "Error reading file while reading atom coordinates.");
      delay (5);
      exit (1);
    }

/* atoms come in groups of three */
    (void) sscanf (input,
		   "%c%c%c%c %7f %7f %7f %c%c%c%c %7f %7f %7f %c%c%c%c %7f %7f %7f",
		   &drvui->atom_l[natom][0], &drvui->atom_l[natom][1], &drvui->atom_l[natom][2],
		   &drvui->atom_l[natom][3], &drvui->atom_xyz[natom][0],
		   &drvui->atom_xyz[natom][1], &drvui->atom_xyz[natom][2],
		   &drvui->atom_l[natom + 1][0], &drvui->atom_l[natom + 1][1],
		   &drvui->atom_l[natom + 1][2], &drvui->atom_l[natom + 1][3],
		   &drvui->atom_xyz[natom + 1][0], &drvui->atom_xyz[natom + 1][1],
		   &drvui->atom_xyz[natom + 1][2], &drvui->atom_l[natom + 2][0],
		   &drvui->atom_l[natom + 2][1], &drvui->atom_l[natom + 2][2],
		   &drvui->atom_l[natom + 2][3], &drvui->atom_xyz[natom + 2][0],
		   &drvui->atom_xyz[natom + 2][1], &drvui->atom_xyz[natom + 2][2]);

/* and the atom list is terminated by numerical data */
    if (drvui->atom_l[natom][0] < 'A') {
      break;
    }

/* scale coordinates */
    for (j = 0; j < 3; j++) {
      drvui->atom_xyz[natom][j] = drvui->atom_xyz[natom][j] / (float) 100000.;
      drvui->atom_xyz[natom + 1][j] = drvui->atom_xyz[natom + 1][j] / (float) 100000.;
      drvui->atom_xyz[natom + 2][j] = drvui->atom_xyz[natom + 2][j] / (float) 100000.;
    }

/* separate numbers from labels and remove trailing identifiers */

    for (l = 0; l < 3; l++) {
      k = 0;
      for (j = 1; j < 4; j++) {
        number[k] = ' ';
        if (drvui->atom_l[natom + l][j] >= '0' && drvui->atom_l[natom + l][j] <= '9') {
          number[k] = drvui->atom_l[natom + l][j];
          k++;
        }
        if (k)
          drvui->atom_l[natom + l][j] = ' ';
      }
      number[k] = '\0';
      drvui->atom_n[natom + l] = atoi (number);
    }

    if (Unique_Atom () == 1) {
/* report converted data of first atom */
      fprintf (drvui->flout,
	       "******      Atom %c%c%c%c%3d %8.5f %8.5f %8.5f\n",
	       drvui->atom_l[natom][0], drvui->atom_l[natom][1],
           drvui->atom_l[natom][2], drvui->atom_l[natom][3],
	       drvui->atom_n[natom], drvui->atom_xyz[natom][0],
           drvui->atom_xyz[natom][1], drvui->atom_xyz[natom][2]);
      natom++;
      if (natom >= MAX_ATOM) {			 /* first unique */
	fprintf (drvui->flout,
		 "Too many atoms in asymmetric unit - increase MAX_ATOM.\n");
	delay (5);
	exit (0);
      }
    }						 /* first unique */
/* return now if no second atom on the line */
    if (natom == nat + natomin) {
      break;					 /* got all expected atoms */
    }
    if (drvui->atom_l[natom][0] == '\0')
      break;
    if (Unique_Atom () == 1) {
/* report converted data of second atom */
      fprintf (drvui->flout,
	       "******      Atom %c%c%c%c%3d %8.5f %8.5f %8.5f\n",
	       drvui->atom_l[natom][0], drvui->atom_l[natom][1], drvui->atom_l[natom][2],
	       drvui->atom_l[natom][3]
	       , drvui->atom_n[natom], drvui->atom_xyz[natom][0], drvui->atom_xyz[natom][1],
	       drvui->atom_xyz[natom][2]);
      natom++;
      if (natom >= MAX_ATOM) {			 /* second unique */
	fprintf (drvui->flout,
		 "Too many atoms in asymmetric unit - increase MAX_ATOM.\n");
	delay (5);
	exit (0);
      }
    }						 /* second unique */
/* return now if no third atom on the line */
    if (natom == nat + natomin) {
      break;					 /* got all expected atoms */
    }
    if (drvui->atom_l[natom][0] == '\0')
      break;
    if (Unique_Atom () == 1) {			 /* third unique */
/* report converted data of third atom */
      fprintf (drvui->flout,
	       "******      Atom %c%c%c%c%3d %8.5f %8.5f %8.5f\n",
	       drvui->atom_l[natom][0], drvui->atom_l[natom][1],
           drvui->atom_l[natom][2], drvui->atom_l[natom][3],
           drvui->atom_n[natom], drvui->atom_xyz[natom][0],
           drvui->atom_xyz[natom][1], drvui->atom_xyz[natom][2]);
      natom++;
      if (natom >= MAX_ATOM) {
	fprintf (drvui->flout,
		 "Too many atoms in asymmetric unit - increase MAX_ATOM.\n");
	delay (5);
	exit (0);
      }
      if (natom == nat + natomin) {
	break;					 /* got all expected atoms */
      }
    }						 /* third unique */
  }						 /*for all expected atom records */
  for (i++; i < (nat + sat) / 3; i++) {		 /* skip remainder of atom records */

    if (!fgets (input, 100, impin)) {		 /* read a line */
      fprintf (drvui->flout,
	       "Error reading file while skipping extra atom records.");
      delay (5);
      exit (1);
    }
  }
  if (!errflag) {				 /* only error-free entries have connectivity records */
    nskip = ncon / 40 + 1;
    if (nat + sat > 100)
      nskip = ncon / 26 + 1;
    for (i = 0; i < nskip; i++) {
      if (!fgets (input, 100, impin)) {		 /* read a line */
	if (!in_line)
	  (void) fclose (impin);
	return;
      }
    }
  }
  if (!in_line)
    (void) fclose (impin);

}						 /* end of import_fdat */

/* ************************************************************** */
/* ************************************************************** */

int Unique_Atom (void)

/* routine to check if atom at position 'natom' is unique. Returns 1 if it
   is. If the position is a duplicate, returns 0. */
{
  int j, k, test;

  for (k = 0; k < natom - 1; k++) {		 /* loop through atoms */
    if (drvui->atom_n[k] == drvui->atom_n[natom]) {
      test = 0;
      for (j = 0; j < 4; j++) {
        if (drvui->atom_l[k][j] != drvui->atom_l[natom][j])
          test = 1;
      }
      if (test == 0) {				 /* name and number the same - check coordinates */
        expand_atom (natom);			 /* find all of test atoms in unit cell */
        for (j = 0; j < ncell; j++) {
          if ((fabs (drvui->atom_xyz[k][1] - drvui->cell_xyz[j][1]) < 0.0001) &&
              (fabs (drvui->atom_xyz[k][2] - drvui->cell_xyz[j][2]) < 0.0001) &&
              (fabs (drvui->atom_xyz[k][0] - drvui->cell_xyz[j][0]) < 0.0001))
            return (0);				 /* duplicate */
        }
      }
    }
  }
  return (1);					 /* unique */
}

/* ************************************************************** */
/* ************************************************************** */

void show_slab(void)
{

#ifdef OPENGL
  if (!gl_view) return;
  glPushMatrix ();

  glLineWidth(3);
  glColor3f (1., 0., 0.);
  glBegin (GL_LINES);

  glVertex3f (slabx1, slaby1, slabz1);
  glVertex3f (slabx2, slaby2, slabz2);

  glVertex3f (slabx3, slaby3, slabz3);
  glVertex3f (slabx1, slaby1, slabz1);

  glVertex3f (slabx3, slaby3, slabz3);
  glVertex3f (slabx2+(slabx3-slabx1), slaby2+(slaby3-slaby1),slabz2+(slabz3-slabz1));

  glVertex3f (slabx2, slaby2, slabz2);
  glVertex3f (slabx2+(slabx3-slabx1), slaby2+(slaby3-slaby1),slabz2+(slabz3-slabz1));

  glVertex3f (slabx1, slaby1, slabz1);
  glVertex3f (slabx4, slaby4, slabz4);


  glVertex3f (slabx3, slaby3, slabz3);
  glVertex3f (slabx3+(slabx4-slabx1), slaby3+(slaby4-slaby1),slabz3+(slabz4-slabz1));

  glVertex3f (slabx4, slaby4, slabz4);
  glVertex3f (slabx3+(slabx4-slabx1), slaby3+(slaby4-slaby1),slabz3+(slabz4-slabz1));

  glVertex3f (slabx2, slaby2, slabz2);
  glVertex3f (slabx2+(slabx4-slabx1),slaby2+(slaby4-slaby1),slabz2+(slabz4-slabz1));


  glVertex3f (slabx2+(slabx4-slabx1),slaby2+(slaby4-slaby1),slabz2+(slabz4-slabz1));
  glVertex3f (slabx4, slaby4, slabz4);

  glVertex3f (slabx2+(slabx4-slabx1),slaby2+(slaby4-slaby1),slabz2+(slabz4-slabz1));
  glVertex3f (slabx2+(slabx4-slabx1)+slabx3-slabx1,
              slaby2+(slaby4-slaby1)+slaby3-slaby1,
              slabz2+(slabz4-slabz1)+slabz3-slabz1);


  glVertex3f (slabx3+(slabx4-slabx1), slaby3+(slaby4-slaby1),slabz3+(slabz4-slabz1));
  glVertex3f (slabx2+(slabx4-slabx1)+slabx3-slabx1,
              slaby2+(slaby4-slaby1)+slaby3-slaby1,
              slabz2+(slabz4-slabz1)+slabz3-slabz1);

  glVertex3f (slabx2+(slabx3-slabx1), slaby2+(slaby3-slaby1),slabz2+(slabz3-slabz1));
  glVertex3f (slabx2+(slabx4-slabx1)+slabx3-slabx1,
              slaby2+(slaby4-slaby1)+slaby3-slaby1,
              slabz2+(slabz4-slabz1)+slabz3-slabz1);
  glEnd ();
  glPopMatrix ();
  glLineWidth(1);
#endif
}

/* ************************************************************** */
/* ************************************************************** */

void push_cylinder (float df[3], float at[3], float radius, char * color)

/* routine to push an openGL cylinder that represents a unit-cell edge */
{
#ifdef OPENGL
    float beta, gamma;
    float glr,glg,glb;
    float d;
    GLUquadricObj *glu_quadObj;

    if (!gl_view) return;
    d = (float)sqrt(df[0]*df[0]+df[1]*df[1]+df[2]*df[2]);
    beta = (float) atan2 (df[0], df[1] + 0.0000001);    /* rotation angle about Z (in radians) */
    gamma = (float) sqrt (df[0] * df[0] + df[1] * df[1]);
    gamma = (float) atan2 (gamma, df[2]);    /* Rotation angle about X */
    glu_quadObj = gluNewQuadric ();
    glPushMatrix ();
    (void)sscanf (color, "%f %f %f", &glr, &glg, &glb);
    glColor3f (glr, glg, glb);
    glTranslatef (at[0], at[1], at[2]);
    glRotatef (-beta * (float) RAD, 0.0f, 0.0f, 1.0f);
    glRotatef (-gamma * (float) RAD, 1.0f, 0.0f, 0.0f);
    gluQuadricDrawStyle (glu_quadObj, GLU_FILL);
    gluCylinder (glu_quadObj,radius,radius, d, 10, 1);
#if !defined (WIN32) && !defined (__APPLE__)
    free(glu_quadObj);
#endif
    glPopMatrix ();
#endif
}

/* ************************************************************** */
/* ************************************************************** */

void generate_gl_texts (void)
/* procedure to render all textual labels in the OpenGL view - this has
   to be done outside the display list so that the labels always face
   toward the viewer during rotations */
{
#ifdef OPENGL
  int i, n;
  float vert[3];
  char *p;
  float scale;
#endif

  if (Display_axes) {
      nvert = 0;
      generate_triple();
#ifdef OPENGL
      if (gl_view)
          draw_GL_triple();
#endif
  }
#ifdef OPENGL
  if (!gl_view) return;
  for (n = 1; n < drvui->nlabel; n++) {
    if (drvui->label_fn[n] != drvui->frame_no) continue;
    nvert = 0;           /* initialize vertex list */
    for (i = 0; i < 3; i++) {
        vert[i] = drvui->label_x[n][i];
        if (Display_axes && ((n==drvui->triple[0]) || (n==drvui->triple[1])
                 || (n == drvui->triple[2]) || (n==drvui->triple[3])))
            vert[i] += origin[i];
    }
    add_vert_nc (vert);

    if (Display_axes && ((n==drvui->triple[0]) || (n==drvui->triple[1])
                 || (n == drvui->triple[2]) || (n==drvui->triple[3])))
        for (i=0; i<3; i++)
            s_vert[i] += offset[i];

    glPushMatrix();
    glLoadName(n);
    glTranslatef(s_vert[0],s_vert[1],s_vert[2]);
    scale = 0.005f;
    if (!strcmp(drvui->label_label[n],"triple_vect"))
        scale *= 0.00001f;                  // make "triple_vect" really small
    glRotatef (-xrot, 1.0f, 0.0f, 0.0f);
    glRotatef (-yrot, 0.0f, 1.0f, 0.0f);
    glRotatef (-zrot, 0.0f, 0.0f, 1.0f);
    glColor3f (0.0f, 0.0f, 0.0f);
    glDisable(GL_LIGHTING);
    glLineWidth(2.0);
    glScalef(drvui->label_scale * scale, drvui->label_scale * scale,
        drvui->label_scale * scale);
    for (p=drvui->label_label[n];*p;p++)
      glutStrokeCharacter(GLUT_STROKE_ROMAN,*p);
    glLineWidth(1.0);
    glEnable(GL_LIGHTING);
    glPopMatrix();
  }             /* end of loop on n   */

#endif
}

/* ************************************************************** */
/* ************************************************************** */

void findsys(void) 
/* routine determine crystal system from symmetry information */
{
  int i,j,k;
  int n,kk=0,nn,symlist[23];
  int yzx=0,zxy=0,ymxmz=0,mxmyz=0,myxmz=0;

  if (drvui->ng==1) { /* only x,y,z => triclinic */
    drvui->sys = 1;
    return;
  }

  if (drvui->ng==2) { /* only two symops => monoclinic */
    drvui->sys = 2;
    return;
  }

  drvui->sys = 3; /* orthorhombic unless we find specific operators below */

  for (i = 0; i < drvui->ng; ++i) {
    for (j = 0; j <= 2; ++j) {
      if(fabs(drvui->ts[i][j]-0.33333333) <0.00001) drvui->sys=5;
      if(fabs(drvui->ts[i][j]-0.16666667) <0.00001) drvui->sys=5;
      if(fabs(drvui->ts[i][j]-0.66666667) <0.00001) drvui->sys=5;
      if(fabs(drvui->ts[i][j]-0.83333333) <0.00001) drvui->sys=5;
    }
    if (drvui->sys==5) return;
  }

  if (fabs(drvui->lat_con[5] - 120.) < 0.00001
      && fabs(drvui->lat_con[4] - 90.) <0.0001) {
    drvui->sys = 5;
    return;
  }

  n=0;
  for (i = 0; i < drvui->ng; ++i) {
    for (j = 0; j <= 2; ++j) {
       if (drvui->ts[i][j] != 0.0) break;
       kk=0;
       for (k = 0; k <= 2; ++k)
         if (drvui->ss[i][j][k] != 0) kk++;
    }
    if (kk >1) {
      drvui->sys = 5;
      return;
    }
    symlist[n++]=i;
  }

  for (nn=0; nn < n; nn++) {
    i=symlist[nn];
    if (drvui->ss[i][0][1]==1 && 
        drvui->ss[i][1][2]==1 &&
        drvui->ss[i][2][0]==1) yzx=1;
    if (drvui->ss[i][0][2]==1 && 
        drvui->ss[i][1][0]==1 &&
        drvui->ss[i][2][1]==1) zxy=1;
    if (drvui->ss[i][0][1]==1 && 
        drvui->ss[i][1][0]==-1 &&
        drvui->ss[i][2][2]==-1) ymxmz=1;
    if (drvui->ss[i][0][0]==-1 && 
        drvui->ss[i][1][1]==-1 &&
        drvui->ss[i][2][2]==1) mxmyz=1;
    if (drvui->ss[i][0][1]==-1 && 
        drvui->ss[i][1][0]==1 &&
        drvui->ss[i][2][2]==-1) myxmz=1;
  }

  if (ymxmz+mxmyz+myxmz==3) drvui->sys=4; /*tetragonal*/
  if (yzx+zxy==2) drvui->sys=6; /*cubic*/
}

/* ************************************************************** */
/* ************************************************************** */

void getsym(char* text, int num, int kk)

/* helper routine to parse SHELX symmetry operator strings */
{ 
    int i,j;
    char *txt;
    char *n;
    char *d;

    txt = (char*)malloc(24*sizeof(char));
 
    j=0;
    for (i=0;i<23;i++)
      if (text[i] != ' ' && text[i] != '\n')
        txt[j++]=toupper(text[i]);

    txt[j]='\0';

    drvui->ts[num][kk] = 0.;

    n = strstr(txt,"/");
    if (n != NULL) {
    n++;
    d = n-3;
    if (atof(d)==0.) d=n-2;
        drvui->ts[num][kk] = (float)(atof(d) / atof(n));
    }

    n= strstr(txt,".");
    if (n != NULL) {
        drvui->ts[num][kk] = (float)atof(n);
        if (n!=txt) {
            if (txt[&txt-&n+1] == '-' )
                drvui->ts[num][kk] *= -1;
        }
    }

    drvui->ss[num][kk][0]=0;
    drvui->ss[num][kk][1]=0;
    drvui->ss[num][kk][2]=0;

    for (i=0;i<(int)strlen(txt);i++) {
        if (txt[i]=='X') {
            drvui->ss[num][kk][0]=1;
            if (i > 0) {
                if (txt[i-1]=='-') drvui->ss[num][kk][0]=-1;
            }
        }
        if (txt[i]=='Y') {
             drvui->ss[num][kk][1]=1;
             if (i > 0) {
                 if (txt[i-1]=='-') drvui->ss[num][kk][1]=-1;
             }
        }     
        if (txt[i]=='Z') {
            drvui->ss[num][kk][2]=1;
            if (i >0) {
                if (txt[i-1]=='-') drvui->ss[num][kk][2]=-1;
            }
        }
    }
    free(txt);
}

