/* ===================================================================== */
/*     DRAWxtl                                                           */
/*                                                                       */
/* Routine to read crystallographic information such as unit cell, space */
/* group, atom positions, and thermal ellipsoid parameters to generate   */
/* POLYHEDRON, SPHERE, ELLIPSOID and BOND geometry POV and VRML scenes.  */
/*                                                                       */
/* Part 3                                                                */
/*                                                                       */
/* Coded by:   L.W. Finger                                               */
/*            Geophysical Laboratory                                     */
/*            Carnegie Institution of Washington                         */
/*            Washington DC 20015                                        */
/*            Larry.Finger@lwfinger.net                                  */
/*                                                                       */
/* Original Version: April 92                                            */
/*            Revised January 1993 to add bonds.                         */
/*            Converted to AVS Module February 1993.                     */
/*            Thermal Ellipsoids added April 1996                        */
/*            Thermal ellipsoids for POV and VRML finished Nov. 1996     */
/*                                                                       */
/*                                                                       */
/* Converted to standalone POV scene file generator Oct. 1994/March 1995 */
/* and VRML generator added September 1995/March 1996                    */
/* by M. Kroeker, Organic Chemistry, Technische Hochschule Darmstadt,    */
/* Petersenstr. 22, D-64287 Darmstadt, Germany                           */
/*                                                                       */
/* Two versions merged  July 1995                                        */
/* ===================================================================== */

/************************************************************************
 *  This file contains routines:
 * build_poly_list - create list of points for polyhedron generation
 * check_color - check that a color is in the POV list and get RGB values for VRML
 * crystal_compute - do the computations for a crystal
 * draw_cell - draw the unit-cell edges
 * generate_spheres - generate the spheres in the drawing
 * import_cif - read structural parameters from a CIF (Crystallographic Information File)    
 * import_schakal - read structural parameters from a Schakal file
 * trim_string - remove any trailing \r, \n or ' ' from the end of a string
 *
 *************************************************************************/

#include "drawxtl.h"
#include "draw_ext.h"
#ifndef WIN32
  #include <unistd.h>
#endif

/* ************************************************************** */
/* ************************************************************** */

void build_poly_list (int numb_list, int polyno)

/* routine to take list of 'numb_list' indices in 's_vert'
   and find each plane on surface of convex polyhedron.  It then builds
   a list of pointers to these vertices for drawing program.  The algorithm
   is as follows:
   
   1.  If there are 4 vertices, the polyhedron is a tetrahedron and all
       combinations of 3 vertices constitute a face.

   2.  For more than 4 vertices, some planes will be interior.  These are
       found by choosing three points (indicated by i,j,k) and forming two
       vectors defined by V1 = P(j) - P(i) and V2 = P(k) - P(i).  For each
       additional point (l), a third vector V3 = P(l) - P(i) is formed and
       the volume enclosed by these three vectors V = V3 . V1 x V2 is
       calculated.  If this volume is zero, point l lies in the plane.  If
       the plane is on the surface of the polyhedron, all non-zero volumes
       must have the same sign, which may be positive or negative.
*/
/*  numb_list - number of candidate vertices */
/*  polyno - number of polygon - for polygon color */
{
  int i, j, k, l, m;             /* local loop counters */
  float v1[3], v2[3], v3[3];     /* vectors used in calculation */
  float p[3];                    /* coordinates of point i */
  float vol, sign_z;
  int ind;                       /* indicator of surface plane */
  int *no;                       /* place to keep vertex id's */
  int imk, jmk, kmk;
  int ii, jj, ll;
  double phi, cosphi;
  float xm, ym, zm, vx, vy, vz, bm, phi0, vxs, vys, vzs, bms;
  float d1, d2, d3, radius;
  int *ns;
  char edgecolor[40];
  char color[128];
#ifdef OPENGL
  float glr, glg, glb;
  float df[3], at[3];
#endif
  int render;
  
  if ((no = (int *) malloc ((unsigned) ((4 * numb_list) * sizeof (int)))) == NULL) {
    printf ("Unable to get no allocation\n");
    return;
  }
  if ((ns = (int *) malloc ((unsigned) ((4 * numb_list) * sizeof (int)))) == NULL) {
    printf ("Unable to get ns allocation\n");
    free (no);
    return;
  }
  radius = rad_edge * 0.005f *Scale;

  if (drvui->p_rad_edge[polyno] != 0.)
    radius = drvui->p_rad_edge[polyno] * 0.005f *Scale;

  if (DepthCue != 0.) {
/*calculate z coordinate after rotation, and scale edge width accordingly*/
    radius = 0.0f;
    for (j = 0; j <= 2; j++)
      radius += (float) G_Rot[j][2] * s_vert[3 * no[0] + j];
    radius = rad_edge * 0.005f *Scale + DepthCue * radius;
  }
  for (render = 0; render < 2; render++) {   /* 1st pass to draw polyhedron, 2nd for edges */
#ifdef OPENGL
    if (gl_view && !render) {
      glPushMatrix ();
      (void)sscanf (drvui->poly_col_v[polyno], "%f %f %f", &glr, &glg, &glb);
      glDisable (GL_LIGHTING);
      glColor3f (glr, glg, glb);
      glBegin (GL_TRIANGLES);
    }
#endif
    for (i = 0; i < numb_list - 2; ++i) {
      no[0] = vertex_list[i];			 /* first point */
      for (m = 0; m <= 2; ++m)
        p[m] = s_vert[3 * no[0] + m];
      for (j = i + 1; j < numb_list - 1; ++j) {
        no[1] = vertex_list[j];			 /* second point */
        for (k = 0; k <= 2; ++k)
          v1[k] = s_vert[3 * no[1] + k] - p[k];
        for (k = j + 1; k < numb_list; ++k) {
          if (numb_list == 4) {             /* tetrahedron */
            poly_list[draw_list++] = 3;         /* number of vertices for triangle */
            poly_list[draw_list++] = vertex_list[i] + 1;
            poly_list[draw_list++] = vertex_list[j] + 1;
            poly_list[draw_list++] = vertex_list[k] + 1;
            if (draw_list > 4 * NvertM) {
              fprintf (drvui->flout,"Overrun of poly_list.\n");
              return;
            }

            imk = vertex_list[i];
            jmk = vertex_list[j];
            kmk = vertex_list[k];
            d1 = (float) fabs (s_vert[3 * imk] - s_vert[3 * jmk]) +
               (float) fabs (s_vert[3 * imk + 1] - s_vert[3 * jmk + 1]) +
               (float) fabs (s_vert[3 * imk + 2] - s_vert[3 * jmk + 2]);
            d2 = (float) fabs (s_vert[3 * imk] - s_vert[3 * kmk]) +
               (float) fabs (s_vert[3 * imk + 1] - s_vert[3 * kmk + 1]) +
               (float) fabs (s_vert[3 * imk + 2] - s_vert[3 * kmk + 2]);
            d3 = (float) fabs (s_vert[3 * kmk] - s_vert[3 * jmk]) +
               (float) fabs (s_vert[3 * kmk + 1] - s_vert[3 * jmk + 1]) +
               (float) fabs (s_vert[3 * kmk + 2] - s_vert[3 * jmk + 2]);
            if ((d1 > 0.01) && (d2 > 0.01) && (d3 > 0.01)) {
              if (!render) { 
                fprintf (drvui->fpoutp,
                   " triangle{< %8.5f, %8.5f, %8.5f>,\n",
                   s_vert[3 * imk], s_vert[3 * imk + 1], s_vert[3 * imk + 2]);
                fprintf (drvui->fpoutp,
                   "<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>  texture{pigment{color %s  }}\n",
                   s_vert[3 * jmk], s_vert[3 * jmk + 1], s_vert[3 * jmk + 2], 
                   s_vert[3 * kmk], s_vert[3 * kmk + 1], s_vert[3 * kmk + 2],
                   drvui->poly_col_p[polyno]);
                fprintf(drvui->fpoutp," }\n");
#ifdef OPENGL
                if (gl_view) {
                  glVertex3f (s_vert[3 * imk], s_vert[3 * imk + 1], s_vert[3 * imk + 2]);
                  glVertex3f (s_vert[3 * jmk], s_vert[3 * jmk + 1], s_vert[3 * jmk + 2]);
                  glVertex3f (s_vert[3 * kmk], s_vert[3 * kmk + 1], s_vert[3 * kmk + 2]);
                }
#endif
              } else {                /* !render */
                if (edges == 1) {
                  if (drvui->p_rad_edge[polyno] != 0.)
                    strcpy (edgecolor, drvui->p_col_edge_p[polyno]);
                  else
                    strcpy (edgecolor, col_edge_p);
                  strcpy(color,edgecolor);
                  check_color(color);
                  fprintf (drvui->fpoutp,
                     " cylinder { <%f,%f,%f>,<%f,%f,%f>,%f\n",
                     s_vert[3 * imk], s_vert[3 * imk + 1], s_vert[3 * imk + 2], 
                     s_vert[3 * jmk], s_vert[3 * jmk + 1], s_vert[3 * jmk + 2], radius);
                  fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n }\n", edgecolor);
                  fprintf (drvui->fpoutp," cylinder { <%f,%f,%f>,<%f,%f,%f>,%f\n",
                     s_vert[3 * jmk], s_vert[3 * jmk + 1], s_vert[3 * jmk + 2], 
                     s_vert[3 * kmk], s_vert[3 * kmk + 1], s_vert[3 * kmk + 2], radius);
                  fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n }\n", edgecolor);
                  fprintf (drvui->fpoutp," cylinder { <%f,%f,%f>,<%f,%f,%f>,%f \n",
                     s_vert[3 * kmk], s_vert[3 * kmk + 1], s_vert[3 * kmk + 2], 
                     s_vert[3 * imk], s_vert[3 * imk + 1], s_vert[3 * imk + 2], radius);
                  fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n }\n", edgecolor);
#ifdef OPENGL
                  if (gl_view) {
                    for (m=0; m<3; m++) {
                      at[m] = s_vert[3 * imk + m];
                      df[m] = s_vert[3 * jmk + m] - s_vert[3 * imk + m];
                    }
                    push_cylinder(df, at, radius, color); 
                    for (m=0; m<3; m++) {
                      at[m] = s_vert[3 * jmk + m];
                      df[m] = s_vert[3 * kmk + m] - s_vert[3 * jmk + m];
                    }
                    push_cylinder(df, at, radius, color); 
                    for (m=0; m<3; m++) {
                      at[m] = s_vert[3 * kmk + m];
                      df[m] = s_vert[3 * imk + m] - s_vert[3 * kmk + m];
                    }
                    push_cylinder(df, at, radius, color); 
                  } 
#endif
                } 
              } 
            } 
          } else {                         /* not tetrahedron */
            no[2] = vertex_list[k];		 /* third point */
            drvui->nplane = 3;
            for (l = 0; l <= 2; ++l)
              v2[l] = s_vert[3 * no[2] + l] - p[l];	/* get v2 */
            sign_z = 0.0f;
            ind = 0;
            v3[0] = v1[1] * v2[2] - v1[2] * v2[1];
            v3[1] = v1[2] * v2[0] - v1[0] * v2[2];
            v3[2] = v1[0] * v2[1] - v1[1] * v2[0];
            for (l = 0; l < numb_list; ++l) {
              vol = 0.0f;
              for (m = 0; m <= 2; ++m)
                vol = vol + (s_vert[3 * vertex_list[l] + m] - p[m]) * v3[m];
              if ((float) fabs (vol) < drvui->polylimit) {
                if (l > k) {
                  no[drvui->nplane++] = vertex_list[l];	 /* add to plane list */
                } else {
                  if ((l != i) && (l != j) && (l != k)) {
                    ind = 1;			 /* we already have this plane */
                    l = numb_list;		 /* skip rest of loop */
                  }
                }
              } else {
/* point is not in plane - check if first out-of-plane point found */
                if (sign_z == 0.0) {
                  sign_z = vol;			 /* yes - set sign */
                } else {
/* if here, not first out of plane - check that vol and sign_z have same sign. 
   If not, set ind=1 to indicate not a plane */
                  if (vol * sign_z < 0.0) {
                    ind = 1;
                    l = numb_list;		 /* skip rest of inner loop */
                  }
                }
              }
          }					 /* end of loop on l */
          if (ind == 0) {			 /* 0 means a surface plane */
            poly_list[draw_list++] = drvui->nplane;	 /* number of vertices in plane */
            for (l = 0; l < drvui->nplane; ++l)
              poly_list[draw_list++] = no[l] + 1;

            fprintf (drvui->fpoutp, "/* %d-sided polygon */\n", drvui->nplane);
/* Make vertices consecutive - find center of polygon */
            xm = ym = zm = 0.0f;
            for (ii = 0; ii < drvui->nplane; ii++) {
              xm = xm + s_vert[3 * no[ii]];
              ym = ym + s_vert[3 * no[ii] + 1];
              zm = zm + s_vert[3 * no[ii] + 2];
            }
            xm = xm / (float) drvui->nplane;
            ym = ym / (float) drvui->nplane;
            zm = zm / (float) drvui->nplane;

/* sort vertices in consecutive order based on minimal angles between vectors */

            for (ii = 0; ii < drvui->nplane; ii++)
              ns[ii] = 0;
            ns[0] = no[0];

            for (ii = 0; ii < drvui->nplane - 1; ii++) {
              phi0 = 1000.0f;
              vxs = s_vert[3 * ns[ii]] - xm;
              vys = s_vert[3 * ns[ii] + 1] - ym;
              vzs = s_vert[3 * ns[ii] + 2] - zm;
              bms = (float) sqrt (vxs * vxs + vys * vys + vzs * vzs);

              for (jj = 0; jj < drvui->nplane; jj++) {
                for (ll = 0; ll < ii; ll++) {
                  if (ns[ll] == no[jj])
                    jj++;
                  if (jj >= drvui->nplane)
                    break;
                }
                if (jj < drvui->nplane) {
                  if (ns[ii] != no[jj]) {
                    vx = s_vert[3 * no[jj]] - xm;
                    vy = s_vert[3 * no[jj] + 1] - ym;
                    vz = s_vert[3 * no[jj] + 2] - zm;
                    bm = (float) sqrt (vx * vx + vy * vy + vz * vz);
                    cosphi = (vxs * vx + vys * vy + vzs * vz) / (bms * bm);
                    if (cosphi < -1.)
                      cosphi = -1.;
                    if (cosphi > 1.)
                      cosphi = 1.;
                    phi = acos (cosphi) * RAD;
                    if (phi < phi0) {
                      phi0 = (float) phi;
                      ns[ii + 1] = no[jj];
                    }
                  }
                }
              }
            } 
            if (numb_list != 4) {
              draw_list = draw_list - drvui->nplane;
              for (l = 0; l < drvui->nplane; ++l)
                poly_list[draw_list++] = ns[l] + 1;
            }


            for (l = 0; l < drvui->nplane - 2; ++l) {
              d1 = (float) fabs (s_vert[3 * ns[0]] - s_vert[3 * ns[l + 1]]) +
                   (float) fabs (s_vert[3 * ns[0] + 1] -  s_vert[3 * ns[l + 1] + 1]) +
                   (float) fabs (s_vert[3 * ns[0] + 2] -  s_vert[3 * ns[l + 1] + 2]);
              d2 = (float) fabs (s_vert[3 * ns[0]] - s_vert[3 * ns[l + 2]]) +
                   (float) fabs (s_vert[3 * ns[0] + 1] - s_vert[3 * ns[l + 2] + 1]) +
                   (float) fabs (s_vert[3 * ns[0] + 2] - s_vert[3 * ns[l + 2] + 2]);
              d3 = (float) fabs (s_vert[3 * ns[l + 2]] - s_vert[3 * ns[l + 1]]) +
                   (float) fabs (s_vert[3 * ns[l + 2] + 1] - s_vert[3 * ns[l + 1] + 1]) +
                   (float) fabs (s_vert[3 * ns[l + 2] + 2] - s_vert[3 * ns[l + 1] + 2]);
              if ((d1 > 0.01) && (d2 > 0.01) && (d3 > 0.01)) {
                if (!render) {
                  fprintf (drvui->fpoutp," triangle {< %8.5f, %8.5f, %8.5f>,\n",
                     s_vert[3 * ns[0]], s_vert[3 * ns[0] + 1], s_vert[3 * ns[0] + 2]);
                  fprintf (drvui->fpoutp,"<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>\n",
                     s_vert[3 * ns[l + 1]], s_vert[3 * ns[l + 1] + 1],
                     s_vert[3 * ns[l + 1] + 2], s_vert[3 * ns[l + 2]],
                     s_vert[3 * ns[l + 2] + 1], s_vert[3 * ns[l + 2] + 2]);
                  fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n}\n",
                     drvui->poly_col_p[polyno]);

#ifdef OPENGL
/* 
      need to add something like
      glNormal3f(xm-xc,ym-yc,zm-zc ); 
      for lighting -  xc,yc,zc being the center of the polyhedron
      (which is not normally available to build_poly_list)
*/
                  if (gl_view) {
                    glVertex3f (s_vert[3 * ns[0]], s_vert[3 * ns[0] + 1],
                        s_vert[3 * ns[0] + 2]);
                    glVertex3f (s_vert[3 * ns[l + 1]], s_vert[3 * ns[l + 1] + 1],
                        s_vert[3 * ns[l + 1] + 2]);
                    glVertex3f (s_vert[3 * ns[l + 2]], s_vert[3 * ns[l + 2] + 1],
                        s_vert[3 * ns[l + 2] + 2]);
                  }
#endif
                }       /* !render */
              }
            }
            if (render) {
              if (edges == 1) {
                if (drvui->p_rad_edge[polyno] != 0.)
                  strcpy (edgecolor, drvui->p_col_edge_p[polyno]);
                else
                  strcpy (edgecolor, col_edge_p);
                strcpy(color,edgecolor);
                check_color(color);
                for (l = 0; l < drvui->nplane - 1; ++l) {
                  d1 = (float) fabs (s_vert[3 * ns[l]] -
                     s_vert[3 * ns[l + 1]]) +
                     (float) fabs (s_vert[3 * ns[l] + 1] -
                     s_vert[3 * ns[l + 1] + 1]) +
                     (float) fabs (s_vert[3 * ns[l] + 2] -
                     s_vert[3 * ns[l + 1] + 2]);

                  if (d1 > 0.01) {
                    fprintf (drvui->fpoutp,
                      " cylinder { <%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f> , %8.5f \n",
                      s_vert[3 * ns[l]], s_vert[3 * ns[l] + 1],
                      s_vert[3 * ns[l] + 2], s_vert[3 * ns[l + 1]],
                      s_vert[3 * ns[l + 1] + 1], s_vert[3 * ns[l + 1] + 2], radius);
                    fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n}\n", edgecolor);

#ifdef OPENGL
                    if (gl_view) {
                      for (m=0; m<3; m++) {
                        at[m] = s_vert[3 * ns[l] + m];
                        df[m] = s_vert[3 * ns[l + 1] + m] - at[m];
                      } 
                      push_cylinder(df, at, radius, color);
                    } 
#endif

                  } 

                } 
                d1 = (float) fabs (s_vert[3 * ns[0]] -
                  s_vert[3 * ns[drvui->nplane - 1]]) +
                  (float) fabs (s_vert[3 * ns[0] + 1] -
                  s_vert[3 * ns[drvui->nplane - 1] + 1]) +
                  (float) fabs (s_vert[3 * ns[0] + 2] -
                  s_vert[3 * ns[drvui->nplane - 1] + 2]);
                if (d1 > 0.01) {
                  fprintf (drvui->fpoutp,
                      " cylinder { <%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f> , %8.5f \n",
                      s_vert[3 * ns[0]], s_vert[3 * ns[0] + 1],
                      s_vert[3 * ns[0] + 2], s_vert[3 * ns[drvui->nplane - 1]],
                      s_vert[3 * ns[drvui->nplane - 1] + 1],
                      s_vert[3 * ns[drvui->nplane - 1] + 2], radius);
                  fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n}\n", edgecolor);

#ifdef OPENGL
                  if (gl_view) {
                    for (m=0; m<3; m++) {
                      at[m] = s_vert[3 * ns[0] + m];
                      df[m] = s_vert[3 * ns[drvui->nplane - 1] + m] - at[m];
                    }
                    push_cylinder(df, at, radius, color);
                  }
#endif
                }             /* d1 > 0 */
              }               /* edges == 1 */ 
            }                 /* render */ 
          }
        }					 /* end of numb_list != 4 */
      }						 /* end of for (k.... */
    }
  }
#ifdef OPENGL
  if (gl_view && !render) {
    glEnd ();
    glEnable (GL_LIGHTING);
    glPopMatrix ();
  }
#endif
  }
  free (no);
  free (ns);
  return;
}						 /* end of build_poly_list */

/* ************************************************************** */
/* ************************************************************** */

void check_color (char *input)
{
  char *transp_start;
  char transp[20];
  float fval;

/*
   VRML routine to check color names and replace them with RGB triples
   derived from 'colors.inc' of POV
*/

  transp[0] = '\0';
  transp_start = strstr (input, "filter");
  if (transp_start != NULL) {
    fval = (float) atof (transp_start + 6);
    sprintf (transp, " transparency %5.2f", fval / 2.);
  }
  if (strncmp (input, "LightGray", 9) == 0) {
    strncpy (input, ".658824 .658824 .658824\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "LightGrey", 9) == 0) {
    strncpy (input, ".658824 .658824 .658824\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "VLightGrey", 10) == 0) {
    strncpy (input, ".8 .8 .8\0", 9);
    strcat (input, transp);
    return;
  }

  if (strncmp (input, "White", 5) == 0) {
    strncpy (input, "1 1 1\0", 6);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Red", 3) == 0) {
    strncpy (input, "1 0 0\0", 6);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "GreenYellow", 11) == 0) {
    strncpy (input, ".576471 .858824 .439216\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Green", 5) == 0) {
    strncpy (input, "0 1 0\0", 6);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "BlueViolet", 10) == 0) {
    strncpy (input, ".62352 .372549 .623529\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Blue", 3) == 0) {
    strncpy (input, "0 0 1\0", 6);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "YellowGreen", 11) == 0) {
    strncpy (input, ".6 .8 .196078\0", 14);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Yellow", 6) == 0) {
    strncpy (input, "1 1 0\0", 6);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Cyan", 3) == 0) {
    strncpy (input, "0 1 1\0", 6);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Magenta", 7) == 0) {
    strncpy (input, "1 0 1\0", 6);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Black", 5) == 0) {
    strncpy (input, "0 0 0\0", 6);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "OrangeRed", 9) == 0) {
    strncpy (input, "1.0 .498039\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Orange", 6) == 0) {
    strncpy (input, "1 .5 .0\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Brown", 5) == 0) {
    strncpy (input, ".647059 .164706 .164706\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray05", 6) == 0) {
    strncpy (input, ".05 .05 .05\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray10", 6) == 0) {
    strncpy (input, ".10 .10 .10\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray15", 6) == 0) {
    strncpy (input, ".15 .15 .15\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray20", 6) == 0) {
    strncpy (input, ".20 .20 .20\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray25", 6) == 0) {
    strncpy (input, ".25 .25 .25\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray30", 6) == 0) {
    strncpy (input, ".30 .30 .30\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray35", 6) == 0) {
    strncpy (input, ".35 .35 .35\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray40", 6) == 0) {
    strncpy (input, ".40 .40 .40\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray45", 6) == 0) {
    strncpy (input, ".45 .45 .45\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray50", 6) == 0) {
    strncpy (input, ".50 .50 .50\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray55", 6) == 0) {
    strncpy (input, ".55 .55 .55\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray60", 6) == 0) {
    strncpy (input, ".60 .60 .60\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray65", 6) == 0) {
    strncpy (input, ".65 .65 .65\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray70", 6) == 0) {
    strncpy (input, ".70 .70 .70\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray75", 6) == 0) {
    strncpy (input, ".75 .75 .75\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray80", 6) == 0) {
    strncpy (input, ".80 .80 .80\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray85", 6) == 0) {
    strncpy (input, ".85 .85 .85\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray90", 6) == 0) {
    strncpy (input, ".90 .90 .90\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray95", 6) == 0) {
    strncpy (input, ".95 .95 .95\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gray", 4) == 0) {
    strncpy (input, ".752941 .752941 .752941\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Grey", 4) == 0) {
    strncpy (input, ".752941 .752941 .752941\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DimGray", 7) == 0) {
    strncpy (input, ".329412 .329412 .329412\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DimGrey", 7) == 0) {
    strncpy (input, ".329412 .329412 .329412\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Aquamarine", 10) == 0) {
    strncpy (input, ".439216 .858824 .576471\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "CadetBlue", 9) == 0) {
    strncpy (input, ".372549 .623529 .623529\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Coral", 5) == 0) {
    strncpy (input, "1.0 .498039 .0\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "CornflowerBlue", 14) == 0) {
    strncpy (input, ".258824 .258824 .435294\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DarkGreen", 9) == 0) {
    strncpy (input, ".184314 .309804 .184314\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DarkOliveGreen", 14) == 0) {
    strncpy (input, ".309804 .309804 .184314\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DarkOrchid", 10) == 0) {
    strncpy (input, ".6 .196078 .8\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DarkSlateBlue", 13) == 0) {
    strncpy (input, ".419608 .137255 .556863\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DarkSlateGray", 13) == 0) {
    strncpy (input, ".184314 .309804 .309804\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DarkSlateGrey", 13) == 0) {
    strncpy (input, ".184314 .309804 .309804\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DarkTurquoise", 13) == 0) {
    strncpy (input, ".439216 .576471 .858824\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Firebrick", 9) == 0) {
    strncpy (input, ".556863 .137255 .137255\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "ForestGreen", 11) == 0) {
    strncpy (input, ".137255 .556863 .137255\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Goldenrod", 9) == 0) {
    strncpy (input, ".858824 .858824 .439216\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Gold", 4) == 0) {
    strncpy (input, ".8 .498039 .196078\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "IndianRed", 9) == 0) {
    strncpy (input, ".309804 .184314 .184314\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Khaki", 5) == 0) {
    strncpy (input, ".623529 .623529 .372549\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "LightBlue", 9) == 0) {
    strncpy (input, ".74902 .847059 .847059\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "LightSteelBlue", 14) == 0) {
    strncpy (input, ".560784 .560784 .737255\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "LimeGreen", 9) == 0) {
    strncpy (input, ".196078 .8 .196078\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Maroon", 6) == 0) {
    strncpy (input, ".556863 .137255 .419608\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "MediumAquamarine", 16) == 0) {
    strncpy (input, ".196078 .8 .6\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "MediumBlue", 10) == 0) {
    strncpy (input, ".196078 .196078 .8\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "MediumForestGreen", 17) == 0) {
    strncpy (input, ".419608 .556863 .137255\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "MediumGoldenrod", 15) == 0) {
    strncpy (input, ".917647 .917647 .678431\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "MediumOrchid", 12) == 0) {
    strncpy (input, ".576471 .439216 .858824\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "MediumSeaGreen", 14) == 0) {
    strncpy (input, ".258824 .435294 .258824\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "MediumSlateBlue", 15) == 0) {
    strncpy (input, ".498039 1.0\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "MediumSpringGreen", 17) == 0) {
    strncpy (input, ".498039 1.0 0\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "MediumTurquoise", 15) == 0) {
    strncpy (input, ".439216 .858824 .858824\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "MediumVioletRed", 15) == 0) {
    strncpy (input, ".858824 .439216 .576471\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "MidnightBlue", 12) == 0) {
    strncpy (input, ".184314 .184314 .309804\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "NavyBlue", 8) == 0) {
    strncpy (input, ".137255 .137255 .556863\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Navy", 4) == 0) {
    strncpy (input, ".137255 .137255 .556863\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Orchid", 6) == 0) {
    strncpy (input, ".858824 .439216 .858824\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "PaleGreen", 9) == 0) {
    strncpy (input, ".560784 .737255 .560784\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Pink", 4) == 0) {
    strncpy (input, ".737255 .560784 .560784\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Plum", 4) == 0) {
    strncpy (input, ".917647 .678431 .917647\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Salmon", 6) == 0) {
    strncpy (input, ".435294 .258824 .258824\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "SeaGreen", 8) == 0) {
    strncpy (input, ".137255 .556863 .419608\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Sienna", 6) == 0) {
    strncpy (input, ".556863 .419608 .137255\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "SkyBlue", 7) == 0) {
    strncpy (input, ".196078 .6 .8\0", 14);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "SlateBlue", 9) == 0) {
    strncpy (input, "0 .498039 1\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "SpringGreen", 11) == 0) {
    strncpy (input, "0 1 .498039\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "SteelBlue", 9) == 0) {
    strncpy (input, ".137255 .419608 .556863\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Tan", 3) == 0) {
    strncpy (input, ".858824 .576471 .439216\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Thistle", 7) == 0) {
    strncpy (input, ".847059 .74902 .847059\0", 23);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Turquoise", 9) == 0) {
    strncpy (input, ".678431 .917647 .917647\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "VioletRed", 9) == 0) {
    strncpy (input, ".8 .196078 .6\0", 14);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Violet", 6) == 0) {
    strncpy (input, ".309804 .184314 .309804\0", 24);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Wheat", 5) == 0) {
    strncpy (input, ".847059 .847059 .74902\0", 23);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "SummerSky", 9) == 0) {
    strncpy (input, ".22 .69 .87\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "RichBlue", 8) == 0) {
    strncpy (input, ".35 .35 .67\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Brass", 5) == 0) {
    strncpy (input, ".71 .65 .26\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Copper", 6) == 0) {
    strncpy (input, ".72 .45 .20\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Bronze2", 7) == 0) {
    strncpy (input, ".65 .49 .24\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Bronze", 6) == 0) {
    strncpy (input, ".55 .47 .14\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Silver", 6) == 0) {
    strncpy (input, ".90 .91 .98\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "BrightGold", 10) == 0) {
    strncpy (input, ".85 .85 .10\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "OldGold", 7) == 0) {
    strncpy (input, ".81 .71 .23\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Feldspar", 8) == 0) {
    strncpy (input, ".82 .57 .46\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Quartz", 6) == 0) {
    strncpy (input, ".85 .85 .95\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Mica", 4) == 0) {
    strncpy (input, "0 0 0\0", 6);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "NeonPink", 8) == 0) {
    strncpy (input, "1 .43 .78\0", 10);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DarkPurple", 10) == 0) {
    strncpy (input, ".53 .12 .47\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "NeonBlue", 8) == 0) {
    strncpy (input, ".3 .3 1\0", 8);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "CoolCopper", 10) == 0) {
    strncpy (input, ".85 .53 .10\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "MandarinOrange", 14) == 0) {
    strncpy (input, ".89 .47 .20\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "LightWood", 9) == 0) {
    strncpy (input, ".91 .76 .65\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "MediumWood", 10) == 0) {
    strncpy (input, ".65 .50 .39\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DarkWood", 8) == 0) {
    strncpy (input, ".52 .37 .26\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "SpicyPink", 9) == 0) {
    strncpy (input, "1 .11 .68\0", 10);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "SemiSweetChoc", 13) == 0) {
    strncpy (input, ".42 .26 .15\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "BakersChoc", 10) == 0) {
    strncpy (input, ".36 .20 .09\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Flesh", 5) == 0) {
    strncpy (input, ".96 .80 .69\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "NewTan", 6) == 0) {
    strncpy (input, ".92 .78 .62\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "NewMidnightBlue", 15) == 0) {
    strncpy (input, "0 0 .61\0", 8);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "VeryDarkBrown", 13) == 0) {
    strncpy (input, ".35 .16 .14\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DarkBrown", 9) == 0) {
    strncpy (input, ".36 .25 .20\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DarkTan", 7) == 0) {
    strncpy (input, ".59 .41 .31\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "GreenCopper", 11) == 0) {
    strncpy (input, ".32 .49 .46\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DkGreenCopper", 13) == 0) {
    strncpy (input, ".29 .46 .43\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "DustyRose", 9) == 0) {
    strncpy (input, ".52 .39 .39\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "HuntersGreen", 12) == 0) {
    strncpy (input, ".13 .37 .31\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Scarlet", 7) == 0) {
    strncpy (input, ".55 .09 .09\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Med_Purple", 10) == 0) {
    strncpy (input, ".73 .16 .96\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Light_Purple", 12) == 0) {
    strncpy (input, ".87 .58 .98\0", 12);
    strcat (input, transp);
    return;
  }
  if (strncmp (input, "Very_Light_Purple", 17) == 0) {
    strncpy (input, ".94 .81 .99\0", 12);
    strcat (input, transp);
    return;
  }
  if (Color_Warning <= 5) {
    fprintf (drvui->flout,
	     " **** Warning - Non-standard color \"%s\" encountered. ****\n",
	     input);
    Color_Warning++;
  }
}

/* ************************************************************** */
/* ************************************************************** */

void crystal_compute (int Quick)
{
  int i;
  char outfil[100];
  float temp;
  float rm[16];

  if (drvui->Str_File == NULL)
    return;					 /* Fail immediately if filename NULL */
/* initialize */
  origin[0] = drvui->X_Origin;
  origin[1] = drvui->Y_Origin;
  origin[2] = drvui->Z_Origin;
  boxlim[0] = drvui->X_Boxlim;
  boxlim[1] = drvui->Y_Boxlim;
  boxlim[2] = drvui->Z_Boxlim;
  docell = Unit_Cell;
  do_ellipsoids = Color_Warning = 0;
  rad_cell = 0.02f;
  Sphere_Mult = SpMult;
  drvui->Bond_Mult = drvui->BndMult;
  drvui->Str_File_Inp = drvui->Str_File;
  natom = 0;
  drvui->nmag = 0;
  drvui->nsphere = drvui->npoly = drvui->nbond = drvui->nplane = 1;
  n_ellips = drvui->ncone = drvui->nlabel = 1;
  drvui->poly_size[0] = 2.5f;
  drvui->poly_min[0] = 0.0f;
  drvui->poly_l[0][0] = drvui->poly_l[0][1] = drvui->poly_l[0][2] = drvui->poly_l[0][3] = ' ';
  drvui->plane_size[0] = 2.5f;
  drvui->plane_l[0][0] = drvui->plane_l[0][1] = drvui->plane_l[0][2] = drvui->plane_l[0][3] = ' ';
  drvui->sphere_size[0] = 0.0f;
  drvui->sphere_l[0][0] = drvui->sphere_l[0][1] = drvui->sphere_l[0][2] = drvui->sphere_l[0][3] = ' ';
  POV_Max[0] = POV_Max[1] = POV_Max[2] = -99999.0f;
  POV_Min[0] = POV_Min[1] = POV_Min[2] = 99999.0f;
  strcpy (drvui->col_bg_p, "White");
  strcpy (drvui->col_bg_v, "1. 1. 1.");
  drvui->frame_no = 0;
  drvui->max_frame = 99999;
  drvui->ambient = 0.0f;                   /* POV lighting options */
  drvui->diffuse = 0.0f;
  drvui->specular = 0.0f;
  drvui->roughness = 0.0f;
#if defined(__APPLE__) && defined(__MACH__)    /* reset working directory in OS X */
  i = chdir(drvui->savedpath);
  if (i != 0) printf ("Warning: change of working directory to %s failed.\n",drvui->savedpath);
#endif
  strcpy (outfil, drvui->Str_File);
  for (i = (int) strlen (outfil); i>0; --i) {
    if (outfil[i] == '.' || outfil[i] == ' ')
      break;
  }
  outfil[i] = '\0';
  strcpy(drvui->File_Root,outfil);
  if (!Quick) {
    strcat (outfil, ".out");
    if ((drvui->flout = fopen (outfil, "w")) == NULL) {
      fprintf (stderr, "Cannot open listing file\n");
      delay (5);
      return;
    }
  }
  if (!Quick) {
    outfil[i] = '\0';
    strcat (outfil, ".wrl");
    if ((drvui->fpoutv = fopen (outfil, "w")) == NULL) {
      fprintf (stderr, "Cannot open VRML output file\n");
      delay (5);
      return;
    }
  }
  if (!Quick) {
    outfil[i] = '\0';
    strcpy(drvui->Pov_File,outfil);
    strcat (drvui->Pov_File, ".pov");
    if ((drvui->fpoutp = fopen (drvui->Pov_File, "w")) == NULL) {
      fprintf (stderr, "Cannot open POV output file\n");
      delay (5);
      return;
    }
  }

  if (!Quick) {
    fprintf (drvui->flout, "\n\n                DRAWxtl\n");
    fprintf (drvui->flout, "  Crystal Structure Drawing Program V%s\n\n", VERSION);
    fprintf (drvui->flout,
	     "  Copyright 1996, 1997, 2000 - 2005 by Larry Finger, Martin Kroeker, and Brian Toby\n");
    fprintf (drvui->flout, "  (all rights reserved).\n\n");
    fprintf (drvui->flout, "     Combined POVxtl and VRMLxtl\n");
    fprintf (drvui->flout,
	     "POVxtl  - Scenes for Persistence of Vision Ray Tracer (POV-Ray) V2.0\n");
    fprintf (drvui->flout,
	     "VRMLxtl - Virtual Reality Modeling Language V1.0/V2.0\n");
    getcwd(outfil,100);
    fprintf (drvui->flout, 
             "DRAWxtl working directory is %s\n",outfil);
    printf ("\n\n                DRAWxtl\n");
    printf ("  Crystal Structure Drawing Program V%s\n\n", VERSION);
    printf ("  Copyright 1996, 1997, 2000 - 2005 by Larry Finger, Martin Kroeker, and Brian Toby\n");
    printf ("  (all rights reserved).\n\n");
    printf ("     Combined POVxtl and VRMLxtl\n");
    printf ("POVxtl  - Scenes for Persistence of Vision Ray Tracer (POV-Ray) V2.0\n");
    printf ("VRMLxtl - Virtual Reality Modeling Language V1.0/V2.0\n");
    getcwd(outfil,100);
    printf ("DRAWxtl working directory is %s\n",outfil);
  }
#ifdef OPENGL
  if (gl_view) {
      glClear ((GLbitfield) (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT));
      if (!glIsList(crystalDL)) crystalDL = glGenLists (1);
      glNewList (crystalDL, GL_COMPILE);
  } 
#endif
  while (drvui->frame_no != drvui->max_frame) {             /* loop through all frames */
    drvui->frame_no++;
    if (!Quick)
      fprintf (stderr, " doing frame %d\n",drvui->frame_no);
    if (drvui->frame_no == 1) {
        clipflag = 0;
        packflag = 0;
        domolcomp = 0;
        origin[0] = drvui->X_Origin;
        origin[1] = drvui->Y_Origin;
        origin[2] = drvui->Z_Origin;
        docell = Unit_Cell;
        do_ellipsoids = Color_Warning = 0;
        drvui->noshadow = 0;
        rad_cell = 0.02f;
        Sphere_Mult = SpMult;
        drvui->Bond_Mult = drvui->BndMult;
    }
    boxlim[0] = drvui->X_Boxlim;
    boxlim[1] = drvui->Y_Boxlim;
    boxlim[2] = drvui->Z_Boxlim;
    natom = 0;
    if ((drvui->fpin = fopen (drvui->Str_File, "r")) == NULL) {
      printf ("Cannot Open Input File");
      return;
    }
    get_input (Quick);
    fclose(drvui->fpin);
    sscanf(drvui->col_bg_v,"%f %f %f",&drvui->glback[0],&drvui->glback[1],&drvui->glback[2]);

    if (boxlim[0] == 100.0) {
      boxlim[0] = (drvui->cryst_lim[drvui->frame_no][3]
                - drvui->cryst_lim[drvui->frame_no][0]) * drvui->lat_con[0];
      boxlim[1] = (drvui->cryst_lim[drvui->frame_no][4]
                - drvui->cryst_lim[drvui->frame_no][1]) * drvui->lat_con[1];
      boxlim[2] = (drvui->cryst_lim[drvui->frame_no][5]
                - drvui->cryst_lim[drvui->frame_no][2]) * drvui->lat_con[2];
    }

/* Calculate Grand rotation matrix */
    if (drvui->frame_no == 1) {
        int m;
        tb_init();
        XYZ_Rot_to_Q(xrot, yrot, zrot);
        quaternion_to_rotmatrix(&Rotq, rm);  /* get rotation matrix */
        for (m = 0; m < 3; m++) {            /*  and copy to G_Rot */
          G_Rot[0][m] = rm[m];
          G_Rot[1][m] = rm[m + 4];
          G_Rot[2][m] = rm[m+8];
        }

/* Initialize POV output limits */

        POV_Max[0] = POV_Max[1] = POV_Max[2] = -99999.0f;
        POV_Min[0] = POV_Min[1] = POV_Min[2] =  99999.0f;
    }
/* generate all possible atoms to get the size of the drawing */

    NvertM = 1;                      /* make master list empty */
    if (!s_vert) {
      if ((s_vert = (float *)malloc ((unsigned) (3 * (long) 
          (MAX_VERTS * sizeof (float))))) ==	NULL) {
        printf ("Unable to get initial s_vert allocation\n");
        delay (5);
        return;
      }
    }  
    get_atom_id ();                  /* rework atom id numbers to pointers */
    build_box_contents ();           /* complete the master atom list */
    free (s_vert);
    printf ("\nMaster atom list contains %5d atoms.\n", NvertM - 1);
    fprintf (drvui->flout, "\nMaster atom list contains %5d atoms.\n", NvertM - 1);
/* allocate space for the various arrays */
    if (o_vert)
      free(o_vert);
    if ((o_vert = (float *)
	 malloc ((unsigned) (6 * (NvertM + 10) * sizeof (float)))) == NULL) {
      printf ("Unable to get o_vert allocation\n");
      delay (5);
      return;
    }
    if ((s_vert = (float *)
	 malloc ((unsigned) (6 * (NvertM + 10) * sizeof (float)))) == NULL) {
      printf ("Unable to get s_vert allocation\n");
      delay (5);
      return;
    }
    if (vert_sym_no)
      free(vert_sym_no);
    if ((vert_sym_no =
	 (int *) malloc ((unsigned) ((2 * NvertM + 10) * sizeof (int)))) ==	NULL) {
      printf ("Unable to get vert_sym_no allocation\n");
      delay (5);
      return;
    }
    if (drvui->orig_atom_no)
      free(drvui->orig_atom_no);
    if ((drvui->orig_atom_no =
	 (int *) malloc ((unsigned) ((2 * NvertM + 10) * sizeof (int)))) ==	NULL) {
      printf ("Unable to get orig_atom_no allocation\n");
      delay (5);
      return;
    }
    if ((poly_list =
	 (int *) malloc ((unsigned) ((4 * (NvertM + 10)) * sizeof (int)))) == NULL) {
      printf ("Unable to get poly_list allocation\n");
      delay (5);
      return;
    }
    if ((vertex_list =
	 (int *) malloc ((unsigned) ((4 * (NvertM + 10)) * sizeof (int)))) == NULL) {
      printf ("Unable to get vertex_list allocation\n");
      delay (5);
      return;
    }
    analyze_bonds ();				 /* print the bond distance table */
    printf ("\nThe 'pov' and 'wrl' files contain:\n");
    if (docell != 0) {				 /* if unit-cell corners drawn, add them for scaling */
      float vert[3];
      int saved_Nvert;

      saved_Nvert = NvertM;
      for (i = 0; i <= 2; ++i)
        vert[i] = 0.0f;
      add_vert (vert, 0, 1);			 /* place 0,0,0 in list */
      vert[2] = 1.0f;
      add_vert (vert, 0, 1);			 /* place 0,0,1 in list */
      vert[1] = 1.0f;
      add_vert (vert, 0, 1);			 /* place 0,1,1 in list */
      vert[2] = 0.0f;
      add_vert (vert, 0, 1);			 /* place 0,1,0 in list */
      vert[0] = 1.0f;
      add_vert (vert, 0, 1);			 /* place 1,1,0 in list */
      vert[1] = 0.0f;
      add_vert (vert, 0, 1);			 /* place 1,0,0 in list */
      vert[2] = 1.0f;
      add_vert (vert, 0, 1);			 /* place 1,0,1 in list */
      vert[1] = 1.0f;
      add_vert (vert, 0, 1);			 /* place 1,1,1 in list */
      NvertM = saved_Nvert;	     /* reset list count */
    }
    if (drvui->frame_no == 1) {			      /* only do for first frame */
      temp = max(POV_Max[0] - POV_Min[0],POV_Max[1] - POV_Min[1]);
      temp = max(temp, POV_Max[2] - POV_Min[2]);
      if (gl_size == 99999.0f) gl_size = temp;   /* only do ONCE */
      Scale = 15.0f * temp / Magnification;

      if (Vrml2)
        fprintf (drvui->fpoutv, "#VRML V2.0 utf8\n\n");
      else
        fprintf (drvui->fpoutv, "#VRML V1.0 ascii\n\n");
      fprintf (drvui->fpoutv, "# Scene file created by DRAWxtl V%s from %s\n",
	       VERSION, drvui->Str_File_Inp);
      fprintf (drvui->fpoutv,
	       "# with command line options: -b %6.2f %6.2f %6.2f -o %5.2f %5.2f %5.2f\n",
	       boxlim[0], boxlim[1], boxlim[2], origin[0], origin[1],
	       origin[2]);
      fprintf (drvui->fpoutv,
	       "#  -p %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f -v %5.2f %5.2f %5.2f\n",
	       drvui->cryst_lim[drvui->frame_no][0], drvui->cryst_lim[drvui->frame_no][3],
           drvui->cryst_lim[drvui->frame_no][1], drvui->cryst_lim[drvui->frame_no][4],
           drvui->cryst_lim[drvui->frame_no][2], drvui->cryst_lim[drvui->frame_no][5],
           xrot, yrot, zrot);

/* Output VRML header */

      if (!Vrml2) {
        fprintf (drvui->fpoutv,"DEF BackgroundColor Info { string \"%s\" }\n", drvui->col_bg_v);
        fprintf (drvui->fpoutv, "Separator {\n");
        fprintf (drvui->fpoutv, " Material {\n");
        fprintf (drvui->fpoutv, "  ambientColor    0.1 0.1 0.1\n");
        fprintf (drvui->fpoutv, "  diffuseColor    1 1 1\n");
        fprintf (drvui->fpoutv, "  specularColor   0.9 0.9 0.9\n");
        fprintf (drvui->fpoutv, "  shininess       0.8\n");
        fprintf (drvui->fpoutv, " }\n");
      }
      Text_Size = 0.0005f *Scale;

      if (!M_cameras && Vrml2) {		 /* Generate Orthographic Projection for VRML97 */
        fprintf (drvui->fpoutv, "NavigationInfo {type \"EXAMINE\"}\n");
        fprintf (drvui->fpoutv, "Background { skyColor %s }\n", drvui->col_bg_v);
        fprintf (drvui->fpoutv,
		 "Viewpoint { fieldOfView 0.1 position 0. 0. %8.3f }\n", Scale);
        fprintf (drvui->fpoutv, "DirectionalLight {direction 0 0 -1}\n");
        fprintf (drvui->fpoutv, "DirectionalLight {direction 0 0 1}\n");
      } else {
        if (!Vrml2) {				 /* simulate orthographic projection for VRML1.0 */
          float fscale, fangle;

          if (!M_cameras) {
            fscale = Scale * (10.0f / 7.0f);
            fangle = 0.07853f;
          } else {
            fscale = Scale / 7.0f;
            fangle = 0.7853f;
          }
          fprintf (drvui->fpoutv, "PerspectiveCamera {\n");
          fprintf (drvui->fpoutv, "  position        0 0 %8.3f\n", fscale);
          fprintf (drvui->fpoutv, "  orientation     0 0 1 0\n");
          fprintf (drvui->fpoutv, "  focalDistance %8.3f\n", fscale);
          fprintf (drvui->fpoutv, "  heightAngle   %8.5f\n", fangle);
          fprintf (drvui->fpoutv, "}\n");
          fprintf (drvui->fpoutv,
		   "DirectionalLight{ direction 0 0 -1 intensity 1.0}\n");
          fprintf (drvui->fpoutv,
		   "DirectionalLight{ direction 0 0 1 intensity 1.0}\n");
        } else {
          fprintf (drvui->fpoutv, "NavigationInfo {type \"EXAMINE\"}\n");
          fprintf (drvui->fpoutv, "Background { skyColor %s }\n", drvui->col_bg_v);
          fprintf (drvui->fpoutv, "Viewpoint { position 0. 0. %8.3f }\n", Scale / 7.);
          fprintf (drvui->fpoutv, "DirectionalLight {direction 0 0 -1}\n");
          fprintf (drvui->fpoutv, "DirectionalLight {direction 0 0 1}\n");
        } 
      }

      if (Vrml2) {
        fprintf (drvui->fpoutv, "Transform {");
        fprintf (drvui->fpoutv, " rotation 0 0 1 %f\n", zrot / RAD);
        fprintf (drvui->fpoutv, " children Transform {");
        fprintf (drvui->fpoutv, " rotation 0 1 0 %f\n", yrot / RAD);
        fprintf (drvui->fpoutv, " children Transform {");
        fprintf( drvui->fpoutv, " rotation 1 0 0 %f\n", xrot / RAD);
        fprintf (drvui->fpoutv, "  children [\n");
      } else {
        fprintf (drvui->fpoutv, "Rotation {");
        fprintf (drvui->fpoutv, " rotation 0 0 1 %f}\n", zrot / RAD);
        fprintf (drvui->fpoutv, " Rotation {");
        fprintf (drvui->fpoutv, " rotation 0 1 0 %f}\n", yrot / RAD);
        fprintf (drvui->fpoutv, " Rotation {");
        fprintf (drvui->fpoutv, " rotation 1 0 0 %f}\n", xrot / RAD);
      } 

/* Output POV header */

        fprintf (drvui->fpoutp, "/*\n\n    POVxtl-Persistence of Vision Ray Tracer\n");
        fprintf (drvui->fpoutp, "    Crystal Structure Drawing Program %s\n\n", VERSION);
        fprintf (drvui->fpoutp,
	       "       Copyright 1996, 1997, 2000 - 2005 by Larry Finger, Martin Kroeker, and Brian Toby\n");
        fprintf (drvui->fpoutp, "          (all rights reserved).\n\n");
        fprintf (drvui->fpoutp, " Scene file created from %s \n", drvui->Str_File_Inp);
        fprintf (drvui->fpoutp,
	       " with command line options: -b %6.2f %6.2f %6.2f -o %5.2f %5.2f %5.2f \n",
	       boxlim[0], boxlim[1], boxlim[2], origin[0], origin[1],
	       origin[2]);
        fprintf (drvui->fpoutp,
	       "  -p %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f -v %5.2f %5.2f %5.2f -m %5.2f\n*/\n\n",
	       drvui->cryst_lim[drvui->frame_no][0], drvui->cryst_lim[drvui->frame_no][3],
           drvui->cryst_lim[drvui->frame_no][1], drvui->cryst_lim[drvui->frame_no][4],
           drvui->cryst_lim[drvui->frame_no][2], drvui->cryst_lim[drvui->frame_no][5],
           xrot, yrot, zrot, Magnification);

        fprintf (drvui->fpoutp, "#include \"colors.inc\"\n");
        fprintf (drvui->fpoutp, "#include \"chars.inc\"\n");
        if (drvui->roughness > 0.) {
          fprintf (drvui->fpoutp, "\ndefault {\n");
          fprintf (drvui->fpoutp, "   finish {\n");
          fprintf (drvui->fpoutp, "     ambient %5.2f\n",drvui->ambient);
          fprintf (drvui->fpoutp, "     diffuse %5.2f\n",drvui->diffuse);
          fprintf (drvui->fpoutp, "     specular %5.2f\n",drvui->specular);
          fprintf (drvui->fpoutp, "     roughness %5.2f\n",drvui->roughness);
          fprintf (drvui->fpoutp, "   }\n");
          fprintf (drvui->fpoutp, " }\n");
        }	
        if (!M_cameras) {
          fprintf (drvui->fpoutp, "camera{ orthographic\n");
        } else {
          fprintf (drvui->fpoutp, "camera{\n");
      }
      Scale *= 0.9f;
      fprintf (drvui->fpoutp, " location< 0 ,0, %f>\n", 1.6 * Scale);
      if (!M_cameras) {
        fprintf (drvui->fpoutp, " up <0,%f,0>\n", 0.1 * Scale);
        fprintf (drvui->fpoutp, " right <%f,0,0>\n", -0.1 * Scale);
      } else {
        fprintf (drvui->fpoutp, " up <0, 0.1, 0>\n");
        fprintf (drvui->fpoutp, " right <-0.1, 0, 0>\n");
      }
      fprintf (drvui->fpoutp, " look_at <0,0,0>\n");
      fprintf (drvui->fpoutp, "}\n");
      fprintf (drvui->fpoutp, "background{color %s}\n", drvui->col_bg_p);
      fprintf (drvui->fpoutp, "light_source {< 0, 0, %f>\n", Scale);
      fprintf (drvui->fpoutp, " color red 4.0 green 4.0 blue 4.0\n}\n");
      fprintf (drvui->fpoutp, "object{\n union{\n");
      fprintf (drvui->fpoutp, "\n");
      fprintf (drvui->fpoutp, "#declare xrot = %5.1f;\n", xrot);
      fprintf (drvui->fpoutp, "#declare yrot = %5.1f;\n", yrot);
      fprintf (drvui->fpoutp, "#declare zrot = %5.1f;\n", zrot);
      fprintf (drvui->fpoutp, "\n");
    } else {             /* stuff for following frames */
      fprintf (drvui->fpoutp, "\n/************ Frame %d **********/\n", drvui->frame_no);
      fprintf (drvui->fpoutv, "\n############# Frame %d ##########\n", drvui->frame_no);
    }    /* only do for first frame */
    do_drawing ();				 /* generate all elements of the drawing */
    free (vertex_list);
    free (poly_list);
  }						 /* while (frame_no != max_frame) - loop through all frames */

#ifdef OPENGL
  if (gl_view)
      glEndList ();
#endif
  fprintf (drvui->fpoutp, " }\n");
  fprintf (drvui->fpoutp, "rotate <xrot, 0, 0>\n");
  fprintf (drvui->fpoutp, "rotate <0, yrot, 0>\n");
  fprintf (drvui->fpoutp, "rotate <0, 0, zrot>\n");
  fprintf (drvui->fpoutp, "}\n");
  (void) fflush (drvui->fpoutp);

  if (Vrml2)
    fprintf (drvui->fpoutv, "]\n");
  fprintf (drvui->fpoutv, " }\n");
  if (Vrml2) {
    fprintf (drvui->fpoutv, " }\n");
    fprintf (drvui->fpoutv, " }\n");
  }

  (void) fflush (drvui->flout);
  (void) fflush (drvui->fpoutv);
}						 /* end of crystal_compute */

/* ************************************************************** */
/* ************************************************************** */

void draw_cell (int docell)

/* procedure to add unit-cell outline to geometry file */
/* docell - non-zero if unit-cell to be drawn */
{
  int i;
  float vert[3];
  float radius;
#ifdef OPENGL
  int j, k;
  float df[3];
  float at[3];
#endif

  nvert = 0;					 /* initialize vertex list */
  radius = rad_cell * 0.01f *Scale;

  for (i = 0; i <= 2; ++i)
    vert[i] = 0.0f;
  add_vert_nc (vert);				 /* place 0,0,0 in list */
  vert[2] = 1.0f;
  add_vert_nc (vert);				 /* place 0,0,1 in list */
  vert[1] = 1.0f;
  add_vert_nc (vert);				 /* place 0,1,1 in list */
  vert[2] = 0.0f;
  add_vert_nc (vert);				 /* place 0,1,0 in list */
  vert[1] = 0.0f;
  add_vert_nc (vert);				 /* place 0,0,0 in list */
  vert[0] = 1.0f;
  add_vert_nc (vert);				 /* place 1,0,0 in list */
  vert[1] = 1.0f;
  add_vert_nc (vert);				 /* place 1,1,0 in list */
  vert[2] = 1.0f;
  add_vert_nc (vert);				 /* place 1,1,1 in list */
  vert[1] = 0.0;
  add_vert_nc (vert);				 /* place 1,0,1 in list */
  vert[2] = 0.0f;
  add_vert_nc (vert);				 /* place 1,0,0 in list */
  vert[1] = 1.0f;
  add_vert_nc (vert);				 /* place 1,1,0 in list */
  vert[0] = 0.0f;
  add_vert_nc (vert);				 /* place 0,1,0 in list */
  vert[2] = 1.0f;
  add_vert_nc (vert);				 /* place 0,1,1 in list */
  vert[0] = 1.0f;
  add_vert_nc (vert);				 /* place 1,1,1 in list */
  vert[1] = 0.0f;
  add_vert_nc (vert);				 /* place 1,0,1 in list */
  vert[0] = 0.0f;
  add_vert_nc (vert);				 /* place 0,0,1 in list */


  if (docell != 0) {
    if (Vrml2) {
      fprintf (drvui->fpoutv, " Shape {\n");
      fprintf (drvui->fpoutv,"  geometry IndexedLineSet{ coord Coordinate{ point[\n");
    } else {
      fprintf (drvui->fpoutv, " Separator{\n");
      fprintf (drvui->fpoutv, "  Material{\n");
      fprintf (drvui->fpoutv, "   diffuseColor %s\n", col_cell_v);
      fprintf (drvui->fpoutv, "  }\n");
      fprintf (drvui->fpoutv, "  Coordinate3{ point[\n");
    }
    fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[0], s_vert[1],s_vert[2]);
    fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[3], s_vert[4],s_vert[5]);
    fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[6], s_vert[7],s_vert[8]);
    fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[9], s_vert[10],s_vert[11]);
    fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[15], s_vert[16],s_vert[17]);
    fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[18], s_vert[19],s_vert[20]);
    fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[21], s_vert[22],s_vert[23]);
    fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f]\n  }\n", s_vert[24],s_vert[25], s_vert[26]);
    if (Vrml2) {
      fprintf (drvui->fpoutv, "  coordIndex[");
    } else {
      fprintf (drvui->fpoutv, "  IndexedLineSet{coordIndex[");
    }
    fprintf (drvui->fpoutv, "0,1,2,3,0,4,5,6,7,1,2,6,5,3,0,4,7,-1]\n");
    if (Vrml2)
    {
      fprintf (drvui->fpoutv, "  color Color { color [%s]}\n  colorIndex[0]\n"
          "  colorPerVertex FALSE\n",col_cell_v);
    }
    fprintf (drvui->fpoutv, " }\n");
    fprintf (drvui->fpoutv, "}\n");

#ifdef OPENGL
    if (gl_view) {
      for (i=0;i<12;i++) {                 /* generate unit-cell edges for openGL */
        j = i;
        if (i == 9) j = 10;
        if (i == 10) j = 12;
        k = j+1;
        if (i == 11) {
           j = 8;
           k = 1;
        } 
        at[0] = s_vert[3*j];                 /* beginning vertex */
        at[1] = s_vert[3*j+1];
        at[2] = s_vert[3*j+2];
        df[0] = s_vert[3*k] - at[0];         /* difference from start to end */
        df[1] = s_vert[3*k+1] - at[1];
        df[2] = s_vert[3*k+2] - at[2];
        push_cylinder(df,at, radius, col_cell_v);  /* draw the cylinder */
      } 
    }
#endif
    fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f \n",
	     s_vert[0], s_vert[1], s_vert[2], s_vert[3], s_vert[4],
	     s_vert[5], radius);
    fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
    fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
    fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
	     s_vert[3], s_vert[4], s_vert[5], s_vert[6], s_vert[7],
	     s_vert[8], radius);
    fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
    fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
    fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
	     s_vert[6], s_vert[7], s_vert[8], s_vert[9], s_vert[10],
	     s_vert[11], radius);
    fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
    fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
    fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>, %f\n",
	     s_vert[9], s_vert[10], s_vert[11], s_vert[12], s_vert[13],
	     s_vert[14], radius);
    fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
    fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
    fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
	     s_vert[12], s_vert[13], s_vert[14], s_vert[15], s_vert[16],
	     s_vert[17], radius);
    fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
    fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
    fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
	     s_vert[15], s_vert[16], s_vert[17], s_vert[18], s_vert[19],
	     s_vert[20], radius);
    fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
    fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
    fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
	     s_vert[18], s_vert[19], s_vert[20], s_vert[21], s_vert[22],
	     s_vert[23], radius);
    fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
    fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
    fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
	     s_vert[21], s_vert[22], s_vert[23], s_vert[24], s_vert[25],
	     s_vert[26], radius);
    fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
    fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
    fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
	     s_vert[24], s_vert[25], s_vert[26], s_vert[27], s_vert[28],
	     s_vert[29], radius);
    fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
    fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
    fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
	     s_vert[30], s_vert[31], s_vert[32], s_vert[33], s_vert[34],
	     s_vert[35], radius);
    fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
    fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
    fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
	     s_vert[36], s_vert[37], s_vert[38], s_vert[39], s_vert[40],
	     s_vert[41], radius);
    fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
    fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
    fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
	     s_vert[24], s_vert[25], s_vert[26], s_vert[3], s_vert[4],
	     s_vert[5], radius);
    fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
    fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
  }
/* LABELS */
  if (Display_axes && Labels) {
    nvert = 0;                     /* initialize vertex list */
    generate_triple();

/* draw vector triple in POV */
    fprintf (drvui->fpoutp, "/* drawing vector triple */");
    fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f \n",s_vert[0], s_vert[1], s_vert[2], 
          s_vert[3], s_vert[4],s_vert[5], radius);
    fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
    fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
    fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f \n",s_vert[0], s_vert[1], s_vert[2],
          s_vert[6], s_vert[7],s_vert[8], radius);
    fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
    fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
    fprintf (drvui->fpoutp, "  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f \n",s_vert[0], s_vert[1], s_vert[2],
          s_vert[9], s_vert[10],s_vert[11], radius);
    fprintf (drvui->fpoutp, "   texture{pigment{color %s }}\n", col_cell_p);
    fprintf (drvui->fpoutp, "   no_shadow\n  }\n");
/* draw vector triple in VRML */
    if (Vrml2) {
      fprintf (drvui->fpoutv, " Shape{\n");
      fprintf (drvui->fpoutv, "  geometry IndexedLineSet { coord Coordinate{ point [\n");
      fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[0], s_vert[1],s_vert[2]);
      fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[3], s_vert[4],s_vert[5]);
      fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[6], s_vert[7],s_vert[8]);
      fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f]\n  }\n", s_vert[9], s_vert[10], s_vert[11]);
      fprintf (drvui->fpoutv, "  coordIndex [");
      fprintf (drvui->fpoutv, "0,3,0,2,0,1,-1]\n");
      fprintf (drvui->fpoutv, "   color Color { color [%s]}\n   colorIndex[0]\n"
          "   colorPerVertex FALSE\n",col_cell_v);
      fprintf (drvui->fpoutv, "  }\n");
      fprintf (drvui->fpoutv, " }\n");
    } else {
      fprintf (drvui->fpoutv, " Separator{\n");
      fprintf (drvui->fpoutv, "  Material{\n");
      fprintf (drvui->fpoutv, "   diffuseColor %s\n", col_cell_v);
      fprintf (drvui->fpoutv, "  }\n");
      fprintf (drvui->fpoutv, "  Coordinate3{ point[\n");
      fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[0], s_vert[1],s_vert[2]);
      fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[3], s_vert[4],s_vert[5]);
      fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f,\n", s_vert[6], s_vert[7],s_vert[8]);
      fprintf (drvui->fpoutv, "   %8.5f %8.5f %8.5f]\n  }\n", s_vert[9], s_vert[10], s_vert[11]);
      fprintf (drvui->fpoutv, "  IndexedLineSet{coordIndex[");
      fprintf (drvui->fpoutv, "0,3,0,2,0,1,-1]\n");
      fprintf (drvui->fpoutv, "  }\n");
      fprintf (drvui->fpoutv, " }\n");
    }
  }

  nvert = 0;                     /* empty vertex list */
}				 /* end of draw_cell */

#ifdef OPENGL
/* ************************************************************** */
/* ************************************************************** */

void draw_GL_triple(void)
{
// draw the lines for the vector triple in openGL
    int i;
    float at[3], df[3];
    float radius = rad_cell * 0.01f *Scale;
    
    for (i=0; i<3; i++) {
        at[i] = s_vert[i];
        df[i] = s_vert[3 + i] - at[i];
    }
    push_cylinder(df, at, radius, col_cell_v);
    for (i=0; i<3; i++) {
        at[i] = s_vert[i];
        df[i] = s_vert[6 + i] - at[i];
    }
    push_cylinder(df, at, radius, col_cell_v);
    for (i=0; i<3; i++) {
        at[i] = s_vert[i];
        df[i] = s_vert[9 + i] - at[i];
    }
    push_cylinder(df, at, radius, col_cell_v);
}
#endif

/* ************************************************************** */
/* ************************************************************** */

void generate_spheres (void)

/* routine to generate spherical objects and add then to edit list */
{
  float *radii;					 /* dynamic array for sphere radii */
  int Sphere_Count;				 /* Counter for number output */
  int i, j, k;					 /* loop counters */

  Sphere_Count = 0;
  for (i = 1; i < drvui->nsphere; ++i) {  /* loop through spheres */
    if (drvui->sphere_fn[i] != drvui->frame_no) continue;  /* skip if not in this frame */
    nvert = 0;                /* clear the vertex list */
    for (j = 0; j < natom; ++j)    /* loop through atoms */
      if ((drvui->atom_n[j] & 255) == i)
        find_all_in_box (j);
    if (nvert > 0) {          /* if spheres present */
/* allocate storage for radii of spheres */
      radii = (float *) malloc ((unsigned) (nvert * sizeof (float)));
      for (k = 0; k < nvert; ++k)  /* copy radii info to storage */
        radii[k] = drvui->sphere_size[i] * Sphere_Mult;
      Output_Spheres (radii, i);
      free (radii);
    }
    Sphere_Count += nvert;
  }
  if (Sphere_Count > 0) {
    printf ("%4d spheres.\n", Sphere_Count);
    fprintf (drvui->flout, "Generated %4d spheres.\n", Sphere_Count);
  }
}						 /* end of generate_spheres */

/* ************************************************************** */
/* ************************************************************** */

void generate_triple(void)
{
// generate the vertices for the vector triple
    
    int i, j;
    float vert[3];
   
    nvert = 0;
    for (i = 0; i < 3; i++)
      vert[i] = 0.0f;
    add_vert_nc (vert);                 /* place origin at 0,0,0 in list */
    vert[2] = 1.0f / drvui->lat_con[2];
    add_vert_nc (vert);                 /* place 1 A vector along 001 */
    vert[2] = 0.0f;
    vert[1] = 1.0f / drvui->lat_con[1];
    add_vert_nc (vert);                 /* place 1 A vector along 010 */
    vert[1] = 0.0f;
    vert[0] = 1.0f / drvui->lat_con[0];
    add_vert_nc (vert);                 /* place 1 A vector along 100 */
    vert[2] = 1.4f / drvui->lat_con[2];
    vert[0] = 0.0f;
    add_vert_nc (vert);                 /* place 1.4 A vector along 001 */
    vert[2] = 0.0f;
    vert[1] = 1.4f / drvui->lat_con[1];
    add_vert_nc (vert);                 /* place 1.4 A vector along 010 */
    vert[1] = 0.0f;
    vert[0] = 1.4f / drvui->lat_con[0];
    add_vert_nc (vert);                 /* place 1.4 A vector along 100 */
    for (i=nvert-1; i>=0; --i) {             /* Translate coordinates */
      for (j = 0; j < 3; j++) {
        s_vert[3 * i + j] += offset[j] - s_vert[j];
      }
    }
}

/* ************************************************************** */
/* ************************************************************** */

void import_cif (char input[100], int in_line, int Quick)

/* routine to extract structural information from a CIF */
{
  FILE *impin;
  char string[100], filename[100], tstring[100];
  int U_pos[6], X_pos[3], Label_pos = 0;
  int i, j;

  if (in_line) {
    fprintf (stderr,
	     "Cannot read inline CIF data. Please use the import instruction\n");
    delay (5);
    exit (0);
  }
  (void) sscanf (input, "%s %s", filename, filename);	/* get file name */

  if ((impin = fopen (filename, "r")) == NULL) {
    fprintf (stderr, "Cannot Open CIF Import File, Run aborted.\n");
    delay (5);
    exit (0);
  }
  while (strncmp (string, "_cell_length_a", 14) != 0) {
    if (!fgets (string, 100, impin)) {;		 /* search for a-axis length */
      fprintf (stderr, "Error reading CIF Import File, Run aborted.\n");
      delay (5);
      exit (0);
    }
    Blank_Strip (string);
  }
  (void) sscanf (string, "%s %f", tstring, &drvui->lat_con[0]);
  rewind (impin);
  while (strncmp (string, "_cell_length_b", 14) != 0) {
    if (!fgets (string, 100, impin)) {;		 /* search for b-axis */
      fprintf (stderr, "Error reading CIF Import File, Run aborted.\n");
      delay (5);
      exit (0);
    }
    Blank_Strip (string);
  }
  (void) sscanf (string, "%s %f", tstring, &drvui->lat_con[1]);
  rewind (impin);
  while (strncmp (string, "_cell_length_c", 14) != 0) {
    if (!fgets (string, 100, impin)) {;		 /* search for c axis */
      fprintf (stderr, "Error reading CIF Import File, Run aborted.\n");
      delay (5);
      exit (0);
    }
    Blank_Strip (string);
  }
  (void) sscanf (string, "%s %f", tstring, &drvui->lat_con[2]);
  rewind (impin);
  while (strncmp (string, "_cell_angle_al", 14) != 0) {
    if (!fgets (string, 100, impin)) {;		 /* search for alpha */
      fprintf (stderr, "Error reading CIF Import File, Run aborted.\n");
      delay (5);
      exit (0);
    }
    Blank_Strip (string);
  }
  rewind (impin);
  (void) sscanf (string, "%s %f", tstring, &drvui->lat_con[3]);
  while (strncmp (string, "_cell_angle_be", 14) != 0) {
    if (!fgets (string, 100, impin)) {;		 /* search for beta */
      fprintf (stderr, "Error reading CIF Import File, Run aborted.\n");
      delay (5);
      exit (0);
    }
    Blank_Strip (string);
  }
  rewind (impin);
  (void) sscanf (string, "%s %f", tstring, &drvui->lat_con[4]);
  while (strncmp (string, "_cell_angle_ga", 14) != 0) {
    if (!fgets (string, 100, impin)) {;		 /* search for gamma */
      fprintf (stderr, "Error reading CIF Import File, Run aborted.\n");
      delay (5);
      exit (0);
    }
    Blank_Strip (string);
  }
  rewind (impin);
  j = 1;
  (void) sscanf (string, "%s %f", tstring, &drvui->lat_con[5]);
  if (!Quick)
    fprintf (drvui->flout,
	     "******      Cell     %8.5f %8.5f %8.5f %8.3f %8.3f %8.3f\n",
	     drvui->lat_con[0], drvui->lat_con[1], drvui->lat_con[2], drvui->lat_con[3],
         drvui->lat_con[4], drvui->lat_con[5]);
  while (strncmp (string, "_symmetry_space", 15) != 0) {
    if (!fgets (string, 100, impin)) {;		 /* search for space group */
      j = 0;					 /* indicate not found */
      break;					 /* no error if not found */
    }
    Blank_Strip (string);
  }
  if (j != 0) {
    tstring[0] = 's';
    tstring[1] = 'p';
    tstring[2] = 'g';
    tstring[3] = 'r';
    tstring[4] = ' ';
    for (i = 0; i < 60; i++) {			 /* change the keyword to spaces */
      if (string[i] == ' ') {
	break;
      }
      else {
	string[i] = ' ';
      }
    }
    Blank_Strip (string);			 /* strip off the spaces */
    if (string[0] != '?') {
      for (i = 0; i < 60; i++)
	if (string[i] == (char) 39)
	  break;
      for (i++, j = 5; i < 60; i++)
	if (string[i] == (char) 39)
	  break;
	else
	  tstring[j++] = string[i];
      tstring[j] = '\0';
      if (!Quick)
	fprintf (drvui->flout, "******      %s\n", tstring);
      for (i = 0; i < 60; i++) {
	if (tstring[i] == 'H')
	  tstring[i] = ' ';			 /* remove any H from rhomb. spgrps */
	if (tstring[i] == 'S')
	  tstring[i] = ' ';			 /* remove any S from monoclinic spgrps */
      }
      symop (tstring);
    }
  }
  rewind (impin);
/* skip through file to atom_site */
  while (strncmp (string, "_atom_site_", 11) != 0) {
    if (!fgets (string, 100, impin)) {		 /* search for phase info */
      fprintf (stderr, "Error reading CIF Import File, Run aborted.\n");
      delay (5);
      exit (0);
    }
    Blank_Strip (string);
  }
/* first _atom_site_ line in string need to find position of label, x, y, and z */
  j = 0;
  while (strncmp (string, "_atom_site_", 11) == 0) {
    if (strncmp (string, "_atom_site_label", 16) == 0)
      Label_pos = j;
    if (strncmp (string, "_atom_site_fract_x", 18) == 0)
      X_pos[0] = j;
    if (strncmp (string, "_atom_site_fract_y", 18) == 0)
      X_pos[1] = j;
    if (strncmp (string, "_atom_site_fract_z", 18) == 0)
      X_pos[2] = j;
    j++;
    if (!fgets (string, 100, impin)) {		 /* search for phase info */
      fprintf (stderr, "Error reading CIF Import File, Run aborted.\n");
      delay (5);
      exit (0);
    }
    Blank_Strip (string);
  }
  for (;;) {
    if (strlen (string) < 5)
      break;
    if (strncmp (string, "loop_", 5) == 0)
      break;
    drvui->atom_n[natom] = 0;
    for (i = 0; i < 4; i++)
      drvui->atom_l[natom][i] = ' ';			 /* initialize atom name */
    strcpy (tstring, string);
    if (Label_pos > 0)
      Token_Strip (tstring, Label_pos);
    j = 0;
    for (i = 0; i < 10; i++) {
      if (tstring[i] >= '0' && tstring[i] <= '9') {
        drvui->atom_n[natom] = 10 * drvui->atom_n[natom] + (int) tstring[i] - 48;
        drvui->atom_l[natom][i] = ' ';
      } else {
        drvui->atom_l[natom][j++] = tstring[i];
        if (j > 3 || tstring[i] == ' ')
          break;
      }
    }
    strcpy (tstring, string);
    if (X_pos[0] > 0)
      Token_Strip (tstring, X_pos[0]);
    (void) sscanf (tstring, "%f", &drvui->atom_xyz[natom][0]);
    strcpy (tstring, string);
    if (X_pos[1] > 0)
      Token_Strip (tstring, X_pos[1]);
    (void) sscanf (tstring, "%f", &drvui->atom_xyz[natom][1]);
    strcpy (tstring, string);
    if (X_pos[2] > 0)
      Token_Strip (tstring, X_pos[2]);
    (void) sscanf (tstring, "%f", &drvui->atom_xyz[natom][2]);
    if (!Quick)
      fprintf (drvui->flout,
	       "******      Atom %c%c%c%c%3d %8.5f %8.5f %8.5f\n",
	       drvui->atom_l[natom][0], drvui->atom_l[natom][1], 
           drvui->atom_l[natom][2], drvui->atom_l[natom][3],
	       drvui->atom_n[natom], drvui->atom_xyz[natom][0],
           drvui->atom_xyz[natom][1], drvui->atom_xyz[natom][2]);
    for (i = 0; i < 100; i++)
      string[i] = ' ';
    if (!fgets (string, 100, impin)) {
      fprintf (stderr, "Error reading CIF Import File, Run aborted.\n");
      delay (5);
      exit (0);
    }
    Blank_Strip (string);
    if (++natom >= MAX_ATOM) {
      fprintf (stderr,
	       "Too many atoms in asymmetric unit - increase MAX_ATOM.\n");
      delay (5);
      exit (0);
    }
  }
  rewind (impin);
/* skip through file to atom_site_aniso */
  while (strncmp (string, "_atom_site_aniso", 16) != 0) {
    if (!fgets (string, 100, impin)) {		 /* search for anisotropic U's */
      (void) fclose (impin);
      return;
    }
    Blank_Strip (string);
  }
/* first _atom_site_aniso line in string
   need to find position of label, U11, etc */
  j = 0;
  while (strncmp (string, "_atom_site_aniso", 16) == 0) {
    if (strncmp (string, "_atom_site_aniso_label", 22) == 0)
      Label_pos = j;
    if (strncmp (string, "_atom_site_aniso_U_11", 21) == 0)
      U_pos[0] = j;
    if (strncmp (string, "_atom_site_aniso_U_22", 21) == 0)
      U_pos[1] = j;
    if (strncmp (string, "_atom_site_aniso_U_33", 21) == 0)
      U_pos[2] = j;
    if (strncmp (string, "_atom_site_aniso_U_12", 21) == 0)
      U_pos[3] = j;
    if (strncmp (string, "_atom_site_aniso_U_13", 21) == 0)
      U_pos[4] = j;
    if (strncmp (string, "_atom_site_aniso_U_23", 21) == 0)
      U_pos[5] = j;
    j++;
    if (!fgets (string, 100, impin)) {		 /* search for phase info */
      fprintf (stderr, "Error reading CIF Import File, Run aborted.\n");
      delay (5);
      exit (0);
    }
    Blank_Strip (string);
  }
  for (;;) {
    if (strlen (string) < 5)
      break;
    if (strncmp (string, "_", 1) == 0)
      break;
    drvui->ellips_n[n_ellips] = 0;
    for (i = 0; i < 4; i++)
      drvui->ellips_l[n_ellips][i] = ' ';		 /* initialize ellipsoid name */
    strcpy (tstring, string);
    if (Label_pos > 0)
      Token_Strip (tstring, Label_pos);
    j = 0;
    for (i = 0; i < 10; i++) {
      if (tstring[i] >= '0' && tstring[i] <= '9') {
	drvui->ellips_n[n_ellips] = 10 * drvui->ellips_n[n_ellips] + (int) tstring[i] - 48;
	drvui->ellips_l[n_ellips][i] = ' ';
      }
      else {
	drvui->ellips_l[n_ellips][j++] = tstring[i];
	if (j > 3 || tstring[i] == ' ')
	  break;
      }
    }
    drvui->ell_type[n_ellips] = 1;			 /* Uij */
    do_ellipsoids = 1;
    drvui->save_el_number[n_ellips] = drvui->ellips_n[n_ellips];
    strcpy (tstring, string);
    if (U_pos[0] > 0)
      Token_Strip (tstring, U_pos[0]);
    (void) sscanf (tstring, "%f", &drvui->ellips[n_ellips][0]);
    strcpy (tstring, string);
    if (U_pos[1] > 0)
      Token_Strip (tstring, U_pos[1]);
    (void) sscanf (tstring, "%f", &drvui->ellips[n_ellips][1]);
    strcpy (tstring, string);
    if (U_pos[2] > 0)
      Token_Strip (tstring, U_pos[2]);
    (void) sscanf (tstring, "%f", &drvui->ellips[n_ellips][2]);
    strcpy (tstring, string);
    if (U_pos[3] > 0)
      Token_Strip (tstring, U_pos[3]);
    (void) sscanf (tstring, "%f", &drvui->ellips[n_ellips][3]);
    strcpy (tstring, string);
    if (U_pos[4] > 0)
      Token_Strip (tstring, U_pos[4]);
    (void) sscanf (tstring, "%f", &drvui->ellips[n_ellips][4]);
    strcpy (tstring, string);
    if (U_pos[5] > 0)
      Token_Strip (tstring, U_pos[5]);
    (void) sscanf (tstring, "%f", &drvui->ellips[n_ellips][5]);
    memset(drvui->ellips_col_p[n_ellips],0,40);
    memset(drvui->ellips_col_v[n_ellips],0,40);
    strcpy(drvui->ellips_col_p[n_ellips],"Gray30");
    strcpy(drvui->ellips_col_v[n_ellips],"0.3 0.3 0.3");
    fprintf (drvui->flout,
	     "******      Uij  %c%c%c%c%3d %8.5f %8.5f %8.5f %8.5f %8.5f %8.5f\n",
	     drvui->ellips_l[n_ellips][0], drvui->ellips_l[n_ellips][1],
	     drvui->ellips_l[n_ellips][2]
	     , drvui->ellips_l[n_ellips][3], drvui->ellips_n[n_ellips], drvui->ellips[n_ellips][0]
	     , drvui->ellips[n_ellips][1], drvui->ellips[n_ellips][2], drvui->ellips[n_ellips][3]
	     , drvui->ellips[n_ellips][4], drvui->ellips[n_ellips][5]);
    if (++n_ellips >= MAX_EL) {
      fprintf (stderr,
	       "Too many ellipsoid types specified - increase MAX_EL.\n");
      delay (5);
      exit (0);
    }
    if (!fgets (string, 100, impin)) {
      fprintf (stderr, "Error reading CIF Import File, Run aborted.\n");
      delay (5);
      exit (0);
    }
    Blank_Strip (string);
  }
  (void) fclose (impin);
}

/* ************************************************************** */
/* ************************************************************** */

void import_schakal (char input[100], int in_line, int Quick)

/* routine to extract structural information from a SCHAKAL data file */
{
  FILE *impin;
  char string[100];				 /* temporary string */
  char filename[100];				 /* string for filename */
  int intype;					 /* type of datum read */
  int i, j;					 /* temporary */

  if (in_line) {
    impin = drvui->fpin;
  }
  else {
    (void) sscanf (input, "%s %s", filename, filename);	/* get file name */
    if ((impin = fopen (filename, "r")) == NULL) {
      fprintf (stderr, "Cannot Open SCHAKAL Import File, Run aborted.\n");
      delay (5);
      exit (0);
    }
  }
  if (!Quick)
    fprintf (drvui->flout, "****** Now interpreting SCHAKAL-style data\n");
  while (!feof (impin)) {
    if (!fgets (input, 100, impin)) {		 /* read a line */
      fprintf (stderr, "Error reading file.");
      delay (5);
      exit (1);
    }
    intype = 0;
    for (i = 0; i < 4; i++) {
      if (input[i] == '\0')
	intype = 3;
    }
    if (strncmp (input, "TITL", 4) == 0)
      intype = 1;				 /* echo and ignore */
    if (strncmp (input, "titl ", 4) == 0)
      intype = 1;				 /* echo and ignore */
    if (strncmp (input, "CELL", 4) == 0)
      intype = 2;
    if (strncmp (input, "cell", 4) == 0)
      intype = 2;
    if (strncmp (input, "ATOM", 4) == 0)
      intype = 3;				 /* atom coordinate card */
    if (strncmp (input, "atom", 4) == 0)
      intype = 3;
    if (strncmp (input, "END", 3) == 0)
      intype = 4;				 /* trigger exit */
    if (strncmp (input, "end", 3) == 0)
      intype = 4;				 /* trigger exit */


    switch (intype) {
    case 1:
      for (i = 0; i < 51; i++)
	if (input[i] == '\n')
	  input[i] = ' ';
      input[50] = '\0';				 /* truncate title */
      if (!Quick)
	fprintf (drvui->flout, "******      %s\n", input);
      break;					 /* ignore title except for listing */
    case 2:
      for (i = 3; i <= 5; ++i)
        drvui->lat_con[i] = (float) 0.;
      (void) sscanf (input, "%s %f %f %f %f %f %f", string, &drvui->lat_con[0],
		     &drvui->lat_con[1], &drvui->lat_con[2], &drvui->lat_con[3],
             &drvui->lat_con[4], &drvui->lat_con[5]);
      if (!Quick)
	fprintf (drvui->flout,
		 "******      Cell     %8.5f %8.5f %8.5f %8.3f %8.3f %8.3f\n",
		 drvui->lat_con[0], drvui->lat_con[1], drvui->lat_con[2], drvui->lat_con[3],
		 drvui->lat_con[4], drvui->lat_con[5]);
      break;
    case 3:					 /* read an atom line */
      (void) sscanf (input, "%s %c%c%c%c %f %f %f", string,
		     &drvui->atom_l[natom][0], &drvui->atom_l[natom][1],
		     &drvui->atom_l[natom][2], &drvui->atom_l[natom][3],
		     &drvui->atom_xyz[natom][0], &drvui->atom_xyz[natom][1],
		     &drvui->atom_xyz[natom][2]);
      drvui->atom_n[natom] = 0;
      for (j = 0; j < 4; j++)			 /* extract atom number from SCHAKAL name */
	if (drvui->atom_l[natom][j] >= '0' && drvui->atom_l[natom][j] <= '9') {
	  drvui->atom_n[natom] = drvui->atom_n[natom] * 10 + (int) drvui->atom_l[natom][j] - 48;
	  drvui->atom_l[natom][j] = ' ';
	}
      if (drvui->atom_n[natom] == 0)
        drvui->atom_n[natom] = 1;
      if (!Quick)
        fprintf (drvui->flout,
		   "******      Atom %c%c%c%c%3d %8.5f %8.5f %8.5f\n",
		   drvui->atom_l[natom][0], drvui->atom_l[natom][1],
           drvui->atom_l[natom][2], drvui->atom_l[natom][3],
           drvui->atom_n[natom], drvui->atom_xyz[natom][0],
		   drvui->atom_xyz[natom][1], drvui->atom_xyz[natom][2]);
      natom++;
      if (natom >= MAX_ATOM) {
        fprintf (stderr,"Too many atoms in asymmetric unit - increase MAX_ATOM.\n");
        delay (5);
        exit (0);
      }
      break;
    case 4:
      if (!Quick)
	fprintf (drvui->flout, "****** END - end of SCHAKAL dataset\n");
      if (!in_line)
	(void) fclose (impin);
      return;
    default:
      break;					 /* end of default */
    }						 /* end of switch */
  }
}						 /* end of import_schakal */

/* ************************************************************** */
/* ************************************************************** */

void trim_string(char string[], int len)
{
/* trim string - remove trailing ^J (Windows) and spaces. 'len' is maximum length */
    int i;
    string[len-1] = 0;
    if (strlen(string) == 0) {
        strcpy(string,"White");
        return;
    }
    for (i=0;i<(int)strlen(string);i++) {
        if (string[i] < ' ') {
            string[i] = 0;
            break;
        }
    }
    for (i=strlen(string);i > 0; --i) {
        if (string[i-1] <= ' ') {
            string[i-1] = 0;
        } else
            break;
    }
}


