/* ===================================================================== */
/*     Read_Input.c                                                      */
/*                                                                       */
/* Routine to read crystallographic information and Fourier Map files    */
/* ===================================================================== */
/* Routines included in this file:
 *
 *   check_atom_name - compare two atom name-type variables
 *   convert_pos - convert string of form n/m to float
 *   end_flp - convert endedness of an integer
 *   end_flp_real - convert endedness of a float
 *   Error_Box - print error message to stderr
 *   read_fcf - read SHELX/CIF Fo/Fc file and compute Fourier Map
 *   read_flp - read FullProf map
 *   read_grd - read GSAS (.grd) Fourier Map
 *   read_inp - read the input (.str) file
 *   read_stf - read JAVA (.stf) Fourier Map file
 *   read_vasp -read VASP (.vsp) Fourier Map file
 *   read_w2k - read WIEN (.w2k) Fourier Map file
 *
 *************************************************************************/

#include "drawxtl.h"
#include "draw_ext.h"
int end_flip(int);
float end_flip_real(float);
void Error_Box(char *);

/* ************************************************************** */
/* ************************************************************** */

int check_atom_name(char * name1, char * name2)
{
// determine if two atom names are the same. Return 1 (TRUE) if they are

    if (name1[0] == name2[0] && name1[1] == name2[1] &&
        name1[2] == name2[2] && name1[3] == name2[3])
        return 1;
    else
        return 0;
}

/* ************************************************************** */
/* ************************************************************** */

float convert_pos(char * string)
{
/* routine to convert an atom string to a float. If only digits, a simple sscanf
 *  is done. If the position is of the form n/m, each part is decoded and the
 * results of the division is returned.
 */

    float result;
    int i, n, m;
    char * pdest;
    char sub_string1[10], sub_string2[10];

    if (!(pdest = strstr(string,"/"))) {
        sscanf(string,"%f",&result);             /* no / in the input */
    } else {
        i = pdest - string +1;                   /* number of chars to copy */
        strncpy(sub_string1, string, i);
        sub_string1[i - 1] = '\0';
        strcpy(sub_string2, pdest + 1);          /* copy part after / */
        sscanf(sub_string1,"%d", &n);
        sscanf(sub_string2,"%d", &m);
        result = (float)n/(float)m;
    }
    return result;
}

/* ************************************************************** */
/* ************************************************************** */

int end_flip(int value)
{
/* routine to flip the endian nature of the input integer
 * This routine was copied from a Web article:
 *   http://www.codeproject.com/cpp/endianness.asp by Juan Carlos Cobas
 */

   return (((value&0x000000FF)<<24)+((value&0x0000FF00)<<8)+
   ((value&0x00FF0000)>>8)+((value&0xFF000000)>>24));
}

/* ************************************************************** */
/* ************************************************************** */

float end_flip_real(float value)
{
/* routine to flip the endian nature of the input float 
 * This routine was copied from a Web article:
 *   http://www.codeproject.com/cpp/endianness.asp by Juan Carlos Cobas
 */
   union u {float vi; unsigned char c[4];}; 
   union v {float ni; unsigned char d[4];};
   union u un; 
   union v vn; 
   un.vi = value;
   vn.d[0]=un.c[3]; 
   vn.d[1]=un.c[2]; 
   vn.d[2]=un.c[1]; 
   vn.d[3]=un.c[0]; 
   return (vn.ni); 

}

/* ************************************************************** */
/* ************************************************************** */

void Error_Box(char * string)
{
    fprintf(stderr,"%s\n",string);
}

/* ************************************************************** */
/* ************************************************************** */

void read_fcf(char * infile, int Quick)
{
/* read CIF-standard Fo/Fc file and compute density */

    FILE * mapin;
    float *fo,*fc,*f2,*ftmp1, *ftmp2, *ftmp3;
    short *ih,*ik,*il,*itmp1, *itmp2, *itmp3;
    float * Cklx;
    float * Dklx;
    float * Elxy;
    float * Flxy;
    double x,y,z;
    char string[100];
    char maptitle[61];
    int squared,phases,ab;
    int nval=200;
    int nr;
    int i,j,k,l;
    int ijk = 0;
    float rhomin=1.0e15f, rhomax=-1.0e15f;
    float factor;
    short ih0,ik0,il0;
    short ih1,ik1,il1;
    float fo0,fc0,fc1,f20,f21;
    int kmin, kmax, nk;
    int lmin, lmax, nl;
    float phase;
    int skip;
    int progress;

    if (FourierPt != NULL) {
        return;
    }

    if ((mapin = fopen (infile, "r")) == NULL) {
        sprintf(string,"Cannot open Fo/Fc (CIF fcf) file %s", infile);
        Error_Box(string);
        return;
    }

    fo=(float*)malloc(200*sizeof(float));
    if (fo==NULL) {
        sprintf(string,"Unable to obtain storage for Fo data");
        Error_Box(string);
        return;
    }
    fc=(float*)malloc(200*sizeof(float));
    if (fc==NULL) {
        sprintf(string,"Unable to obtain storage for Fc data");
        Error_Box(string);
        return;
    }
    f2=(float*)malloc(200*sizeof(float));
    if (f2==NULL) {
        sprintf(string,"Unable to obtain storage for B or phase data");
        Error_Box(string);
        return;
    }
    ih=(short*)malloc(200*sizeof(short));
    if (ih==NULL) {
        sprintf(string,"Unable to obtain storage for h data");
        Error_Box(string);
        return;
    }
    ik=(short*)malloc(200*sizeof(short));
    if (ik==NULL) {
        sprintf(string,"Unable to obtain storage for k data");
        Error_Box(string);
        return;
    }
    il=(short*)malloc(200*sizeof(short));
    if (il==NULL) {
        sprintf(string,"Unable to obtain storage for l data");
        Error_Box(string);
        return;
    }
 
    do {
      fgets(string,80,mapin);
      if (strstr(string,"title")) {
        memset(maptitle,0,61); 
        sscanf(string,"%*s %60c",maptitle); 
        trim_string(maptitle,60);
        strcpy(Map_Info.title,maptitle);
        Map_Info.info_valid = 1;
      }
    } while (strncmp(string," _refln",6) ); /* search for start of _refln group */

    if (strncmp(string," _refln_index_h",14)) {
      sprintf(string,"Unsupported field sequence in fcf file - need h k l first");
      Error_Box(string);
      return;
    }

    squared = 0; 
    phases = 0;
    ab = 0;
    while (!strncmp(string," _refln",6) ) {
      fgets(string,80,mapin);
      if (strstr(string,"squared")) squared=1;
      if (strstr(string,"_phase_")) phases=1;
      if (strstr(string,"A_calc")) ab = 1;
    }

    if (phases == 0 && ab == 0) {
      sprintf(string,"fcf file does not contain phase angle or A,B data");
      Error_Box(string);
      return;
    }

    nr = 0;
    do {
      i=sscanf(string,"%hd %hd %hd %f %*f %f %f",&ih0,&ik0,&il0,&fo0,&fc0,&f20);
      if (i<6) break;
      if (squared==1) fo0=(float)sqrt(fo0);
      if (ab) {       /* change all data to 'phase' type */
          fc1 = (float) sqrt(fc0 * fc0 + f20 * f20);
          f20 = (float)atan2(f20/fc1, fc0/fc1);
          fc0 = fc1;
      } else {
          f20 *= (float)(PI/180.0);     /* phase in radians */
      }

/* apply spacegroup symmetry (including xyz - catch duplicates in input list) */
      for (k = 0; k < drvui->ng; ++k) {
          skip=0;

          ih1=drvui->ss[k][0][0]*ih0 + drvui->ss[k][1][0]*ik0 + drvui->ss[k][2][0]*il0;
          ik1=drvui->ss[k][0][1]*ih0 + drvui->ss[k][1][1]*ik0 + drvui->ss[k][2][1]*il0;
          il1=drvui->ss[k][0][2]*ih0 + drvui->ss[k][1][2]*ik0 + drvui->ss[k][2][2]*il0;
          phase = 2.0f * (float)PI * (drvui->ts[k][0]*ih0 + drvui->ts[k][1]*ik0 
              + drvui->ts[k][2]*il0);
          f21 = f20 + phase;
          f21 = (float)atan2(sin(f21),cos(f21));    /* get f21 in range -PI to PI */
       
          for (j=0;j<nr;j++) {
              if (ih[j] == ih1 && ik[j] == ik1 && il[j] == il1) { 
                skip=1;
                break;
              }
          }
          if (skip==0) {
            ih[nr] = ih1;
            ik[nr] = ik1;
            il[nr] = il1;
            fo[nr] = fo0;
            fc[nr] = fc0;
            f2[nr] = f21;
            nr++;

          }
          ih1 *= -1;               /* add Friedel opposite */
          ik1 *= -1;
          il1 *= -1;
          f21 *= -1.0;
          skip = 0;

          for (j=0;j<nr;j++) {
              if (ih[j] == ih1 && ik[j] == ik1 && il[j] == il1) {
                skip=1;
                break;
              }
          }
          if (skip==0) {
            ih[nr]=ih1;
            ik[nr]=ik1;
            il[nr]=il1;
            fo[nr]=fo0;
            fc[nr] = fc0;
            f2[nr] = f21;
            nr++;
          }

          if (nr > nval-10) {
            nval +=100;
            ftmp1=(float*)realloc(fo,nval*sizeof(float));
            ftmp2=(float*)realloc(fc,nval*sizeof(float));
            ftmp3=(float*)realloc(f2,nval*sizeof(float));
            itmp1=(short*)realloc(ih,nval*sizeof(short));
            itmp2=(short*)realloc(ik,nval*sizeof(short));
            itmp3=(short*)realloc(il,nval*sizeof(short));
            if (!ftmp1 || !ftmp2 || !ftmp3 || !itmp1 || !itmp2 || !itmp3) {
              sprintf(string,"Unable to expand storage for h, k, l, fo, fc or phase data");
              free(fo);
              free(fc);
              free(f2);
              free(ih);
              free(ik);
              free(il);
              Error_Box(string);
              return;
            } else {
              fo = ftmp1;
              fc = ftmp2;
              f2 = ftmp3;
              ih = itmp1;
              ik = itmp2;
              il = itmp3;
            }
          } /* if we need to realloc */
      } /* loop over all symmetry operators */
      fgets(string,80,mapin);
    } while (i>0 && !feof(mapin));

    fclose(mapin);


    factor = 1.0f/(drvui->lat_con[0]*drvui->lat_con[1]*drvui->lat_con[2]
        * (float)sqrt(1.0 - cos(drvui->lat_con[3] * PI / 180.0) *
                     cos(drvui->lat_con[3] * PI / 180.0) *
                     cos(drvui->lat_con[4] * PI / 180.0) *
                     cos(drvui->lat_con[4] * PI / 180.0) *
                     cos(drvui->lat_con[5] * PI / 180.0) *
                     cos(drvui->lat_con[5] * PI / 180.0)));
    kmin = kmax = ik[0];
    lmin = lmax = il[0];
    mapin = fopen("refl.out","w");

    for (l=0; l < nr; l++) {
        if (ik[l] < kmin) kmin = ik[l];     /* find range of k and l */
        if (ik[l] > kmax) kmax = ik[l];
        if (il[l] < lmin) lmin = il[l];
        if (il[l] > lmax) lmax = il[l];
        switch (Map_Info.map_type) {        /* get map coefficients/V */
        default:
        case 0:                             /* Fo map */
            break;
        case 1:                             /* Fc map */
            fo[l] = fc[l];
            break;
        case 2:                             /* Fo - Fc map */
            fo[l] = fo[l] - fc[l];
            break;
        case 3:                             /* 2Fo - Fc map */
            fo[l] = 2.0f * fo[l] - fc[l];
            break;
        case 4:                             /* Fo2 (Patterson) */
            fo[l] = fo[l] * fo[l];
            f2[l] = 0.0f;
            break;
        }
        fprintf(mapin,"%4d %3d %3d %8.3f %8.3f %8.3f\n",ih[l],ik[l],il[l],fo[l],fc[l],f2[l]);
        fo[l] *= factor;                          /* update coefficients times 1/V */
    }

    fclose(mapin);
    nk = kmax - kmin +1;
    nl = lmax - lmin +1;
    Cklx = (float *)malloc(nk * nl * sizeof(float));
    Dklx = (float *)malloc(nk * nl * sizeof(float));
    Elxy = (float *)malloc(nl * sizeof(float));
    Flxy = (float *)malloc(nl * sizeof(float));

    if (!Cklx || !Dklx || !Elxy || !Flxy) {
        Error_Box("Unable to allocate Beevers-Lipson arrays.");
        return;
    }
    mapstep_a = (int) (4.0f * drvui->lat_con[0] + 0.5f);
    mapstep_b = (int) (4.0f * drvui->lat_con[1] + 0.5f);
    mapstep_c = (int) (4.0f * drvui->lat_con[2] + 0.5f);

    FourierPt = (float *) malloc( mapstep_a*mapstep_b*mapstep_c * sizeof(float));
    if (FourierPt == NULL) {
      Error_Box("ERROR -- Unable to allocate space for map.");
      return;
    }
    memset(FourierPt, 0, mapstep_a*mapstep_b*mapstep_c * sizeof(float));

/* start the Beevers-Lipson expansion */

    Progress_Window(-1,"Computing Density",(float)mapstep_a*mapstep_b);
    
    progress=0;
    for(i=0; i < mapstep_a; i++) {
        x = (float)i/(float)mapstep_a;
        memset(Cklx, 0, nk * nl * sizeof(float));
        memset(Dklx, 0, nk * nl * sizeof(float));
        for (l=0; l<nr; l++) {
            ijk = (il[l] - lmin) * nk + ik[l] - kmin;
            Cklx[ijk] += fo[l] * (float)cos(2.0 * PI * ih[l] * x - f2[l]);
            Dklx[ijk] -= fo[l] * (float)sin(2.0 * PI * ih[l] * x - f2[l]);
        }
        for(j=0; j < mapstep_b; j++) {
            Progress_Window(0,NULL,(float)++progress);
            y = (float)j/(float)mapstep_b;
            memset(Elxy, 0, nl * sizeof(float));
            memset(Flxy, 0, nl * sizeof(float));
            for (l=0; l<nl; l++) {
                for (k=0; k<nk; k++) {
                    ijk = l * nk + k;
                    Elxy[l] += Cklx[ijk] * (float)cos(2.0 * PI * (k+kmin) * y)
                             + Dklx[ijk] * (float)sin(2.0 * PI * (k+kmin) * y);
                    Flxy[l] += Dklx[ijk] * (float)cos(2.0 * PI * (k+kmin) * y)
                             - Cklx[ijk] * (float)sin(2.0 * PI * (k+kmin) * y);
                }
            }
            for(k=0; k < mapstep_c; k++) {
              z = (float)k/(float)mapstep_c;
              ijk =  i*mapstep_b*mapstep_c + j*mapstep_c + k;

              for (l=0; l<nl; l++) {
                FourierPt[ijk] += Elxy[l] * (float)cos(2.0 * PI * (l + lmin) * z)
                                + Flxy[l] * (float)sin(2.0 * PI * (l + lmin) * z);
              }
              if (FourierPt[ijk] < rhomin) 
                  rhomin = FourierPt[ijk];
              if (FourierPt[ijk] > rhomax) 
                  rhomax = FourierPt[ijk];
            }
        }
    }
    print_map_info(Quick);

    Progress_Window(-2, NULL, 0.0f);
    ReadFourMap = 1;
    Map_Info.rhomn = rhomin;
    Map_Info.rhomx = rhomax;
    Map_Info.map_int[0] = mapstep_a;
    Map_Info.map_int[1] = mapstep_b;
    Map_Info.map_int[2] = mapstep_c;
    Map_Info.lat_con[0] = drvui->lat_con[0];
    Map_Info.lat_con[1] = drvui->lat_con[1];
    Map_Info.lat_con[2] = drvui->lat_con[2];
    Map_Info.lat_con[3] = drvui->lat_con[3];
    Map_Info.lat_con[4] = drvui->lat_con[4];
    Map_Info.lat_con[5] = drvui->lat_con[5];
    Map_Info.xlim[0] = 0.0f;
    Map_Info.xlim[1] = 1.0f;
    Map_Info.ylim[0] = 0.0f;
    Map_Info.ylim[1] = 1.0f;
    Map_Info.zlim[0] = 0.0f;
    Map_Info.zlim[1] = 1.0f;
	
    free(Cklx);
    free(Dklx);
    free(Elxy);
    free(Flxy);
    free(fo);
    free(fc);
    free(f2);
    free(ih);
    free(ik);
    free(il);
}

/* ************************************************************** */
/* ************************************************************** */

void read_flp(char * infile, int Quick)
{
/* read FullProf binary map file */

#include "read_flp.h"                /* define all the struct's */

    char titulo[81];
    char version[6];
    struct crystal_cell_type celda;
    FILE * mapin;
    float rhomin=1.0e+15f, rhomax=-1.0e+15f;
    int ireslt = 0;
    int  icount;
    struct space_group spg;
    char temp[16384];
    int i, j, k, ijk=0;
    int natoms;
    struct FFT_param map_param;
    struct Atom_Type atm[200];
    float dens[200];
    int new_mapstep_a, astart, aend;
    int new_mapstep_b, bstart, bend;
    int new_mapstep_c, cstart, cend;
    int wrong_end = 0;

    memset(version, 0, 6);
    memset(titulo, 0, 81);
    if ((mapin = fopen (infile, "r+b")) == NULL) {
        if (!Quick)fprintf(drvui->flout,"Cannot open Fourier map (FulProf) file %s\n", infile);
        return;
    }
    ireslt = fread(&icount, sizeof(int), 1, mapin);         /* read record length */
	if (!ireslt) {
          if (!Quick)fprintf(drvui->flout, "Error reading Fourier map file %s\n",infile);
          fclose(mapin);
    }
    if (icount != 80) {
        icount = end_flip(icount);
        wrong_end = 1;
    }
    if (icount != 80) {
        if (!Quick)fprintf(drvui->flout,"Map does not have the correct format. Job aborted.\n");
        return;
    }
    fread(titulo, sizeof(char), icount, mapin);    /* read title */
    fread(&icount, sizeof(int), 1, mapin);         /* skip trailer */
    trim_string(titulo,80);
    strcpy(Map_Info.title,titulo);

    fread(&icount, sizeof(int), 1, mapin);         /* read header for next record */
    if (wrong_end) icount = end_flip(icount);
    fread(version, sizeof(char), icount, mapin);
    fread(&icount, sizeof(int), 1, mapin);         /* skip trailer */

    fread(&icount, sizeof(int), 1, mapin);         /* read header for next record */
    if (wrong_end) icount = end_flip(icount);
    fread(&celda, sizeof(char), icount, mapin);
    if (wrong_end) {
        for (i=0; i<3;i++) {
            celda.ang[i] = end_flip_real(celda.ang[i]);
            celda.cell[i] = end_flip_real(celda.cell[i]);
        }
    }
    fread(&icount, sizeof(int), 1, mapin);         /* skip trailer */
    for (i=0;i<3;i++) {
        Map_Info.lat_con[i] = celda.cell[i];
        Map_Info.lat_con[i+3] = celda.ang[3];
    }

    fread(&icount, sizeof(int), 1, mapin);         /* read header for next record */
    if (wrong_end) icount = end_flip(icount);
    fread(&spg, sizeof(char), icount, mapin);
    fread(&icount, sizeof(int), 1, mapin);         /* skip trailer */

    fread(&icount, sizeof(int), 1, mapin);         /* read header for next record */
    if (wrong_end) icount = end_flip(icount);
    fread(&natoms, sizeof(char), icount, mapin);
    fread(&icount, sizeof(int), 1, mapin);         /* skip trailer */

    fread(&icount, sizeof(int), 1, mapin);         /* read header for next record */
    if (wrong_end) icount = end_flip(icount);
    fread(&atm, sizeof(char), icount, mapin);
    fread(&icount, sizeof(int), 1, mapin);         /* skip trailer */

	for (i=0; i<2; i++) {                          /* skip next 2 records */
        fread(&icount, sizeof(int), 1, mapin);     /* read header for next record */
        if (wrong_end) icount = end_flip(icount);
        fread(&temp, sizeof(char), icount, mapin);
        fread(&icount, sizeof(int), 1, mapin);     /* skip trailer */
	}

    fread(&icount, sizeof(int), 1, mapin);         /* read header for next record */
    if (wrong_end) icount = end_flip(icount);
    fread(&map_param, sizeof(char), icount, mapin);
    fread(&icount, sizeof(int), 1, mapin);         /* skip trailer */
    if (wrong_end) {
        for (i=0; i<3; i++) {
            map_param.ngrid[i] = end_flip(map_param.ngrid[i]);
        }
    }
    Map_Info.map_int[0] = mapstep_a = map_param.ngrid[0];
    Map_Info.map_int[1] = mapstep_b = map_param.ngrid[1];
    Map_Info.map_int[2] = mapstep_c = map_param.ngrid[2];
    for (i=0; i<2; i++) {
        if (wrong_end) {
            map_param.xlim[i] = end_flip_real(map_param.xlim[i]);
            map_param.ylim[i] = end_flip_real(map_param.ylim[i]);
            map_param.zlim[i] = end_flip_real(map_param.zlim[i]);
        }
        Map_Info.xlim[i] = map_param.xlim[i];
        Map_Info.ylim[i] = map_param.ylim[i];
        Map_Info.zlim[i] = map_param.zlim[i];
    }
    Map_Info.info_valid = 1;

    fread(&icount, sizeof(int), 1, mapin);         /* read header for next record */
    if (wrong_end) icount = end_flip(icount);
    fread(&temp, sizeof(char), icount, mapin);
    fread(&icount, sizeof(int), 1, mapin);         /* skip trailer */

	if ((fabs(Map_Info.xlim[0]) > 0.001) || (fabs(Map_Info.xlim[1] - 1.0f) > 0.001)) {
/* not full cell along x */
		new_mapstep_a = (int)(1.0f/(Map_Info.xlim[1] - Map_Info.xlim[0])+0.5f) * mapstep_a;
		astart = (int)(Map_Info.xlim[0] * new_mapstep_a);
		aend = (int)(Map_Info.xlim[1] * new_mapstep_a);
    } else {
		new_mapstep_a = mapstep_a;
		astart = 0;
		aend = mapstep_a;
	}
	if ((fabs(Map_Info.ylim[0]) > 0.001) || (fabs(Map_Info.ylim[1] - 1.0f) > 0.001)) {
/* not full cell along y */
		new_mapstep_b = (int)(1.0f/(Map_Info.ylim[1] - Map_Info.ylim[0])+0.5f) * mapstep_b;
		bstart = (int)(Map_Info.ylim[0] * new_mapstep_b);
		bend = (int)(Map_Info.ylim[1] * new_mapstep_b);
    } else {
		new_mapstep_b = mapstep_b;
		bstart = 0;
		bend = mapstep_b;
	}
	if ((fabs(Map_Info.zlim[0]) > 0.001) || (fabs(Map_Info.zlim[1] - 1.0f) > 0.001)) {
/* not full cell along z */
		new_mapstep_c = (int)(1.0f/(Map_Info.zlim[1] - Map_Info.zlim[0])+0.5f) * mapstep_c;
		cstart = (int)(Map_Info.zlim[0] * new_mapstep_c);
		cend = (int)(Map_Info.zlim[1] * new_mapstep_c);
    } else {
		new_mapstep_c = mapstep_c;
		cstart = 0;
		cend = mapstep_c;
	}
    if (FourierPt) free(FourierPt);
    FourierPt = (float *) malloc( new_mapstep_a*new_mapstep_b*new_mapstep_c * sizeof(float));
    if (FourierPt == NULL) {
        ireslt = 0;
        if (!Quick) fprintf(drvui->flout,"ERROR -- Unable to allocate space for map.\n");
        return;
    }
    memset(FourierPt,0, new_mapstep_a*new_mapstep_b*new_mapstep_c * sizeof(float));

/* read the Rho values */

     for (i=astart; i< aend; i++) {
        for (j=bstart; j<bend; j++) {
            fread(&icount, sizeof(int), 1, mapin); /* read header for next record */
            if (wrong_end) icount = end_flip(icount);
            fread(&dens, sizeof(char), icount, mapin);
            fread(&icount, sizeof(int), 1, mapin); /* skip trailer */
            for (k=cstart; k<cend; k++) {
				ijk = i * new_mapstep_c * new_mapstep_b + j * new_mapstep_c + k;
                if (wrong_end) dens[k] = end_flip_real(dens[k]);
                FourierPt[ijk] = dens[k];
                if (FourierPt[ijk] < rhomin) rhomin = FourierPt[ijk];
                if (FourierPt[ijk] > rhomax) rhomax = FourierPt[ijk];
            }
        }
    }
    fclose(mapin);
    ReadFourMap = 1;
	Map_Info.rhomn = rhomin;
	Map_Info.rhomx = rhomax;
    mapstep_a = new_mapstep_a;
    mapstep_b = new_mapstep_b;
    mapstep_c = new_mapstep_c;
	Map_Info.map_int[0] = mapstep_a;
	Map_Info.map_int[1] = mapstep_b;
	Map_Info.map_int[2] = mapstep_c;
    print_map_info(Quick);
}

/* ************************************************************** */
/* ************************************************************** */

void read_grd(char * infile, int Quick)
{
      FILE * mapin;
      char line[81];
      char * reslt;
      float rhomin=999999.0f, rhomax=-9999999.0f;
      int ireslt = 0;
      int new_mapstep_a = 0, astart, aend;
	  int new_mapstep_b = 0, bstart, bend;
	  int new_mapstep_c = 0, cstart, cend;

      if ((mapin = fopen (infile, "r")) == NULL) {
        fprintf (stderr,
           "Cannot open Fourier map (.grd) file %s\n", infile);
      }
      reslt = fgets(line, sizeof(line), mapin);
      if (reslt != NULL) 
        ireslt = fscanf(mapin, "%f %f %f %f %f %f", 
 			&map_a, &map_b, &map_c, &map_alpha, &map_beta, &map_gamma);
        Map_Info.lat_con[0] = map_a;
        Map_Info.lat_con[1] = map_b;
        Map_Info.lat_con[2] = map_c;
        Map_Info.lat_con[3] = map_alpha;
        Map_Info.lat_con[4] = map_beta;
        Map_Info.lat_con[5] = map_gamma;
      if (ireslt > 0)
        ireslt = fscanf(mapin, "%d %d %d", 
 			&mapstep_a, &mapstep_b, &mapstep_c);
      if (ireslt > 0) {
        int i, j , k, ijk;
        if ((fabs(Map_Info.xlim[0]) > 0.001) || (fabs(Map_Info.xlim[1] - 1.0f) > 0.001)) {
/* not full cell along x */
            new_mapstep_a = (int)(1.0f/(Map_Info.xlim[1] - Map_Info.xlim[0])+0.5f) * mapstep_a;
            astart = (int)(Map_Info.xlim[0] * new_mapstep_a);
            aend = (int)(Map_Info.xlim[1] * new_mapstep_a);
        } else {
            new_mapstep_a = mapstep_a;
            astart = 0;
            aend = mapstep_a;
        }
        if ((fabs(Map_Info.ylim[0]) > 0.001) || (fabs(Map_Info.ylim[1] - 1.0f) > 0.001)) {
/* not full cell along y */
            new_mapstep_b = (int)(1.0f/(Map_Info.ylim[1] - Map_Info.ylim[0])+0.5f) * mapstep_b;
            bstart = (int)(Map_Info.ylim[0] * new_mapstep_b);
            bend = (int)(Map_Info.ylim[1] * new_mapstep_b);
        } else {
            new_mapstep_b = mapstep_b;
            bstart = 0;
            bend = mapstep_b;
        }
        if ((fabs(Map_Info.zlim[0]) > 0.001) || (fabs(Map_Info.zlim[1] - 1.0f) > 0.001)) {
/* not full cell along z */
            new_mapstep_c = (int)(1.0f/(Map_Info.zlim[1] - Map_Info.zlim[0])+0.5f) * mapstep_c;
            cstart = (int)(Map_Info.zlim[0] * new_mapstep_c);
            cend = (int)(Map_Info.zlim[1] * new_mapstep_c);
        } else {
            new_mapstep_c = mapstep_c;
            cstart = 0;
            cend = mapstep_c;
        }
        FourierPt = malloc( mapstep_a*mapstep_b*mapstep_c * sizeof(float));
        if (FourierPt == NULL) {
          ireslt = 0;
          fprintf(stderr, "ERROR -- Unable to allocate space for map.\n");
        }
        memset(FourierPt,0,new_mapstep_a*new_mapstep_b*new_mapstep_c*sizeof(float));

/* read the Rho values */
        for(i=astart; i < aend; i++) {
          for(j=bstart; j < bend; j++) {
            for(k=cstart; k < cend; k++) {
              ijk =  i*new_mapstep_b*new_mapstep_c + j*new_mapstep_c + k;
              ireslt = fscanf(mapin,"%f", &FourierPt[ijk]);
              if (FourierPt[ijk] < rhomin) rhomin = FourierPt[ijk];
              if (FourierPt[ijk] > rhomax) rhomax = FourierPt[ijk];
            }
          }
        }
      }
      Map_Info.rhomn = rhomin;
      Map_Info.rhomx = rhomax;
      Map_Info.map_int[0] = mapstep_a = new_mapstep_a;
      Map_Info.map_int[1] = mapstep_b = new_mapstep_b;
      Map_Info.map_int[2] = mapstep_c = new_mapstep_c;
       
      if (ireslt > 0 && !Quick) {
        print_map_info(Quick);
        ReadFourMap = 1;
      } else {
        fprintf(stderr, "Error reading Fourier map file %s\n\tTitle= %s\n", 
 	       infile, line);
      }
      fclose(mapin);
}

/* ************************************************************** */
/* ************************************************************** */

void read_inp (int Quick)

/* routine to read and process input file */
{
  char input[100];				 /* input buffer for data read from file */
  char input1[100];              /* temporary buffer used for printing */
  char string[100];				 /* temporary string */
  char filename[100];            /* used for map file name */
  int intype;					 /* type of datum read */
  int i, j, k;					 /* temporary */
  char t_name[4];				 /* temporary atom name storage */
  char t_color[40];				 /* temporary ellipsoid color storage */
  int in_line;					 /* flag for inline import of foreign files */
  float t_rad;
  char el_color_tmp[MAX_EL][100];
  int n_el_tmp = 0;
  int done;

  float slab_con[6],slab_off[3],slab_rot[3],rot_mat[3][3];
  float csal,snal,csbe,snbe,csga,snga;
  float vert[3];
  float rad;
  float bmat[3][3];
  int tmp_frame_no = 1;
  int MapRegionRead = 0;
  char atom_pos[3][30];

/* ************************************************************** */
/* ************************************************************** */
/*
    Commands Implemented

    arrow
    atom
    axislines
    betaij
    bij,Bij
    bond
    bounds
    box
    cell
    clip
    cutout
    dash
    depthcue
    edges
    ellipcolor
    ellipsoids
    finish
    frame
    import
    inline
    labelscale
    labeltext
    list
    lonepair
    lookat
    mag_trans
    magnification
    mapcalclimits
    mapcontour
    mapread
    mapregion
    molcomp
    nolabels
    noshadow
    origin
    orthographic
    pack
    phong
    plane
    polyedge
    polysz
    polytolerance
    polylimit
    polyfudge
    polyvert
    rem, REM
    sgrp, spgp, spgr
    shell
    slab
    sphere
    special
    title, titl
    uij,Uij
    vectors    
    view
    vrml1
    vrml2
    vrml97
    xyzoff
    end, END
                                                                  */
/* ************************************************************** */
/* ************************************************************** */

  Map_Info.xlim[0] = 0.0f;
  Map_Info.xlim[1] = 1.0f;
  Map_Info.ylim[0] = 0.0f;
  Map_Info.ylim[1] = 1.0f;
  Map_Info.zlim[0] = 0.0f;
  Map_Info.zlim[1] = 1.0f;
  drvui->origin1_flag = 0;

  for (i = 0; i < 3; i++)
    drvui->xyzoff[i] = 0.0f;
  if (drvui->frame_no > 1) {
      for (i=0; i<6; i++) {
          drvui->cryst_lim[drvui->frame_no][i] = drvui->cryst_lim[drvui->frame_no-1][i];
          drvui->clip_lim[drvui->frame_no][i] = drvui->clip_lim[drvui->frame_no-1][i];
      }
  }
  while (!feof (drvui->fpin)) {
    memset(input,0,100);
    memset(input1,0,100);
    memset(string,0,100);
    memset(t_color,0,40);
    if (fgets (input, 100, drvui->fpin)) {		 /* read a line */
      trim_string(input, 100);
      Blank_Strip (input);			 /* suppress any input blanks */
    } else {
      fprintf (stderr, "Error reading file.");
      delay (5);
      exit (1);
    }
    intype = 0;
    strcpy(input1,input);
    if (strncmp (input, "title", 5) == 0) {
      if(1 == drvui->frame_no)
          intype = 1;
        else
          intype = 0;                       /* ignore title command if not frame 1 */
    }
    if (strncmp (input, "titl", 4) == 0) {
      if(1 == drvui->frame_no)
          intype = 1;
      else
          intype = 0;                       /* ignore title command if not frame 1 */
    }
    if (strncmp (input, "TITL", 4) == 0) {
      if(1 == drvui->frame_no)
          intype = 1;
      else
          intype = 0;                       /* ignore title command if not frame 1 */
    }
    if (strncmp (input, "rem", 3) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 1;
      else
          intype = 0;                       /* ignore rem command if not current frame */
    }
    if (strncmp (input, "REM", 3) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 1;
      else
          intype = 0;                       /* ignore rem command if not current frame */
    }
    if (strncmp (input, "cell", 4) == 0)
      intype = 2;
    if (strncmp (input, "sgrp", 4) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 3;
      else
          intype = 0;                       /* ignore sgrp command if not current frame */
    }
    if (strncmp (input, "spgp", 4) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 3;
      else
          intype = 0;                       /* ignore spgp command if not current frame */
    }
    if (strncmp (input, "spgr", 4) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 3;
      else
          intype = 0;                       /* ignore spgr command if not current frame */
    }
    if (strncmp (input, "ellipsoids", 10) == 0)
      intype = 4;
    if (strncmp (input, "lookat", 6) == 0)
      intype = 5;
    if (strncmp (input, "atom", 4) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 6;
      else
          intype = 0;                     /* ignore atom command if not current frame */
    }
    if (strncmp (input, "polysz", 6) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 7;
      else
          intype = 0;                       /* ignore polysz command if not current frame */
    }
    if (strncmp (input, "sphere", 6) == 0) {
      if(tmp_frame_no == drvui->frame_no) 
          intype = 8;
      else
          intype = 0;                       /* ignore sphere command if not current frame */
    }
    if (strncmp (input, "end", 3) == 0)
      intype = 9;
    if (strncmp (input, "END", 3) == 0)
      intype = 9;
    if (strncmp (input, "xyzoff", 6) == 0)
      intype = 10;
    if (strncmp (input, "bond", 4) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 11;
      else
          intype = 0;                       /* ignore bond command if not current frame */
    }
    if (strncmp (input, "box", 3) == 0)
      intype = 12;
    if (strncmp (input, "edges", 5) == 0)
      intype = 13;
    if (strncmp (input, "view", 4) == 0)
      intype = 14;
    if (strncmp (input, "plane", 5) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 15;
      else
          intype = 0;                       /* ignore plane command if not current frame */
    }
    if (strncmp (input, "bij", 3) == 0)
      intype = 16;
    if (strncmp (input, "Bij", 3) == 0)
      intype = 16;
    if (strncmp (input, "betaij", 6) == 0)
      intype = 17;
    if (strncmp (input, "uij", 3) == 0)
      intype = 18;
    if (strncmp (input, "Uij", 3) == 0)
      intype = 18;
    if (strncmp (input, "bounds", 6) == 0)
      intype = 19;
    if (strncmp (input, "pack", 4) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 20;
      else
          intype = 0;                       /* ignore pack command if not current frame */
    }
    if (strncmp (input, "magnification", 13) == 0)
      intype = 21;
    if (strncmp (input, "origin", 6) == 0)
      intype = 22;
    if (strncmp (input, "ellipcolor", 10) == 0)
      intype = 23;
    if (strncmp (input, "import", 6) == 0)
      intype = 24;
    if (strncmp (input, "inline", 6) == 0)
      intype = 25;
    if (strncmp (input, "orthographic", 12) == 0)
      intype = 26;
    if (strncmp (input, "vectors", 7) == 0)
      intype = 27;
    if (strncmp (input, "phong", 5) == 0)
      intype = 28;
    if (strncmp (input, "molcomp", 7) == 0)
      intype = 29;
    if (strncmp (input, "cutout", 6) == 0)
      intype = 30;
    if (strncmp (input, "depthcue", 8) == 0)
      intype = 31;
    if (strncmp (input, "vrml1", 5) == 0)
      intype = 32;
    if (strncmp (input, "vrml2", 5) == 0)
      intype = 1;
    if (strncmp (input, "vrml97", 6) == 0)
      intype = 1;
    if (strncmp (input, "polytol", 7) == 0)
      intype = 33;
    if (strncmp (input, "polylimit", 9) == 0)
      intype = 33;
    if (strncmp (input, "polyfudge", 9) == 0)
      intype = 33;
    if (strncmp (input, "clip", 4) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 34;
      else
          intype = 0;                       /* ignore clip command if not current frame */
    }
    if (strncmp (input, "list", 4) == 0)
      intype = 35;
    if (strncmp (input, "polyedge", 8) == 0)
      intype = 36;
    if (strncmp (input, "frame", 5) == 0)
      intype = 37;
    if (strncmp (input, "axislines", 9) == 0)
      intype = 38;
    if (strncmp (input, "shell", 5) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 39;
      else
          intype = 0;                       /* ignore shell command if not current frame */
    }
    if (strncmp (input, "nolabels", 8) == 0)
      intype = 40;
    if (strncmp (input, "lonepair", 8) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 41;
      else
          intype = 0;                       /* ignore lonepair command if not current frame */
    }
    if (strncmp (input, "labeltext", 9) == 0)
      intype = 42;
    if (strncmp (input, "polyvert", 8) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 43;
      else
          intype = 0;                       /* ignore polyvert command if not current frame */
    }
    if (strncmp (input, "background", 10) == 0)
      intype = 44;
    if (strncmp (input, "slab", 4) == 0)
      intype = 45;
    if (strncmp (input, "mag_trans", 9) == 0)
      intype = 46;
    if (strncmp (input, "special", 7) == 0)
      intype = 47;
    if (strncmp (input, "noshadow", 8) == 0)
      intype = 48;
    if (strncmp (input, "finish", 6) == 0)
      intype = 49;
    if (strncmp (input, "arrow", 5) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 50;
        else
          intype = 0;                     /* ignore arrow command if not current frame */
    }
    if (strncmp (input, "dash", 4) == 0) {
      if(tmp_frame_no == drvui->frame_no)
          intype = 51;
      else
          intype = 0;                     /* ignore dash command if not current frame */
    }
    if (strncmp (input, "mapcontour", 10) == 0)
      intype = 52;
    if (strncmp (input, "mapread", 7) == 0)
      intype = 53;
    if (strncmp (input, "mapregion", 9) == 0)
      intype = 54;
    if (strncmp (input, "mapcalclimits", 13) == 0)
      intype = 55;
    if (strncmp (input, "labelscale", 10) == 0)
      intype = 56;

    switch (intype) {
    case 0:
      break;
    case 1:
      break;					 /* ignore title except for listing */
    case 2:                      /* cell command */
      for (i = 3; i <= 5; ++i)
          drvui->lat_con[i] = 0.0f;
      (void) sscanf (input, "%s %f %f %f %f %f %f", string, &drvui->lat_con[0]
		     , &drvui->lat_con[1], &drvui->lat_con[2], &drvui->lat_con[3], &drvui->lat_con[4]
		     , &drvui->lat_con[5]);
      break;
    case 3:
      symop (input);				 /* process space group information */
      break;
    case 4:					 /* ellipsoids command */
      do_ellipsoids = 1;
      (void) sscanf (input, "%s %f", string, &Ellipsoid_Prob);
      if (Ellipsoid_Prob > 1)
        Ellipsoid_Prob *= 0.01f;		 /* Probability to fraction */
      break;
    case 5:					 /* read vectors for lookat command */
      (void) sscanf (input, "%s %f %f %f %f %f %f", string, &drvui->lookat_v1[0]
		     , &drvui->lookat_v1[1], &drvui->lookat_v1[2], &drvui->lookat_v2[0],
             &drvui->lookat_v2[1], &drvui->lookat_v2[2]);
      Options = Options | L_OPT;
      break;
    case 6:					 /* atom command */
      get_label (input, &drvui->atom_l[natom][0], &drvui->atom_l[natom][1],
		     &drvui->atom_l[natom][2], &drvui->atom_l[natom][3], 1);
      (void)sscanf (input, "%d %s %s %s",&drvui->atom_n[natom],atom_pos[0],atom_pos[1],atom_pos[2]);
      for (i=0; i<3; i++)
          drvui->atom_xyz[natom][i] = convert_pos(atom_pos[i]);
      natom++;
      if (natom >= MAX_ATOM) {
        fprintf (stderr,
		    "Too many atoms in asymmetric unit - increase MAX_ATOM.\n");
        delay (5);
        exit (0);
      }
      break;
    case 7:					 /* polysz command */
    case 39:				 /* shell command */
    case 43:				 /* polyvert command */
      get_label (input, &drvui->poly_l[drvui->npoly][0], &drvui->poly_l[drvui->npoly][1],
            &drvui->poly_l[drvui->npoly][2], &drvui->poly_l[drvui->npoly][3], 1);
      if (intype == 7) {
          (void) sscanf (input, "%f %39c", &drvui->poly_size[drvui->npoly], t_color);
      } else if (intype == 39) {
          (void) sscanf (input, "%f %f %39c",
	             &drvui->poly_min[drvui->npoly], &drvui->poly_size[drvui->npoly], t_color);
      } else if (intype == 43) {
          get_label (input, &drvui->poly_t[drvui->npoly][0], &drvui->poly_t[drvui->npoly][1],
                 &drvui->poly_t[drvui->npoly][2], &drvui->poly_t[drvui->npoly][3], 0);
          (void) sscanf (input, "%f %39c", &drvui->poly_size[drvui->npoly], t_color);
      }
      trim_string(t_color, 40);
      strcpy(drvui->poly_col_p[drvui->npoly],t_color);
      strcpy (drvui->poly_col_v[drvui->npoly], drvui->poly_col_p[drvui->npoly]);
      check_color (drvui->poly_col_v[drvui->npoly]);
      if (intype == 7 || intype == 43)
          drvui->poly_min[drvui->npoly] = 0.0f;
      drvui->poly_fn[drvui->npoly] = tmp_frame_no;
      if (intype == 7 || intype == 39)
          strcpy(drvui->poly_t[drvui->npoly],"");
      drvui->npoly++;
      if (drvui->npoly >= MAX_POLY) {
          fprintf (stderr,"Too many polyhedra specified - increase MAX_POLY.\n");
          delay (5);
          exit (0);
      }
      break;
    case 8:					 /* sphere command */
    { char str[5][40];
      memset(str,0,200*sizeof(char));
      get_label (input, &drvui->sphere_l[drvui->nsphere][0], &drvui->sphere_l[drvui->nsphere][1]
         , &drvui->sphere_l[drvui->nsphere][2], &drvui->sphere_l[drvui->nsphere][3], 1);
      i = sscanf(input,"%s %s %s %s %s", str[0], str[1], str[2], str[3], str[4]);
      if (i==2 || (i == 4 && strstr(str[2],"filter"))) {       // old-style input
          (void)sscanf(str[0],"%f",&drvui->sphere_size[drvui->nsphere]);
          drvui->sphere_n[drvui->nsphere] = -1;
          strcpy(string,str[1]);
          if (i >= 3) {
              strcat(string," ");
              strcat(string,str[2]);
          }
      } else if (i>=3) {                   // new-style input with atom number
          (void)sscanf(str[0],"%d",&drvui->sphere_n[drvui->nsphere]);
          (void)sscanf(str[1],"%f",&drvui->sphere_size[drvui->nsphere]);
          strcpy(string,str[2]);
      } else {
          Error_Box("Sphere input line malformed.");
      }
      if (i>=4) {
          strcat(string," ");
          strcat(string,str[3]);
      }
      if (i==5) {
          strcat(string," ");
          strcat(string,str[4]);
      }
      trim_string(string,40);
      strcpy(drvui->sphere_col_p[drvui->nsphere],string);
      strcpy (drvui->sphere_col_v[drvui->nsphere], drvui->sphere_col_p[drvui->nsphere]);
      check_color (drvui->sphere_col_v[drvui->nsphere]);
      drvui->sphere_fn[drvui->nsphere] = drvui->frame_no;
      drvui->nsphere++;
      if (drvui->nsphere >= MAX_SPHERE) {
          Error_Box("Too many sphere types specified - increase MAX_SPHERE.");
          return;
      }
      break; }
    case 9:					 /* end or END */
      if (!drvui->lat_con[0]) {
        fprintf(stderr,"There is no 'cell' line in your input file.\n");
        exit(0);
      }
      for (i = 0; i < natom; i++) {		 /* shift atom coordinates for origin offset */
          for (j = 0; j < 3; j++) {
            drvui->atom_xyz[i][j] -= drvui->xyzoff[j];
          }
      }
      drvui->max_frame = tmp_frame_no;
      if (!Quick)
          fprintf(drvui->flout,"End of Frame %d\n",drvui->frame_no);
      if (!Quick && drvui->frame_no == drvui->max_frame)
          fprintf (drvui->flout, "Final Frame done\n");
      if (n_el_tmp > 0) {
          for (i = 0; i < n_el_tmp; i++) {
            get_label (el_color_tmp[i], &t_name[0], &t_name[1],
		       &t_name[2], &t_name[3], 0);
            if (el_color_tmp[i][0] == '*') {
              (void) sscanf (el_color_tmp[i], "%s %s", string, t_color);
              j = -1;
            } else {
              (void) sscanf (el_color_tmp[i], "%d %s", &j, t_color);
            }
            for (k = 0; k < n_ellips; k++) {	 /* find this atom in ellipsoid list */
              if ((t_name[0] == drvui->ellips_l[k][0]) &&
                  (t_name[1] == drvui->ellips_l[k][1]) &&
                  (t_name[2] == drvui->ellips_l[k][2]) &&
                  (t_name[3] == drvui->ellips_l[k][3])) {
                if ((j == -1) || (j == drvui->ellips_n[k])) {
                  strcpy (drvui->ellips_col_p[k], t_color);	/* copy color */
                  strcpy (drvui->ellips_col_v[k], drvui->ellips_col_p[k]);
                  check_color (drvui->ellips_col_v[k]);
                }
              }
            }
          }
      }
      for (i=1; i<drvui->nsphere; i++) {        // remove any auto-generated ellipsoids that
          for (j=1; j<n_ellips; j++) {          //     interfere with spheres
              if (check_atom_name(drvui->sphere_l[i],drvui->ellips_l[j])) {
                  if (drvui->sphere_n[i] == -1
                      || drvui->sphere_n[i] == drvui->save_el_number[j]) {
                      drvui->ell_type[j] = -100;
                  }
              }
          }
      }
      switch (FourierMapType) {
      case 1:
          read_grd( FourierFileName, Quick);
          break;
      case 2:
          read_stf( FourierFileName, Quick);
          break;
      case 3:
          read_w2k( FourierFileName, Quick);
          break;
      case 4:
          read_vasp( FourierFileName, Quick);
          break;
      case 5:
          read_flp( FourierFileName, Quick);
          break;
      case 6:
          read_fcf( FourierFileName, Quick);
          break;
      default:
          break;
      }
      if (!MapRegionRead) {             /* mapregion line NOT read */
          xMin = Map_Info.xlim[0];
          xMax = Map_Info.xlim[1];
          yMin = Map_Info.ylim[0];
          yMax = Map_Info.ylim[1];
          zMin = Map_Info.zlim[0];
          zMax = Map_Info.zlim[1];
      }
      return;					 /* thru with end */
    case 10:					 /* xyzoff command */
      (void)sscanf (input, "%s %s %s %s",string,atom_pos[0],atom_pos[1],atom_pos[2]);
      for (i=0; i<3; i++)
          drvui->xyzoff[i] = convert_pos(atom_pos[i]);
      break;
    case 51:                     /* dash command */
    case 11:					 /* bond command */
      get_label (input, &drvui->bond_l1[drvui->nbond][0], &drvui->bond_l1[drvui->nbond][1],
		   &drvui->bond_l1[drvui->nbond][2], &drvui->bond_l1[drvui->nbond][3], 1);
      get_label (input, &drvui->bond_l2[drvui->nbond][0], &drvui->bond_l2[drvui->nbond][1],
		   &drvui->bond_l2[drvui->nbond][2], &drvui->bond_l2[drvui->nbond][3], 0);
      (void) sscanf (input, "%f %f %f %39c",
		     &drvui->bond_size[drvui->nbond], &drvui->bond_min[drvui->nbond],
             &drvui->bond_max[drvui->nbond],t_color);
      trim_string(t_color, 40);
      strcpy(drvui->col_bond_p[drvui->nbond],t_color);
      if (strlen (drvui->col_bond_p[drvui->nbond]) == 0)
          strcpy (drvui->col_bond_p[drvui->nbond], "Grey");
      strcpy (drvui->col_bond_v[drvui->nbond], drvui->col_bond_p[drvui->nbond]);
      check_color (drvui->col_bond_v[drvui->nbond]);
      drvui->bond_fn[drvui->nbond] = tmp_frame_no;
      if (intype == 51)
          drvui->bond_style[drvui->nbond] = 1;
      else if (intype == 11)
          drvui->bond_style[drvui->nbond] = 0;
      drvui->nbond++;
      if (drvui->nbond >= MAX_BOND) {
          fprintf (stderr,"Too many bond types specified - increase MAX_BOND.\n");
          delay (5);
          exit (0);
      }
      break;
    case 12:					 /* box command */
      (void) sscanf (input, "%s %f %39c", string, &rad_cell, t_color);
      trim_string(t_color, 40);
      strcpy(col_cell_p,t_color);
      if (strlen (col_cell_p) == 0)
          strcpy (col_cell_p, "Black");
      strcpy (col_cell_v, col_cell_p);
      check_color (col_cell_v);
      if (rad_cell == 0.0f) {
          docell = 0;
          Labels = 0;
      }
      break;
    case 13:					 /* edges command */
      memset(col_edge_p,0,40);
      (void) sscanf (input, "%s %f %39c", string, &rad_edge, col_edge_p);
      trim_string(col_edge_p, 40);
      edges = 1;
      if (rad_edge < 0.01f)
          rad_edge = 0.01f;
      if (strlen (col_edge_p) == 0)
          strcpy (col_edge_p, "Black");
      strcpy (col_edge_v, col_edge_p);
      check_color (col_edge_v);
      break;
    case 14:					 /* view command */
      if ((Options & V_OPT) == 0)		 /* only if no -v switch */
          (void) sscanf (input, "%s %f %f %f", string, &xrot, &yrot, &zrot);
      xrot_i = xrot;
      yrot_i = yrot;
      zrot_i = zrot;
      break;
    case 15:					 /* plane command */
      get_label (input, &drvui->plane_l[drvui->nplane][0], &drvui->plane_l[drvui->nplane][1]
		 , &drvui->plane_l[drvui->nplane][2], &drvui->plane_l[drvui->nplane][3], 1);
      (void) sscanf (input, "%f %39c", &drvui->plane_size[drvui->nplane],t_color);
      trim_string(t_color, 40);
      strcpy(drvui->plane_col_p[drvui->nplane],t_color);
      strcpy (drvui->plane_col_v[drvui->nplane], drvui->plane_col_p[drvui->nplane]);
      check_color (drvui->plane_col_v[drvui->nplane]);
      drvui->plane_fn[drvui->nplane] = drvui->frame_no;
      drvui->nplane++;
      if (drvui->nplane >= MAX_PLANE) {
         fprintf (stderr,"Too many plane types specified - increase MAX_PLANE.\n");
         delay (5);
         exit (0);
      }
      break;
    case 16:					 /* bij or Bij command */
      drvui->ell_type[n_ellips] = -1;
    case 17:					 /* betaij command */
      if (intype == 17)
          drvui->ell_type[n_ellips] = 0;
    case 18:					 /* uij or Uij command */
      if (intype == 18)
          drvui->ell_type[n_ellips] = 1;
      get_label (input, &drvui->ellips_l[n_ellips][0], &drvui->ellips_l[n_ellips][1]
		 , &drvui->ellips_l[n_ellips][2], &drvui->ellips_l[n_ellips][3], 1);
      (void) sscanf (input, "%d %f %f %f %f %f %f %39c", &drvui->ellips_n[n_ellips], 
          &drvui->ellips[n_ellips][0], &drvui->ellips[n_ellips][1], &drvui->ellips[n_ellips][2], 
          &drvui->ellips[n_ellips][3], &drvui->ellips[n_ellips][4], &drvui->ellips[n_ellips][5], t_color);
      trim_string(t_color, 40);
      strcpy(drvui->ellips_col_p[n_ellips],t_color);	/* read unique coefficients and color */
      if (!strlen (drvui->ellips_col_p[n_ellips]))
          strcpy (drvui->ellips_col_p[n_ellips], "Gray80");
      strcpy (drvui->ellips_col_v[n_ellips], drvui->ellips_col_p[n_ellips]);
      check_color (drvui->ellips_col_v[n_ellips]);
      n_ellips++;
      if (n_ellips >= MAX_EL) {
          fprintf (stderr,"Too many ellipsoid types specified - increase MAX_EL.\n");
          delay (5);
          exit (0);
      }
      break;
    case 19:					 /* bounds command */
      if ((Options & B_OPT) == 0) {		 /* only do this if no -b switch */
          (void) sscanf (input, "%s %f %f %f", string, &boxlim[0], &boxlim[1],
		       &boxlim[2]);
          boxflag = 1;
      }
      break;
    case 20:					 /* pack command */
      if ((Options & P_OPT) == 0) {		 /* only if no -p switch */
          (void) sscanf (input, "%s %f %f %f %f %f %f", string,
              &drvui->cryst_lim[drvui->frame_no][0], &drvui->cryst_lim[drvui->frame_no][3],
              &drvui->cryst_lim[drvui->frame_no][1], &drvui->cryst_lim[drvui->frame_no][4],
              &drvui->cryst_lim[drvui->frame_no][2], &drvui->cryst_lim[drvui->frame_no][5]);
          if (!clipflag) {
              for (i=0; i<6; i++) {
                  drvui->clip_lim[drvui->frame_no][i] = drvui->cryst_lim[drvui->frame_no][i];
              }
          }
          packflag = 1;
      }
      break;
    case 21:					 /* magnification command */
      if ((Options & M_OPT) == 0)		 /* only if no -m switch */
          (void) sscanf (input, "%s %f", string, &Magnification);
      break;
    case 22:					 /* origin command */
      if ((Options & O_OPT) == 0) {		 /* only if no -o switch */
          (void) sscanf (input, "%s %f %f %f", string, &origin[0], &origin[1],
		       &origin[2]);
      }
      break;
    case 23:					 /* ellipcolor command */
      Token_Strip (input, 1);
      trim_string(input, 100);
      strcpy (el_color_tmp[n_el_tmp++], input);
      break;
    case 24:					 /* inline command */
    case 25:					 /* import command */
      in_line = 0;
      done = 0;
      if (strncmp (input, "inline", 6) == 0)
          in_line = 1;
      Token_Strip (input, 1);			 /* strip command */
      if (strncmp (input, "gsas", 4) == 0) {
          import_gsas (input, in_line, Quick);
          done = 1;
      }
      if (strncmp (input, "cif", 3) == 0) {
          import_cif (input, in_line, Quick);
          done = 1;
      }
      if (strncmp (input, "schakal", 7) == 0) {
          import_schakal (input, in_line, Quick);
          done = 1;
      }
      if (strncmp (input, "shakal", 6) == 0) {
          import_schakal (input, in_line, Quick);
          done = 1;
      }
      if (strncmp (input, "shelx", 5) == 0) {
          import_shelx (input, in_line, Quick);
          done = 1;
      }
      if (strncmp (input, "fdat", 4) == 0) {
          import_fdat (input, in_line, Quick);
          done = 1;
      }
      if (strncmp (input, "csd", 3) == 0) {
          import_fdat (input, in_line, Quick);
          done = 1;
      }
      if (strncmp (input, "ccdf", 4) == 0) {
          import_fdat (input, in_line, Quick);
          done = 1;
      }
      if (done == 0) {
          fprintf (stderr,
              "Import/inline source type '%s' not recognized.\n", input);
      }
      break;
    case 26:					 /* orthographic */
      M_cameras = 0;
#ifdef OPENGL
      gl_ortho=1;
#endif
      break;
    case 27: {                   /* unit cell direction vectors */
      float v[3];
      Display_axes = 1;
      offset[0] = offset[1] = offset[2] = 0.0f;
      i = sscanf( input, "%s %f %f %f", string, &v[0], &v[1], &v[2]);
      if (i == 4)
          for (i=0; i<3; i++)
              offset[i] = v[i];
      }
      break;
    case 28:					 /* phong */
      (void) sscanf (input, "%s %f %f", string, &drvui->Phong_Value, &drvui->Phong_Size);
      break;
    case 29:					 /* molcomp */
      (void) sscanf (input, "%s %f", string, &drvui->mol_d);
      domolcomp = 1;
      break;
    case 30:					 /* cutout */
      El_Cutout = 1;
      memset(drvui->Cutout_color,0,40);
      (void) sscanf (input, "%s %39c", string, drvui->Cutout_color);
      trim_string(drvui->Cutout_color, 40);
      break;
    case 31:					 /* depthcue - thickness of edges scaled by Z coordinate */
      (void) sscanf (input, "%s %f", string, &DepthCue);
      break;
    case 32:					 /* vrml1 - turn on VRML1 output */
      Vrml2 = 0;
      break;
    case 33:					 /* polylimit (formerly polyfudge) */
      (void) sscanf (input, "%s %f", string, &drvui->polylimit);
      if (!Quick)
          fprintf (drvui->flout,"Tolerance factor for non-planarity of polyhedra faces "
              "set to %f. This option should be used with care\n",drvui->polylimit);
      break;
    case 34:					 /* clip command */
      (void) sscanf (input, "%s %f %f %f %f %f %f", string, &drvui->clip_lim[drvui->frame_no][0],
		     &drvui->clip_lim[drvui->frame_no][3], &drvui->clip_lim[drvui->frame_no][1],
             &drvui->clip_lim[drvui->frame_no][4], &drvui->clip_lim[drvui->frame_no][2],
             &drvui->clip_lim[drvui->frame_no][5]);
      clipflag = 1;
      break;
    case 35:					 /* list - limit for printed distance table */
      (void) sscanf (input, "%s %f", string, &printdist);
      if (!Quick)
          fprintf (drvui->flout,"Interatomic distances up to %f Angstroms will be "
               "tabulated in the logfile\n",printdist);
      break;
    case 36:					 /* polyedge - enables individual coloring of polyhedral edges */
      get_label (input, &t_name[0], &t_name[1], &t_name[2], &t_name[3], 1);
      (void) sscanf (input, "%f %39c", &t_rad, t_color);
      trim_string(t_color, 40);
      for (i = 1; i < drvui->npoly; i++) {		 /* find this atom in poly list */
          if ((t_name[0] == drvui->poly_l[i][0]) &&
              (t_name[1] == drvui->poly_l[i][1]) &&
              (t_name[2] == drvui->poly_l[i][2]) && (t_name[3] == drvui->poly_l[i][3])) {
              edges = 1;
              drvui->p_rad_edge[i] = t_rad;
              strcpy (drvui->p_col_edge_p[i], t_color);
              strcpy (drvui->p_col_edge_v[i], drvui->p_col_edge_p[i]);
              check_color (drvui->p_col_edge_v[i]);
          } 
      }
      break;
    case 37:					 /* frame command */
      tmp_frame_no++;
      break;
    case 38:					 /* axislines */
      memset(Ellipaxis_color,0,40);
      i = sscanf (input, "%s %f %39c", string, &Ellipaxis_width,Ellipaxis_color);
      trim_string(Ellipaxis_color, 40);
      if (i < 3)
          strcpy (Ellipaxis_color, "Gray20");
      if (Ellipaxis_color[strlen (Ellipaxis_color) - 1] == '\n')
          Ellipaxis_color[strlen (Ellipaxis_color) - 1] = '\0';
      break;
    case 40:					 /* nolabels command */
      Labels = 0;				 /* set value false */
      break;
    case 41:					 /* lonepair command */
      get_label (input, &drvui->cone_l1[drvui->ncone][0], &drvui->cone_l1[drvui->ncone][1]
		 , &drvui->cone_l1[drvui->ncone][2], &drvui->cone_l1[drvui->ncone][3], 1);
      (void) sscanf (input, "%d %f %f %f %39c",
		     &drvui->numlonepairs[drvui->ncone], &drvui->cone_height[drvui->ncone],
		     &drvui->cone_min[drvui->ncone], &drvui->cone_max[drvui->ncone], t_color);
      trim_string(t_color, 40);
      strcpy(drvui->col_cone_p[drvui->ncone],t_color);
      if (strlen (drvui->col_cone_p[drvui->ncone]) == 0)
          strcpy (drvui->col_cone_p[drvui->ncone], "Grey");
      strcpy (drvui->col_cone_v[drvui->ncone], drvui->col_cone_p[drvui->ncone]);
      drvui->cone_fn[drvui->ncone] = drvui->frame_no;
      check_color (drvui->col_cone_v[drvui->ncone]);
      drvui->ncone++;
      if (drvui->ncone >= MAX_CONE) {
          fprintf (drvui->flout,"Too many lp bond types specified - increase MAX_CONE.\n");
          delay (5);
          exit (0);
      }
      break;
    case 42:					 /* labeltext command */
      memset (drvui->label_label[drvui->nlabel], 0, 64);
      Token_Strip (input, 1);
      (void) sscanf (input, "%f %f %f %63c",
		     &(drvui->label_x[drvui->nlabel][0]),
		     &(drvui->label_x[drvui->nlabel][1]), &(drvui->label_x[drvui->nlabel][2]),
		     drvui->label_label[drvui->nlabel]);
      trim_string(drvui->label_label[drvui->nlabel], 64);
      drvui->label_fn[drvui->nlabel] = tmp_frame_no;
      drvui->nlabel++;
      if (drvui->nlabel >= MAX_LABELS) {
          fprintf (stderr,"Too many labels specified - increase MAX_LABELS.\n");
          delay (5);
          exit (0);
      }
      break;
    case 44:					 /* background (color) command */
      Token_Strip (input, 1);
      trim_string(input, 100);
      strcpy (drvui->col_bg_p, input);
      strcpy (drvui->col_bg_v, input);
      check_color (drvui->col_bg_v);
      break;
    case 45:
      (void)sscanf (input,"%*s %f %f %f %f %f %f %f %f %f %f %f %f %d",
          &slab_con[0],&slab_con[1],&slab_con[2],&slab_con[3],
          &slab_con[4],&slab_con[5],
          &slab_off[0],&slab_off[1],&slab_off[2],
          &slab_rot[0],&slab_rot[1],&slab_rot[2],&slabmode);

      rad = 45.0f / (float) atan (1.0);
      for (i = 3; i <= 5; ++i)
        if (slab_con[i] <= 0)
          slab_con[i] = 90.0f;
      for (i = 0; i <= 2; ++i)
        for (j = 0; j <= 2; ++j)
          bmat[j][i] = 0.0f;         /* initialize matrix */
      bmat[0][0] = 1.0f;
      csal = (float) cos (slab_con[3] / rad);
      snal = (float) sin (slab_con[3] / rad);
      csbe = (float) cos (slab_con[4] / rad);
      snbe = (float) sin (slab_con[4] / rad);
      csga = (float) cos (slab_con[5] / rad);
      snga = (float) sin (slab_con[5] / rad);
      bmat[0][1] = csga;
      bmat[1][1] = snga;
      bmat[0][2] = csbe;
      bmat[1][2] = (csal - csbe * csga) / snga;
      bmat[2][2] =    (float) sqrt (1.0 - bmat[0][2] * bmat[0][2] -
              bmat[1][2] * bmat[1][2]);
      for (i = 0; i <= 2; ++i)
        for (j = 0; j <= 2; ++j)
          bmat[j][i] *= slab_con[i];

      csal = (float) cos (slab_rot[0] / rad);
      snal = (float) sin (slab_rot[0] / rad);
      csbe = (float) cos (slab_rot[1] / rad);
      snbe = (float) sin (slab_rot[1] / rad);
      csga = (float) cos (slab_rot[2] / rad);
      snga = (float) sin (slab_rot[2] / rad);
      rot_mat[0][0] = csbe * csga;
      rot_mat[0][1] = snal * snbe * csga + csal * snga;
      rot_mat[0][2] = -csal * snbe * csga + snal * snga;
      rot_mat[1][0] = -csbe * snga;
      rot_mat[1][1] = -snal * snbe * snga + csal * csga;
      rot_mat[1][2] = csal * snbe * snga + snal * csga;
      rot_mat[2][0] = snbe;
      rot_mat[2][1] = -snal * csbe;
      rot_mat[2][2] = csal * csbe;

/* 0 0 0 */
      for (i = 0; i <= 2; ++i)
        vert[i] = 0.0f;
      slabx1=0.;
      slaby1=0.;
      slabz1=0.;

/* 0 0 1 */
      vert[2] = 1.0f;
      slabx2=0.;
      for (i = 0; i <= 2; ++i)
        slabx2 += (float) (bmat[0][i] * vert[i]);
      slaby2=0.;
      for (i = 0; i <= 2; ++i)
        slaby2 += (float) (bmat[1][i] * vert[i]);
      slabz2=0.;
      for (i = 0; i <= 2; ++i)
        slabz2 += (float) (bmat[2][i] * vert[i]);
 
/* 0 1 0 */
      vert[1] = 1.0f;
      vert[2] = 0.0f;
      slabx3=0.;
      for (i = 0; i <= 2; ++i)
        slabx3 += (float) (bmat[0][i] * vert[i]);
      slaby3=0.;
      for (i = 0; i <= 2; ++i)
        slaby3 += (float) (bmat[1][i] * vert[i]);
      slabz3=0.;
      for (i = 0; i <= 2; ++i)
        slabz3 += (float) (bmat[2][i] * vert[i]);

/* 1 0 0 */
      vert[0] = 1.0f;
      vert[1] = 0.0f;
      slabx4=0.;
      for (i = 0; i <= 2; ++i)
        slabx4 += (float) (bmat[0][i] * vert[i]);
      slaby4=0.;
      for (i = 0; i <= 2; ++i)
        slaby4 += (float) (bmat[1][i] * vert[i]);
      slabz4=0.;
      for (i = 0; i <= 2; ++i)
        slabz4 += (float) (bmat[2][i] * vert[i]);


      slabx1 -= slab_off[0];
      slabx2 -= slab_off[0];
      slabx3 -= slab_off[0];
      slabx4 -= slab_off[0];
      slaby1 -= slab_off[1];
      slaby2 -= slab_off[1];
      slaby3 -= slab_off[1];
      slaby4 -= slab_off[1];

      slabz1 -= slab_off[2];
      slabz2 -= slab_off[2];
      slabz3 -= slab_off[2];
      slabz4 -= slab_off[2];

      vert[0]=slabx1 * rot_mat[0][0]+slaby1 * rot_mat[1][0]+slabz1 * rot_mat[2][0];
      vert[1]=slabx1 * rot_mat[0][1]+slaby1 * rot_mat[1][1]+slabz1 * rot_mat[2][1];
      vert[2]=slabx1 * rot_mat[0][2]+slaby1 * rot_mat[1][2]+slabz1 * rot_mat[2][2];
      slabx1=vert[0];
      slaby1=vert[1];
      slabz1=vert[2];
      vert[0]=slabx2 * rot_mat[0][0]+slaby2 * rot_mat[1][0]+slabz2 * rot_mat[2][0];
      vert[1]=slabx2 * rot_mat[0][1]+slaby2 * rot_mat[1][1]+slabz2 * rot_mat[2][1];
      vert[2]=slabx2 * rot_mat[0][2]+slaby2 * rot_mat[1][2]+slabz2 * rot_mat[2][2];
      slabx2=vert[0];
      slaby2=vert[1];
      slabz2=vert[2];
      vert[0]=slabx3 * rot_mat[0][0]+slaby3 * rot_mat[1][0]+slabz3 * rot_mat[2][0];
      vert[1]=slabx3 * rot_mat[0][1]+slaby3 * rot_mat[1][1]+slabz3 * rot_mat[2][1];
      vert[2]=slabx3 * rot_mat[0][2]+slaby3 * rot_mat[1][2]+slabz3 * rot_mat[2][2];
      slabx3=vert[0];
      slaby3=vert[1];
      slabz3=vert[2];
      vert[0]=slabx4 * rot_mat[0][0]+slaby4 * rot_mat[1][0]+slabz4 * rot_mat[2][0];
      vert[1]=slabx4 * rot_mat[0][1]+slaby4 * rot_mat[1][1]+slabz4 * rot_mat[2][1];
      vert[2]=slabx4 * rot_mat[0][2]+slaby4 * rot_mat[1][2]+slabz4 * rot_mat[2][2];
      slabx4=vert[0];
      slaby4=vert[1];
      slabz4=vert[2];
      break;

    case 46:					 /* magnetic command */
      Token_Strip (input, 1);
      (void) sscanf(input,"%f %f %f %f %f %f %f %f %f",&drvui->mag_matrix[0][0],
                    &drvui->mag_matrix[0][1],&drvui->mag_matrix[0][2],
                    &drvui->mag_matrix[1][0],&drvui->mag_matrix[1][1],
                    &drvui->mag_matrix[1][2],&drvui->mag_matrix[2][0],
                    &drvui->mag_matrix[2][1],&drvui->mag_matrix[2][2]);
      break;
    case 47:
      (void) sscanf (input, "%*7c %d %d %d",
       	&omit1[nomits],&omit2[nomits],&omit3[nomits]);	
      nomits++;
      break;
    case 48:            /* noshadow */
      drvui->noshadow = 1;
      break;
    case 49:			/* finish command */
      (void) sscanf (input, "%s %f %f %f %f",string, &drvui->ambient,
            &drvui->diffuse, &drvui->specular, &drvui->roughness);	
      break;
    case 50:            /* arrow command */
      Token_Strip (input, 1);
      (void) sscanf(input,"%f %f %f %f %f %f %f %f %39c",&drvui->mag_xp[drvui->nmag][0],
                         &drvui->mag_xp[drvui->nmag][1],&drvui->mag_xp[drvui->nmag][2],
                         &drvui->mag_xc[drvui->nmag][0],&drvui->mag_xc[drvui->nmag][1],
                         &drvui->mag_xc[drvui->nmag][2],&drvui->arrow_length[drvui->nmag],
                         &drvui->arrow_diam[drvui->nmag],t_color);
      trim_string(t_color,40);
      strcpy(drvui->col_arrow_p[drvui->nmag],t_color);
      strcpy(drvui->col_arrow_v[drvui->nmag],drvui->col_arrow_p[drvui->nmag]);
      check_color(drvui->col_arrow_v[drvui->nmag]);
      drvui->arrow_fn[drvui->nmag] = tmp_frame_no;
      drvui->nmag++;
      break;
    case 52:            /* mapcontour command */
      numOfFourierContours++;
      if (numOfFourierContours > MAX_MAP_CONTOUR) {
          fprintf (stderr,"Too many contours specified - increase MAX_MAP_CONTOUR.\n");
          delay (5);
          exit (0);
      }
      Token_Strip (input, 1);
      memset(FourierContourColor[numOfFourierContours],0,40);
      (void) sscanf(input,"%f %s %39c",&FourierContourLevel[numOfFourierContours],
          string,FourierContourColor[numOfFourierContours]);
      trim_string(FourierContourColor[numOfFourierContours],40);
      if(strncmp(string, "mesh", 4) == 0)
          FourierContourSolid[numOfFourierContours] = 0;
      else
          FourierContourSolid[numOfFourierContours] = 1;
      break;
    case 53:              /* mapread command */
      Token_Strip (input, 1);
      strcpy(Map_Info.title," Title not given in map file");
      if (Quick) break;
      memset(filename,0,sizeof(filename));
      memset(t_name,0,sizeof(t_name));
      (void) sscanf(input, "%s %s %s", string, filename, t_name);
      trim_string(filename, 100);
      trim_string(t_name,100);

      if (strncmp(string, "grd", 3) == 0) {
          strcpy(FourierFileName,filename);
          FourierMapType=1;
      } else if (strncmp(string, "stf", 3) == 0) {
          strcpy(FourierFileName,filename);
          FourierMapType=2;
      } else if (strncmp(string, "w2k", 3) == 0) {
          strcpy(FourierFileName,filename);
          FourierMapType=3;
      } else if (strncmp(string, "vsp", 4) == 0) {
          strcpy(FourierFileName,filename);
          FourierMapType=4;
      } else if (strncmp(string, "flp", 4) == 0) {
          strcpy(FourierFileName,filename);
          FourierMapType=5;
      } else if (strncmp(string, "fcf", 4) == 0) {
          strcpy(FourierFileName,filename);
          FourierMapType=6;
          Map_Info.map_type = 0;
          if (!strcmp(t_name,"Fc")) Map_Info.map_type = 1;
          else if (!strcmp(t_name,"Fo-Fc")) Map_Info.map_type = 2;
          else if (!strcmp(t_name,"2Fo-Fc")) Map_Info.map_type = 3;
          else if (!strcmp(t_name,"Fo2")) Map_Info.map_type = 4;
      } else {
          fprintf(drvui->flout,"mapread command does not specify a legal file type.\n");
      }
      break;
    case 54:              /* mapregion command */
      Token_Strip(input, 1);
      (void) sscanf(input, "%f %f %f %f %f %f",&xMin, &xMax, &yMin,
          &yMax, &zMin, &zMax);
	  MapRegionRead = 1;
      break;
    case 55:              /* mapcalclimits command */
      Token_Strip(input, 1);
      (void) sscanf(input, "%f %f %f %f %f %f",&Map_Info.xlim[0], &Map_Info.xlim[1],
          &Map_Info.ylim[0], &Map_Info.ylim[1], &Map_Info.zlim[0], &Map_Info.zlim[1]);
      break;
    case 56:              /* labelscale command */
      Token_Strip(input, 1);
      (void) sscanf(input, "%f", &drvui->label_scale);
      break;

    default:
      if (!Quick) {
          fprintf (drvui->flout,
		     "Next line does not start with a recognizable keyword - ignored\n");
          fprintf (drvui->flout, "*         %s\n", input1);	 /* echo to listing file */
      }
      break;
    }
    if (!Quick && intype)
      fprintf (drvui->flout, "*         %s\n", input1);	 /* echo to listing file */
  }
  if (n_ellips > 1)
      do_ellipsoids = 1;
  if (packflag != 0) {
      boxlim[0] = boxlim[1] = boxlim[2] = 100.0f;
  }
}						 /* end of read_inp */

/* ************************************************************** */
/* ************************************************************** */

void read_stf(char * infile, int Quick)
{
    FILE * mapin;
    char line[81];
    char * reslt;
    float rhomin=999999.0f, rhomax=-9999999.0f;
    int ireslt = 0;
    int ijk = 0;
    int i,j,k;
    float bx1,bx2,by1,by2,bz1,bz2;
            
    if ((mapin = fopen (infile, "r")) == NULL) {
        fprintf (stderr,
           "Cannot open Fourier map (.stf) file %s\n", infile);
        return;
    }
    reslt = fgets(line, sizeof(line), mapin);
    reslt = fgets(line, sizeof(line), mapin);
    reslt = fgets(line, sizeof(line), mapin);
    if (reslt){
        ireslt = sscanf(line, "%*s %d %d %d", 
 			&mapstep_a, &mapstep_b, &mapstep_c);
    }		
    if (ireslt > 0) {
      FourierPt = (float*)malloc( mapstep_a*mapstep_b*mapstep_c * sizeof(float));
      if (FourierPt == NULL) {
        ireslt = 0;
        fprintf(stderr,"ERROR -- Unable to allocate space for map.\n");
        return;
      }
      reslt = fgets(line, sizeof(line), mapin);
      sscanf (line,"%*s %f %f %f %f %f %f",&bx1,&bx2,&by1,&by2,&bz1,&bz2);
      map_a=(bx2-bx1);
      map_b=(by2-by1);
      map_c=(bz2-bz1);
      map_alpha=drvui->lat_con[3];
      map_beta=drvui->lat_con[4];
      map_gamma=drvui->lat_con[5];
      reslt = fgets(line, sizeof(line), mapin);
      reslt = fgets(line, sizeof(line), mapin);
      reslt = fgets(line, sizeof(line), mapin);

      if (map_alpha == 90. && map_beta == 90. && map_gamma == 90.) {
        /* read the Rho values */ /* C blocks of A*B values */
        for(k=0; k < mapstep_c; k++) {
          for(j=0; j < mapstep_b-1; j++) {
            for(i=0; i < mapstep_a-1; i++) {
              ijk =  k*(mapstep_a-1)*(mapstep_b-1) + j*(mapstep_a-1) + i;
              ireslt = fscanf(mapin,"%f", &FourierPt[ijk]);
              if (FourierPt[ijk] < rhomin) rhomin = FourierPt[ijk];
              if (FourierPt[ijk] > rhomax) rhomax = FourierPt[ijk];
            }
            fscanf(mapin,"%*f");
          }
          for (i=0;i<mapstep_a;i++) fscanf(mapin,"%*f");
        }

        mapstep_a--;
        mapstep_b--;
        mapstep_c--;
      } else {
        for(i=0; i < mapstep_a; i++) {
          for(j=0; j < mapstep_b-1; j++) {
            for(k=0; k < mapstep_c-1; k++) {
              ijk =  k*(mapstep_a-1)*(mapstep_b-1) + j*(mapstep_a-1) + i;
              ireslt = fscanf(mapin,"%f", &FourierPt[ijk]);
              if (ijk == 0) rhomin = rhomax = FourierPt[ijk];
              if (FourierPt[ijk] < rhomin) rhomin = FourierPt[ijk];
              if (FourierPt[ijk] > rhomax) rhomax = FourierPt[ijk];
            }
            fscanf(mapin,"%*f");
          }
          for (k=0;k<mapstep_c;k++)fscanf(mapin,"%*f");
        }
        mapstep_a--;
        mapstep_b--;
        mapstep_c--;
      } /* orthorhombic or oblique */ 

    } /* if mapstep line present */  
       
    if (ijk > 0) {
      print_map_info(Quick);
      ReadFourMap = 1;
    } else {
      fprintf(drvui->flout, "Error reading Fourier map file %s\n", 
 	       infile);
    }
    fclose(mapin);
}

/* ************************************************************** */
/* ************************************************************** */

void read_vasp(char * infile, int Quick)
{
      FILE * mapin;
      char line[200];
      char * reslt;
      float rhomin=1.0e15f, rhomax=-1.0e15f;
      int ireslt = 0;
      int ijk = 0;
      int i,j,k,l,m;
	  float tmp_rho[10];
      float scale;
      float cmat[3][3];
      int new_mapstep_a = 0, astart, aend;
      int new_mapstep_b = 0, bstart, bend;
      int new_mapstep_c = 0, cstart, cend;
            
      map_a = drvui->lat_con[0];
      map_b = drvui->lat_con[1];
      map_c = drvui->lat_con[2];
      map_alpha = drvui->lat_con[3];
      map_beta = drvui->lat_con[4];
      map_gamma = drvui->lat_con[5];
	  if ((mapin = fopen (infile, "r")) == NULL) {
          if (!Quick)
              fprintf (drvui->flout,
           "Cannot open VASP file %s\n", infile);
        return;
      }
	  fgets(line, sizeof(line), mapin);
      trim_string(line,80);
      strcpy(Map_Info.title,line);
	  fgets(line, sizeof(line), mapin);
      sscanf(line, "%f", &scale);
	  for(i=0; i<3; i++) {   /* read the cell matrix */
		  if (!(reslt = fgets(line, sizeof(line), mapin))) {
              if (!Quick)
                  fprintf(drvui->flout,"Error reading VASP file.\n");
			  return;
		  }
          sscanf(line,"%f %f %f",&cmat[i][0],&cmat[i][1],&cmat[i][2]);
	  }
      for (i=0; i<3; i++)
          for (j=0; j<3; j++)
              cmat[i][j] *= scale;
	  fgets(line, sizeof(line), mapin);
      Map_Info.lat_con[0] = (float) sqrt(cmat[0][0]);   /* this section needs fixing */
      Map_Info.lat_con[1] = (float) sqrt(cmat[1][1]);
      Map_Info.lat_con[2] = (float) sqrt(cmat[2][2]);
      Map_Info.lat_con[3] = 90.0f;            
      Map_Info.lat_con[4] = 90.0f;
      Map_Info.lat_con[5] = 90.0f;
	  j = 0;
	  i = strlen(line)/4;
	  for (k=0; k<i; k++) {
		  sscanf(line + 4 * k,"%4d",&l);
		  j += l;
	  }
	  for(i=0; i<j+2; i++) {   /* set to skip k-point stuff */
		  if (!(reslt = fgets(line, sizeof(line), mapin))) {
              if (!Quick)
                  fprintf(drvui->flout,"Error reading VASP file.\n");
			  return;
		  }
	  }
	  fgets(line, sizeof(line), mapin);
      ireslt = sscanf(line, "%d %d %d", 
 			&mapstep_a, &mapstep_b, &mapstep_c);
      if (ireslt > 0) {
        if ((fabs(Map_Info.xlim[0]) > 0.001) || (fabs(Map_Info.xlim[1] - 1.0f) > 0.001)) {
/* not full cell along x */
            new_mapstep_a = (int)(1.0f/(Map_Info.xlim[1] - Map_Info.xlim[0])+0.5f) * mapstep_a;
            astart = (int)(Map_Info.xlim[0] * new_mapstep_a);
            aend = (int)(Map_Info.xlim[1] * new_mapstep_a);
        } else {
            new_mapstep_a = mapstep_a;
            astart = 0;
            aend = mapstep_a;
        }
        if ((fabs(Map_Info.ylim[0]) > 0.001) || (fabs(Map_Info.ylim[1] - 1.0f) > 0.001)) {
/* not full cell along y */
            new_mapstep_b = (int)(1.0f/(Map_Info.ylim[1] - Map_Info.ylim[0])+0.5f) * mapstep_b;
            bstart = (int)(Map_Info.ylim[0] * new_mapstep_b);
            bend = (int)(Map_Info.ylim[1] * new_mapstep_b);
        } else {
            new_mapstep_b = mapstep_b;
            bstart = 0;
            bend = mapstep_b;
        }
        if ((fabs(Map_Info.zlim[0]) > 0.001) || (fabs(Map_Info.zlim[1] - 1.0f) > 0.001)) {
/* not full cell along z */
            new_mapstep_c = (int)(1.0f/(Map_Info.zlim[1] - Map_Info.zlim[0])+0.5f) * mapstep_c;
            cstart = (int)(Map_Info.zlim[0] * new_mapstep_c);
            cend = (int)(Map_Info.zlim[1] * new_mapstep_c);
        } else {
            new_mapstep_c = mapstep_c;
            cstart = 0;
            cend = mapstep_c;
        }

        if (FourierPt) free(FourierPt);

        FourierPt = (float*)malloc( new_mapstep_a*new_mapstep_b*new_mapstep_c * sizeof(float));
        if (FourierPt == NULL) {
          ireslt = 0;
          if (!Quick) fprintf(drvui->flout,"ERROR -- Unable to allocate space for map.\n");
          return;
        }
        memset(FourierPt,0,new_mapstep_a*new_mapstep_b*new_mapstep_c*sizeof(float));

/* read the Rho values */

		m = 11;
        for(i=cstart; i < cend; i++) {
          for(j=bstart; j < bend; j++) {
            for(k=astart; k < aend; k++) {
			    if (m > 9) {
					reslt = fgets(line, sizeof(line), mapin);
					sscanf(line,"%f %f %f %f %f %f %f %f %f %f",&tmp_rho[0],
						&tmp_rho[1],&tmp_rho[2],&tmp_rho[3],&tmp_rho[4],
						&tmp_rho[5],&tmp_rho[6],&tmp_rho[7],&tmp_rho[8],
						&tmp_rho[9]);
					m = 0;
				}
                ijk =  k * new_mapstep_b * new_mapstep_c + j * new_mapstep_c + i;
				FourierPt[ijk] = tmp_rho[m++];
                if (FourierPt[ijk] < rhomin) rhomin = FourierPt[ijk];
                if (FourierPt[ijk] > rhomax) rhomax = FourierPt[ijk];
            }
          }
        }
      }
       
      if (reslt) {
          ReadFourMap = 1;
      } else {
          if (!Quick)fprintf(drvui->flout, "Error reading Fourier map file %s",infile);
      }
      Map_Info.rhomn = rhomin;
      Map_Info.rhomx = rhomax;
      Map_Info.map_int[0] = mapstep_a = new_mapstep_a;
      Map_Info.map_int[1] = mapstep_b = new_mapstep_b;
      Map_Info.map_int[2] = mapstep_c = new_mapstep_c;
	  Map_Info.info_valid = 1;
      fclose(mapin);
      print_map_info(Quick);

}

/* ************************************************************** */
/* ************************************************************** */

void read_w2k(char * infile, int Quick)
{
      FILE * mapin;
      char line[81];
      char * reslt;
      float rhomin=1.0e15f, rhomax=-1.0e15f;
      int ireslt = 0;
      int dimen=0;
      int ijk = 0;
      int i,j,k;
      int new_mapstep_a = 0, astart, aend;
      int new_mapstep_b = 0, bstart, bend;
      int new_mapstep_c = 0, cstart, cend;

      Map_Info.lat_con[0] = map_a = drvui->lat_con[0];
      Map_Info.lat_con[1] = map_b = drvui->lat_con[1];
      Map_Info.lat_con[2] = map_c = drvui->lat_con[2];
      Map_Info.lat_con[3] = map_alpha = drvui->lat_con[3];
      Map_Info.lat_con[4] = map_beta = drvui->lat_con[4];
      Map_Info.lat_con[5] = map_gamma = drvui->lat_con[5];
            
      if ((mapin = fopen (infile, "r")) == NULL) {
        if (!Quick)
            fprintf (drvui->flout,"Cannot open Fourier map (.w2k) file %s\n", infile);
        return;
      }

      do {
        reslt = fgets(line, sizeof(line), mapin);
        if (!strncmp(line,"(@3 [3]",7)) dimen=1;
      } while (!feof(mapin) && dimen==0);
 
      if (dimen==0) {
        if (!Quick)
            fprintf (drvui->flout,"Cannot read Fourier map (.w2k) file %s\n", infile);
        return;
      }

      ireslt=sscanf(line, "%*s %*s %d %d %d",&mapstep_a,&mapstep_b,&mapstep_c);

      if (ireslt > 0) {
        if ((fabs(Map_Info.xlim[0]) > 0.001) || (fabs(Map_Info.xlim[1] - 1.0f) > 0.001)) {
/* not full cell along x */
            new_mapstep_a = (int)(1.0f/(Map_Info.xlim[1] - Map_Info.xlim[0])+0.5f) * mapstep_a;
            astart = (int)(Map_Info.xlim[0] * new_mapstep_a);
            aend = (int)(Map_Info.xlim[1] * new_mapstep_a);
        } else {
            new_mapstep_a = mapstep_a;
            astart = 0;
            aend = mapstep_a;
        }
        if ((fabs(Map_Info.ylim[0]) > 0.001) || (fabs(Map_Info.ylim[1] - 1.0f) > 0.001)) {
/* not full cell along y */
            new_mapstep_b = (int)(1.0f/(Map_Info.ylim[1] - Map_Info.ylim[0])+0.5f) * mapstep_b;
            bstart = (int)(Map_Info.ylim[0] * new_mapstep_b);
            bend = (int)(Map_Info.ylim[1] * new_mapstep_b);
        } else {
            new_mapstep_b = mapstep_b;
            bstart = 0;
            bend = mapstep_b;
        }
        if ((fabs(Map_Info.zlim[0]) > 0.001) || (fabs(Map_Info.zlim[1] - 1.0f) > 0.001)) {
/* not full cell along z */
            new_mapstep_c = (int)(1.0f/(Map_Info.zlim[1] - Map_Info.zlim[0])+0.5f) * mapstep_c;
            cstart = (int)(Map_Info.zlim[0] * new_mapstep_c);
            cend = (int)(Map_Info.zlim[1] * new_mapstep_c);
        } else {
            new_mapstep_c = mapstep_c;
            cstart = 0;
            cend = mapstep_c;
        }

        if (FourierPt) free(FourierPt);

        FourierPt = (float*)malloc( new_mapstep_a*new_mapstep_b*new_mapstep_c * sizeof(float));
        if (FourierPt == NULL) {
          ireslt = 0;
          if (!Quick)fprintf(drvui->flout,"ERROR -- Unable to allocate space for map.\n");
          return;
        }
        memset(FourierPt,0,new_mapstep_a*new_mapstep_b*new_mapstep_c*sizeof(float));

        reslt = fgets(line, sizeof(line), mapin);

        for(k=cstart; k < cend; k++) {
          for(j=bstart; j < bend-1; j++) {
            for(i=astart; i < aend-1; i++) {
              ijk =  j*(new_mapstep_a-1)*(new_mapstep_b-1) + k*(new_mapstep_a-1) + i;
              ireslt = fscanf(mapin,"%f", &FourierPt[ijk]);
              if (FourierPt[ijk] < rhomin) rhomin = FourierPt[ijk];
              if (FourierPt[ijk] > rhomax) rhomax = FourierPt[ijk];
            }
            fscanf(mapin,"%*f");
          }
          for (i=0;i<mapstep_a;i++) fscanf(mapin,"%*f");
        }

     }

     if (ijk > 0) {
        ReadFourMap = 1;
      } else {
        if (!Quick)
            fprintf(drvui->flout, "Error reading Fourier map file %s\n",infile);
      }
      fclose(mapin);
      Map_Info.rhomn = rhomin;
      Map_Info.rhomx = rhomax;
	  Map_Info.map_int[0] = mapstep_a = --new_mapstep_a;
	  Map_Info.map_int[1] = mapstep_b = --new_mapstep_b;
	  Map_Info.map_int[2] = mapstep_c = --new_mapstep_c;
	  Map_Info.info_valid = 1;
      print_map_info(Quick);
}

