/* ===================================================================== */
/*     DRAWxtl                                                           */
/*                                                                       */
/* Routine to read crystallographic information such as unit cell, space */
/* group, atom positions, and thermal ellipsoid parameters to generate   */
/* POLYHEDRON, SPHERE, ELLIPSOID and BOND geometry POV and VRML scenes.  */
/*                                                                       */
/* Part 1                                                                */
/*                                                                       */
/* Coded by:  L.W. Finger                                                */
/*            Geophysical Laboratory                                     */
/*            Carnegie Institution of Washington                         */
/*            Washington DC 20015                                        */
/*            Larry.Finger@lwfinger.net                                  */
/*                                                                       */
/* Original Version: April 92                                            */
/*            Revised January 1993 to add bonds.                         */
/*            Converted to AVS Module February 1993.                     */
/*            Thermal Ellipsoids added April 1996                        */
/*            Thermal ellipsoids for POV and VRML finished Nov. 1996     */
/*                                                                       */
/*                                                                       */
/* Converted to standalone POV scene file generator Oct. 1994/March 1995 */
/* and VRML generator added September 1995/March 1996                    */
/* by M. Kroeker, Organic Chemistry, Technische Hochschule Darmstadt,    */
/* Petersenstr. 22, D-64287 Darmstadt, Germany                           */
/*                                                                       */
/* Two versions merged  July 1995                                        */
/* ===================================================================== */
/* Routines included in this file:
 *
 *   main - entry point
 *   add_vert - add a vertex to the list with checking
 *   add_vert_nc - add a vertex without checking
 *   add_to_list - add an atom to the unit-cell list
 *   analyze_bonds - print the bond table
 *   axeqb - solve the matrix equation AX = B
 *   Blank_Strip - strip leading blanks
 *   build_box_contents - get all the atoms in the display box
 *   Calc_Rot - calculate the rotation matrix from lookat vectors
 *   Convert_ellipsoids - change all ellipsoids to betaij notation
 *   Conv_Sym_Mat - change hexagonal and trigonal symmetry rotations to cartesian
 *   delay - wait for a time
 *   dist - calculate a bond distance
 *   do_drawing - calls all the individual generation routines
 *   expand_atoms - uses symmetry to expand an atom in the unit cell
 *   eigen - calculate eigen values and vectors
 *   find_all_in_box - get all the atoms in the display box
 *   generate_bonds - draw the appropriate bonds
 *   generate_cones - find and draw the lone-pair cones
 *   generate_arrows - draw the arrows
 *   get_atom_id - encode drawing info into atom number
 *   get_input - preliminary routine for reading input file
 *   get_label - get an atom label
 *   label_cell - convert cell axis labels to labeltext commands
 *   Locate_Triple - find the "best" location for the vector triple
 *   make_bmat - form the unit-cell matrices
 *   Output_Spheres - generate the spheres
 *   P_to_C - convert ellipsoid probability to appropriate constant
 *   plot_vrml_poly - output VRML polyhedra
 *   print_sym - format the symmetry matrices
 *   Token_Strip - remove an item from the beginning of the input line
 *   usage - print out the minimal help information
 *   resizeWindow - GLUT callback when openGL window is resized
 *   display_crystal - GLUT callback for openGL display
 *   Dump_View - prepare EPS file
 *   key - GLUT callback to interpret keyboard commands
 *   specialkey - GLUT callback to interpret various control keys
 *   mouse - GLUT callback for mouse
 *   mousemotion - GLUT callback function for motion
 *   start_picking - used in dynamic edit of screen contents
 *   Process_hits - process the omit list
 *   pick_label - used in dynamic positioning of text labels
 *   write_header - output the POV and VRML headers
 *   not_in_slab - return info regarding whether an atom is inside the slab
 *   dot0_3d - perform a dot product
 *   enormsq0_3d - square of the Euclidean norm
 *   tb_init - initialize trackball
 *   trackball - implements the virtual trackball
 *   window_to_tb - converts from window coordinates to trackball
 *   tb_project_to_sphere - does the projection from trackball to sphere
 *   axis_to_quaternion - changes a single axial rotation to quaternion form
 *   qmult - multiplies two quaternions
 *   qnormalize - normalizes a quaternion
 *   quaternion_to_rotmatrix - generates the equivalent rotation matrix
 *   vnew - creates a new vector and sets its components
 *   vset - loads the components into an existing vector
 *   vcopy - copies a vector
 *   vzero - zeros a vector
 *   vnormalize - normalizes a vector to unit length
 *   vlength - returns the length of a vector
 *   vscale - multiplies a vector by a scalar
 *   vmult - multiplies two vectors
 *   vadd - adds two vectors
 *   vsub - subtracts a vector from another
 *   vdot - takes the dot product of two vectors
 *   vcross - calculates the cross product
 *   XYZ_Rot_to_Q - forms a quaternion from rotation matrices about X, Y, and Z
 *   print_map_info - prints the information about a Fourier map
 *   Progress_Window - simulates a progress meter on stderr
 *
 *************************************************************************/

#include "drawxtl.h"
#include "draw_gbl.h"
#include "gl2ps.h"
#ifndef WIN32
  #include <unistd.h>
#endif

#ifdef OPENGL
#ifdef WIN32
  #define AMBLEV 0.2f
#else
  #define AMBLEV 0.2f
#endif
  static void display_crystal (void);
  static void key (unsigned char, int, int);
  static void mouse (int, int, int, int);
  static void mousemotion (int, int);
  static void process_hits(GLint, GLuint*);
  static void resizeWindow (int, int);
  static void specialkey (int, int, int);
  static void start_picking(int,int);
  static int pick_label(int,int,int,int,float*);
  static void write_header(void);
  static void close_and_exit(void);
#endif

/* ************************************************************** */
/* ************************************************************** */
int main (int argc, char *argv[])
{
  int i, Quick = 0;
  char temp[200];

#ifdef DELAY
  time_t start, end;
#endif

#ifdef OPENGL
  GLfloat mat_specular[] = { 1.0f, 1.0f, 1.0f, 1.0f };
  GLfloat light_ambient[] =	{ AMBLEV, AMBLEV, AMBLEV, 1.0f };
  GLfloat light_specular[] = { 0.1f, 0.1f, 0.1f, 1.0f };
  GLfloat light_diffuse[] = { 1.0f, 1.0f, 1.0f, 1.0f };
  GLfloat mat_shininess[] = { 0.8f };
#endif

#ifdef DELAY
  start = time (NULL);
#endif

/* Initialize */

  drvui = (CRYSTAL *)malloc(sizeof(CRYSTAL));
  memset(drvui,0,sizeof(CRYSTAL));
  drvui->X_Origin = drvui->Y_Origin = drvui->Z_Origin = 0.5f;
  drvui->X_Boxlim = drvui->Y_Boxlim = drvui->Z_Boxlim = 100.0f;
  xrot = yrot = zrot = 0.0f;
  boxflag = packflag = domolcomp = 0;
  do_ellipsoids = edges = 0;
  Ellipsoid_Prob = 0.5f;
  drvui->label_scale = 1.0f;
  drvui->labels_moved = 0;
  drvui->autolabel = 0;
  drvui->acentric = 1;
  Unit_Cell = M_cameras = no_comment = Labels = Vrml2 = 1;
  gl_view = 1;
  Display_axes = El_Cutout = 0;
  strcpy (drvui->Cutout_color, "Gray20");
  strcpy (Ellipaxis_color, "Gray20");
  SpMult = drvui->BndMult = Magnification = 1.0f;
  strcpy (col_cell_v, "0 0 0");			 /* preset unit-cell colors to Black */
  strcpy (col_cell_p, "Black");
  Ellipsoid_Scale = P_to_C (0.5f);
  Options = 0;
  drvui->polylimit = 0.1f;
  drvui->Phong_Value = 0.2f;
  drvui->Phong_Size = 1.0f;
  DepthCue = 0.0f;
  printdist = 3.5f;
  numOfFourierContours = 0;
  memset(omit1,0,1000);
  memset(omit2,0,1000);
  memset(omit3,0,1000);
/* Parse arguments */

  if (argc < 2) {
    usage ();
    delay (3);
  }
  for (i = 1; i < argc; i++) {
    if (!strcmp (argv[i], "-n"))
      Unit_Cell = 0;
    else if (!strcmp (argv[i], "-h"))
      usage ();
    else if (!strcmp (argv[i], "-?"))
      usage ();
    else if (!strcmp (argv[i], "-o")) {
      drvui->X_Origin = (float) atof (argv[++i]);
      drvui->Y_Origin = (float) atof (argv[++i]);
      drvui->Z_Origin = (float) atof (argv[++i]);
      Options = Options | O_OPT;
    }
    else if (!strcmp (argv[i], "-O")) {
      M_cameras = 0;
#ifdef OPENGL
      gl_ortho=1;
#endif
    }
    else if (!strcmp (argv[i], "-C")) {
      M_cameras = 0;
#ifdef OPENGL
      gl_ortho=1;
#endif
    }
    else if (!strcmp (argv[i], "-c")) {
      no_comment = 0;
    }
    else if (!strcmp (argv[i], "-b")) {
      drvui->X_Boxlim = (float) atof (argv[++i]);
      drvui->Y_Boxlim = (float) atof (argv[++i]);
      drvui->Z_Boxlim = (float) atof (argv[++i]);
      boxflag = 1;
      Options = Options | B_OPT;
    } else if (!strcmp (argv[i], "-p")) {
      drvui->cryst_lim[1][0] = (float) atof (argv[++i]);
      drvui->cryst_lim[1][3] = (float) atof (argv[++i]);
      drvui->cryst_lim[1][1] = (float) atof (argv[++i]);
      drvui->cryst_lim[1][4] = (float) atof (argv[++i]);
      drvui->cryst_lim[1][2] = (float) atof (argv[++i]);
      drvui->cryst_lim[1][5] = (float) atof (argv[++i]);
      Options = Options | P_OPT;
      packflag = 1;
    } else if (!strcmp (argv[i], "-v")) {
      xrot = (float) atof (argv[++i]);
      yrot = (float) atof (argv[++i]);
      zrot = (float) atof (argv[++i]);
      Options = Options | V_OPT;
    } else if (!strcmp (argv[i], "-m")) {
      Magnification = (float) atof (argv[++i]);
      Options = Options | M_OPT;
    } else if (!strcmp (argv[i], "-u")) {
      Display_axes = 1;
    } else if (!strcmp (argv[i], "-Q")) {
      Quick = 1;
    } else if (!strcmp (argv[i], "-H")) {
      view_inst = 0;
    } else if (!strcmp (argv[i], "-G")) {
      gl_view = 0;
    } else if (!strcmp (argv[i], "-a")) {
        drvui->autolabel = 2;
    } else
      drvui->Str_File = argv[i];
  }
 
/* next section will change working directory to that of the data file */
  (void)getcwd(drvui->savedpath,100);           /* save current working directory */
  strcpy(temp,drvui->Str_File);
  for (i=strlen(temp); i>0; --i) {
    if(temp[i] == '/' || temp[i] == '\\') {
        temp[i] = '\0';
        break;\
    }
  }
  if (i > 0) chdir(temp);
  (void)getcwd(drvui->savedpath,100);           /* save current working directory */
  for (i=strlen(drvui->Str_File); i>0; --i) {
    if (drvui->Str_File[i] == '/' || drvui->Str_File[i] == '\\') {
      i++;
      break;
    }
  }
  if (i > 0) {
    strcpy(temp,drvui->Str_File +i);
    strcpy(drvui->Str_File,temp);
  }
  if (!(drvui->fpin = fopen(drvui->Str_File,"r"))) {
      printf("Input file '%s' cannot be opened. Run aborted!\n",drvui->Str_File);
      exit(0);
  }
  fclose(drvui->fpin);
  if (packflag != 0) {
    drvui->X_Boxlim = drvui->Y_Boxlim = drvui->Z_Boxlim = 500.0f;
  }
  if ((drvui->atom_no = (int *)
      malloc ((unsigned) ((long) (MAX_VERTS * sizeof (int))))) == NULL) {
    printf ("Unable to get initial atom_no allocation\n");
    return 0;
  }
  if ((xypos = (float *)
       malloc ((unsigned) (3 * (long) (MAX_VERTS * sizeof (float))))) == NULL) {
    printf ("Unable to get initial xypos allocation\n");
    return 0;
  }
#ifdef OPENGL
  if (gl_view) {
    glutInit (&argc, argv);
    glutInitDisplayMode ((GLbitfield) (GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGB));
    glutInitWindowPosition (100, 50);
    glutInitWindowSize (gl_width, gl_height);
    (void) glutCreateWindow ("DRAWxtl V4.2");
    glutDisplayFunc (display_crystal);
    glutReshapeFunc (resizeWindow);

/*************************************************************************************
 *
 *   The glutWMCloseFunc call below makes closing the openGL window by clicking on the X
 *   in the upper right corner the same as pressing the <esc>, 'q', or 'Q' keys. This
 *   function is not part of the standard GLUT API and may not compile on all systems.
 *   It is OK to comment it out. Without it, the structure orientation at time of exit
 *   is not written to the POV file, nor is the appropriate view statement appended to 
 *   the 'str' file, otherwise the program operates normally.
 *
 *   LWF 08/08/05
 **************************************************************************************/

#if defined (WIN32) || defined(FREEGLUT) || (defined(GLUT_MACOSX_IMPLEMENTATION) && GLUT_MACOSX_IMPLEMENTATION>1)
    glutWMCloseFunc (close_and_exit);
#endif
    glutKeyboardFunc (key);
    glutMouseFunc (mouse);
    glutMotionFunc (mousemotion);
    glutSpecialFunc (specialkey);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glMaterialfv (GL_FRONT, GL_SPECULAR, mat_specular);

    glMaterialfv (GL_FRONT, GL_SHININESS, mat_shininess);

    glLightfv (GL_LIGHT0, GL_SPECULAR, light_specular);
    glLightfv (GL_LIGHT0, GL_DIFFUSE, light_diffuse);
    glLightfv (GL_LIGHT0, GL_AMBIENT, light_ambient);
    glLightModeli (GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
    glEnable (GL_LIGHTING);
    glEnable (GL_LIGHT0);
    glEnable (GL_COLOR_MATERIAL);

    glColorMaterial (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
    glEnable (GL_DEPTH_TEST);
  }
#endif
  crystal_compute (Quick);			 /* generate POV and VRML outputs */
#ifdef DELAY
  end = time (NULL);
  printf ("Elapsed time: %10.1f secs.\n", difftime (end, start));
  start = time (NULL);
  delay (2);
#endif
#ifdef OPENGL
  if (gl_view) {
    glutMainLoop ();
  }
#endif
  if (FourierPt) free(FourierPt);
  return 1;
}						 /* end main for POV and VRML versions */

/* ************************************************************** */
/* ************************************************************** */

void add_vert (float vert[3], int no, int check)

/* procedure to add the vertex in 'vert' to the current list if the
   vertex is contained within the limits of the box, within the limits
   of a polyhedral center, or needed to complete a molecule */
{
  float c_vert[3];				 /* vertex in cartesian coords */
  float r_vert[3];				 /* vertex in rotated coordinates */
  int i, j;                                      /* loop variables */
  float temp;					 /* temporary storage */
  int outside;					 /* temporary for molecular completion */
  int number;					 /* decoded atom number */
  int poly_center;				 /* true if this atom is center of a polyhedron */
  int consider_extra;
  float d=0.0f;

  for (i = 0; i < 3; ++i) {			 /* convert vertex coordinates to Cartesian */
    c_vert[i] = 0.0f;
    for (j = 0; j < 3; ++j)
      c_vert[i] += (float) drvui->b_mat[i][j] * (vert[j] - origin[j]);
  }
  number = no / 100;
  poly_center = (((drvui->atom_n[number] >> 8) & 255) > 0);
  consider_extra = (domolcomp != 0 || (drvui->npoly > 1 && !poly_center));
  outside = 0;
  if (check == 0) {
    for (i = 0; i <= 2; i++) {			 /* check packing range */
      if ((vert[i] < drvui->cryst_lim[drvui->frame_no][i])
          || (vert[i] > drvui->cryst_lim[drvui->frame_no][i + 3]))
        outside = 1;
    }
    if (outside == 0 && slabmode == 1)
      outside = not_in_slab(c_vert[0], c_vert[1],c_vert[2]);
    if (((outside == 0) && (fabs (c_vert[0]) > boxlim[0])) ||
              (fabs (c_vert[1]) > boxlim[1]) || (fabs (c_vert[2]) > boxlim[2]))
        outside = 1;        		 /* check box limits */
    if (outside != 0 && !consider_extra)
        return;
  }
  d = 0.0f;
  if (outside == 1) {				 /* molecule or polyhedral completion is in progress */
    float dmin, dmax;
    int polyno;					 /* number of the polyhedron */

    for (i = 1; i < NvertM; i++) {
      dmax = drvui->mol_d * drvui->mol_d;
      dmin = 0.0f;
      d = 0.0f;
      for (j = 0; j < 3; j++) {
        temp = (s_vert[3 * i + j] - c_vert[j]);
        d += temp * temp;
      }
      j = drvui->atom_no[i] / 100;
      poly_center = (((drvui->atom_n[j] >> 8) & 255) > 0);
      if (poly_center) {
        polyno = (drvui->atom_n[j] >> 8 ) & 255;
        dmax = drvui->poly_size[polyno] * drvui->poly_size[polyno];
        dmin = drvui->poly_min[polyno] * drvui->poly_min[polyno];
      }
      if ((domolcomp != 0 || poly_center) && (d <= dmax && d > 0.02)) {
        outside = 2;
        if (poly_center && d < dmin)
          outside = 1;
        if ((fabs (c_vert[0]) > 10. * boxlim[0]) ||
            (fabs (c_vert[1]) > 10. * boxlim[1]) ||
            (fabs (c_vert[2]) > 10. * boxlim[2])) {
          fprintf (stderr,
		     "mol'completing beyond 10*lattice constant - aborting run\n");
	      outside = 1;				 /* far out of check box limits - combinatorial explosion */
	      exit(0);
        }
      }
    }
  }
  if (outside == 1)
    return;
  i = (int) (no / 100);
  for (j = 1; j < NvertM; ++j) {
    if (fabs (s_vert[3 * j] - c_vert[0]) < 0.01 &&
	fabs (s_vert[3 * j + 1] - c_vert[1]) < 0.01 &&
	fabs (s_vert[3 * j + 2] - c_vert[2]) < 0.01 &&
    check_atom_name(drvui->atom_l[i], drvui->atom_l[drvui->atom_no[j] / 100]) &&
    (number == i))
      return;					 /* eliminate duplicates */
  }
  for (i = 0; i <= 2; ++i) {			 /* add atom to Master List */
    xypos[3 * NvertM + i] = vert[i];
    s_vert[3 * NvertM + i] = c_vert[i];
    if (offset[i] > c_vert[i])
      offset[i] = c_vert[i];
  }
  drvui->atom_no[NvertM++] = no;
  if (NvertM > MAX_VERTS) {
    if (domolcomp != 0) {
      fprintf (stdout, "Too many vertices for dimensions:\n");
      fprintf (stdout, " Did you use a 'molcomp' command incorrectly?\n");
      fprintf (stdout, " If not, increase MAX_VERTS.\n");
    } else {
      fprintf (stdout,
	       "Too many vertices for dimensions - increase MAX_VERTS.\n");
    }
    delay (5);
    exit (0);
  }
  for (i = 0; i <= 2; ++i) {			 /* calculate position of point after POV rotation */
    r_vert[i] = 0.0f;
    for (j = 0; j <= 2; ++j)
      r_vert[i] += (float) G_Rot[j][i] * c_vert[j];
  }
  for (i = 0; i <= 2; ++i) {			 /* Update Min and Max of output */
    if (POV_Max[i] < r_vert[i])
      POV_Max[i] = r_vert[i];
    if (POV_Min[i] > r_vert[i])
      POV_Min[i] = r_vert[i];
  }
}

/* ************************************************************** */
/* ************************************************************** */

void add_vert_nc (float vert[3])

/* procedure to add the vertices in 'vert' to the current working list
   (NOT Master List ) without checking */
{

  int i, j;					 /* loop variables */

  if (!s_vert) {
      if ((s_vert = (float *)malloc ((unsigned) (3 * (long) 
          (MAX_VERTS * sizeof (float))))) ==	NULL) {
        printf ("Unable to get initial s_vert allocation\n");
        delay (5);
        return;
      }
  }  
  for (i = 0; i <= 2; ++i) {			 /* convert vertex coordinates to Cartesian */
    s_vert[3 * nvert + i] = 0.0f;
    for (j = 0; j <= 2; ++j)
      s_vert[3 * nvert + i] += (float) drvui->b_mat[i][j] * (vert[j] - origin[j]);
  }
  nvert++;
}

/* ************************************************************** */
/* ************************************************************** */

void add_to_list (float xp[3], int no)

/* add atom position to unit-cell list if unique */
/* no - symmetry operator number */
{
  int i, ind;

  for (i = 0; i <= 2; ++i) {			 /* get coordinates between 0 and 1 */
    if (xp[i] >= 1.0)
      xp[i] -= 1.0f;
    if (xp[i] < 0.0)
      xp[i] += 1.0f;
  }
  ind = 0;
  for (i = 0; i < ncell; ++i) {
    if ((fabs (drvui->cell_xyz[i][0] - xp[0]) <= 0.00002) &&
	(fabs (drvui->cell_xyz[i][1] - xp[1]) <= 0.00002) &&
	(fabs (drvui->cell_xyz[i][2] - xp[2]) <= 0.00002))
      ind = 1;					 /*set if same */
  }
  if (ind == 0) {				 /* true for new position */
    for (i = 0; i <= 2; ++i)			 /*copy to list */
      drvui->cell_xyz[ncell][i] = xp[i];
    drvui->sym_op_no[ncell++] = no;
  }
}

/* ************************************************************** */
/* ************************************************************** */

void analyze_bonds (void)

/* routine to print bond distances up to a maximum of 'printdist' Angstrom */
{
  int i, j, k, l, nvert1, *itype;
  float *d, dmin;

  if ((itype = (int *) malloc ((unsigned) (MAX_VERTS * sizeof (int)))) == NULL)
    return;
  fprintf (drvui->flout, "\n\n Bond Distance Analysis\n");
/* Find all atoms of this type in master list */
  l = nvert = nvert1 = 1;
  for (j = 0; j < natom; ++j) {
    l = nvert;
    find_all_in_box (j);
    for (k = l; k < nvert; ++k)
      itype[k] = j;				 /* itype[k] has atom number j */
  }
  for (i = 0; i < natom; ++i) {			 /* loop through atom types */
/* add the base entry to the beginning of the atom list */
    for (j = 0; j < 3; ++j) {			 /* convert vertex coordinates to Cartesian */
      s_vert[j] = 0.0f;
      for (k = 0; k <= 2; ++k)
        s_vert[j] += (float) drvui->b_mat[j][k] * (drvui->atom_xyz[i][k] - origin[k]);
    }
    if ((d = (float *) malloc ((unsigned) (nvert * sizeof (float)))) == NULL) {
      free (itype);
      printf ("Unable to get bond distance allocation\n");
      return;					 /* get space for distances */
    }
    for (j = 1; j < nvert; ++j)			 /* calculate distances */
      d[j] = dist (0, j);
    fprintf (drvui->flout,
	     "\nDistances from %c%c%c%c%3d at %8.5f%8.5f%8.5f    to\n\n",
	     drvui->atom_l[i][0], drvui->atom_l[i][1], drvui->atom_l[i][2], drvui->atom_l[i][3],
	     drvui->sv_atom_n[i], drvui->atom_xyz[i][0], drvui->atom_xyz[i][1],
         drvui->atom_xyz[i][2]);
    for (k = nvert1; k < nvert; ++k) {
      dmin = (float) 999999.9;
      for (j = nvert1; j < nvert; ++j) {	 /* loop through distances */
        if (d[j] < dmin) {
          l = j;
          dmin = d[j];
        }
      }
      if (dmin == 0.0 && itype[l] == i) {
        d[l] = 999999.9f;
      } else {
        if (dmin <= printdist) {
          fprintf (drvui->flout,
              " %c%c%c%c%3d at %9.5f%9.5f%9.5f  %8.3f\n",
              drvui->atom_l[itype[l]][0], drvui->atom_l[itype[l]][1],
              drvui->atom_l[itype[l]][2], drvui->atom_l[itype[l]][3],
              drvui->sv_atom_n[itype[l]], o_vert[3 * l],
              o_vert[3 * l + 1], o_vert[3 * l + 2], dmin);
          d[l] = 999999.9f;
        } else
          break;
      }
    }
    free (d);
  }
  free (itype);
  fprintf (drvui->flout, "\n\n");
}

/* ************************************************************** */
/* ************************************************************** */

void axeqb (double a1[3][3], double x[3], double b[3])

/* solve matrix equation ax = b for x, where a is 3x3 and x and b are
   vectors */
{
  int i, j;
  double a[3][3], det;

  det = determinant (a1);
  if (fabs (det) < 1.0e-6)
    det = 1.0e-6;
  a[0][0] = (a1[1][1] * a1[2][2] - a1[1][2] * a1[2][1]) / det;
  a[1][0] = -(a1[1][0] * a1[2][2] - a1[1][2] * a1[2][0]) / det;
  a[2][0] = (a1[1][0] * a1[2][1] - a1[1][1] * a1[2][0]) / det;
  a[0][1] = -(a1[0][1] * a1[2][2] - a1[0][2] * a1[2][1]) / det;
  a[1][1] = (a1[0][0] * a1[2][2] - a1[0][2] * a1[2][0]) / det;
  a[2][1] = -(a1[0][0] * a1[2][1] - a1[0][1] * a1[2][0]) / det;
  a[0][2] = (a1[0][1] * a1[1][2] - a1[0][2] * a1[1][1]) / det;
  a[1][2] = -(a1[0][0] * a1[1][2] - a1[0][2] * a1[1][0]) / det;
  a[2][2] = (a1[0][0] * a1[1][1] - a1[0][1] * a1[1][0]) / det;
  for (i = 0; i < 3; i++) {
    x[i] = 0.0;
    for (j = 0; j < 3; j++)
      x[i] += a[i][j] * b[j];
  }
}

/* ************************************************************** */
/* ************************************************************** */

void Blank_Strip (char input[100])

/* routine to suppress any leading spaces in the input string */
{
  int i, j;

  for (i = 0; i < 100; i++)
    if (input[i] != ' ')
      break;
  for (j = 0; j < 100; j++) {
    input[j] = input[i];
    if (input[i++] == '\0')
      break;
  }
}

/* ************************************************************** */
/* ************************************************************** */

void build_box_contents (void)

/* routine to locate all atoms in the display parallelopiped plus
   any atoms needed to complete molecules or polyhedra */
{
  float vert[3];				 /* storage for center */
  int ix[3];					 /* array for lattice offset */
  int j, k, saved_nvert;
  int atoms, loop;
  float saved_xyz[MAX_ATOM][3];

  loop = 0;
  while (1) {                        /* infinite loop here */
    saved_nvert = NvertM;      /* save number currently in list */
    for (atoms = 0; atoms < natom; atoms++) {     /* loop through atoms */
      for (j=0;j<3;j++) {            /* put the base atom in the range 0-1 */
        saved_xyz[atoms][j] = drvui->atom_xyz[atoms][j];
        if (drvui->atom_xyz[atoms][j] < 0.0f) drvui->atom_xyz[atoms][j] += 1.0f;
        if (drvui->atom_xyz[atoms][j] > 1.0f) drvui->atom_xyz[atoms][j] -= 1.0f;
      }
      expand_atom (atoms);          /* generate all atoms of this flavor in unit cell */
      for (j = 0; j < ncell; ++j) { /*atoms in cell */
        for (k = 0; k <= 2; ++k)
          vert[k] = drvui->cell_xyz[j][k];
        add_vert (vert, 100 * atoms + drvui->sym_op_no[j], 0);	/* add point to cell if inside box */
      }
    }
    if (saved_nvert == NvertM) {
      break;
    }
  }
  while (1) {                        /* infinite loop here */
    saved_nvert = NvertM;			 /* save number currently in list */
    for (atoms = 0; atoms < natom; atoms++) {	 /* loop through atoms */
      expand_atom (atoms);			 /* generate all atoms of this flavor in unit cell */
      for (j=0;j<ncell;++j) {        /*atoms in cell*/
        for (k=0;k<=2;++k) 
          vert[k] = drvui->cell_xyz[j][k];
        add_vert(vert,100*atoms+drvui->sym_op_no[j],0);  /* add point to cell if inside box*/
      }
      for (ix[0] = (int) drvui->cryst_lim[drvui->frame_no][0] - 1; 
               ix[0] <= (int) drvui->cryst_lim[drvui->frame_no][3] + 1; ++ix[0]) {	/* step through cells parallel a */
        for (ix[1] = (int) drvui->cryst_lim[drvui->frame_no][1] - 1; 
               ix[1] <= (int) drvui->cryst_lim[drvui->frame_no][4] + 1; ++ix[1]) {	/* parallel b */
          for (ix[2] = (int) drvui->cryst_lim[drvui->frame_no][2] - 1; 
               ix[2] <= (int) drvui->cryst_lim[drvui->frame_no][5] + 1; ++ix[2]) {	/* parallel c */
            if (abs (ix[0]) + abs (ix[1]) + abs (ix[2]) != 0) {
              for (j = 0; j < ncell; ++j) {	 /*atoms in cell */
                for (k = 0; k <= 2; ++k)
                  vert[k] = ix[k] + drvui->cell_xyz[j][k];
                add_vert (vert, 100 * atoms + drvui->sym_op_no[j], 0);	/* add point to cell if inside box */
              }
            }
          }
        }
      }
    }
    if ((domolcomp == 0 && drvui->npoly == 1) || saved_nvert == NvertM) {
      for (atoms = 0; atoms < natom; atoms++) {     /* loop through atoms */
          for (j=0;j<3;j++) {            /* restore base atom position */
              drvui->atom_xyz[atoms][j] = saved_xyz[atoms][j];
          }
      }
      return;
    }
    if (loop == 0)
      printf ("\n");
    loop++;
    printf ("Recursive Atom Addition, Loop No. %3d, %4d Atoms in List.\n",
	    loop, NvertM - 1);
  }
}						 /* end of build_box_contents */

/* ************************************************************** */
/* ************************************************************** */

void Calc_Rot (float v1[3], float v2[3])

/* routine to calculate rotation axes needed to put v1 parallel to Z
   and the projection of v2 parallel to Y */
{
  float u1[3], u2[3], u3[3], temp;
  double U[3][3];
  int i, j;

  for (i = 0; i < 3; i++)
    for (j = 0; j < 3; j++)
      U[i][j] = 0.0f;			 /* clear U */
  temp = 0.0f;
  for (i = 0; i < 3; i++) {			 /* convert v1 to Cartesian */
    u1[i] = 0.0f;
    for (j = 0; j < 3; j++)
      u1[i] += (float) drvui->b_mat[i][j] * v1[j];
    temp += u1[i] * u1[i];
  }
  if (temp <= 0.0) {
    u1[2] = 1.0f;			 /* if null input, make u1 = 0,0,1 */
    U[2][2] = 1.0f;
  } else {
    temp = (float) sqrt (temp);
    for (i = 0; i < 3; i++) {			 /* make a unit vector */
      u1[i] /= temp;
      U[i][2] = u1[i];				 /* put column in U */
    }
  }
  for (i = 0; i < 3; i++) {			 /* convert v2 to Cartesian */
    u2[i] = 0.0f;
    for (j = 0; j < 3; j++)
      u2[i] += (float) drvui->b_mat[i][j] * v2[j];
  }
  u3[0] = u2[1] * u1[2] - u1[1] * u2[2];	 /* calculate u2 X u1 */
  u3[1] = u2[2] * u1[0] - u2[0] * u1[2];
  u3[2] = u2[0] * u1[1] - u1[0] * u2[1];
  temp = 0.0f;
  for (i = 0; i < 3; i++)
    temp += u3[i] * u3[i];
  if (temp <= 0.0) {
    u3[0] = 1.0f;			 /* if null input, make u3 = 1,0,0 */
    U[0][0] = 1.0f;
  } else {
    temp = (float) sqrt (temp);
    for (i = 0; i < 3; i++) {			 /* make a unit vector */
      u3[i] /= temp;
      U[i][0] = u3[i];				 /* add column to U */
    }
  }
  u2[0] = u1[1] * u3[2] - u3[1] * u1[2];	 /* calculate u2 = u1 X u3 */
  u2[1] = u1[2] * u3[0] - u1[0] * u3[2];
  u2[2] = u1[0] * u3[1] - u3[0] * u1[1];
  for (i = 0; i < 3; i++)
    U[i][1] = u2[i];
/* make determinant of U > 0 */
  if (determinant (U) < 0.0) {
    for (i = 0; i < 3; i++)
      U[i][0] = -U[i][0];
  }

/* The rotation required to get the desired orientation, which is described
    by G_Rot, has been set up so G_Rot * U = I. Thus G_Rot = U(inv),
    U is unitary and U(inv) = U(transpose). */

  for (i = 0; i < 3; i++)
    for (j = 0; j < 3; j++)
      G_Rot[i][j] = (float) U[j][i];
/* calculate rotations */
  temp = (float) (G_Rot[1][2] * G_Rot[1][2] + G_Rot[2][2] * G_Rot[2][2]);
  if (temp > 0.001) {
    float temp2;

    yrot = (float) (atan (G_Rot[0][2] / sqrt (temp)) * RAD);
    temp2 = (float) (G_Rot[1][2] / cos (yrot / RAD));
    if (temp2 > 1.0)
      temp2 = 1.0f;
    if (temp2 < -1.0)
      temp2 = -1.0f;
    xrot = -(float) asin (temp2) * (float) RAD;
    temp2 = -(float) (G_Rot[0][1] / cos (yrot / RAD));
    if (temp2 > 1.0)
      temp2 = 1.0f;
    if (temp2 < -1.0)
      temp2 = -1.0f;
    zrot = (float) asin (temp2) * (float) RAD;
  } else {
    yrot = -(float) asin (G_Rot[0][2]) * (float) RAD;
    xrot = 0.0f;				 /* XXXXXXXXXXX  - May not be right */
    zrot = 0.0f;				 /* XXXXXXXXXXX  - May not be right */
  }
}

/* ************************************************************** */
/* ************************************************************** */


void convert_ellipsoid (void)

/* routine to convert ellipsoid coefficients Bij or Uij to betaij form */
{
  int i, j;

  for (i = 1; i < n_ellips; ++i) {
    if (drvui->ell_type[i] != 0) {			 /* multiply Uij and Bij by a*(i)*a*(j)/4 */
      for (j = 0; j < 3; ++j)
        drvui->ellips[i][j] *= drvui->rec_lat_con[j] * drvui->rec_lat_con[j] * 0.25f;
      drvui->ellips[i][3] *= drvui->rec_lat_con[0] * drvui->rec_lat_con[1] * 0.25f;
      drvui->ellips[i][4] *= drvui->rec_lat_con[0] * drvui->rec_lat_con[2] * 0.25f;
      drvui->ellips[i][5] *= drvui->rec_lat_con[1] * drvui->rec_lat_con[2] * 0.25f;
      if (drvui->ell_type[i] > 0)			 /* multiply Uij by 8pi^2 */
        for (j = 0; j < 6; ++j)
	  drvui->ellips[i][j] *= 78.9568f;
    }
  }
}						 /* end of convert_ellipsoid */

/* ************************************************************** */
/* ************************************************************** */

void Conv_Sym_Mat (void)

/* routine to convert the rotational part of a symmetry matrix to a Cartesian rotation */
{
  double sum;
  float angle;
  int j, k, l;
  double mat[3][3];

  for (l = 0; l < drvui->ng; l++) {
    for (j = 0; j < 3; j++) {
      for (k = 0; k < 3; k++) {
        mat[j][k] = drvui->rss[l][j][k];
      }
    }
    if (determinant (mat) < 0.0) {		 /* if improper rotation, negate all elements */
      for (j = 0; j < 3; j++) {
        for (k = 0; k < 3; k++) {
          mat[j][k] *= -1.0f;
        }
      }
    }
    for (j = 0, sum = 0.0; j < 3; j++)		 /* compute sum of squares of elements */
      for (k = 0; k < 3; k++)
        sum += mat[j][k] * mat[j][k];
    if (fabs (sum - 3.0) > 0.1) {		 /* if sum = 3, rotation matrix OK */
/* at this point only 4 operators in trigonal or hexagonal system left */
/*    we need to find which we have */
      if (fabs (mat[0][0]) <= 0.1) {
        if (fabs (mat[0][1] - 1.0) < 0.0001) {
          angle = -1.0471976f;		 /* -60 degrees (in radians) */
        } else {
          angle = 2.0943951f;		 /* 120 degrees */
        }
      } else {
        if (fabs (mat[0][0] - 1.0) < 0.0001) {
          angle = 1.0471976f;		 /* 60 degrees (in radians) */
        } else {
          angle = -2.0943951f;		 /* -120 degrees */
        }
      }
      mat[0][0] = mat[1][1] = (float) cos (angle);
      mat[0][1] = (float) sin (angle);
      mat[1][0] = -mat[0][1];
    }
    if (determinant (mat) < 0.0) {		 /* if improper rotation, negate all elements */
      for (j = 0; j < 3; j++) {
        for (k = 0; k < 3; k++) {
          mat[j][k] *= -1.0f;
        }
      }
    }
    for (j = 0; j < 3; j++) {			 /* restore matrix */
      for (k = 0; k < 3; k++) {
        drvui->rss[l][j][k] = (float) mat[j][k];
      }
    }
  }
}

/* ************************************************************** */
/* ************************************************************** */

void delay (int i)

/* delays for 'i' seconds if DELAY is defined, otherwise it is a noop */
{
#ifdef DELAY
  time_t start, end;

  end = time (NULL);
  start = time (NULL);
  while (difftime (start, end) < (float) i)
    start = time (NULL);
#endif
}

/* ************************************************************** */
/* ************************************************************** */

float dist (int j, int k)

/* calculates distance between vertices at s_vert[j] and s_vert[k] */
{
  float d, t;
  int i;

  d = 0.0f;
  for (i = 0; i <= 2; ++i) {
    t = s_vert[3 * j + i] - s_vert[3 * k + i];
    d += t * t;
  }
  return ((float) sqrt (d));
}

/* ************************************************************** */
/* ************************************************************** */

void do_drawing (void)

/* routine to put together the elements of the drawing */
{
    int i;

    label_cell();
    if (drvui->frame_no == 1) 
        draw_cell (docell);	         /* generate unit-cell corners and draw it */
    nvert = 0;                       /* empty vertex list */
    generate_spheres ();             /* generate all spheres to be drawn within box */
    if (do_ellipsoids && n_ellips > 1 && drvui->frame_no == 1) {
        generate_ellipsoids ();		 /* now the ellipsoids */
    }
    if (drvui->nmag >0)
        generate_arrows();
    generate_bonds ();				 /* generate all bonds to be drawn within box */
    generate_cones ();				 /* generate all lonepairs to be drawn within box */
    generate_poly ();                /* generate all polyhedra to be drawn */
    generate_planes ();				 /* now the planes */
    generate_texts ();               /* any textual elements */
    if (slabmode > 1 && drvui->frame_no == 1)
        show_slab();                 /* show cutout section */
    if (ReadFourMap) {
      for (i = 1; i <= numOfFourierContours; ++i) {
/* generate & display a Fourier map */
        generate_map (FourierContourLevel[i], FourierContourSolid[i],
            FourierContourColor[i]);
      }
    }
}


/* ************************************************************** */
/* ************************************************************** */

void expand_atom (int natom)

/* routine to expand an atom position and obtain all equipoints within the unit cell */
/* natom - number of atom in list to expand */
{
  float xp[3], xpp[3];           /* user in generating position */
  int i, j, k, l;                /* loop variables */

  ncell = 0;                     /* clear list */
  for (i = 0; i < drvui->ng; ++i) {     /* symmetry operations */
    for (j = 0; j <= 2; ++j) {
      xp[j] = drvui->ts[i][j];
      for (k = 0; k <= 2; ++k)
        xp[j] = xp[j] + drvui->ss[i][j][k] * drvui->atom_xyz[natom][k];
    }
    for (k = 0; k < drvui->nlat; ++k) { /*lattice points */
      for (l = 0; l <= 2; ++l)
        xpp[l] = xp[l] + drvui->lat_pos[k][l];
      add_to_list (xpp, i);
      if (!drvui->acentric) {
        for (l = 0; l <= 2; ++l)
          xpp[l] = -xpp[l];
        add_to_list (xpp, i);
      }
    }
  }
}

/* ************************************************************** */
/* ************************************************************** */

int eigen (float *biso, float beta[6], float valu[3], float vect[3][3])

/* routine to calculate eigenvalues 'valu' and vectors 'v' for array of anisotropic
   thermal parameters 'beta', 'biso' is the equivalent isotropic B  */
/* routine copied from ORTEP */
{
  static double errnd = 0.000007;
  double b0, b1, b2, c0, c1;
  double a[3][3], b[3][3], w[3][3], u[3], v[3], z[3];
  double x, y, tem, sigma, vnew, vold, smax;
  double Vect[3][3];
  int i, j, l, ii, iii, i1, imax = 0;
  static double tpi2 = 19.7392088;		 /* 1/2 pi^2 */

#define phif(z) ((b2 - z) * z + b1) * z + b0

/* put beta into square symmetric matrix b (called M in ORTEP) */

  for (i = 0; i < 3; i++)
    b[i][i] = beta[i];
  b[0][1] = b[1][0] = beta[3];
  b[0][2] = b[2][0] = beta[4];
  b[1][2] = b[2][1] = beta[5];
/* multiply b * ginv to get w */
  for (i = 0; i < 3; i++) {
    for (j = 0; j < 3; j++) {
      w[i][j] = 0.0;
      for (l = 0; l < 3; l++) {
        w[i][j] += b[l][i] * drvui->ginv[j][l];
      }
    }
  }
  sigma = 0.0;
  for (i = 0; i < 3; i++) {
    for (j = 0; j < 3; j++) {
      a[i][j] = w[i][j];
      sigma += a[i][j] * a[i][j];
    }
  }
  if (sigma <= 0.0)
    return (0);					 /* error on null 'w' */
  sigma = sqrt (sigma);
/* get coefficients of third-order characteristic equation 
        Equation is -L^3 + b2 L^2 + b1 L + b0 = 0
*/
  b2 =   a[0][0] + a[1][1] + a[2][2];
  b1 = - a[0][0] * a[1][1] - a[0][0] * a[2][2] - a[1][1] * a[2][2]
       + a[0][2] * a[2][0] + a[0][1] * a[1][0] + a[1][2] * a[2][1];
  b0 =   a[0][0] * a[1][1] * a[2][2] + a[1][0] * a[2][1] * a[0][2]
	   + a[2][0] * a[1][2] * a[0][1] - a[2][0] * a[0][2] * a[1][1]
	   - a[0][0] * a[2][1] * a[1][2] - a[1][0] * a[0][1] * a[2][2];

  *biso = (float) (4.0 * b2 / 3.0);

/* first root by bisection */

  x = 0.0;
  y = sigma;
  tem = phif (sigma);
  vnew = 0.0;
  if (b0 != 0.0) {
    if ((b0 < 0.0) && (tem <= 0.0))
      y = -y;
    if (b0 > 0.0) {
      y = 0.0;
      x = sigma;
      if (tem > 0.0)
        x = -x;
    }

/* phif(y) > 0 and phif(x) < 0  */

    for (j = 0; j < 80; ++j) {
      if ((tem = phif (vnew)) == 0.0)
        break;
      if (tem < 0.0) {
        x = vnew;
      } else
        y = vnew;
      vold = vnew;
      vnew = (x + y) * 0.5;
      tem = fabs (vold - vnew);
      if (tem <= errnd)
        break;
      if (vold != 0.0) {
        if (fabs (tem / vold) <= errnd)
          break;
      }
    }
    if (j == 80) {
      fprintf (drvui->flout, "No convergence.\n");
      return (0);
    }
  }
  u[2] = vnew;
  c1 = b2 - vnew;
  c0 = b1 + c1 * vnew;
  tem = c1 * c1 + 4.0 * c0;
  if (tem < -0.0001) {
    fprintf (drvui->flout, "Complex Roots! tem = %f\n", tem);
    return (0);					 /* error if complex roots */
  }
  if (tem < 0.0)
    tem = 0.0;
  tem = sqrt (tem);
  u[0] = 0.5 * (c1 - tem);
  u[1] = 0.5 * (c1 + tem);
  for (j = 0; j < 2; j++) {			 /* sort roots in increasing order */
    if (u[j] > u[2]) {
      tem = u[2];
      u[2] = u[j];
      u[j] = tem;
    }
  }
  if (u[0] < 0.0) {
    fprintf (drvui->flout, "Imaginary Root, squares are: %8.5f %8.5f %8.5f\n",
	     u[0], u[1], u[2]);
    return (0);
  }

/* count multiple roots */
  tem = 100.0 * errnd;
  j = 0;
  if ((u[1] - u[0] - tem) < 0)
    j++;
  if ((u[2] - u[1] - tem) < 0)
    j += 2;
  if (j == 3) {
/* 3 equal roots */
    for (ii = 0; ii < 3; ii++) {
      for (i = 0; i < 3; i++)
        Vect[i][ii] = 0.0;
      Vect[ii][ii] = 1.0;
      valu[ii] = (float) sqrt (u[ii] / tpi2);
    }
    return (1);
  }
  for (ii = 0; ii < 3; ii++) {
    for (iii = 0; iii < 3; iii++) {
      for (i = 0; i < 3; i++)
        a[i][iii] = w[i][iii];
      a[iii][iii] = w[iii][iii] - u[ii];
    }
    smax = 0.0;
    for (i = 0; i < 3; i++) {
      i1 = (i < 2) ? i + 1 : 0;
      b[0][i] = a[1][i] * a[2][i1] - a[2][i] * a[1][i1];
      b[1][i] = a[2][i] * a[0][i1] - a[0][i] * a[2][i1];
      b[2][i] = a[0][i] * a[1][i1] - a[1][i] * a[0][i1];
      tem = b[0][i] * b[0][i] + b[1][i] * b[1][i] + b[2][i] * b[2][i];
      if (tem > smax) {
        smax = tem;
        imax = i;
      }
    }
    if (smax <= 0.0)
      return (0);
    smax = (float) sqrt (smax);
    for (i = 0; i < 3; i++)
      v[i] = b[i][imax] / smax;
    axeqb (a, z, v);
/* convert direct-space vector z into Cartesian system */
    for (i = 0; i < 3; i++) {
      v[i] = 0.0;
      for (j = 0; j < 3; j++)
        v[i] += drvui->b_mat[i][j] * z[j];
    }
    tem = v[0] * v[0] + v[1] * v[1] + v[2] * v[2];
    tem = (float) sqrt (tem);
    for (i = 0; i < 3; i++)
      Vect[ii][i] = v[i] / tem;			 /* unit vector */
  }
  for (i = 0; i < 3; i++)
    valu[i] = (float) sqrt (u[i] / tpi2);
/* look for degenerate vectors */
  if (fabs (Vect[0][0] * Vect[1][0] + Vect[0][1] * Vect[1][1]
	    + Vect[0][2] * Vect[1][2]) > 0.001) {	/* set #1 = 2 x 3 */
    Vect[0][0] = Vect[1][1] * Vect[2][2] - Vect[2][1] * Vect[1][2];
    Vect[0][1] = Vect[2][0] * Vect[1][2] - Vect[1][0] * Vect[2][2];
    Vect[0][2] = Vect[1][0] * Vect[2][1] - Vect[2][0] * Vect[1][1];
  }
  if (fabs (Vect[2][0] * Vect[1][0] + Vect[2][1] * Vect[1][1]
	    + Vect[2][2] * Vect[1][2]) > 0.001) {	/* set #3 = 1 x 2 */
    Vect[2][0] = Vect[0][1] * Vect[1][2] - Vect[1][1] * Vect[0][2];
    Vect[2][1] = Vect[1][0] * Vect[0][2] - Vect[0][0] * Vect[1][2];
    Vect[2][2] = Vect[0][0] * Vect[1][1] - Vect[1][0] * Vect[0][1];
  }
  if (determinant (Vect) < 0.0) {
    for (i = 0; i < 3; i++)
      for (j = 0; j < 3; j++)
        Vect[i][j] *= -1.0;
  }
  for (i = 0; i < 3; i++)
    for (j = 0; j < 3; j++)
      vect[i][j] = (float) Vect[i][j];
  return (1);
}

/* ************************************************************** */
/* ************************************************************** */

void find_all_in_box (int i)

/* routine to locate all atoms of type 'i' in the display box */
/* i - atom number */
{
  int n, j, k, l;				 /* loop indices */
  int sym_no;					 /* symmetry number */

  for (n = 1; n < NvertM; n++) {		 /* loop through the Master list */
    j = (int) drvui->atom_no[n] / 100;			 /* extract atom number */
    if (j == i) {
      drvui->orig_atom_no[nvert] = i;
      sym_no = drvui->atom_no[n] - 100 * j;		 /* symmetry operator number */
      for (k = 0; k < 3; k++) {
        o_vert[3 * nvert + k] = xypos[3 * n + k];
        s_vert[3 * nvert + k] = 0.0f;
        for (l = 0; l < 3; l++) {		 /* calculate cartesian coordinates */
          s_vert[3 * nvert + k] +=
                 (float) drvui->b_mat[k][l] * (xypos[3 * n + l] - origin[l]);
        }
      }
      vert_sym_no[nvert++] = sym_no;
      if (nvert > 2 * NvertM + 10) {
        vert_sym_no=realloc(vert_sym_no, (nvert+10)*sizeof(int));
        drvui->orig_atom_no=realloc(drvui->orig_atom_no, (nvert+10)*sizeof(int));
      }
      if (nvert > MAX_VERTS) {
        fprintf (stdout,"Too many vertices for dimensions - increase MAX_VERTS.\n");
        delay (5);
        exit (0);
      }
    }
  }
}						 /* end of find_all_in_box */

/* ************************************************************** */
/* ************************************************************** */
void generate_bonds (void)

/* routine to generate bond descriptions and add then to edit list */
{
  int start, same_type;
  int i, j, k;        			 /* loop counters */
  int N_Bond;        			 /* Counter for number of bonds of a type */
  int l1;        			 /* loop counter for matrix operations */
  int nvert1;        			 /* vertex counter */
  int Bond_Count;        		 /* Counter for bonds output */
  char bnd1[4], bnd2[4];        	 /* temporary storage of bond labels */
  float d;        			 /* bond distance */
  float df[3];        			 /* vector difference */
  float beta;        			 /* rotation angle */
  float gamma;        			 /* another rotation angle */
  float at1[3], at2[3];        		 /*initial coordinates of bond endpoints */
  float clip0, clip1;        		 /* clipping scale factors */
  int m, l2, l3, out = 0;
  int ell[2];        			 /* save ellipsoid numbers */

  int is_ellipsoid[2];        		 /* ellipsoid markers - for moving bond endpoint */
  int o1, p, p1, q, r;        		 /* loop variables for rotation matrix calculation */
  float elrot0[3][3];        		 /* rotation matrices of ellipsoids */
  float Z0[3], Z1[3];        		 /* components of rms displacement along bond */
  int *ellips_id;
  float factor, rms0, rms1;
#ifdef OPENGL
  float glr, glg, glb;
  GLUquadricObj *glu_quadObj;
#endif
  int omit;

  if (drvui->nbond == 1)
    return;        			 /* exit if no work to be done */

  Bond_Count = 0;
  if (!(ellips_id = (int *) malloc ((unsigned) ((long) (MAX_VERTS * sizeof (int)))))) {
    fprintf(drvui->flout,"Unable to get ellips_id allocation!");
    exit(0);
  }
  for (i = 1; i < drvui->nbond; ++i) {        	 /* loop through bond types */
    if (drvui->bond_fn[i] != drvui->frame_no) continue;  /* skip if not in this frame */
    nvert = 0;        			         /* empty the vertex list */
    for (j = 0; j < natom; j++) {        /* loop through atom types */
      int o = drvui->atom_n[j];
      if (((o & 255) > 0) | (((o >> 24) & 255) > 0)) { /* sphere or ellipsoid */

        o = (drvui->atom_n[j] >> 24) & 255;
        if (o > 0) {                      /* ellipsoid */
          if ((drvui->bond_l1[i][0] == drvui->ellips_l[o][0]) &&
              (drvui->bond_l1[i][1] == drvui->ellips_l[o][1]) &&
              (drvui->bond_l1[i][2] == drvui->ellips_l[o][2]) &&
              (drvui->bond_l1[i][3] == drvui->ellips_l[o][3])) {
            if (do_ellipsoids == 1 && El_Cutout == 1) {    /* skip if B iso */
              is_ellipsoid[0] = 1;         /* or no cutout */
              ell[0] = o;
            }
            o = nvert;
            find_all_in_box (j);          /* get all atoms of type j */
            for (p = o; p < nvert; p++)
              ellips_id[p] = ell[0];      /* save ellipsoid id for these entries */
          }
        } else {                          /* sphere */
          o = drvui->atom_n[j] & 255;     /* number of sphere */
          is_ellipsoid[0] = 0;
          if ((drvui->bond_l1[i][0] == drvui->sphere_l[o][0]) &&
              (drvui->bond_l1[i][1] == drvui->sphere_l[o][1]) &&
              (drvui->bond_l1[i][2] == drvui->sphere_l[o][2]) &&
              (drvui->bond_l1[i][3] == drvui->sphere_l[o][3])) {
            find_all_in_box (j);         /* get all atoms of type j */
          }
        }        			 /* o > 0 */
      }
    }        				 /* for (j=0;j<natom;++j) */

    bnd1[0] = drvui->bond_l1[i][0];
    bnd1[1] = drvui->bond_l1[i][1];
    bnd1[2] = drvui->bond_l1[i][2];
    bnd1[3] = drvui->bond_l1[i][3];
    nvert1 = nvert;
    if ((drvui->bond_l1[i][0] == drvui->bond_l2[i][0]) &&
        (drvui->bond_l1[i][1] == drvui->bond_l2[i][1]) &&
        (drvui->bond_l1[i][2] == drvui->bond_l2[i][2]) &&
        (drvui->bond_l1[i][3] == drvui->bond_l2[i][3])) {
      bnd2[0] = drvui->bond_l1[i][0];        	 /* atoms are same type */
      bnd2[1] = drvui->bond_l1[i][1];
      bnd2[2] = drvui->bond_l1[i][2];
      bnd2[3] = drvui->bond_l1[i][3];
      is_ellipsoid[1] = is_ellipsoid[0];
      ell[1] = ell[0];
      same_type = 1;
    } else {
      bnd2[0] = drvui->bond_l2[i][0];
      bnd2[1] = drvui->bond_l2[i][1];
      bnd2[2] = drvui->bond_l2[i][2];
      bnd2[3] = drvui->bond_l2[i][3];
      same_type = 0;
/* atoms i and j are not the same type */
      for (j = 0; j < natom; ++j) {         /* loop through atom types */
        if (((drvui->atom_n[j] & 255) > 0) | (((drvui->atom_n[j] >> 24) & 255) > 0)) { /* sphere or ellipsoid */
          int o;

          o = (drvui->atom_n[j] >> 24) & 255;
          if (o > 0) {                      /* ellipsoid */
            if ((drvui->bond_l2[i][0] == drvui->ellips_l[o][0]) &&
                (drvui->bond_l2[i][1] == drvui->ellips_l[o][1]) &&
                (drvui->bond_l2[i][2] == drvui->ellips_l[o][2]) &&
                (drvui->bond_l2[i][3] == drvui->ellips_l[o][3])) {
              if (do_ellipsoids == 1 && El_Cutout == 1) {    /* skip if B iso */
                is_ellipsoid[1] = 1;         /* or no cutout */
                ell[1] = o;
              }
              o = nvert;
              find_all_in_box (j);         /* get all atoms of type j */
              for (p = o; p < nvert; p++)
                ellips_id[p] = ell[1];         /* save ellipsoid type */
            }
          } else {
            o = drvui->atom_n[j] & 255;               /* sphere number */
            if ((drvui->bond_l2[i][0] == drvui->sphere_l[o][0]) &&
                (drvui->bond_l2[i][1] == drvui->sphere_l[o][1]) &&
                (drvui->bond_l2[i][2] == drvui->sphere_l[o][2]) &&
                (drvui->bond_l2[i][3] == drvui->sphere_l[o][3])) {
              find_all_in_box (j);         /* get all atoms of type j */
              is_ellipsoid[1] = 0;
            }
          }        			 /* o > 0 */
        }
      }        				 /* for (j=0;j<natom;++j) */
    }        				 /* if atoms are same type */

    if (((nvert1 > 0) && (nvert > nvert1)) || (same_type == 1)) {
      N_Bond = 0;
      for (j = 0; j < nvert1; ++j) {         /* loop through atoms of type 1 */
        start = nvert1;
        if (same_type == 1)
          start = j;
        for (k = start; k < nvert; ++k) {     /* loop on type 2 atoms  */
          d = dist (j, k);        	 /* get atomic distance */
          if ((d >= drvui->bond_min[i]) && (d <= drvui->bond_max[i])) {    /* here if distance within range */
            df[0] = df[1] = drvui->Bond_Mult * drvui->bond_size[i];    /* scaling */
            df[2] = 0.5f *d;

            Bond_Count++;        	 /* update count */
            for (p = 0; p < 3; p++) {         /* initialize endpoints of bond */
              Z0[p] = s_vert[3 * j + p];
              Z1[p] = s_vert[3 * k + p];
            }

            if (do_ellipsoids == 1 && El_Cutout == 1) {

/* Elipsoid has octant cutout, calculate rms displacement along bond  */

              factor = (float) (Ellipsoid_Scale * 0.9f / (PI * d * 1.414));    /* calculate multiplier */

/* Mean-square displacement of a scaled thermal ellipsoid along the direction of a bond is given by: */
/*                                                                                      */
/*      D = factor^2 * (X'G'R'BRGX)                                                     */
/*                                                                                      */
/*  where X is the interatomic vector given in triclinic fractional coordinates,        */
/*        G is the lattice metric with the ij'th element given by a[i].a[j],            */
/*        R is the rotational part of the symmetry operation that transform atom i,     */
/*        B is a square matrix of the betaij thermal factors,                           */
/*    and a prime indicates the transpose of the matrix.                                */
/*                                                                                      */
              if (is_ellipsoid[0] == 1) {     /* at one end of bond */
                float B[3][3], BP[3][3];

                o1 = vert_sym_no[j];
                p1 = ellips_id[j];         /* have original ellipsoid number */
                for (p = 0; p < 3; p++)
                  B[p][p] = drvui->ellips[p1][p];   /* extract betaij's into square matrix */
                B[0][1] = B[1][0] = drvui->ellips[p1][3];
                B[0][2] = B[2][0] = drvui->ellips[p1][4];
                B[1][2] = B[2][1] = drvui->ellips[p1][5];
                for (p = 0; p < 3; p++) {
                  df[p] = 0.0;
                  for (q = 0; q < 3; ++q) {
                    elrot0[p][q] = 0.0;
                    df[p] += drvui->ginv[q][p] * (o_vert[3 * k + q] - o_vert[3 * j + q]);
                    for (r = 0; r < 3; r++)
                      elrot0[p][q] += drvui->rss[o1][r][p] * B[r][q];
                  }
                }
                for (p = 0; p < 3; p++) {
                  for (q = 0; q < 3; q++) {
                    BP[p][q] = 0.0;
                    for (r = 0; r < 3; r++)
                      BP[p][q] += (float) elrot0[p][r] * drvui->rss[o1][r][q];    /* BP is rotated betij matrix */
                  }
                }
                for (p = 0; p < 3; p++) {
                  Z0[p] = 0.0;
                  for (q = 0; q < 3; q++)
                    Z0[p] += df[q] * BP[q][p];
                }
/* rms0 is the scaled rms displacement value of the ellipsoid in the direction of the bond */
                rms0 = factor * (float) sqrt (Z0[0] * df[0] +
                          Z0[1] * df[1] + Z0[2] * df[2]);
/* calculate components of end point at intersection of ellipsoid and bond */
                for (q = 0; q < 3; q++)
                  Z0[q] = s_vert[3 * j + q] + (s_vert[3 * k + q] - s_vert[3 * j + q]) * rms0 / d;
              }        			 /* if is_ellipsoid[0] */

              if (is_ellipsoid[1] == 1) {     /* ellipsoid at other end of bond */
                float B[3][3], BP[3][3];

                o1 = vert_sym_no[k];
                p1 = ellips_id[k];         /* have original ellipsoid number */
                for (p = 0; p < 3; p++)
                  B[p][p] = drvui->ellips[p1][p];     /* extract betaij's into square matrix */
                B[0][1] = B[1][0] = drvui->ellips[p1][3];
                B[0][2] = B[2][0] = drvui->ellips[p1][4];
                B[1][2] = B[2][1] = drvui->ellips[p1][5];
                for (p = 0; p < 3; p++) {
                  df[p] = 0.0;
                  for (q = 0; q < 3; q++) {
                    elrot0[p][q] = 0.0;
                    df[p] += drvui->ginv[q][p] * (o_vert[3 * k + q] - o_vert[3 * j + q]);
                    for (r = 0; r < 3; ++r)
                      elrot0[p][q] += drvui->rss[o1][r][p] * B[r][q];
                  }
                }
                for (p = 0; p < 3; p++) {
                  for (q = 0; q < 3; q++) {
                    BP[p][q] = 0.0;
                    for (r = 0; r < 3; r++)
                      BP[p][q] += (float) elrot0[p][r] * drvui->rss[o1][r][q];    /* BP is rotated betij matrix */
                  }
                }

                for (p = 0; p < 3; p++) {
                  Z1[p] = 0.0;
                  for (q = 0; q < 3; q++)
                    Z1[p] += df[q] * BP[q][p];
                }
/* rms1 is the scaled rms displacement value of the ellipsoid in the direction of the bond */

                rms1 = factor * (float) sqrt (Z1[0] * df[0] +
        			      Z1[1] * df[1] + Z1[2] * df[2]);

/* calculate components of end point at intersection of ellipsoid and bond */
                for (q = 0; q < 3; q++)
                  Z1[q] = s_vert[3 * k + q] - (s_vert[3 * k + q]
                      - s_vert[3 * j + q]) * rms1 / d;
              }        			 /* if is_ellipsoid[1] */
            }        			 /* if any segmented ellipsoids present */

            for (l1 = 0; l1 < 3; l1++) {
              at1[l1] = (float) Z0[l1];
              at2[l1] = (float) Z1[l1];
              df[l1] = (float) (Z1[l1] - Z0[l1]);
            }

            if (clipflag != 0) {
              out = 0;
              l2 = 0;
              l3 = 0;
              clip0 = 0.0f;
              clip1 = 1.0f;
              for (l1 = 0; l1 < 3; l1++) {
                if (o_vert[3 * j + l1] < drvui->clip_lim[drvui->frame_no][l1] - 0.01
                      || o_vert[3 * j + l1] > drvui->clip_lim[drvui->frame_no][l1 + 3] + 0.01)
                  l2 = 1;
                if (o_vert[3 * k + l1] < drvui->clip_lim[drvui->frame_no][l1] - 0.01
                      || o_vert[3 * k + l1] > drvui->clip_lim[drvui->frame_no][l1 + 3] + 0.01)
                  l3 = 1;
              }
              if (l2 && l3)
                out = 1;        	 /* both atoms outside */
              if (l3)
                clip1 = 0.5f;
              if (l2)
                clip0 = 0.5f;

              at2[0] = at1[0] + clip1 * df[0];
              at2[1] = at1[1] + clip1 * df[1];
              at2[2] = at1[2] + clip1 * df[2];

              at1[0] = at1[0] + clip0 * df[0];
              at1[1] = at1[1] + clip0 * df[1];
              at1[2] = at1[2] + clip0 * df[2];
            }        			 /* if clipflag */

            omit = 0;
            for (l1 = 0; l1 < nomits; l1++) {
              if (omit1[l1]== i*1000 && omit2[l1] == Bond_Count) omit = 1;
            }

            if ((clipflag == 0 || out == 0) && omit == 0) {
              fprintf (drvui->fpoutp," /* Bond: %c%c %8.5f %8.5f %8.5f TO %c%c %8.5f %8.5f %8.5f*/\n",
                 bnd1[0], bnd1[1], o_vert[3 * j],o_vert[3 * j + 1], o_vert[3 * j + 2],
                 bnd2[0], bnd2[1], o_vert[3 * k],o_vert[3 * k + 1], o_vert[3 * k + 2]);
              if (drvui->bond_style[i] == 0) {
                fprintf (drvui->fpoutp," cylinder{<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>, %8.5f \n",
                   at1[0], at1[1], at1[2], at2[0], at2[1],at2[2], drvui->Bond_Mult * drvui->bond_size[i]);
                fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n", drvui->col_bond_p[i]);
                fprintf(drvui->fpoutp," }\n");
              } else {
                for (m = 0; m < 3; m++)
                  df[m] = (float) (at2[m] - at1[m]);
                fprintf (drvui->fpoutp," cylinder{<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>, %8.5f \n",
                   at1[0], at1[1], at1[2], at1[0]+df[0]/5.,at1[1]+df[1]/5.,at1[2]+df[2]/5., drvui->Bond_Mult * drvui->bond_size[i]);
                fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n", drvui->col_bond_p[i]);
                fprintf(drvui->fpoutp," }\n");
                fprintf (drvui->fpoutp," cylinder{<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>, %8.5f \n",
                   at1[0]+df[0]*2./5., at1[1]+df[1]*2./5, at1[2]+df[2]*2./5., 
                   at1[0]+df[0]*3./5.,at1[1]+df[1]*3./5.,at1[2]+df[2]*3/5., 
                   drvui->Bond_Mult * drvui->bond_size[i]);
                fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n", 
                    drvui->col_bond_p[i]);
                fprintf(drvui->fpoutp," }\n");
                fprintf (drvui->fpoutp," cylinder{<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>, %8.5f \n",
                   at1[0]+df[0]*4./5., at1[1]+df[1]*4./5, at1[2]+df[2]*4./5., 
                   at2[0],at2[1],at2[2], drvui->Bond_Mult * drvui->bond_size[i]);
                fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n", drvui->col_bond_p[i]);
                fprintf(drvui->fpoutp," }\n");
              }
              N_Bond++;
              if (no_comment == 0)
                fprintf (drvui->fpoutv,"# Bond: %c%c %8.5f %8.5f %8.5f TO %c%c %8.5f %8.5f %8.5f\n",
        	  bnd1[0], bnd1[1], o_vert[3 * j],o_vert[3 * j + 1], o_vert[3 * j + 2],
        	  bnd2[0], bnd2[1], o_vert[3 * k],o_vert[3 * k + 1], o_vert[3 * k + 2]);
              for (m = 0; m < 3; m++)
                df[m] = (float) (at2[m] - at1[m]);
              d = (float) sqrt (df[0] * df[0] + df[1] * df[1] +    df[2] * df[2]);
/*
   calculate rotation matrices needed to rotate bond about Z onto Y-Z plane and then about X so that
   bond is parallel to Z.  Finally, apply inverse transformation to cylinder, which started parallel
   to X.  Resulting cylinder will be parallel to vector between atoms.
*/
              beta = (float) atan2 (df[0], df[1] + 0.0000001);    /* rotation angle about Z (in radians) */
              gamma = (float) sqrt (df[0] * df[0] + df[1] * df[1]);
              gamma = (float) atan2 (gamma, df[2]);    /* Rotation angle about X */
#ifdef OPENGL
              if (gl_view) {
                glu_quadObj = gluNewQuadric ();
                glPushMatrix ();
                glLoadName(i*1000);
                glPushName(Bond_Count);
                (void)sscanf (drvui->col_bond_v[i], "%f %f %f", &glr, &glg, &glb);
                glColor3f (glr, glg, glb);
                if (drvui->bond_style[i] == 0) {
                  glPushMatrix ();
                  glTranslatef (at1[0], at1[1], at1[2]);
                  glRotatef (-beta * (float) RAD, 0.0f, 0.0f, 1.);
                  glRotatef (-gamma * (float) RAD, 1., 0.0f, 0.0f);
                  gluQuadricDrawStyle (glu_quadObj, GLU_FILL);
                  gluCylinder (glu_quadObj,drvui->Bond_Mult * drvui->bond_size[i],
                         drvui->Bond_Mult * drvui->bond_size[i], d, 10, 1);
                  glPopMatrix ();
                } else {
                  glPushMatrix ();
                  glTranslatef (at1[0], at1[1], at1[2]);
                  glRotatef (-beta * (float) RAD, 0.0f, 0.0f, 1.);
                  glRotatef (-gamma * (float) RAD, 1., 0.0f, 0.0f);
                  gluQuadricDrawStyle (glu_quadObj, GLU_FILL);
                  gluCylinder (glu_quadObj,drvui->Bond_Mult * drvui->bond_size[i],
                           drvui->Bond_Mult * drvui->bond_size[i], d/5., 10, 1);
                  glPopMatrix ();
                  glPushMatrix ();
                  glTranslatef (at1[0]+df[0]*0.4f, at1[1]+df[1]*0.4f,at1[2]+df[2]*0.4f);
                  glRotatef (-beta * (float) RAD, 0.0f, 0.0f, 1.);
                  glRotatef (-gamma * (float) RAD, 1., 0.0f, 0.0f);
                  gluQuadricDrawStyle (glu_quadObj, GLU_FILL);
                  gluCylinder (glu_quadObj,drvui->Bond_Mult * drvui->bond_size[i],
                           drvui->Bond_Mult * drvui->bond_size[i], d/5., 10, 1);
                  glPopMatrix ();
                  glPushMatrix ();
                  glTranslatef (at1[0]+df[0]*0.8f,at1[1]+df[1]*0.8f,at1[2]+df[2]*0.8f);
                  glRotatef (-beta * (float) RAD, 0.0f, 0.0f, 1.);
                  glRotatef (-gamma * (float) RAD, 1., 0.0f, 0.0f);
                  gluQuadricDrawStyle (glu_quadObj, GLU_FILL);
                  gluCylinder (glu_quadObj,drvui->Bond_Mult * drvui->bond_size[i],
                           drvui->Bond_Mult * drvui->bond_size[i], d/5., 10, 1);
                  glPopMatrix ();
                }
                glPopName();
                glPopMatrix ();               /* SPECIAL */
#if !defined (WIN32) && !defined (__APPLE__)
                free(glu_quadObj);
#endif
              }
#endif                /* ifdef OPENGL */
              if (Vrml2) {
                if (drvui->bond_style[i] == 0) {
                  fprintf (drvui->fpoutv, "  Transform{\n ");
                  fprintf (drvui->fpoutv,"   translation %5.3f %5.3f %5.3f\n",
                    (at1[0] + at2[0]) / 2.,(at1[1] + at2[1]) / 2., (at1[2] + at2[2]) / 2.);
                  fprintf (drvui->fpoutv, "   rotation 0 0 1 %f\n", -beta);
                  fprintf (drvui->fpoutv,"   children Transform { rotation 1 0 0 %f\n",1.5708 - gamma);
                  fprintf (drvui->fpoutv, " children [ Shape {\n");
                  fprintf (drvui->fpoutv," appearance Appearance { material Material {   diffuseColor %s} }\n",
                            drvui->col_bond_v[i]);
                  fprintf (drvui->fpoutv, "  geometry Cylinder {");
                  fprintf (drvui->fpoutv, "radius %f", drvui->Bond_Mult * drvui->bond_size[i]);
                  fprintf (drvui->fpoutv, " height %f", d);
                  fprintf (drvui->fpoutv, "}}]}}\n");
              } else {
                  fprintf (drvui->fpoutv, "  Transform{\n ");
                  fprintf (drvui->fpoutv,"   translation %5.3f %5.3f %5.3f\n",
                      at1[0] + df[0] / 5.,at1[1] + df[1] / 5., at1[2] + df[2] / 5.);
                  fprintf (drvui->fpoutv, "   rotation 0 0 1 %f\n", -beta);
                  fprintf (drvui->fpoutv,"   children Transform { rotation 1 0 0 %f\n",1.5708 - gamma);
                  fprintf (drvui->fpoutv, " children [ Shape {\n");
                  fprintf (drvui->fpoutv," appearance Appearance { material Material {   diffuseColor %s} }\n",
                      drvui->col_bond_v[i]);
                  fprintf (drvui->fpoutv, "  geometry Cylinder {");
                  fprintf (drvui->fpoutv, "radius %f", drvui->Bond_Mult * drvui->bond_size[i]);
                  fprintf (drvui->fpoutv, " height %f", d/5.);
                  fprintf (drvui->fpoutv, "}}]}}\n");
                  fprintf (drvui->fpoutv, "  Transform{\n ");
                  fprintf (drvui->fpoutv,"   translation %5.3f %5.3f %5.3f\n",
                      (at1[0] + at2[0]) / 2.,(at1[1] + at2[1]) / 2., (at1[2] + at2[2]) / 2.);
                  fprintf (drvui->fpoutv, "   rotation 0 0 1 %f\n", -beta);
                  fprintf (drvui->fpoutv,"   children Transform { rotation 1 0 0 %f\n",1.5708 - gamma);
                  fprintf (drvui->fpoutv, " children [ Shape {\n");
                  fprintf (drvui->fpoutv," appearance Appearance { material Material {   diffuseColor %s} }\n",
                      drvui->col_bond_v[i]);
                  fprintf (drvui->fpoutv, "  geometry Cylinder {");
                  fprintf (drvui->fpoutv, "radius %f", drvui->Bond_Mult * drvui->bond_size[i]);
                  fprintf (drvui->fpoutv, " height %f", d/5.);
                  fprintf (drvui->fpoutv, "}}]}}\n");
                  fprintf (drvui->fpoutv, "  Transform{\n ");
                  fprintf (drvui->fpoutv,"   translation %5.3f %5.3f %5.3f\n",
                      at1[0] + df[0] *4. / 5., at1[1] + df[1] * 4. / 5., at1[2] + df[2] *4. / 5.);
                  fprintf (drvui->fpoutv, "   rotation 0 0 1 %f\n", -beta);
                  fprintf (drvui->fpoutv,"   children Transform { rotation 1 0 0 %f\n",1.5708 - gamma);
                  fprintf (drvui->fpoutv, " children [ Shape {\n");
                  fprintf (drvui->fpoutv," appearance Appearance { material Material {   diffuseColor %s} }\n",
                      drvui->col_bond_v[i]);
                  fprintf (drvui->fpoutv, "  geometry Cylinder {");
                  fprintf (drvui->fpoutv, "radius %f", drvui->Bond_Mult * drvui->bond_size[i]);
                  fprintf (drvui->fpoutv, " height %f", d/5.);
                  fprintf (drvui->fpoutv, "}}]}}\n");
                }
              } else {
                if (drvui->bond_style[i] == 0) {
                  fprintf (drvui->fpoutv, " Separator{\n");
                  fprintf (drvui->fpoutv, "  Transform{\n ");
                  fprintf (drvui->fpoutv,"   translation %5.3f %5.3f %5.3f\n  }\n",
        	     (at1[0] + at2[0]) / 2.,(at1[1] + at2[1]) / 2., (at1[2] + at2[2]) / 2.);
                  fprintf (drvui->fpoutv, "  Rotation {\n");
                  fprintf (drvui->fpoutv, "   rotation 0 0 1 %f\n  }\n", -beta);
                  fprintf (drvui->fpoutv, "  Rotation {\n");
                  fprintf (drvui->fpoutv, "   rotation 1 0 0 %f\n  }\n", 1.5708 - gamma);
                  fprintf (drvui->fpoutv,"  Material {\n   diffuseColor %s\n", drvui->col_bond_v[i]);
                  fprintf (drvui->fpoutv, "  }\n");
                  fprintf (drvui->fpoutv, "  Cylinder {\n");
                  fprintf (drvui->fpoutv, "   parts SIDES\n");
                  fprintf (drvui->fpoutv, "   radius %f\n", drvui->Bond_Mult * drvui->bond_size[i]);
                  fprintf (drvui->fpoutv, "   height %f\n", d);
                  fprintf (drvui->fpoutv, "  }\n }\n");
                } else {
                  fprintf (drvui->fpoutv, " Separator{\n");
                  fprintf (drvui->fpoutv, "  Transform{\n ");
                  fprintf (drvui->fpoutv,"   translation %5.3f %5.3f %5.3f\n  }\n",
                       at1[0] + df[0] / 5., at1[1] + df[1] / 5., at1[2] + df[2] / 5.);
                  fprintf (drvui->fpoutv, "  Rotation {\n");
                  fprintf (drvui->fpoutv, "   rotation 0 0 1 %f\n  }\n", -beta);
                  fprintf (drvui->fpoutv, "  Rotation {\n");
                  fprintf (drvui->fpoutv, "   rotation 1 0 0 %f\n  }\n", 1.5708 - gamma);
                  fprintf (drvui->fpoutv,"  Material {\n   diffuseColor %s\n", drvui->col_bond_v[i]);
                  fprintf (drvui->fpoutv, "  }\n");
                  fprintf (drvui->fpoutv, "  Cylinder {\n");
                  fprintf (drvui->fpoutv, "   parts SIDES\n");
                  fprintf (drvui->fpoutv, "   radius %f\n", drvui->Bond_Mult * drvui->bond_size[i]);
                  fprintf (drvui->fpoutv, "   height %f\n", d/5.);
                  fprintf (drvui->fpoutv, "  }\n }\n");
                  fprintf (drvui->fpoutv, " Separator{\n");
                  fprintf (drvui->fpoutv, "  Transform{\n ");
                  fprintf (drvui->fpoutv,"   translation %5.3f %5.3f %5.3f\n  }\n",
                       (at1[0] + at2[0]) / 2.,(at1[1] + at2[1]) / 2., (at1[2] + at2[2]) / 2.);
                  fprintf (drvui->fpoutv, "  Rotation {\n");
                  fprintf (drvui->fpoutv, "   rotation 0 0 1 %f\n  }\n", -beta);
                  fprintf (drvui->fpoutv, "  Rotation {\n");
                  fprintf (drvui->fpoutv, "   rotation 1 0 0 %f\n  }\n", 1.5708 - gamma);
                  fprintf (drvui->fpoutv,"  Material {\n   diffuseColor %s\n", drvui->col_bond_v[i]);
                  fprintf (drvui->fpoutv, "  }\n");
                  fprintf (drvui->fpoutv, "  Cylinder {\n");
                  fprintf (drvui->fpoutv, "   parts SIDES\n");
                  fprintf (drvui->fpoutv, "   radius %f\n", drvui->Bond_Mult * drvui->bond_size[i]);
                  fprintf (drvui->fpoutv, "   height %f\n", d/5.);
                  fprintf (drvui->fpoutv, "  }\n }\n");
                  fprintf (drvui->fpoutv, " Separator{\n");
                  fprintf (drvui->fpoutv, "  Transform{\n ");
                  fprintf (drvui->fpoutv,"   translation %5.3f %5.3f %5.3f\n  }\n",
                      at1[0] + df[0] * 4./ 5., at1[1] + df[1] * 4./ 5., at1[2] + df[2] *4. / 5.);
                  fprintf (drvui->fpoutv, "  Rotation {\n");
                  fprintf (drvui->fpoutv, "   rotation 0 0 1 %f\n  }\n", -beta);
                  fprintf (drvui->fpoutv, "  Rotation {\n");
                  fprintf (drvui->fpoutv, "   rotation 1 0 0 %f\n  }\n", 1.5708 - gamma);
                  fprintf (drvui->fpoutv,"  Material {\n   diffuseColor %s\n", drvui->col_bond_v[i]);
                  fprintf (drvui->fpoutv, "  }\n");
                  fprintf (drvui->fpoutv, "  Cylinder {\n");
                  fprintf (drvui->fpoutv, "   parts SIDES\n");
                  fprintf (drvui->fpoutv, "   radius %f\n", drvui->Bond_Mult * drvui->bond_size[i]);
                  fprintf (drvui->fpoutv, "   height %f\n", d/5.);
                  fprintf (drvui->fpoutv, "  }\n }\n");
                }
              }
            }
          }        			 /* if distance OK */
        }        			 /* for k */
      }        				 /* for j */
    }        				 /* if nvert */
  }        				 /* for i (bond types) */
  free (ellips_id);
  printf ("%4d bonds.\n", Bond_Count);
  fprintf (drvui->flout, "\nGenerated %4d bonds.\n\n", Bond_Count);
}        				 /* end of generate_bonds */

/* ************************************************************** */
/* ************************************************************** */

void generate_cones (void)

/* routine to generate lone pair cone descriptions and add them to edit list */
{
  int start, same_type;
  int i, j, k, l;				 /* loop counters */
  int N_Bond;					 /* Counter for number of bonds of a type */
  int l1;					 /* loop counter for matrix operations */
  int nvert1;					 /* vertex counter */
  int Bond_Count;				 /* Counter for bonds output */
  char bnd1[4];					 /* temporary storage of bond labels */
  float d = 0;					 /* bond distance */
  float df[3];					 /* vector difference */
  float beta;					 /* rotation angle */
  float gamma;					 /* another rotation angle */
  float at1[3], at2[3];				 /*initial coordinates of bond endpoints */
  float clip0, clip1;				 /* clipping scale factors */
  int m, l2, l3, out = 0;

  int p;					 /* loop variables for rotation matrix calculation */
  float Z0[3], Z1[3];				 /* components of rms displacement along bond */
  float mybonds[20][20];			 /*temporary storage for atoms bonded to lonepair site */
  int nbnds = 0;
  float vecsum[3], vecpro[3];
#ifdef OPENGL
  float glr, glg, glb;
  GLUquadricObj *glu_quadObj;
#endif
  
  if (drvui->ncone == 1)
    return;					 /* exit if no work to be done */

  Bond_Count = 0;
  for (i = 1; i < drvui->ncone; ++i) {        	 /* loop through cone types */
    if (drvui->cone_fn[i] != drvui->frame_no) continue;  /* skip if wrong frame */
    nvert = 0;        			         /* empty the vertex list */
    for (j = 0; j < natom; ++j) {        /* loop through atom types */
      if (((drvui->atom_n[j] & 255) > 0) | (((drvui->atom_n[j] >> 24) & 255) > 0)) { /* sphere or ellipsoid */
        int o;

        o = drvui->atom_n[j] >> 24;
        if (o > 0) {                     /* ellipsoid */
          if ((drvui->cone_l1[i][0] == drvui->ellips_l[o][0]) &&
              (drvui->cone_l1[i][1] == drvui->ellips_l[o][1]) &&
              (drvui->cone_l1[i][2] == drvui->ellips_l[o][2]) &&
              (drvui->cone_l1[i][3] == drvui->ellips_l[o][3]))
            find_all_in_box (j);         /* get all atoms of type j */
        } else {
          o = drvui->atom_n[j] & 255;
          if ((drvui->cone_l1[i][0] == drvui->sphere_l[o][0]) &&
              (drvui->cone_l1[i][1] == drvui->sphere_l[o][1]) &&
              (drvui->cone_l1[i][2] == drvui->sphere_l[o][2]) &&
              (drvui->cone_l1[i][3] == drvui->sphere_l[o][3])) {
            find_all_in_box (j);         /* get all atoms of type j */
          }
        }        			 /* o > 0 */
      }
    }        				 /* for (j=0;j<natom;++j) */

    bnd1[0] = drvui->cone_l1[i][0];
    bnd1[1] = drvui->cone_l1[i][1];
    bnd1[2] = drvui->cone_l1[i][2];
    bnd1[3] = drvui->cone_l1[i][3];
    nvert1 = nvert;
    same_type = 0;

/* atoms i and j are not the same type */
    for (j = 0; j < natom; ++j) {         /* loop through atom types */
      if (((drvui->atom_n[j] & 255) > 0) | (((drvui->atom_n[j] >> 24) & 255) > 0)) { /* sphere or ellipsoid */
        find_all_in_box (j);        	 /* get all atoms of type j */
      }
    }        				 /* for (j=0;j<natom;++j) */

    if (((nvert1 > 0) && (nvert > nvert1)) || (same_type == 1)) {
      N_Bond = 0;
      nbnds = 0;
      for (j = 0; j < nvert1; ++j) {         /* loop through atoms of type 1 */
        nbnds = 0;
        start = nvert1;
        if (same_type)
          start = j;
        start = 0;
        for (k = start; k < nvert; ++k) {     /* loop on type 2 atoms  */
          d = dist (j, k);        	 /* get atomic distance */
          if (d > 0.2 && d <= drvui->cone_height[i]) {     /* here if distance within range */
            mybonds[0][nbnds] = s_vert[3 * k];
            mybonds[1][nbnds] = s_vert[3 * k + 1];
            mybonds[2][nbnds] = s_vert[3 * k + 2];
            nbnds++;
          }        			 /* if distance ok */
        }        			 /* for k */
    /*# */
        if (nbnds < 4 - drvui->numlonepairs[i]) {
          continue;
    }
        df[0] = df[1] = drvui->Bond_Mult * drvui->bond_size[i];    /* scaling */
        df[2] = 0.5f *d;

        Bond_Count++;        		 /* update count */
        for (p = 0; p < 3; p++) {         /* initialize endpoints of bond */
          Z0[p] = s_vert[3 * j + p];
        }

        for (l = 0; l < drvui->numlonepairs[i]; l++) {
          for (p = 0; p < 3; p++) {         /* initialize endpoints of bond */
            Z1[p] = 0;

            if (drvui->numlonepairs[i] == 1) {
              for (k = 0; k < nbnds; k++)
                Z1[p] -= mybonds[p][k] - s_vert[3 * j + p];
              Z1[p] = Z0[p] + Z1[p];
            } else {
              if (drvui->numlonepairs[i] == 2) {
                vecsum[p] = 0.;
                vecpro[p] = 0.;
                for (k = 0; k < nbnds; k++) {
                  vecsum[p] += mybonds[p][k] - s_vert[3 * j + p];
                }

                vecpro[0] = (s_vert[3 * j + 1] - mybonds[1][0])
                     * (s_vert[3 * j + 2] - mybonds[2][1])
                     - (s_vert[3 * j + 1] - mybonds[1][1])
                     * (s_vert[3 * j + 2] - mybonds[2][0]);
                vecpro[1] = (s_vert[3 * j + 2] - mybonds[2][0])
                     * (s_vert[3 * j] - mybonds[0][1])
                     - (s_vert[3 * j + 2] - mybonds[2][1])
                     * (s_vert[3 * j] - mybonds[0][0]);
                vecpro[2] = (s_vert[3 * j] - mybonds[0][0])
                     * (s_vert[3 * j + 1] - mybonds[1][1])
                     - (s_vert[3 * j] - mybonds[0][1])
                     * (s_vert[3 * j + 1] - mybonds[1][0]);
                switch (l) {
                case 0:
                  Z1[p] = -vecsum[p] + vecpro[p];
                  break;
                case 1:
                  Z1[p] = -vecsum[p] - vecpro[p];
                  break;
                }
                Z1[p] = Z0[p] + Z1[p];
              }
            }
          }        			 /* for p */

          d = (float) sqrt ((Z0[0] - Z1[0]) * (Z0[0] - Z1[0])
                          + (Z0[1] - Z1[1]) * (Z0[1] - Z1[1])
                          + (Z0[2] - Z1[2]) * (Z0[2] - Z1[2]));
          d = drvui->cone_height[i] / d;
          Z1[0] = Z0[0] + d * (Z1[0] - Z0[0]);
          Z1[1] = Z0[1] + d * (Z1[1] - Z0[1]);
          Z1[2] = Z0[2] + d * (Z1[2] - Z0[2]);

          for (l1 = 0; l1 < 3; l1++) {
            at1[l1] = (float) Z0[l1];
            at2[l1] = (float) Z1[l1];
            df[l1] = (float) (Z1[l1] - Z0[l1]);
          }

          if (clipflag != 0) {
            out = 0;
            l2 = 0;
            l3 = 0;
            clip0 = 0.0f;
            clip1 = 1.0f;
            for (l1 = 0; l1 < 3; l1++) {
              if (o_vert[3 * j + l1] < drvui->clip_lim[drvui->frame_no][l1] - 0.01
                 || o_vert[3 * j + l1] > drvui->clip_lim[drvui->frame_no][l1 + 3] + 0.01)
                l2 = 1;
              if (o_vert[3 * k + l1] < drvui->clip_lim[drvui->frame_no][l1] - 0.01
                 || o_vert[3 * k + l1] > drvui->clip_lim[drvui->frame_no][l1 + 3] + 0.01)
                l3 = 1;
            }
            if (l2 && l3)
              out = 1;        		 /* both atoms outside */
            if (l2)
              out = 1;
            if (l3)
              clip1 = 0.5f;
            if (l2)
              clip0 = 0.5f;

            at1[0] = at1[0] + clip0 * df[0];
            at1[1] = at1[1] + clip0 * df[1];
            at1[2] = at1[2] + clip0 * df[2];

            at2[0] = at1[0] + clip1 * df[0];
            at2[1] = at1[1] + clip1 * df[1];
            at2[2] = at1[2] + clip1 * df[2];

          }        			 /* if clipflag */

          if (clipflag == 0 || out == 0) {
            fprintf (drvui->fpoutp," /* LonePair on %c%c %8.5f %8.5f %8.5f */\n",
               bnd1[0], bnd1[1], o_vert[3 * j],o_vert[3 * j + 1], o_vert[3 * j + 2]);
            fprintf (drvui->fpoutp," cone{<%8.5f, %8.5f, %8.5f>, %8.5f, <%8.5f, %8.5f, %8.5f>,  %8.5f \n",
               at1[0], at1[1], at1[2],drvui->Bond_Mult * drvui->cone_min[i], at2[0], at2[1],
               at2[2], drvui->Bond_Mult * drvui->cone_max[i]);
            fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n }\n", drvui->col_cone_p[i]);
            fprintf (drvui->fpoutp," sphere{<%8.5f, %8.5f, %8.5f> %8.5f \n",
               at2[0], at2[1], at2[2], drvui->Bond_Mult * drvui->cone_max[i]);
            fprintf (drvui->fpoutp,"  texture{pigment{color %s  }}\n}\n", drvui->col_cone_p[i]);
            N_Bond++;
            if (no_comment == 0)
              fprintf (drvui->fpoutv,"# LonePair on: %c%c %8.5f %8.5f %8.5f\n",
                 bnd1[0], bnd1[1], o_vert[3 * j],o_vert[3 * j + 1], o_vert[3 * j + 2]);
            for (m = 0; m < 3; m++)
              df[m] = (float) (at2[m] - at1[m]);
            d = (float) sqrt (df[0] * df[0] + df[1] * df[1] + df[2] * df[2]);
/*
   calculate rotation matrices needed to rotate bond about Z onto Y-Z plane and then about X so that
   bond is parallel to Z.  Finally, apply inverse transformation to cylinder, which started parallel
   to X.  Resulting cylinder will be parallel to vector between atoms.
*/
            beta = (float) atan2 (df[0], df[1] + 0.0000001);    /* rotation angle about Z (in radians) */
            gamma = (float) sqrt (df[0] * df[0] + df[1] * df[1]);
            gamma = (float) atan2 (gamma, df[2]);    /* Rotation angle about X */
#ifdef OPENGL
	        if (gl_view) {
              glPushMatrix ();
              sscanf (drvui->col_cone_v[i], "%f %f %f", &glr, &glg, &glb);
              glColor3f (glr, glg, glb);
              glTranslatef (at1[0], at1[1], at1[2]);
              glRotatef (-beta * (float) RAD, 0.0f, 0.0f, 1.0f);
              glRotatef (-gamma * (float) RAD, 1.0f, 0.0f, 0.0f);
              glu_quadObj = gluNewQuadric ();
              gluQuadricDrawStyle (glu_quadObj, GLU_FILL);
              gluCylinder (glu_quadObj, drvui->Bond_Mult * drvui->cone_min[i],
                 drvui->Bond_Mult * drvui->cone_max[i], d, 10, 1);
              glPopMatrix ();
              glPushMatrix ();
              glTranslatef (at2[0], at2[1], at2[2]);
              glutSolidSphere (drvui->cone_max[i], 10, 10);
              glPopMatrix ();
            }
#endif
            if (Vrml2) {
              fprintf (drvui->fpoutv, "  Transform{\n ");
              fprintf (drvui->fpoutv,"   translation %5.3f %5.3f %5.3f\n",
                 (at1[0] + at2[0]) / 2.,(at1[1] + at2[1]) / 2., (at1[2] + at2[2]) / 2.);
              fprintf (drvui->fpoutv, "   rotation 0 0 1 %f\n", -beta);
              fprintf (drvui->fpoutv,"   children Transform { rotation 0 1 0 %f\n", 3.14);
              fprintf (drvui->fpoutv,"   children Transform { rotation 1 0 0 %f\n",1.57 + gamma);
              fprintf (drvui->fpoutv, " children [ Shape {\n");
              fprintf (drvui->fpoutv, " appearance Appearance { material Material {   diffuseColor %s} }\n",
                 drvui->col_cone_v[i]);
              fprintf (drvui->fpoutv, "  geometry Cone {");
              fprintf (drvui->fpoutv, "bottomRadius %f", drvui->Bond_Mult * drvui->cone_max[i]);
              fprintf (drvui->fpoutv, " height %f", d);
              fprintf (drvui->fpoutv, "}}]}}}\n");
              fprintf (drvui->fpoutv, "  Transform {");
              fprintf (drvui->fpoutv," translation %8.5f %8.5f %8.5f\n",at2[0], at2[1], at2[2]);
              fprintf (drvui->fpoutv, " children [\n Shape {");
              fprintf (drvui->fpoutv,"appearance Appearance {material Material { diffuseColor %s} }\n",
                 drvui->col_cone_v[i]);
              fprintf (drvui->fpoutv,"  geometry Sphere { radius %8.5f}", drvui->cone_max[i]);
              fprintf (drvui->fpoutv, "}]}");
            } else {
              fprintf (drvui->fpoutv, " Separator{\n");
              fprintf (drvui->fpoutv, "  Transform{\n ");
              fprintf (drvui->fpoutv,"   translation %5.3f %5.3f %5.3f\n  }\n",(at1[0] + at2[0]) / 2.,
                 (at1[1] + at2[1]) / 2., (at1[2] + at2[2]) / 2.);
              fprintf (drvui->fpoutv, "  Rotation {\n");
              fprintf (drvui->fpoutv, "   rotation 0 0 1 %f\n  }\n", -beta);
              fprintf (drvui->fpoutv, "  Rotation {\n");
              fprintf (drvui->fpoutv, "   rotation 0 1 0  %f\n  }\n", 3.14);
              fprintf (drvui->fpoutv, "  Rotation {\n");
              fprintf (drvui->fpoutv, "   rotation 1 0 0 %f\n  }\n", 1.57 + gamma);
              fprintf (drvui->fpoutv, "  Material {\n   diffuseColor %s\n", drvui->col_cone_v[i]);
              fprintf (drvui->fpoutv, "  }\n");
              fprintf (drvui->fpoutv, "  Cone {\n");
              fprintf (drvui->fpoutv, "   parts SIDES\n");
              fprintf (drvui->fpoutv, "   bottomRadius %f\n",drvui->Bond_Mult * drvui->cone_max[i]);
              fprintf (drvui->fpoutv, "   height %f\n", d);
              fprintf (drvui->fpoutv, "  }\n }\n");
              fprintf (drvui->fpoutv, " Separator {");
              fprintf (drvui->fpoutv, "  Transform {");
              fprintf (drvui->fpoutv, " translation %8.5f %8.5f %8.5f}\n", at2[0], at2[1], at2[2]);
              fprintf (drvui->fpoutv, "  Material {");
              fprintf (drvui->fpoutv, " shininess 0.3\n");
              fprintf (drvui->fpoutv, " diffuseColor %s}\n", drvui->col_cone_v[i]);
              fprintf (drvui->fpoutv, "  Sphere { radius %8.5f}\n", drvui->cone_max[i]);
              fprintf (drvui->fpoutv, " }\n");
            }        			 /* if (Vrml2) */
          }        			 /* if not clipped */

        }        			 /* for l */
      }        				 /* for j */
    }        				 /* if nvert */
  }        				 /* for i */
  fprintf (stderr, "%4d cones.\n", Bond_Count);
  fprintf (drvui->flout, "\nGenerated %4d cones.\n\n", Bond_Count);
}						 /* end of generate_bonds */

void generate_arrows (void)

/* routine to generate arrow objects for magnetic moments */
{
  int Arrow_Count;				 /* Counter for number output */
  int i, k, l;				 /* loop counters */
#ifdef OPENGL
  float glr,glb,glg;
  GLUquadricObj *glu_quadObj;
#endif
  float base;
  float x[3], xp[3], yp[3];
  int xi,xj,xk;
  float phi, chi;
  int savedNvertM = NvertM;
     
  Arrow_Count = 0;
  for (i = 0; i < drvui->nmag; ++i) {	   /* loop through magnetic arrows */
    if (drvui->arrow_fn[i] != drvui->frame_no) continue; /* skip if not right frame */
    for (k=0;k<3;k++) {
        x[k] = drvui->mag_xp[i][k] - drvui->xyzoff[k];
    }
    for (xi = -3; xi<=3; xi++) {
        xp[0] = (float)xi;
        for (xj = -3; xj<=3; xj++) {
            xp[1] = (float)xj;
            for (xk=-3; xk<=3; xk++) {
                xp[2] = (float)xk;         /* xp is a magnetic cell translation */
                for (k=0;k<3;k++) {
                    yp[k]=x[k];            /* yp will be position in nuclear cell  */
                    for (l=0;l<3;l++) {
                        yp[k] += drvui->mag_matrix[k][l] * xp[l];
                    }
                }
                add_vert(yp,100*i,0);     /* add vector to vertex list */
            }
        }
    }
  }                    /* end of loop through magnetic arrows */
  for (k=savedNvertM; k< NvertM; k++) {
    float tmp, tmp1;
    double t;
    float d1, d2, d3, alpha, temp;
    float cosalp;
    i = drvui->atom_no[k]/100;         /* get number of magnetic item */
    for (l = 0; l<3; l++) {
        x[l] = drvui->mag_xc[i][l];   /* get magnetic components */
    }
    t = sqrt(x[0] * x[0] + x[1] * x[1] + x[2] * x[2]);
    if (t < 0.002) continue;      /* skip this arrow if 0 length */
    for (l=0; l<3; l++) {
        x[l] /= (float)t;
    }
    t = sqrt(x[0] * x[0] + x[1] * x[1]);
    chi = (float) atan2(t, x[2]);
    if (t > 0.001)
        phi = (float) atan2(x[0]/t,x[1]/t);
    else
        phi = 0.0f;
#ifdef OPENGL
    if (gl_view) {
      glLoadName(200000+i);
      glPushName(k);
      glPushMatrix ();
      (void) sscanf (drvui->col_arrow_v[i], "%f %f %f", &glr, &glg, &glb);
      glColor3f (glr, glg, glb);
      glTranslatef (s_vert[3*k], s_vert[3*k+1], 
            s_vert[3*k+2]);
      glRotatef (-phi* (float) RAD, 0.0, 0.0, 1.0f);
      glRotatef (-chi* (float) RAD, 1.0f, 0., 0.);
      glTranslatef(0.0,0.0,-drvui->arrow_length[i]/2.0f);
      glu_quadObj = gluNewQuadric ();
      gluQuadricDrawStyle (glu_quadObj, GLU_FILL);
      gluCylinder (glu_quadObj, drvui->arrow_diam[i],
              drvui->arrow_diam[i], 0.9 * drvui->arrow_length[i], 10, 1);
#if !defined (WIN32) && !defined (__APPLE__)
      free(glu_quadObj);
#endif
      glPopMatrix();
      glPushMatrix();
      glTranslatef (s_vert[3*k], s_vert[3*k+1], 
              s_vert[3*k+2]);
      glRotatef (-phi* (float) RAD, 0.0, 0.0, 1.0f);
      glRotatef (-chi* (float) RAD, 1.0f, 0., 0.);
      glTranslatef(0.0f,0.0f, drvui->arrow_length[i] * 0.27f);
	
      glu_quadObj = gluNewQuadric ();
      gluQuadricDrawStyle (glu_quadObj, GLU_FILL);
      gluCylinder (glu_quadObj, drvui->arrow_diam[i] * 2.0,
   	   	       0.01, 0.4*drvui->arrow_length[i], 10, 1);
#if !defined (WIN32) && !defined (__APPLE__)
      free(glu_quadObj);
#endif
      glPopMatrix ();
      glPopName();
    }
#endif

/* arrow in POV */

    tmp = 0.5f * drvui->arrow_length[i];
    base = 2.0f * drvui->arrow_diam[i];
    fprintf (drvui->fpoutp, " object{\n");
    fprintf (drvui->fpoutp, "   union{\n");
    fprintf (drvui->fpoutp,"     object{\n");
    fprintf (drvui->fpoutp,"      cylinder{ < 0,0,%8.5f > , < 0,0,%8.5f >,%8.5f\n",
         -0.5f * drvui->arrow_length[i], 0.27f * drvui->arrow_length[i],drvui->arrow_diam[i]);
    fprintf (drvui->fpoutp,
         "         texture{pigment{color %s}}}\n", drvui->col_arrow_p[i]);
    fprintf (drvui->fpoutp, "    finish{phong %5.2f phong_size %5.2f reflection 0.1}\n     }\n",
         drvui->Phong_Value, drvui->Phong_Size);
    tmp1 = 0.27f * drvui->arrow_length[i];
    tmp = 0.77f * drvui->arrow_length[i];
    fprintf (drvui->fpoutp, "     object{\n");
    fprintf (drvui->fpoutp, "      cone{<0, 0, %8.5f>, %8.5f, <0, 0, %8.5f>,  0 \n",
             tmp1, base, tmp); 
    fprintf (drvui->fpoutp,
		     "         texture{pigment{color %s}}}\n", drvui->col_arrow_p[i]);
    fprintf (drvui->fpoutp, "     finish{phong %5.2f phong_size %5.2f reflection 0.1}\n     }\n   }\n",
         drvui->Phong_Value, drvui->Phong_Size);
    fprintf (drvui->fpoutp,"   rotate <%8.5f,0,0>\n", -chi * RAD);
    fprintf (drvui->fpoutp,"   rotate <0,0,%8.5f>\n", -phi * RAD);
    fprintf (drvui->fpoutp,"   translate <%8.5f, %8.5f, %8.5f>\n }\n",
             s_vert[3 * k], s_vert[3 * k + 1], s_vert[3 * k + 2]);


/* arrow in VRML - rotation vector is v cross z */
    d1 = x[1];
    d2 = -x[0];
    d3 = 0.0f;
    temp = (float) sqrt(d1*d1 + d2 * d2 + d3*d3 );
    if (temp < 0.001f) {        /* v is parallel to z - make cross be x */
        d1 = 1.0;
        temp = 1.0;
    }
    d1 /= temp;
    d2 /= temp;
    d3 /= temp;
    cosalp = x[2]/(float)sqrt(x[0] * x[0] + x[1] * x[1] + x[2] * x[2]);;
    alpha = -(float) acos (cosalp);
    if (Vrml2) {
/* cylinder for shaft of arrow */
      fprintf (drvui->fpoutv, "  Transform{\n");
      fprintf (drvui->fpoutv, "    rotation %6.3f %6.3f %6.3f %7.3f\n",
                d1, d2, d3, alpha);
      fprintf (drvui->fpoutv, "    translation %5.3f %5.3f %5.3f\n",
              s_vert[3*k],s_vert[3*k+1],s_vert[3*k+2]);
      fprintf (drvui->fpoutv, "    children [\n");
      fprintf (drvui->fpoutv, "      Transform{\n");
      fprintf (drvui->fpoutv, "        rotation 1 0 0 1.5708\n");
      fprintf (drvui->fpoutv, "        translation 0 0 %5.3f\n",-0.05f * drvui->arrow_length[i]);
      fprintf (drvui->fpoutv, "        children [\n");
      fprintf (drvui->fpoutv, "          Shape {\n");
      fprintf (drvui->fpoutv, "            geometry Cylinder {\n");
      fprintf (drvui->fpoutv, "              radius %f\n", drvui->arrow_diam[i]);
      fprintf (drvui->fpoutv, "              height %f\n", 0.9f * drvui->arrow_length[i]);
      fprintf (drvui->fpoutv, "            }\n");
      fprintf (drvui->fpoutv, "            appearance Appearance { material Material "
                                               "{ diffuseColor %s}}\n",
                                               drvui->col_arrow_v[i]);
      fprintf (drvui->fpoutv, "          }\n");
      fprintf (drvui->fpoutv, "        ]\n");
      fprintf (drvui->fpoutv, "      }\n");
      fprintf (drvui->fpoutv, "    ]\n");
      fprintf (drvui->fpoutv, "  }\n");
/* Cone at end of arrow */
      fprintf (drvui->fpoutv, "  Transform{\n");
      fprintf (drvui->fpoutv, "    rotation %6.3f %6.3f %6.3f %7.3f\n",
                d1, d2, d3, alpha);
      fprintf (drvui->fpoutv, "    translation %5.3f %5.3f %5.3f\n",
              s_vert[3*k],s_vert[3*k+1],s_vert[3*k+2]);
      fprintf (drvui->fpoutv, "    children Transform {\n");
      fprintf (drvui->fpoutv, "      children [\n");
      fprintf (drvui->fpoutv, "        Transform{\n");
      fprintf (drvui->fpoutv, "          rotation 1 0 0 1.5708\n");
      fprintf (drvui->fpoutv, "          translation 0 0 %5.3f\n",
              0.5f * drvui->arrow_length[i]);
      fprintf (drvui->fpoutv, "          children [\n");
      fprintf (drvui->fpoutv, "            Shape {\n");
      fprintf (drvui->fpoutv, "              geometry Cone {\n");
      fprintf (drvui->fpoutv, "                bottomRadius %f\n", 2.0 * drvui->arrow_diam[i]);
      fprintf (drvui->fpoutv, "                height %f\n", drvui->arrow_length[i]*0.4);
      fprintf (drvui->fpoutv, "              }\n");
      fprintf (drvui->fpoutv, "              appearance Appearance { material Material "
                                               "{   diffuseColor %s}}\n",drvui->col_arrow_v[i]);
      fprintf (drvui->fpoutv, "            }\n");
      fprintf (drvui->fpoutv, "          ]\n");
      fprintf (drvui->fpoutv, "        }\n");
      fprintf (drvui->fpoutv, "      ]\n");
      fprintf (drvui->fpoutv, "    }\n");
      fprintf (drvui->fpoutv, "  }\n");
    } else { /* VRML 1 */
      fprintf (drvui->fpoutv, " Separator{\n");
      fprintf (drvui->fpoutv, "  Transform{\n");
      fprintf (drvui->fpoutv, "   translation %5.3f %5.3f %5.3f\n",
              s_vert[3*k],s_vert[3*k+1],s_vert[3*k+2]);
      fprintf (drvui->fpoutv, "   rotation %6.3f %6.3f %6.3f %7.3f\n  }\n",
                d1, d2, d3, alpha);
      fprintf (drvui->fpoutv, "  Material {diffuseColor %s}\n", drvui->col_arrow_v[i]);
      fprintf (drvui->fpoutv, "  Rotation{rotation 1 0 0 1.5708}\n");
      fprintf (drvui->fpoutv, "  Cylinder {\n");
      fprintf (drvui->fpoutv, "   parts SIDES\n");
      fprintf (drvui->fpoutv, "   radius %f\n", drvui->arrow_diam[i]);
      fprintf (drvui->fpoutv, "   height %f\n", 0.9f * drvui->arrow_length[i]);
      fprintf (drvui->fpoutv, "  }\n }\n");
      fprintf (drvui->fpoutv, " Separator{\n");
      fprintf (drvui->fpoutv, "  Transform{\n");
      fprintf (drvui->fpoutv, "   translation %5.3f %5.3f %5.3f\n",
              s_vert[3*k],s_vert[3*k+1],s_vert[3*k+2]);
      fprintf (drvui->fpoutv, "   rotation %6.3f %6.3f %6.3f %7.3f\n  }\n",
                d1, d2, d3, alpha);
      fprintf (drvui->fpoutv, "  Material {diffuseColor %s}\n", drvui->col_arrow_v[i]);
      fprintf (drvui->fpoutv, "  Transform{\n");
      fprintf (drvui->fpoutv, "   rotation 1 0 0 1.5708\n");
      fprintf (drvui->fpoutv, "   translation 0 0 %5.3f\n  }\n",
              0.5f * drvui->arrow_length[i]);
      fprintf (drvui->fpoutv, "  Cone {\n");
      fprintf (drvui->fpoutv, "   parts SIDES\n");
      fprintf (drvui->fpoutv, "   bottomRadius %f\n",2.0 * drvui->arrow_diam[i]);
      fprintf (drvui->fpoutv, "   height %f\n", drvui->arrow_length[i]*0.4);
      fprintf (drvui->fpoutv, "  }\n }\n");
    }
    Arrow_Count++;
  }
  NvertM = savedNvertM;

  if (Arrow_Count > 0) {
    fprintf (stderr,"%4d magnetic moment arrows.\n", Arrow_Count);
    fprintf (drvui->flout, "Generated %4d arrows.\n", Arrow_Count);
  }
}						 /* end of generate_arrows */

/* ************************************************************** */
/* ************************************************************** */

void get_atom_id (void)

/* routine to rework atom numbers to point at polyhedron, sphere
   plane, or ellipsoid characteristics list */
{
  int i, j;                              /* local variables */

  for (i = 0; i < natom; ++i) {
    drvui->sv_atom_n[i] = drvui->atom_n[i];    /* save original atom number */
    drvui->atom_n[i] = 0;                     /* clear encoded location */
    for (j = 1; j < drvui->nsphere; ++j) {
      if (check_atom_name(drvui->atom_l[i], drvui->sphere_l[j])
      && (drvui->sv_atom_n[i] == drvui->sphere_n[j] || drvui->sphere_n[j] == -1)) {
          drvui->atom_n[i] = j;        	    /* atom is in sphere list */
      }
    }
    for (j = 1; j < drvui->npoly; ++j) {
        if (check_atom_name(drvui->atom_l[i], drvui->poly_l[j])) {
	      drvui->atom_n[i] += (j << 8);			 /* atom is in poly. list */
        }
    }
    for (j = 1; j < drvui->nplane; ++j) {
        if (check_atom_name(drvui->atom_l[i],drvui->plane_l[j])) {
	      drvui->atom_n[i] += (j << 16);		 /* atom is in plane list */
        } 
    }
    if (do_ellipsoids) {
      for (j = 1; j < n_ellips; ++j) {
        if (check_atom_name(drvui->atom_l[i],drvui->ellips_l[j])
            && (drvui->sv_atom_n[i] == drvui->ellips_n[j])) {
          drvui->atom_n[i] += (j << 24);         /* atom is in ellipsoid list */
        }
      }
    }
  }
}						 /* end of get_atom_id  */

/* ************************************************************** */
/* ************************************************************** */

void get_input (int Quick)

/* Routine to process input file, make lattice metric and do preliminary
   ellipsoid processing */
{
  int i, j, k, npd;
  float biso;

  if (!Quick)
    fprintf (drvui->flout,
	     "\n********************* Input: ************************\n");
  read_inp (Quick);
  if (Quick)
    return;
  fprintf (drvui->flout,
	   "*****************************************************\n\n");
  if (!Labels)
    Display_axes = 0;				 /* If no labels, no vectors */
  if (Vrml2)
    printf ("\nGenerating VRML97 Output.\n");
  else
    printf ("\nGenerating VRML1.0 Output.\n");
  print_sym ();                                                 /* print symmetry information */
  make_bmat (drvui->lat_con);                                          /* create the lattice metric */
/* set limits of search regions */
  if (packflag == 0 && boxflag == 0 && drvui->frame_no == 1) {		    /* neither given and frame 1 */
      for (i=0; i<3; i++) {
          drvui->cryst_lim[drvui->frame_no][i] = origin[i] - 0.5f;     /* set to generate one cell */
          drvui->cryst_lim[drvui->frame_no][i + 3] = origin[i] + 0.5f;
      }
      packflag = 1;
  }
  if (packflag != 0) {				 /* pack command given - calculate approximate bounds */
    float temp, temp1;

    for (i = 0; i < 3; i++) {
      temp = origin[i] - drvui->cryst_lim[drvui->frame_no][i];
      temp1 = drvui->cryst_lim[drvui->frame_no][i + 3] - origin[i];
      if (temp1 > temp)
          temp = temp1;
      boxlim[i] = (temp + 0.2f) * drvui->lat_con[i] / 0.5f;
    }
  } else {					 /* bounds command given - calculate approximate pack limits */
    for (i = 0; i < 3; i++) {
      drvui->cryst_lim[drvui->frame_no][i] = origin[i] - boxlim[i] / drvui->lat_con[i] - (float) 0.2;
      drvui->cryst_lim[drvui->frame_no][i + 3] = boxlim[i] / drvui->lat_con[i] - origin[i] + (float) 0.2;
    }
  }
  if ((Options & L_OPT) != 0)
    Calc_Rot (drvui->lookat_v1, drvui->lookat_v2);		 /* Get rotation angles for "lookat" */
  if (n_ellips > 1) {
    convert_ellipsoid ();                    /* change Bij and Uij forms to betaij */
    npd = 0;
    fprintf (drvui->flout, "\nAnalysis of Thermal Ellipsoids:\n");
    for (i = 1; i < n_ellips; i++) {
      if (eigen (&biso, drvui->ellips[i], drvui->ellips_RMS[i], drvui->ellips_EV[i]) == 0) {
        fprintf (drvui->flout,
		   "Ellipsoid for %c%c%c%c %d is non-positive definite.\n",
		   drvui->ellips_l[i][0], drvui->ellips_l[i][1], drvui->ellips_l[i][2],
		   drvui->ellips_l[i][3], drvui->ellips_n[i]);
        npd = 1;
        drvui->ellips_RMS[i][0] = (float) sqrt (biso / 78.957);	/* convert Biso to mu */
      } else {
        fprintf (drvui->flout,
		   "\n%c%c%c%c %3d Equivalent Isotropic B: %6.2f, U: %6.3f\n",
		   drvui->ellips_l[i][0], drvui->ellips_l[i][1], drvui->ellips_l[i][2],
		   drvui->ellips_l[i][3], drvui->ellips_n[i], biso, biso / 78.957);
        fprintf (drvui->flout, "      RMS Amplitudes: %6.3f %6.3f %6.3f\n",
		   drvui->ellips_RMS[i][0], drvui->ellips_RMS[i][1], drvui->ellips_RMS[i][2]);
        fprintf (drvui->flout, "      Eigen Vectors (by row):\n");
        for (j = 0; j < 3; j++) {
          fprintf (drvui->flout, "     ");
          for (k = 0; k < 3; k++)
            fprintf (drvui->flout, "     %8.5f", drvui->ellips_EV[i][j][k]);
            fprintf (drvui->flout, "\n");
        }
      }
    }
    if (npd == 1) {
      n_ellips = 1;
      fprintf (drvui->flout, "\n***** Ellipsoid generation is turned off.\n");
    } else {
      Ellipsoid_Scale = P_to_C (Ellipsoid_Prob);
    }
  }
}						 /* end of get_input */

/* ************************************************************** */
/* ************************************************************** */

void get_label (char input[100], char *c1, char *c2, char *c3, char *c4,
		int strip)

/* Routine to get 1 to 4 character atom ID from input line, and shift input
   line to left  */
{

  int i, j;

  i = 0;
  if (strip != 0)
    while (input[i] != ' ')
      input[i++] = ' ';              /* skip over command */
  while (input[i] == ' ')
    i++;                             /* skip whitespace after command */
  *c1 = input[i];                    /* return first atom label character */
  input[i++] = ' ';
  *c2 = input[i];                    /* return 2nd atom label char */
  if (input[i] != ' ')
    input[i++] = ' ';                /* if 2nd not blank, rub it out */
  *c3 = input[i];                    /* return 3rd character */
  if (input[i] != ' ')
    input[i++] = ' ';                /* if 3rd not blank, rub it out */
  *c4 = input[i];                    /* return 4th character */
  if (input[i] != ' ')
    input[i++] = ' ';                /* if 4th not blank, rub it out */
  while (input[i] == ' ')
    i++;                             /* skip white space after atom label */
  for (j = 0; i < 100; i++) {
    input[j++] = input[i];           /* copy rest of command */
  }
}						 /* end of get_label */

/* ************************************************************** */
/* ************************************************************** */

void Locate_Triple (void)

/* routine to try to place the vector triple where it will be seen in the
POV diagram */
{
  int i, j;
  float temp, pos[3], cpos[3];

  for (i = 0; i < 3; ++i) {			 /* convert origin to Cartesian */
    pos[i] = 0.0f;
    for (j = 0; j < 3; ++j)
      pos[i] -= (float) drvui->b_mat[i][j] * origin[j];
  }
  for (i = 0; i <= 2; ++i) {			 /* calculate position of point after POV rotation */
    cpos[i] = 0.0f;
    for (j = 0; j <= 2; ++j)
      cpos[i] += (float) G_Rot[j][i] * pos[j];
  }
  for (i = 0; i < 3; i++) {
    temp = 0.04f *(POV_Max[i] - POV_Min[i]);

    if (cpos[i] > 0)
      pos[i] = POV_Max[i] + temp;
    else
      pos[i] = POV_Min[i] - temp;
  }
/* calculate coordinates of pos before rotation */
  for (i = 0; i < 3; i++) {
    offset[i] = 0.0f;
    for (j = 0; j < 3; j++) {
      offset[i] += (float) G_Rot[i][j] * pos[j];
    }
  }
}

/* ************************************************************** */
/* ************************************************************** */

void make_bmat (float lat_con[6])
{
  float snal, snbe, snga, csal, csbe, csga, vol, det;
  int i, j;

  for (i = 3; i <= 5; ++i)
    if (drvui->lat_con[i] <= 0)
      drvui->lat_con[i] = 90.0f;
  if (drvui->sys == 5)
    drvui->lat_con[5] = 120.0f;
  for (i = 0; i <= 2; ++i)
    for (j = 0; j <= 2; ++j)
      drvui->b_mat[j][i] = 0.0f;		 /* initialize matrix */
  drvui->b_mat[0][0] = 1.0f;
  csal = (float) cos (drvui->lat_con[3] / RAD);
  snal = (float) sin (drvui->lat_con[3] / RAD);
  csbe = (float) cos (drvui->lat_con[4] / RAD);
  snbe = (float) sin (drvui->lat_con[4] / RAD);
  csga = (float) cos (drvui->lat_con[5] / RAD);
  snga = (float) sin (drvui->lat_con[5] / RAD);
  for (i = 0; i < 3; i++)
    drvui->ginv[i][i] = drvui->lat_con[i] * drvui->lat_con[i];
  drvui->ginv[0][1] = drvui->ginv[1][0] = drvui->lat_con[0] * drvui->lat_con[1] * csga;
  drvui->ginv[0][2] = drvui->ginv[2][0] = drvui->lat_con[0] * drvui->lat_con[2] * csbe;
  drvui->ginv[1][2] = drvui->ginv[2][1] = drvui->lat_con[1] * drvui->lat_con[2] * csal;
  drvui->b_mat[0][1] = csga;
  drvui->b_mat[1][1] = snga;
  drvui->b_mat[0][2] = csbe;
  drvui->b_mat[1][2] = (csal - csbe * csga) / snga;
  drvui->b_mat[2][2] =	(float) sqrt (1.0 - drvui->b_mat[0][2] * drvui->b_mat[0][2] -
		      drvui->b_mat[1][2] * drvui->b_mat[1][2]);
  for (i = 0; i <= 2; ++i)
    for (j = 0; j <= 2; ++j)
      drvui->b_mat[j][i] *= drvui->lat_con[i];
/* calculate cell volume and reciprocal lengths */
  vol = (float) (drvui->b_mat[0][0] *
		 (drvui->b_mat[1][1] * drvui->b_mat[2][2] - drvui->b_mat[1][2] * drvui->b_mat[2][1])
		 - drvui->b_mat[0][1] * (drvui->b_mat[1][0] * drvui->b_mat[2][2] - drvui->b_mat[1][2] * drvui->b_mat[2][0])
		 + drvui->b_mat[0][2] * (drvui->b_mat[1][0] * drvui->b_mat[2][1] - drvui->b_mat[1][1] * drvui->b_mat[2][0]));
  drvui->rec_lat_con[0] = drvui->lat_con[1] * drvui->lat_con[2] * snal / vol;
  drvui->rec_lat_con[1] = drvui->lat_con[0] * drvui->lat_con[2] * snbe / vol;
  drvui->rec_lat_con[2] = drvui->lat_con[1] * drvui->lat_con[0] * snga / vol;

/* get binv */

  det = determinant (drvui->b_mat);
  drvui->b_inv_mat[0][0] = (float) (drvui->b_mat[1][1] * drvui->b_mat[2][2] - drvui->b_mat[1][2] * drvui->b_mat[2][1]) / det;
  drvui->b_inv_mat[1][0] =	-(float) (drvui->b_mat[1][0] * drvui->b_mat[2][2]
      - drvui->b_mat[1][2] * drvui->b_mat[2][0]) / det;
  drvui->b_inv_mat[2][0] =	(float) (drvui->b_mat[1][0] * drvui->b_mat[2][1]
      - drvui->b_mat[1][1] * drvui->b_mat[2][0]) / det;
  drvui->b_inv_mat[0][1] =	-(float) (drvui->b_mat[0][1] * drvui->b_mat[2][2]
      - drvui->b_mat[0][2] * drvui->b_mat[2][1]) / det;
  drvui->b_inv_mat[1][1] =	(float) (drvui->b_mat[0][0] * drvui->b_mat[2][2]
      - drvui->b_mat[0][2] * drvui->b_mat[2][0]) / det;
  drvui->b_inv_mat[2][1] =	-(float) (drvui->b_mat[0][0] * drvui->b_mat[2][1]
      - drvui->b_mat[0][1] * drvui->b_mat[2][0]) / det;
  drvui->b_inv_mat[0][2] =	(float) (drvui->b_mat[0][1] * drvui->b_mat[1][2]
      - drvui->b_mat[0][2] * drvui->b_mat[1][1]) / det;
  drvui->b_inv_mat[1][2] =	-(float) (drvui->b_mat[0][0] * drvui->b_mat[1][2]
      -  drvui->b_mat[0][2] * drvui->b_mat[1][0]) / det;
  drvui->b_inv_mat[2][2] =	(float) (drvui->b_mat[0][0] * drvui->b_mat[1][1]
      - drvui->b_mat[0][1] * drvui->b_mat[1][0]) / det;
}						 /* end of make_bmat */

/* ************************************************************** */
/* ************************************************************** */

void Output_Spheres (float *radii, int i)

/* routine to output the spheres with vertices defined in s_vert and o_vert
   The radius of the sphere is in radii */
{
#ifdef OPENGL
  float glr, glg, glb;
#endif
  int k;
  int outside = 0;
  int l1;
  int omit;

  for (k = 0; k < nvert; ++k) {
    if (clipflag != 0) {

      outside = 0;
      for (l1 = 0; l1 < 3; l1++)
        if (o_vert[3 * k + l1] < drvui->clip_lim[drvui->frame_no][l1] - 0.01 ||
            o_vert[3 * k + l1] > drvui->clip_lim[drvui->frame_no][l1 + 3] + 0.01)
          outside = 1;
    }
    
    omit = 0;
    for (l1=0;l1<nomits;l1++) {
      if (omit1[l1]==i && omit2[l1] ==k ) omit = 1;
    }

        
    if (omit == 0 && (clipflag == 0 || outside == 0)) {
      fprintf (drvui->fpoutp,
	       " /* Sphere for %c%c%c%c at %8.5f %8.5f %8.5f */ \n",
	       drvui->sphere_l[i][0], drvui->sphere_l[i][1], drvui->sphere_l[i][2], drvui->sphere_l[i][3]
	       , o_vert[3 * k], o_vert[3 * k + 1], o_vert[3 * k + 2]);
      fprintf (drvui->fpoutp, " sphere{<%8.5f, %8.5f, %8.5f>, %8.5f\n",
	       s_vert[3 * k], s_vert[3 * k + 1], s_vert[3 * k + 2], radii[k]);
      fprintf (drvui->fpoutp, "  texture{pigment{color %s  }}\n", drvui->sphere_col_p[i]);
      fprintf (drvui->fpoutp,
	       "   finish{phong %5.2f phong_size %5.2f reflection 0.1}\n",
	       drvui->Phong_Value, drvui->Phong_Size);
      fprintf (drvui->fpoutp, " }\n");
#ifdef OPENGL
      if (gl_view) {
        glPushMatrix ();
        glLoadName(i);
        glPushName(k);
        glTranslatef (s_vert[3 * k], s_vert[3 * k + 1], s_vert[3 * k + 2]);
        (void) sscanf (drvui->sphere_col_v[i], "%f %f %f", &glr, &glg, &glb);
        glColor3f (glr, glg, glb);
        glutSolidSphere (radii[k], 10, 10);
        glPopName();
        glPopMatrix ();
      }
#endif
      if (no_comment == 0)
        fprintf (drvui->fpoutv,
             "# Sphere for %c%c%c%c at %8.5f %8.5f %8.5f \n",
             drvui->sphere_l[i][0], drvui->sphere_l[i][1], drvui->sphere_l[i][2],
             drvui->sphere_l[i][3], o_vert[3 * k], o_vert[3 * k + 1],
             o_vert[3 * k + 2]);
      if (Vrml2) {
        fprintf (drvui->fpoutv, "  Transform {");
        fprintf (drvui->fpoutv, " translation %8.5f %8.5f %8.5f\n",
            s_vert[3 * k], s_vert[3 * k + 1], s_vert[3 * k + 2]);
        fprintf (drvui->fpoutv, " children [\n Shape {");
        fprintf (drvui->fpoutv,
             "appearance Appearance {material Material { diffuseColor %s} }\n",
             drvui->sphere_col_v[i]);
        fprintf (drvui->fpoutv, "  geometry Sphere { radius %8.5f}", radii[k]);
        fprintf (drvui->fpoutv, "}]}");
      } else {
        fprintf (drvui->fpoutv, " Separator {");
        fprintf (drvui->fpoutv, "  Transform {");
        fprintf (drvui->fpoutv, " translation %8.5f %8.5f %8.5f}\n",
            s_vert[3 * k], s_vert[3 * k + 1], s_vert[3 * k + 2]);
        fprintf (drvui->fpoutv, "  Material {");
        fprintf (drvui->fpoutv, " shininess 0.3\n");
        fprintf (drvui->fpoutv, " diffuseColor %s}\n", drvui->sphere_col_v[i]);
        fprintf (drvui->fpoutv, "  Sphere { radius %8.5f}\n", radii[k]);
        fprintf (drvui->fpoutv, " }\n");
      }
    }
  }
}

/* ************************************************************** */
/* ************************************************************** */

float P_to_C (float prob)

/* Routine to convert probability to ellipsoid scaling, C.  Given probability
   'prob', calculate value of C required for ellipsoid,

   C^2 = (delx/sigx)^2 + (dely/sigy)^2 + (delz/sigz)^2,

   where 'prob' is the probability that a random point in the distribution
   will fall inside the ellipsoid.  Table values taken from D.B. Owen,
   "Handbook of Statistical Tables", Addison-Wesley, Reading, Mass., 1962.

   N.B. The present version of the tables only handles probabilities from
   0.01 to 0.99.
*/
#define N_C 99
{
  static double C_table[N_C + 1] =
	{ 0.3389, 0.4299, 0.4951, 0.5479, 0.5932, 0.6334, 0.6699, 0.7035,
    0.7349, 0.7644,
    0.7924, 0.8192, 0.8447, 0.8694, 0.8932, 0.9162, 0.9386, 0.9605,
    0.9818, 1.0026,
    1.0230, 1.0430, 1.0627, 1.0821, 1.1012, 1.1200, 1.1386, 1.1570,
    1.1751, 1.1932,
    1.2110, 1.2288, 1.2464, 1.2638, 1.2812, 1.2985, 1.3158, 1.3330,
    1.3501, 1.3672,
    1.3842, 1.4013, 1.4183, 1.4354, 1.4524, 1.4695, 1.4866, 1.5037,
    1.5209, 1.5382,
    1.5555, 1.5729, 1.5904, 1.6080, 1.6257, 1.6436, 1.6616, 1.6797,
    1.6980, 1.7164,
    1.7351, 1.7540, 1.7730, 1.7924, 1.8119, 1.8318, 1.8519, 1.8724,
    1.8932, 1.9144,
    1.9360, 1.9580, 1.9804, 2.0034, 2.0269, 2.0510, 2.0757, 2.1012,
    2.1274, 2.1544,
    2.1824, 2.2114, 2.2416, 2.2730, 2.3059, 2.3404, 2.3767, 2.4153,
    2.4563, 2.5003,
    2.5478, 2.5997, 2.6571, 2.7216, 2.7955, 2.8829, 2.9912, 3.1365,
    3.3682, 5.9503
  };

  int i;
  double C;

  i = (int) (100.0 * prob);			 /* find correct table entry */
  if (i > N_C)
    i = N_C;
/* interpolate (linear) */
  C = C_table[i - 1] + (prob - 0.01 * i) * (C_table[i] - C_table[i - 1]);
  return ((float) C);
}

/* ************************************************************** */
/* ************************************************************** */

void plot_vrml_poly (int polyno)
{
  int i, j, k, l, m, n;
  float *out;
  int *point, equal;
  char edgecolor[40];

/* get space for unique vertices and a list to point to them */

  if ((out = (float *) malloc ((unsigned) (draw_list * 3 
      * sizeof (float)))) == NULL) {
    fprintf (stderr, "Unable to allocate space in plot_vrml_poly.\n");
    return;
  }
  if ((point =  (int *) malloc ((unsigned) (draw_list * sizeof (int)))) == NULL) {
    free (out);
    fprintf (stderr, "Unable to allocate space in plot_vrml_poly.\n");
    return;
  }
  if (!Vrml2)
    fprintf (drvui->fpoutv, "Coordinate3 { point [\n");
  n = 0;
  for (i = 0; i < draw_list; i++) {		 /* build list of unique points */
    j = poly_list[i];
    for (k = i + 1; k <= i + j; k++) {
      l = poly_list[k] - 1;
      for (m = 0; m < 3; m++)
        out[3 * n + m] = s_vert[3 * l + m];
      equal = 0;
      for (m = 0; m < n; m++) {			 /* check if new */
        if ((fabs (out[3 * m] - out[3 * n]) < 0.0002)
           && (fabs (out[3 * m + 1] - out[3 * n + 1]) < 0.0002)
           && (fabs (out[3 * m + 2] - out[3 * n + 2]) < 0.0002)) {
          point[k] = m;				 /* point previously found */
          equal = 1;
          break;
        }
      }
      if (equal == 0)
        point[k] = n++;				 /* new point - save it by incrementing n */
    }
    i = i + j;
  }
  --n;
  if (Vrml2)
    fprintf (drvui->fpoutv,
	     "}\n geometry IndexedFaceSet { coord Coordinate{ point [\n");
  for (l = 0; l < n; l++)			 /* output the unique set of coordinates */
    fprintf (drvui->fpoutv, " %5.3f %5.3f %5.3f,\n", out[3 * l], out[3 * l + 1],
	     out[3 * l + 2]);
  fprintf (drvui->fpoutv, " %5.3f %5.3f %5.3f]\n", out[3 * n], out[3 * n + 1],
	   out[3 * n + 2]);
  if (Vrml2) {
    fprintf (drvui->fpoutv, " }\n");
    fprintf (drvui->fpoutv, " coordIndex [\n");
  } else {
    fprintf (drvui->fpoutv, "}\n IndexedFaceSet { coordIndex [\n");
  }
  l = 0;
  for (i = 0; i < draw_list - 1; i++) {
    j = poly_list[i];
    l++;
/*    fprintf(drvui->fpoutv,"# %d-sided polygon:\n",j);     */
    for (k = 0; k < j; k++)
      fprintf (drvui->fpoutv, "%d,", point[l++]);
    if (i + j < draw_list - 2) {
      fprintf (drvui->fpoutv, "-1,\n");
    } else {
      if (Vrml2) {
        fprintf (drvui->fpoutv, "-1]\n");
      } else {
        fprintf (drvui->fpoutv, "-1]\n  }\n");
      }
    }
    i = i + j;
  }
  if (Vrml2) {
    fprintf (drvui->fpoutv, " solid FALSE\n");
    fprintf (drvui->fpoutv, " convex TRUE\n");
    fprintf (drvui->fpoutv, " creaseAngle 1.57075\n");
    fprintf (drvui->fpoutv, "  }\n }\n");
  }
  if (edges == 1) {
    if (drvui->p_rad_edge[polyno] != 0.)
      strcpy (edgecolor, drvui->p_col_edge_v[polyno]);
    else
      strcpy (edgecolor, col_edge_v);

    if (Vrml2) {
      fprintf (drvui->fpoutv, " Shape {\n");
      fprintf (drvui->fpoutv,
	       "appearance Appearance {\n  material Material { diffuseColor %s} \n }\n",
	       edgecolor);
      fprintf (drvui->fpoutv,
	       "geometry IndexedLineSet {  coord Coordinate{ point [\n");
      for (l = 0; l < n; l++)			 /* output the unique set of coordinates */
        fprintf (drvui->fpoutv, " %5.3f %5.3f %5.3f,\n", out[3 * l],
             out[3 * l + 1], out[3 * l + 2]);
      fprintf (drvui->fpoutv, " %5.3f %5.3f %5.3f]\n", out[3 * n],
           out[3 * n + 1], out[3 * n + 2]);
      fprintf (drvui->fpoutv, "  }\n");
      fprintf (drvui->fpoutv, "  coordIndex [\n");
    } else {
      fprintf (drvui->fpoutv, " Material {\n   diffuseColor %s \n }\n", edgecolor);
      fprintf (drvui->fpoutv, "IndexedLineSet { coordIndex [\n");
    }
    l = 0;
    for (i = 0; i < draw_list - 1; i++) {
      j = poly_list[i];
      l++;
      for (k = 0; k < j; k++)
        fprintf (drvui->fpoutv, "%d,", point[l++]);
      if (i + j < draw_list - 2) {
        fprintf (drvui->fpoutv, "%d,-1,\n", point[l - j]);
      } else {
        if (Vrml2) {
          fprintf (drvui->fpoutv, "%d,-1]\n  ", point[l - j]);
        } else {
          fprintf (drvui->fpoutv, "%d,-1]\n  }\n }\n", point[l - j]);
        }
      }
      i = i + j;
    }
    if (Vrml2)
      fprintf (drvui->fpoutv, "}\n }\n");
  } else {
    if (!Vrml2)
      fprintf (drvui->fpoutv, " }\n");
  }
  free (out);					 /* release the space */
  free (point);
  return;
}

/* ************************************************************** */
/* ************************************************************** */

void print_sym (void)

/* print symmetry information */
{
  int i, j, k;

 if (drvui->origin1_flag) {
      fprintf(drvui->flout, "\nThe current space group has multiple origins. The generator \n"
          "used by DRAWxtl always selects the origin at a center of inversion.\n"
          "If your structure does not display correctly, add the following line\n"
          "to your input (str) file\n"
          "  xyzoff  %.3f %.3f %.3f\n\n",-drvui->origin_offset[0],
            -drvui->origin_offset[1], -drvui->origin_offset[2]);
      fprintf(stderr, "\n**** \aCheck the listing (.out) file for important\n"
          "      info regarding your space group.\n");
  }
  switch (drvui->sys) {
  case 1:
    fprintf (drvui->flout, "Triclinic - ");
    break;
  case 2:
    fprintf (drvui->flout, "Monoclinic - ");
    break;
  case 3:
    fprintf (drvui->flout, "Orthorhombic - ");
    break;
  case 4:
    fprintf (drvui->flout, "Tetragonal - ");
    break;
  case 5:
    fprintf (drvui->flout, "Hexagonal - ");
    break;
  case 6:
    fprintf (drvui->flout, "Cubic - ");
  }
  if (!drvui->acentric) {
    fprintf (drvui->flout, "Centric - ");
  }
  else {
    fprintf (drvui->flout, "Acentric - ");
  }
  switch (drvui->nbr) {
  case 1:
    fprintf (drvui->flout, "P lattice");
    break;
  case 2:
    fprintf (drvui->flout, "A-centered");
    break;
  case 3:
    fprintf (drvui->flout, "B-centered");
    break;
  case 4:
    fprintf (drvui->flout, "C-centered");
    break;
  case 5:
    fprintf (drvui->flout, "F-centered");
    break;
  case 6:
    fprintf (drvui->flout, "I-centered");
    break;
  case 7:
    fprintf (drvui->flout, "R-centered");
    break;
  default:
    fprintf (drvui->flout, "**** Unknown Lattice ***");
  }
  fprintf (drvui->flout, " with %d Symmetry Operators\n\n", drvui->ng);
  for (i = 1; i <= drvui->ng; ++i) {

    for (j = 1; j <= 3; ++j) {
      for (k = 1; k <= 3; ++k) {
	drvui->rss[i - 1][j - 1][k - 1] = (float) drvui->ss[i - 1][j - 1][k - 1];
	if (drvui->ss[i - 1][j - 1][k - 1] != 0) {
	  if (drvui->ss[i - 1][j - 1][k - 1] > 0) {
	    fprintf (drvui->flout, "+");
	  } else {
	    fprintf (drvui->flout, "-");
	  }
	  switch (k) {
	  case 1:
	    fprintf (drvui->flout, "x");
	    break;
	  case 2:
	    fprintf (drvui->flout, "y");
	    break;
	  case 3:
	    fprintf (drvui->flout, "z");
	  }
	}
      }
      if (drvui->ts[i - 1][j - 1] != 0.0) {
        if (drvui->ts[i - 1][j - 1] == 0.5)
          fprintf (drvui->flout, "+1/2");
        if (drvui->ts[i - 1][j - 1] == 0.25)
          fprintf (drvui->flout, "+1/4");
        if (drvui->ts[i - 1][j - 1] == 0.75)
          fprintf (drvui->flout, "+3/4");
        if (fabs (drvui->ts[i - 1][j - 1] - 0.33333333) < 0.00001)
          fprintf (drvui->flout, "+1/3");
        if (fabs (drvui->ts[i - 1][j - 1] - 0.66666667) < 0.00001)
          fprintf (drvui->flout, "+2/3");
        if (fabs (drvui->ts[i - 1][j - 1] - 0.16666667) < 0.00001)
          fprintf (drvui->flout, "+1/6");
        if (fabs (drvui->ts[i - 1][j - 1] - 0.83333333) < 0.00001)
          fprintf (drvui->flout, "+5/6");
      }
      if (j != 3)
        fprintf (drvui->flout, ",");
    }
    if (i != drvui->ng) {
      if (i % 4 != 0) {
        fprintf (drvui->flout, " ; ");
      } else {
        fprintf (drvui->flout, "\n");
      }
    }
  }
  Conv_Sym_Mat ();
  fprintf (drvui->flout, "\n");
}						 /* end of print_sym */

/* ************************************************************** */
/* ************************************************************** */

void Token_Strip (char string[100], int no)

/*  string[100] - input output string */
/*  no - number of tokens to strip */
/* routine to strip tokens from the leading part of a string */
{
  int i, j;

  for (i = 0, j = 0; i < no; i++) {		 /* loop through tokens to be removed */
    while (string[j++] != ' ')
      string[j - 1] = ' ';			 /* characters to space */
    while (string[j] == ' ')
      j++;					 /* skip all spaces */
  }
  Blank_Strip (string);				 /* move stuff to left */
}

/* ************************************************************** */
/* ************************************************************** */

void usage (void)
{
  fprintf (stderr, "\n****************************************\n\n");
  fprintf (stderr, "  Crystal Structure Drawing Program V%s\n\n", VERSION);
  fprintf (stderr,
	   "Not enough command-line arguments have been specified.\n\n");
  fprintf (stderr, "The full usage is:\n\n");
  fprintf (stderr,
	   "DRAWxtl [-b Xb Yb Zb] [-c] [-C] [-m M] [-n] [-O] [-o Xo Yo Zo]\n");
  fprintf (stderr,
	   "        [-p xlo xhi ylo yhi zlo zhi] [-u] [-v Xr Yr Zr] <file>\n\n");
  fprintf (stderr,
	   "-b x-, y-, z-limits of view box in angstroms (20 20 20)\n");
  fprintf (stderr, "-c include comments in the VRML output\n");
  fprintf (stderr, "-C output multiple cameras to the VRML file\n");
  fprintf (stderr, "-F <map-file> read GSAS Fourier .grd file <map-file>\n");
  fprintf (stderr, "-G do NOT generate openGL graphical view\n");
  fprintf (stderr,
	   "-H do NOT display help information at startup of openGL\n");
  fprintf (stderr, "-m magnification factor for image (1.)\n");
  fprintf (stderr, "-n no display of unit cell\n");
  fprintf (stderr, "-O create orthographic view\n");
  fprintf (stderr, "-o x-, y-, z-coordinates of origin (0.5 0.5 0.5)\n");
  fprintf (stderr,
	   "-p limits of fract. coord. along a, b, and c in display box (0 1 0 1 0 1)\n");
  fprintf (stderr,
	   "-Q quick read of input information but no graphics output\n");
  fprintf (stderr, "-u display axial orientation vector triple\n");
  fprintf (stderr, "-v xrot yrot zrot : View rotation angles (0. 0. 0.)\n");
  fprintf (stderr,
	   "<file> is the name of the file describing the structure.\n\n");
  fprintf (stderr, "Arguments enclosed in brackets [] are optional.\n");
  fprintf (stderr, "<file> is required.\n\n");
  exit (1);
}

#ifdef OPENGL
void resizeWindow (int w, int h)
{

  float ratio;
  GLfloat light_position[] = { 0.0f, 0.0f, 100.0f, 0.0f };
  GLfloat light_direction[] = { 0.0f, 0.0f, 1.0f };

  /* Prevent a divide by zero, when window is too short
     (you cannot make a window of zero width). */
  if (h == 0)
    h = 1;

  ratio = 1.0f *w / h;

/* Set the viewport to be the entire window */
  glViewport (0, 0, w, h);			 /* urx,ury,llx,lly */

/* Reset the coordinate system before modifying */
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();				 /*initialize rotation matrix */

/* Set the correct perspective. */
  if (!gl_ortho)
      gluPerspective (45., ratio, 1., 1000.);	 /* view angle, aspect,near/far clip */
  else {
      if (w<=h)
          glOrtho(-gl_size,gl_size,-gl_size/ratio,gl_size/ratio,-0.f,10000.f);
      else
          glOrtho(-gl_size*ratio,gl_size*ratio,-gl_size,gl_size,-0.f,10000.f);
  }
  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();
  gluLookAt (0.0, 0.0, 60.,			 /* camera position */
	     0.0, 0.0, -1.0,			 /* camera lookat point */
	     0.0f, 1.0f, 0.0f);			 /* camera "up" vector */
  glLightfv (GL_LIGHT0, GL_POSITION, light_position);
  glLightfv (GL_LIGHT0, GL_SPOT_DIRECTION, light_direction);

  glClearColor (drvui->glback[0], drvui->glback[1], drvui->glback[2],0.0f);
/* Update size information */
  gl_height = h;
  gl_width = w;
}


void display_crystal (void)
{
  int i;
  char * p;
  GLdouble objx;
  GLdouble objy;
  GLdouble objz;
  int ypos = gl_height - 20;
  
  char Text1[] = { "Use Arrow keys to move image. "
      "Use q or <esc> to exit, <HOME> to reset position" };
  char Text2[] = { "Left mouse button to rotate about x,y, z - right to zoom in/out."
                   "POV output file will match last orientation."};
  char Text3[] = { "'c' turns graphics cursor on, 'p' moves it to atom at mouse position" };
  char Text4[] = { "'a' moves graphics cursor to nearest atom, 'l' labels atom, 'b' labels bond" };
  char Text5[] = { "'o' toggles between orthographic and perspective mode"};

  GLfloat light_position[] = {0.0f, 0.0f, 100.0f, 0.0f};
  GLfloat light_direction[] = { 0.0f, 0.0f, 1.0f };
  GLdouble modelMatrix1[16];
  GLdouble projMatrix1[16];
  GLint viewport1[4];

  glClear ((GLbitfield) (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT));
  glLoadIdentity ();
  if (!gl_ortho)
      gluLookAt ((POV_Max[0] + POV_Min[0]) / 2., (POV_Max[1] + POV_Min[1]) / 2.,   Scale*.2,	/* camera position */
	     (POV_Max[0] + POV_Min[0]) / 2., (POV_Max[1] + POV_Min[1]) / 2., -1.0,	/* camera lookat point */
	     0.0f, 1.0f, 0.0f);			 /* camera "up" vector */
  else
      gluLookAt ((POV_Max[0] + POV_Min[0]) / 2., (POV_Max[1] + POV_Min[1]) / 2.,   Scale*2,	/* camera position */
	     (POV_Max[0] + POV_Min[0]) / 2., (POV_Max[1] + POV_Min[1]) / 2., -1.0,	/* camera lookat point */
	     0.0f, 1.0f, 0.0f);			 /* camera "up" vector */
  glLightfv (GL_LIGHT0, GL_POSITION, light_position);
  glLightfv (GL_LIGHT0, GL_SPOT_DIRECTION, light_direction);
  glColor3f (0.0f, 0.0f, 0.0f);
  glGetDoublev (GL_MODELVIEW_MATRIX, modelMatrix1);
  glGetDoublev (GL_PROJECTION_MATRIX, projMatrix1);
  glGetIntegerv (GL_VIEWPORT, viewport1);
  if (!view_inst) {
    (void) gluUnProject (10.0, (float) ypos, 0.0, modelMatrix1,
			 projMatrix1, viewport1, &objx, &objy, &objz);
    glRasterPos3d (objx, objy, objz);		 /* put Text1 at 10,gl_height-20 pixels */
    i = 0;
    while (Text1[i] != '\0') {
      glutBitmapCharacter (GLUT_BITMAP_HELVETICA_12, (int) Text1[i++]);
    }
    ypos -= 12;
    (void) gluUnProject (10.0, (float) ypos, 0.0, modelMatrix1,
			 projMatrix1, viewport1, &objx, &objy, &objz);
    glRasterPos3d (objx, objy, objz);		 /* put Text2 at 10,gl_height-32 pixels */
    i = 0;
    while (Text2[i] != '\0') {
      glutBitmapCharacter (GLUT_BITMAP_HELVETICA_12, (int) Text2[i++]);
    }
    ypos -= 12;
    (void) gluUnProject (10.0, (float) ypos, 0.0, modelMatrix1,
			 projMatrix1, viewport1, &objx, &objy, &objz);
    glRasterPos3d (objx, objy, objz);		 /* put Text3 at 10,gl_height-44 pixels */
    i =  0;
    while (Text3[i] != '\0') {
      glutBitmapCharacter (GLUT_BITMAP_HELVETICA_12, (int) Text3[i++]);
    }
    ypos -= 12;
    (void) gluUnProject (10.0, (float) ypos, 0.0, modelMatrix1,
			 projMatrix1, viewport1, &objx, &objy, &objz);
    glRasterPos3d (objx, objy, objz);		 /* put Text4 at 10,gl_height-56 pixels */
    i =  0;
    while (Text4[i] != '\0') {
      glutBitmapCharacter (GLUT_BITMAP_HELVETICA_12, (int) Text4[i++]);
    }
    ypos -= 12;
    (void) gluUnProject (10.0, (float) ypos, 0.0, modelMatrix1,
			 projMatrix1, viewport1, &objx, &objy, &objz);
    glRasterPos3d (objx, objy, objz);		 /* put Text5 at 10,gl_height-68 pixels */
    i =  0;
    while (Text5[i] != '\0') {
      glutBitmapCharacter (GLUT_BITMAP_HELVETICA_12, (int) Text5[i++]);
    }
    ypos -= 20;
  }
/*     display_cursor(); */
  if (cur_show) {
    char Text1[100];
    char Text2[100];
    char Text3[100];
    strcpy(Text3,"");
    if (! ReadFourMap)
        sprintf (Text1,"Cursor: %.4f,%.4f,%.4f (step=%.2fA)", 
				cur_cen[0], cur_cen[1], cur_cen[2], cur_step);
    else
        sprintf (Text1,"Cursor: %.4f,%.4f,%.4f (step=%.2fA). Rho=%.3f",
		  cur_cen[0], cur_cen[1], cur_cen[2], cur_step,
		  InterpolateMap (cur_cen[0], cur_cen[1], cur_cen[2]));
    if (cur_atom[0] > 0) {
        sprintf(Text2," Atoms 1.%s",cur_name[0]);
        strcat(Text1, Text2);
    }
    if (cur_atom[1] > 0) {
      sprintf(Text2," 2.%s",cur_name[1]);
      strcat(Text1,Text2);
      sprintf(Text3,"d12=%6.3fA",dist12);
      if (cur_atom[2] > 0) {
        sprintf(Text2," 3.%s",cur_name[2]);
        strcat(Text1,Text2);
        sprintf(Text2,", d23=%6.3fA, a123=%5.2f",dist23,ang123);
        strcat(Text3,Text2);
        if (cur_atom[3] > 0) {
            sprintf(Text2," 4.%s",cur_name[3]);
            strcat(Text1,Text2);
            sprintf(Text2,", d34=%6.3fA, a234=%5.2f, tor1234=%5.2f",dist34,ang234,torsion_ang);
            strcat(Text3,Text2);
        }
      }
    }
    glPushMatrix();
    (void)gluUnProject (10.0, (float) ypos, 0.0, modelMatrix1,
			projMatrix1, viewport1, &objx, &objy, &objz);
    glRasterPos3d (objx, objy, objz);		 /* put Text1 at 10,ypos pixels */
    for (p=Text1;*p;p++)
      glutBitmapCharacter (GLUT_BITMAP_HELVETICA_12, *p);
    ypos -= 12;
    (void)gluUnProject (10.0, (float) ypos, 0.0, modelMatrix1,
			projMatrix1, viewport1, &objx, &objy, &objz);
    glRasterPos3d (objx, objy, objz);		 /* put Text3 12 pixels lower */
    if (strlen(Text3)) {
      for (p=Text3;*p;p++)
        glutBitmapCharacter (GLUT_BITMAP_HELVETICA_12, *p);
    }
    glPopMatrix();
  }
  glPushMatrix ();
  glTranslatef (gl_pos_x, gl_pos_y, gl_pos_z);
  glRotatef (zrot, 0.0f, 0.0f, 1.0f);
  glRotatef (yrot, 0.0f, 1.0f, 0.0f);
  glRotatef (xrot, 1.0f, 0.0f, 0.0f);
  glTranslatef (-(POV_Max[0] + POV_Min[0]) / 2.0f,
		-(POV_Max[1] + POV_Min[1]) / 2.0f,
		-(POV_Max[2] + POV_Min[2]) / 2.0f);
  draw_cursor ();          /* display openGL cursor */
  glCallList (crystalDL);
  generate_gl_texts();
  glGetIntegerv(GL_VIEWPORT,viewport);             /* save matrices for later */
  glGetDoublev(GL_MODELVIEW_MATRIX,modelMatrix);
  glGetDoublev(GL_PROJECTION_MATRIX,projMatrix);
  glPopMatrix ();
  glutSwapBuffers ();
}

#ifdef OPENGL
void Dump_View(void)
{
  FILE *fp;
  char file[256];
  int state = GL2PS_OVERFLOW, buffsize = 0;
  int format = GL2PS_EPS;
  int sort;
  int options = GL2PS_DRAW_BACKGROUND|GL2PS_OCCLUSION_CULL|GL2PS_BEST_ROOT|GL2PS_SILENT;
  int nbcol = 16;
  GLint viewport[4];

  float cpx,cpy,cpz;
  float m[16];
  GLfloat mat_specular[] = {  1.0f, 1.0f, 1.0f, 1.0f };
  GLfloat light_ambient[] = { AMBLEV, AMBLEV, AMBLEV, 1.0f };
  GLfloat light_specular[] = { 0.1f, 0.1f, 0.1f, 1.0f };
  GLfloat light_diffuse[] = { 1.0f, 1.0f, 1.0f, 1.0f };
  GLfloat mat_shininess[] = { 0.8f };
  GLfloat light_position[] = {0.0f, 1.0f, 1.0f, 0.0f};
  GLfloat light_direction[] = { 0.0f, -1.0f, 0.0f };

  if (ReadFourMap)
     sort = GL2PS_SIMPLE_SORT;
   else
     sort = GL2PS_BSP_SORT;

  glGetIntegerv(GL_VIEWPORT,viewport);
  strcpy(file, drvui->File_Root);
  strcat(file,".ps");

  fp = fopen(file, "wb");

  while(state == GL2PS_OVERFLOW){
    buffsize += 1024*1024;
    gl2psBeginPage(file, "DRAWxtl 4.2", viewport, format, sort, options,
                   GL_RGBA, 0, NULL, nbcol, nbcol, nbcol,
                   buffsize, fp, file);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    if (M_cameras==0) {
        glOrtho(-gl_size,gl_size,-gl_size,gl_size,0.f,10000.f);
    } else {
        gluPerspective (45., (GLdouble)viewport[2]/(GLdouble)viewport[3], 1., 1000.);    
    }
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    cpx = (POV_Max[0] + POV_Min[0]) / 2.0f;
    cpy = (POV_Max[1] + POV_Min[1]) / 2.0f;
    cpz = (POV_Max[2] + POV_Min[2]) / 2.0f;
    glMaterialfv (GL_FRONT, GL_SPECULAR, mat_specular);
    glMaterialfv (GL_FRONT, GL_SHININESS, mat_shininess);
    glLightfv (GL_LIGHT0, GL_SPECULAR, light_specular);
    glLightfv (GL_LIGHT0, GL_DIFFUSE, light_diffuse);
    glLightfv (GL_LIGHT0, GL_AMBIENT, light_ambient);
    glLightModeli (GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
    glEnable (GL_LIGHTING);
    glEnable (GL_LIGHT0);
    glEnable (GL_COLOR_MATERIAL);
    glColorMaterial (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
    glEnable (GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glClearColor (drvui->glback[0], drvui->glback[1], drvui->glback[2],0.0f);
    glClear ((GLbitfield)(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT));
    gluLookAt (cpx, cpy, Scale * .50,    /* camera position */
               cpx, cpy, -1.0,           /* camera lookat point */
               0.0f, 1.0f, 0.0f);        /* camera "up" vector */
    glLightfv (GL_LIGHT0, GL_POSITION, light_position);
    glLightfv (GL_LIGHT0, GL_SPOT_DIRECTION, light_direction);
    glColor3f(0.0f, 0.0f, 0.0f);
    glPushMatrix ();
    glTranslatef (gl_pos_x, gl_pos_y, gl_pos_z);
    quaternion_to_rotmatrix(&Rotq, m);

    glMultMatrixf(m);
    glTranslatef (-cpx,-cpy,-cpz);
    draw_cursor();
    glCallList (crystalDL);
    generate_gl_texts();
    glPopMatrix ();


    state = gl2psEndPage();
  }

  fclose(fp);
  resizeWindow(gl_width, gl_height);
}
#endif
  
static void key (unsigned char k, int x, int y)
{
  int Sense = 1;
  int atomno;

  switch (k) {
  case 27:					 /* Escape */
  case 'Q':
  case 'q':
    close_and_exit();
  case 'D':
  case 'd':                           /* d - dump Postscript view */
#ifdef OPENGL
      if (gl_view) {
          Dump_View();
      }
      break;
#endif
  case 'H':
  case 'h':                           /* h - toggle display of help text */
    view_inst = 1 - view_inst;
    break;
  case 'C':
  case 'c':                           /* c - if cursor off, turn it on */
      if (cur_show) {                 /*     if cursor on, divide step by 5 */
      cur_step *= 0.2f;               /*     if resulting step < 0.01, turn cursor off */
      if (cur_step < 0.01) {
          cur_show = 0;
          cur_step = 0.5;
      }
    } else {
          cur_show = 1;
    }
    break;
  case 'x':                            /* x - move cursor along +x */
    move_cursor(0,1.0f);
    break;
  case 'y':                            /* y - move cursor along +y */
    move_cursor(1,1.0f);
    break;
  case 'z':                            /* z - move cursor along +z */
    move_cursor(2,1.0f);
    break;
  case 'X':                            /* X - move cursor along -x */
    move_cursor(0,-1.0f);
    break;
  case 'Y':                            /* Y - move cursor along -y */
    move_cursor(1,-1.0f);
    break;
  case 'Z':                            /* Z - move cursor along -z */
    move_cursor(2,-1.0f);
    break;
  case 'o':
  case 'O':                           /* O - toggle between orthographic and perspective projection */
    if (!gl_ortho) {
	    gl_ortho = 1;
        gl_size = max(POV_Max[0] - POV_Min[0],POV_Max[1] - POV_Min[1]);
        gl_size = max(gl_size, POV_Max[2] - POV_Min[2]);
    } else {
        gl_ortho = 0;
    }
    resizeWindow(gl_width,gl_height);
    break;
  case 'P':
  case 'p':
    atomno = find_proj_atom(x, y);
    Update_Cursor_List(cur_cen, atomno); /* add this position to cursor list */
    break;
  case 'a':
  case 'A':			/* find nearest atom, calculate distance */
    atomno = find_atom();
    Update_Cursor_List(cur_cen, atomno); /* add this position to cursor list */
    break;
  case 'M':
      Sense = -1;
  case 'm':
      atomno=Maximize_rho(Sense);      /* find local maximum (minimum if Sense < 0) */
      Update_Cursor_List(cur_cen, atomno); // add this position to cursor list
      break;
  case 'L':
  case 'l': {
      unsigned int j, nn, theatom;
      if (strlen(cur_name[0])==0) break;
      if (cur_atom[0]<1) break;
      theatom=0;			    

      if (cur_atom[1]>0) theatom=1;
      if (cur_atom[2]>0) theatom=2;
      if (cur_atom[3]>0) theatom=3;

      drvui->label_fn[drvui->nlabel]= drvui->frame_no;	  
      drvui->label_x[drvui->nlabel][0]=cur_cen[0];
      drvui->label_x[drvui->nlabel][1]=cur_cen[1];
      drvui->label_x[drvui->nlabel][2]=cur_cen[2];
      nn=0;
      for (j=0;j<strlen(cur_name[theatom]);j++) {
        drvui->label_label[drvui->nlabel][nn++]=cur_name[theatom][j];
      }
      drvui->label_label[drvui->nlabel][nn]='\0';
      drvui->nlabel++;
      drvui->labels_moved = 1;
      break; }
  case 'B':
  case 'b': {
      int first,second;
      if (dist12<0.01) break;
//fprintf(stderr,"atoms in cache : %d %d %d\n",cur_atom[0],cur_atom[1],cur_atom[2]);	  
      if (cur_atom[0]<1) break;
      if (cur_atom[1]<1) break;
      first=cur_atom[0];
      second=cur_atom[1];
  
      drvui->label_fn[drvui->nlabel] = drvui->frame_no;	  
      drvui->label_x[drvui->nlabel][0]=(o_vert[3*second]+o_vert[3*first])/2.0f;
      drvui->label_x[drvui->nlabel][1]=(o_vert[3*second+1]+o_vert[3*first+1])/2.0f;
      drvui->label_x[drvui->nlabel][2]=(o_vert[3*second+2]+o_vert[3*first+2])/2.0f;
      sprintf(drvui->label_label[drvui->nlabel],"%.3f",dist12);
      drvui->nlabel++;
      drvui->labels_moved = 1;
      break; }
  case 'U':                            /* U Undo */
  case 'u':
      if (nomits>0) nomits--;
      fflush(drvui->fpoutp);
      rewind(drvui->fpoutp);
      fflush(drvui->fpoutv);
      rewind(drvui->fpoutv);
      write_header();
      glNewList(crystalDL,GL_COMPILE);
      do_drawing();                /* regenerate drawing */
      glEndList ();
      if (drvui->noshadow == 1 ) 
          fprintf (drvui->fpoutp, " no_shadow}\n");
      else
          fprintf (drvui->fpoutp, " }\n");
      fprintf (drvui->fpoutp, "rotate <xrot, 0, 0>\n");
      fprintf (drvui->fpoutp, "rotate <0, yrot, 0>\n");
      fprintf (drvui->fpoutp, "rotate <0, 0, zrot>\n");
      fprintf (drvui->fpoutp, "}\n");
      (void) fflush (drvui->fpoutp);
      (void) fflush (drvui->fpoutp);
      if (Vrml2)
          fprintf (drvui->fpoutv, "]\n");
      fprintf (drvui->fpoutv, " }\n");
      if (Vrml2) {
          if (yrot != 0.0)
              fprintf (drvui->fpoutv, " }\n");
          if (zrot != 0.0)
              fprintf (drvui->fpoutv, " }\n");
      }
      (void) fflush (drvui->flout);
      break;
  case 'R':                              /* Redo */
  case 'r':
      fprintf(stderr,"redo: nomits %d, omit1[nomits+1]=%d, omit2 %d\n",nomits,
           omit1[nomits+1],omit2[nomits+1]);
      if (omit2[nomits]+omit1[nomits] != 0) nomits++;
      fflush(drvui->fpoutp);
      rewind(drvui->fpoutp);
      fflush(drvui->fpoutv);
      rewind(drvui->fpoutv);
      write_header();
      glNewList(crystalDL,GL_COMPILE);
      do_drawing();                /* regenerate drawing */
      glEndList ();

      fprintf (drvui->fpoutp, " }\n");
      fprintf (drvui->fpoutp, "rotate <xrot, 0, 0>\n");
      fprintf (drvui->fpoutp, "rotate <0, yrot, 0>\n");
      fprintf (drvui->fpoutp, "rotate <0, 0, zrot>\n");
      fprintf (drvui->fpoutp, "}\n");
      (void) fflush (drvui->fpoutp);
      (void) fflush (drvui->fpoutp);
      if (Vrml2)
          fprintf (drvui->fpoutv, "]\n");
      fprintf (drvui->fpoutv, " }\n");
      if (Vrml2) {
          if (yrot != 0.0)
              fprintf (drvui->fpoutv, " }\n");
          if (zrot != 0.0)
              fprintf (drvui->fpoutv, " }\n");
      }
      (void) fflush (drvui->fpoutv);
      (void) fflush (drvui->flout);
      break;
  }
  glutPostRedisplay ();
}

static void specialkey (int k, int x, int y)
{
  switch (k) {
  case GLUT_KEY_LEFT:
    gl_pos_x -= 2.0f *gl_offset;

    break;
  case GLUT_KEY_RIGHT:
    gl_pos_x += 2.0f *gl_offset;

    break;
  case GLUT_KEY_DOWN:
    gl_pos_y -= 2.0f *gl_offset;

    break;
  case GLUT_KEY_UP:
    gl_pos_y += 2.0f *gl_offset;

    break;
  case GLUT_KEY_HOME:
    gl_pos_x = gl_pos_y = gl_pos_z = 0.0f;
    gl_offset = 1.0;
    break;
  default:
    break;
  }
  glutPostRedisplay ();
}

static void mouse (int button, int state, int x, int y)
{
  float trans[3];

  if (button == GLUT_LEFT_BUTTON) {
    if (state == GLUT_DOWN) {
      if (glutGetModifiers() & GLUT_ACTIVE_CTRL) {
		  start_picking(x,y);
      } else if (glutGetModifiers() & GLUT_ACTIVE_SHIFT) {
          if (drvui->nlabel == 1) return;
          Event = DRAG;
          if (c_pick == 0 ) {
              trans[0] = gl_pos_x;
              trans[1] = gl_pos_y;
              trans[2] = gl_pos_z;
              c_pick=pick_label(x,y,gl_width,gl_height, trans);
              Oldx = x;
              Oldy = y;
          }
      } else {
           Event = SPIN;
           Oldx = x;
           Oldy = y;
           c_pick = 0;
           return;
      }
    } else {
        Event = 0;
        c_pick = 0;
    }
  }

  if (button == GLUT_RIGHT_BUTTON) {
     if (state == GLUT_DOWN) {
         Event = ZOOM;
         Oldx = x;
         Oldy = y;
     } else {
         Event = 0;
     }
  }
}

static void mousemotion (int x, int y)
{
    double cosy;
    float m[16];
	float *v, dx, dy;
    float ratio;
    GLdouble objx, objy, objz, objx1, objy1, objz1;

    Mx = x;
    My = y;
    dx = (float)(x - Oldx) * Mscale / gl_width;
    dy = (float)(Oldy - y) * Mscale / gl_height;
    if (!Event || (dx == 0.0f && dy == 0.0f)) return;
    v = Trans;
    if (Event == SPIN ) {
         trackball();
         qmult(&Spin, &Rotq, &Rotq);
    } else if (Event == ZOOM) {
         gl_size -= (dx+dy)/2.0f;
         glMatrixMode(GL_PROJECTION);
         glLoadIdentity();
         ratio = 1.0f*gl_width/gl_height;
         if (gl_width<=gl_height)
             glOrtho(-gl_size,gl_size,-gl_size/ratio,gl_size/ratio,0.f,10000.f);
         else
             glOrtho(-gl_size*ratio,gl_size*ratio,-gl_size,gl_size,0.f,10000.f);
         glMatrixMode(GL_MODELVIEW);
         glutPostRedisplay ();
         return;
    } else if (Event == DRAG) {
      if (c_pick != 0) {
          float factor;
          if (drvui->nlabel == 1) return;
          if (gl_ortho)
              factor = gl_size * Scale * 0.0002f;
          else
              factor = (float)gl_width * Scale * 0.00003f;
          drvui->labels_moved = 1;
          gluUnProject(Oldx, viewport[3]-Oldy, 0, modelMatrix, projMatrix, viewport,
                      &objx, &objy, &objz);
          gluUnProject(x, viewport[3]-y, 0, modelMatrix, projMatrix, viewport,
                      &objx1, &objy1, &objz1);
          if (c_pick != drvui->triple[0]) {
              drvui->label_x[c_pick][0] += factor * (float)(objx1 - objx);
              drvui->label_x[c_pick][1] += factor * (float)(objy1 - objy);
              drvui->label_x[c_pick][2] += factor * (float)(objz1 - objz);
          } else {
              offset[0] += 5.0f * factor * (float)(objx1 - objx);
              offset[1] += 5.0f * factor * (float)(objy1 - objy);
              offset[2] += 5.0f * factor * (float)(objz1 - objz);
          }
          Oldx=x;
          Oldy=y;
          glutPostRedisplay();
          return;
       }
    }  

/* recalculate rotation angles. */

    quaternion_to_rotmatrix(&Rotq, m);

    yrot = -(float)(asin(m[2]) * RAD);
    cosy = cos(yrot/RAD);
    if (fabs(cosy) > 1.0e-3) {
        xrot = (float)(atan2(m[6]/cosy,m[10]/cosy)*RAD);
        zrot = (float)(atan2(m[1]/cosy,m[0]/cosy)*RAD);
    } else {
        zrot = 0.0f;
        xrot = (float)(atan2(-m[9],m[5])*RAD);
    }
/* redraw the window */
    glutPostRedisplay ();
}

static void start_picking (int x, int y)
{
    GLint viewport[4];
    int hits;
    int i, n;
    float  ratio = 1.0f *gl_width / gl_height;
 
    memset(selectBuf,0,BUFSIZE);
    glGetIntegerv(GL_VIEWPORT,viewport);

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glRenderMode(GL_SELECT);
    glSelectBuffer(BUFSIZE,selectBuf);
    glInitNames();
    glPushName(~0);
    glMatrixMode(GL_PROJECTION);
         
    glPushMatrix();
    glLoadIdentity();
    gluPickMatrix(x,viewport[3]-y,7,7,viewport);

    if (!gl_ortho)
      gluPerspective(45.,(GLdouble)ratio,1.,1000.);
    else {
      if (gl_width <= gl_height)
          glOrtho(-gl_size,gl_size,-gl_size/ratio,gl_size/ratio,0.f,10000.f);
      else
          glOrtho(-gl_size*ratio,gl_size*ratio,-gl_size,gl_size,0.f,10000.f);
    } 
    glMatrixMode(GL_MODELVIEW);

    glTranslatef (gl_pos_x, gl_pos_y, gl_pos_z);
    glRotatef (zrot, 0.0f, 0.0f, 1.0f);
    glRotatef (yrot, 0.0f, 1.0f, 0.0f);
    glRotatef (xrot, 1.0f, 0.0f, 0.0f);
    glTranslatef (-(POV_Max[0] + POV_Min[0]) * 0.5f,
		-(POV_Max[1] + POV_Min[1]) * 0.5f,
		-(POV_Max[2] + POV_Min[2]) * 0.5f);
    glCallList(crystalDL);

    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();
    
    hits = glRenderMode(GL_RENDER);
    glFlush();
  
    if (hits > 0) {
         process_hits (hits,selectBuf);
    } else {
        float trans[3];
        trans[0] = gl_pos_x;
        trans[1] = gl_pos_y;
        trans[2] = gl_pos_z;
        hits = pick_label(x, y, gl_width, gl_height, trans);
        if (hits > 0 && hits <= drvui->nlabel) {
            if (hits != drvui->triple[0] && hits != drvui->triple[1]
                && hits != drvui->triple[2] && hits != drvui->triple[3]) {
                for (n=hits; n<drvui->nlabel; n++) {         /* remove label, but not a, b, c, or origin */
                    drvui->label_fn[n] = drvui->label_fn[n+1];
                    strcpy(drvui->label_label[n],drvui->label_label[n+1]);
                    for (i=0; i<3; i++)
                        drvui->label_x[n][i] = drvui->label_x[n+1][i];
                    for (i=0; i<4; i++)
                        if (drvui->triple[i] == n+1)
                            drvui->triple[i] = n;
                }
                drvui->nlabel--;
            }
        }
    }
    glutPostRedisplay();

}

void process_hits (GLint hits, GLuint buffer[])
{
    unsigned int i,j;
    GLuint names,*ptr,minZ, *ptrNames=NULL,numberOfNames;
    int handled=0;

    printf ("hits = %d\n", (int) hits);
    numberOfNames=0;
    
    ptr = (GLuint *) buffer;
    minZ = 0xffffffff;
    for (i = 0; i < (unsigned)hits; i++) { 
        names = *ptr;
        ptr++;
        if (*ptr < minZ && *(ptr+2) != -1) {
            numberOfNames = names;
            minZ = *ptr;
            ptrNames = ptr+2;
        }   
        ptr += names+2;
    } 
    if (numberOfNames==0) return;
	
    ptr = ptrNames;
    if (*ptr >200000) {
   	    for (j=0;j<(unsigned)nomits;++j) {
   		    if ((unsigned)omit1[j]==*ptr && (unsigned)omit2[j]==*(ptr+1)){
		        omit3[j]++;
		        if (omit3[j]>6) omit3[j]=1;
                handled=1;
            }
        }	
   	} 
    if (handled==0) {	
       if (*ptr >200000) omit3[nomits]=1;
       omit1[nomits]=*ptr;
       ptr++;
       omit2[nomits++]=*ptr;
    }

    fflush(drvui->fpoutp);
    rewind(drvui->fpoutp);
    fflush(drvui->fpoutv);
    rewind(drvui->fpoutv);
    write_header();
    glNewList(crystalDL,GL_COMPILE);
    do_drawing();                /* regenerate drawing */
    glEndList ();

    fprintf (drvui->fpoutp, " }\n");
    fprintf (drvui->fpoutp, "rotate <xrot, 0, 0>\n");
    fprintf (drvui->fpoutp, "rotate <0, yrot, 0>\n");
    fprintf (drvui->fpoutp, "rotate <0, 0, zrot>\n");
    fprintf (drvui->fpoutp, "}\n");
    (void) fflush (drvui->fpoutp);
    (void) fflush (drvui->fpoutp);
    if (Vrml2)
        fprintf (drvui->fpoutv, "]\n");
    fprintf (drvui->fpoutv, " }\n");
    if (Vrml2) {
        if (yrot != 0.0)
            fprintf (drvui->fpoutv, " }\n");
        if (zrot != 0.0)
            fprintf (drvui->fpoutv, " }\n");
    }

    (void) fflush (drvui->flout);
}

static int pick_label (int x, int y, int w, int h, float trans[3])
{
// picks the label at the current cursor position

    GLint viewport[4];
    int i;
    GLuint names,*ptr,minZ;
    GLuint *ptrNames = 0, numberOfNames=0;
    int hits;
    int label;
    float  ratio = 1.0f *(float)w / (float)h;

    glGetIntegerv(GL_VIEWPORT,viewport);
    memset(selectBuf,0,BUFSIZE);
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glSelectBuffer(BUFSIZE,selectBuf);
    glRenderMode(GL_SELECT);

    gluPickMatrix(x,viewport[3]-y, 25, 25,viewport);
    if (!gl_ortho)
        gluPerspective(45.,ratio,1.,1000.);
    else {
        if (gl_width <= gl_height)
          glOrtho(-gl_size,gl_size,-gl_size/ratio,gl_size/ratio,0.f,10000.f);
        else
          glOrtho(-gl_size*ratio,gl_size*ratio,-gl_size,gl_size,0.f,10000.f);
         }
                                            
    glMatrixMode(GL_MODELVIEW);
    glInitNames();
    glPushName(~0);
    glTranslatef (trans[0], trans[1], trans[2]);
    glRotatef (zrot, 0.0f, 0.0f, 1.0f);
    glRotatef (yrot, 0.0f, 1.0f, 0.0f);
    glRotatef (xrot, 1.0f, 0.0f, 0.0f);
    glTranslatef (-(POV_Max[0] + POV_Min[0]) / 2.0f,
		-(POV_Max[1] + POV_Min[1]) / 2.0f,
		-(POV_Max[2] + POV_Min[2]) / 2.0f);
    generate_gl_texts();

    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode (GL_MODELVIEW);
    glFlush();

    hits = glRenderMode(GL_RENDER);
    if (hits == 0) return(0);
    ptr = (GLuint *) selectBuf;
    minZ = 0xffffffff;
    for (i = 0; i < hits; i++) { 
        names = *ptr;
        ptr++;
        if (*ptr < minZ && *(ptr+2) != (GLuint)-1) {
            numberOfNames = names;
            minZ = *ptr;
            ptrNames = ptr+2;
        }  
        ptr += names+2;
    }
    if (numberOfNames==0) return(0);
    ptr = ptrNames;
    label=*ptr;
    if (label >= drvui->nlabel) return(0); // not a label ?
    return(label);
}

void write_header(void)
{
      float Text_Size;

      if (Vrml2)
	fprintf (drvui->fpoutv, "#VRML V2.0 utf8\n\n");
      else
	fprintf (drvui->fpoutv, "#VRML V1.0 ascii\n\n");
      fprintf (drvui->fpoutv, "# Scene file created by DRAWxtl V%s from %s\n",
	       VERSION, drvui->Str_File_Inp);
      fprintf (drvui->fpoutv,
	       "# with command line options: -b %6.2f %6.2f %6.2f -o %5.2f %5.2f %5.2f\n",
	       boxlim[0], boxlim[1], boxlim[2], origin[0], origin[1],
	       origin[2]);
      fprintf (drvui->fpoutv,
	       "#  -p %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f -v %5.2f %5.2f %5.2f\n",
	       drvui->cryst_lim[drvui->frame_no][0], drvui->cryst_lim[drvui->frame_no][3],
           drvui->cryst_lim[drvui->frame_no][1], drvui->cryst_lim[drvui->frame_no][4],
           drvui->cryst_lim[drvui->frame_no][2], drvui->cryst_lim[drvui->frame_no][5],
           xrot, yrot, zrot);

/* Output VRML header */

      if (!Vrml2) {
        fprintf (drvui->fpoutv,
		 "DEF BackgroundColor Info { string \"%s\" }\n", drvui->col_bg_v);
        fprintf (drvui->fpoutv, "Separator {\n");
        fprintf (drvui->fpoutv, " Material {\n");
        fprintf (drvui->fpoutv, "  ambientColor    0.1 0.1 0.1\n");
        fprintf (drvui->fpoutv, "  diffuseColor    1 1 1\n");
        fprintf (drvui->fpoutv, "  specularColor   0.9 0.9 0.9\n");
        fprintf (drvui->fpoutv, "  shininess       0.8\n");
        fprintf (drvui->fpoutv, " }\n");
      }
      Text_Size = 0.0005f *Scale;

      if (!M_cameras && Vrml2) {		 /* Generate Orthographic Projection for VRML97 */
        fprintf (drvui->fpoutv, "NavigationInfo {type \"EXAMINE\"}\n");
        fprintf (drvui->fpoutv, "Background { skyColor %s }\n", drvui->col_bg_v);
        fprintf (drvui->fpoutv,
		 "Viewpoint { fieldOfView 0.1 position 0. 0. %8.3f }\n", Scale);
        fprintf (drvui->fpoutv, "DirectionalLight {direction 0 0 -1}\n");
        fprintf (drvui->fpoutv, "DirectionalLight {direction 0 0 1}\n");
      } else {
        if (!Vrml2) {				 /* simulate orthographic projection for VRML1.0 */
          float fscale, fangle;

          if (!M_cameras) {
            fscale = Scale * (10.0f / 7.0f);
            fangle = 0.07853f;
          } else {
            fscale = Scale / 7.0f;
            fangle = 0.78537f;
          }
          fprintf (drvui->fpoutv, "PerspectiveCamera {\n");
          fprintf (drvui->fpoutv, "  position        0 0 %8.3f\n", fscale);
          fprintf (drvui->fpoutv, "  orientation     0 0 1 0\n");
          fprintf (drvui->fpoutv, "  focalDistance %8.3f\n", fscale);
          fprintf (drvui->fpoutv, "  heightAngle   %8.5f\n", fangle);
          fprintf (drvui->fpoutv, "}\n");
          fprintf (drvui->fpoutv,
		   "DirectionalLight{ direction 0 0 -1 intensity 1.0}\n");
          fprintf (drvui->fpoutv,
		   "DirectionalLight{ direction 0 0 1 intensity 1.0}\n");
        } else {
          fprintf (drvui->fpoutv, "NavigationInfo {type \"EXAMINE\"}\n");
          fprintf (drvui->fpoutv, "Background { skyColor %s }\n", drvui->col_bg_v);
          fprintf (drvui->fpoutv, "Viewpoint { position 0. 0. %8.3f }\n", Scale / 7.);
          fprintf (drvui->fpoutv, "DirectionalLight {direction 0 0 -1}\n");
          fprintf (drvui->fpoutv, "DirectionalLight {direction 0 0 1}\n");
        }
      }

      if (Vrml2) {
        fprintf (drvui->fpoutv, "Transform {");
        fprintf (drvui->fpoutv, " rotation 0 0 1 %f\n", zrot / RAD);
        fprintf (drvui->fpoutv, " children Transform {");
        fprintf (drvui->fpoutv, " rotation 0 1 0 %f\n", yrot / RAD);
        fprintf (drvui->fpoutv, " children Transform {");
        fprintf (drvui->fpoutv, " rotation 1 0 0 %f\n", xrot / RAD);
        fprintf (drvui->fpoutv, "  children [\n");
      } else {
        fprintf (drvui->fpoutv, "Rotation {");
        fprintf (drvui->fpoutv, " rotation 0 0 1 %f}\n", zrot / RAD);
        fprintf (drvui->fpoutv, " Rotation {");
        fprintf (drvui->fpoutv, " rotation 0 1 0 %f}\n", yrot / RAD);
        fprintf (drvui->fpoutv, " Rotation {");
        fprintf (drvui->fpoutv, " rotation 1 0 0 %f}\n", xrot / RAD);
      }

/* Output POV header */

      fprintf (drvui->fpoutp, "/*\n\n    POVxtl-Persistence of Vision Ray Tracer\n");
      fprintf (drvui->fpoutp, "    Crystal Structure Drawing Program V%s\n\n",
	       VERSION);
      fprintf (drvui->fpoutp,
	       "       Copyright 1996, 1997, 2000-2005 by Larry Finger, Martin Kroeker, and Brian Toby\n");
      fprintf (drvui->fpoutp, "          (all rights reserved).\n\n");
      fprintf (drvui->fpoutp, " Scene file created from %s \n", drvui->Str_File_Inp);
      fprintf (drvui->fpoutp,
	       " with command line options: -b %6.2f %6.2f %6.2f -o %5.2f %5.2f %5.2f \n",
	       boxlim[0], boxlim[1], boxlim[2], origin[0], origin[1],
	       origin[2]);
      fprintf (drvui->fpoutp,
	       "  -p %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f -v %5.2f %5.2f %5.2f -m %5.2f\n*/\n\n",
	       drvui->cryst_lim[drvui->frame_no][0], drvui->cryst_lim[drvui->frame_no][3],
           drvui->cryst_lim[drvui->frame_no][1], drvui->cryst_lim[drvui->frame_no][4],
           drvui->cryst_lim[drvui->frame_no][2], drvui->cryst_lim[drvui->frame_no][5],
           xrot, yrot, zrot, Magnification);

      fprintf (drvui->fpoutp, "#include \"colors.inc\"\n");
      fprintf (drvui->fpoutp, "#include \"chars.inc\"\n");
      if (!M_cameras) {
        fprintf (drvui->fpoutp, "camera{ orthographic\n");
      } else {
        fprintf (drvui->fpoutp, "camera{\n");
      }
      fprintf (drvui->fpoutp, " location< 0 ,0, %f>\n", 1.1 * Scale);
      if (!M_cameras) {
        fprintf (drvui->fpoutp, " up <0,%f,0>\n", 0.1 * Scale);
        fprintf (drvui->fpoutp, " right <%f,0,0>\n", -0.1 * Scale);
      } else {
        fprintf (drvui->fpoutp, " up <0, 0.1, 0>\n");
        fprintf (drvui->fpoutp, " right <-0.1, 0, 0>\n");
      }
      fprintf (drvui->fpoutp, " look_at <0,0,0>\n");
      fprintf (drvui->fpoutp, "}\n");
      fprintf (drvui->fpoutp, "background{color %s}\n", drvui->col_bg_p);
      fprintf (drvui->fpoutp, "light_source {< 0, 0, %f>\n", Scale);
      fprintf (drvui->fpoutp, " color red 4.0 green 4.0 blue 4.0\n}\n");
      fprintf (drvui->fpoutp, "object{\n union{\n");
      fprintf (drvui->fpoutp, "\n");
      fprintf (drvui->fpoutp, "#declare xrot = %5.1f;\n", xrot);
      fprintf (drvui->fpoutp, "#declare yrot = %5.1f;\n", yrot);
      fprintf (drvui->fpoutp, "#declare zrot = %5.1f;\n", zrot);
      fprintf (drvui->fpoutp, "\n");
      fflush(drvui->fpoutp);
}
#endif

int not_in_slab(float x, float y, float z)
{
/* BOX_CONTAINS_POINT_3D from John Burkardt's geometry.c */
  float dot;
  float normsq;

  dot = dot0_3d (slabx1,slaby1,slabz1,slabx2,slaby2,slabz2,x,y,z);
  if (dot <0.0) return 1;
	
  normsq = enormsq0_3d(slabx1,slaby1,slabz1,slabx2,slaby2,slabz2);
  if (normsq <dot) return 1;
	
  dot = dot0_3d (slabx1,slaby1,slabz1,slabx3,slaby3,slabz3,x,y,z);
  if (dot <0.0) return 1;

  normsq = enormsq0_3d(slabx1,slaby1,slabz1,slabx3,slaby3,slabz3);
  if (normsq <dot) return 1;

  dot = dot0_3d (slabx1,slaby1,slabz1,slabx4,slaby4,slabz4,x,y,z);
  if (dot <0.0) return 1;

  normsq = enormsq0_3d(slabx1,slaby1,slabz1,slabx4,slaby4,slabz4);
  if (normsq <dot) return 1;

  return 0;
}

float dot0_3d (float x0, float y0, float z0, float x1, float y1, float z1,
	       float x2, float y2, float z2)
{
/* 3D dot product of (P1-P0) and (P2-P0), John Burkardt */

  return (x1 - x0) * (x2 - x0) + (y1 - y0) * (y2 - y0) +
          (z1 - z0) * (z2 - z0);
}

float enormsq0_3d (float x0, float y0, float z0, float x1, float y1, float z1)
{
/* square of the Euclidean norm of (P1-P0) in 3D, John Burkardt */

  return (x1 - x0) * (x1 - x0) + (y1 - y0) * (y1 - y0) +
          (z1 - z0) * (z1 - z0);
}

/* Initialize the trackball interface. */

void tb_init(void)
{

	/* Set the default trackball size and mouse scale factor. */

	Tbsize = 0.5f;
	Mscale = 2.5f;

	/* Default is no rotation or translation. */

	Spin.s = 1.0;
	vzero(Spin.v);
	Rotq.s = 1.0;
	vzero(Rotq.v);
	vzero(Trans);

	/* Initialize the other instance variables.  The rest are set the
	first time the user clicks the mouse. */

	Event = 0;
	idle_redraw_ = FALSE;
	changed = TRUE;
    Rotq = XYZ_Rot_to_Q(xrot, yrot, zrot);
}

void trackball(void)
{
	float p1x, p1y, p2x, p2y;
	float theta, t;
	XYZ p1, p2, axis, d;

	QUAT *q = &Spin;

	if (Mx == Oldx && My == Oldy) {

		/* Zero rotation. */

		q->s = 1.;
		vzero(q->v);
		return;
	}
	/* Scale the old and new mouse positions to (-1, 1). */

	window_to_tb((float)Oldx, (float)Oldy, &p1x, &p1y);
	window_to_tb((float)Mx, (float)My, &p2x, &p2y);

	/*First, figure out z-coordinates for projection of P1 and P2 to
	  the deformed sphere.                                            */

	vset(p1, p1x, p1y, tb_project_to_sphere(Tbsize, p1x, p1y));
	vset(p2, p2x, p2y, tb_project_to_sphere(Tbsize, p2x, p2y));

	/* Now the axis of rotation. */

	vcross(p1, p2, axis);

	/* Figure out how much to rotate around that axis. */

	vsub(p1, p2, d);
	t = vlength(d) * Mscale / 20.0f;
	if (t > 1.0) t = 1.0;
	if (t < -1.0) t = -1.0;
	theta = (float)asin(t);

	/* Return a rotation quaternion */

	axis_to_quaternion(axis, theta, q);
}

/* Map mouse click mx, my to a more convenient (-1.0, 1.0) range, based on
window size. */

void window_to_tb(float mx, float my, float *x, float *y)
{
	*x = (2.0f * mx) / (float)gl_width - 1.f;
	*y = 1.0f - (2.0f * my) / (float)gl_height;
}

/* Project an x, y pair onto a sphere of radius r OR a hyperbolic sheet if we
are away from the center of the sphere.  */

float tb_project_to_sphere(float r, float x, float y)
{
	float d, t, z;

	d = (float)sqrt(x * x + y * y);
	if (d < r * 0.70710678118654752440) {        /* inside sphere */
		z = (float)sqrt(r * r - d * d);
	} else {                        /* on hyperbola */
		t = r / (float)sqrt(2.0);
		z = t * t / d;
	}
	return (z);
}

/* Given an axis and an angle, compute a rotation quaternion. */

void axis_to_quaternion(float *axis, float theta, QUAT *quat)
{
	quat->s = (float)cos(theta / 2.0);
	vcopy(axis, quat->v);
	vnormalize(quat->v);
	vscale(quat->v, (float)sin(theta / 2.0));
}

/* Given two rotations, q1 and q2, expressed as quaternions,
quaternion multiplication yields the equivalent product rotation.
This routine also normalizes the result every COUNT times it is
called, to keep error from creeping in. */

#define COUNT 100

void qmult(QUAT *q1, QUAT *q2, QUAT *dest)
{
	XYZ v1, v2;
	QUAT q;
	static int count = 0;

	vcopy(q1->v, v1);
	vcopy(q2->v, v2);
	vscale(v1, q2->s);
	vscale(v2, q1->s);
	vcross(q1->v, q2->v, q.v);
	vadd(v1, q.v, q.v);
	vadd(v2, q.v, q.v);
	q.s = q1->s * q2->s - vdot(q1->v, q2->v);

	*dest = q;

	if (++count > COUNT) {
		count = 0;
		qnormalize(dest);
	}
}

/* Normalize greatest component, to avoid problems that occur when the
component we're normalizing gets close to zero (and the other components may
add up to more than 1.0 because of rounding error).  */

void qnormalize(QUAT *q)
{
	int which, i;
	float gr, v[4];

	vcopy(q->v, v);
	v[3] = q->s;

	which = 0;
	gr = v[0];
	for (i = 1; i < 4; i++) {
		if (fabs(v[i]) > fabs(gr)) {
			gr = v[i];
			which = i;
		}
	}

	v[which] = 0.;
	v[which] = (float)sqrt(1.0 - (v[0] * v[0] + v[1] * v[1] +
		v[2] * v[2] + v[3] * v[3]));

	/* Check to see if we need negative square root. */

	if (gr < 0.0) {
		v[which] = -v[which];
	}
	vcopy(v, q->v);
	q->s = v[3];
}

/* Build a rotation matrix from a rotation quaternion. */

void quaternion_to_rotmatrix(QUAT *q, float *m)
{
	float w, x, y, z;

	w = q->s;
	x = q->v[0];
	y = q->v[1];
	z = q->v[2];

	/* Column major order, used to left multiply points, i.e., P'=M*P */

	m[0]  = 1.f - 2.f * (y * y + z * z);
	m[1]  = 2.f * (x * y + w * z);
	m[2]  = 2.f * (x * z - w * y);
	m[3]  = 0.;

	m[4]  = 2.f * (x * y - w * z);
	m[5]  = 1.f - 2.f * (x * x + z * z);
	m[6]  = 2.f * (y * z + w * x);
	m[7]  = 0.;

	m[8]  = 2.f * (x * z + w * y);
	m[9]  = 2.f * (y * z - w * x);
	m[10] = 1.f - 2.f * (x * x + y * y);
	m[11] = 0.;

	m[12] = 0.;
	m[13] = 0.;
	m[14] = 0.;
	m[15] = 1.;
}

float *vnew(float x, float y, float z)
{
	float *v;

	v = (float *)malloc(3 * sizeof(float));
	v[0] = x;
	v[1] = y;
	v[2] = z;

	return (v);
}

void vset(float *v, float x, float y, float z)
{
	v[0] = x;
	v[1] = y;
	v[2] = z;
}

void vcopy(float *vsrc, float *vdst)
{
	vdst[0] = vsrc[0];
	vdst[1] = vsrc[1];
	vdst[2] = vsrc[2];
}

void vzero(float *v)
{
	v[0] = 0.0;
	v[1] = 0.0;
	v[2] = 0.0;
}

void vnormalize(float *v)
{
#if defined(_IEEE) || defined(_IEEE_FP)
	vscale(v, 1.0 / vlength(v));
#else
	float l;

	l = vlength(v);
	if (l > EPS) {
		vscale(v, 1.0f / l);
	}
#endif
}

float vlength(float *v)
{
	return (float)(sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2]));
}

void vscale(float *v, float f)
{
	v[0] *= f;
	v[1] *= f;
	v[2] *= f;
}

void vmult(float *src1, float *src2, float *dst)
{
	dst[0] = src1[0] * src2[0];
	dst[1] = src1[1] * src2[1];
	dst[2] = src1[2] * src2[2];
}

void vadd(float *src1, float *src2, float *dst)
{
	dst[0] = src1[0] + src2[0];
	dst[1] = src1[1] + src2[1];
	dst[2] = src1[2] + src2[2];
}

void vsub(float *src1, float *src2, float *dst)
{
	dst[0] = src1[0] - src2[0];
	dst[1] = src1[1] - src2[1];
	dst[2] = src1[2] - src2[2];
}

float vdot(float *v1, float *v2)
{
	return (v1[0] * v2[0] + v1[1] * v2[1] + v1[2] * v2[2]);
}

void vcross(float *v1, float *v2, float *cross)
{
	XYZ temp;

	  /* x        y       z         z       y */
	temp[0] = (v1[1] * v2[2]) - (v1[2] * v2[1]);
	temp[1] = (v1[2] * v2[0]) - (v1[0] * v2[2]);
	temp[2] = (v1[0] * v2[1]) - (v1[1] * v2[0]);
	vcopy(temp, cross);
}

QUAT XYZ_Rot_to_Q(float Xrot, float Yrot, float Zrot)
{
/* routine to convert the X,Y,Z rotations into a quaternion */
    float * axis;
    QUAT q1,q2,q3,qtemp,Rotq;

    axis = (float *)malloc(3 * sizeof(float));
    axis[2] = axis[1] = 0.0;
    axis[0] = 1.0f;
    axis_to_quaternion(axis, (float)(Xrot/RAD), &q1);
    axis[0] = axis[2] = 0.0;
    axis[1] = 1.0f;
    axis_to_quaternion(axis, (float)(Yrot/RAD), &q2);
    axis[2] = 1.0f;
    axis[1] = axis[0] = 0.0;
    axis_to_quaternion(axis, (float)(Zrot/RAD), &q3);
    qmult(&q3, &q2, &qtemp);
    qmult(&qtemp, &q1, &Rotq);
    free(axis);
    return Rotq;
}

void print_map_info(int Quick)
{
    if (Quick) return;
    fprintf(drvui->flout,"Map Title: %s\n",Map_Info.title);
	fprintf(drvui->flout,"Map Cell Parameters: %7.3f %7.3f %7.3f %7.2f %7.2f %7.2f\n",
		Map_Info.lat_con[0], Map_Info.lat_con[1], Map_Info.lat_con[2],
		Map_Info.lat_con[3], Map_Info.lat_con[4], Map_Info.lat_con[5]);
	fprintf(drvui->flout,"Minimum and Maximum density: %8.1f %8.1f\n",Map_Info.rhomn, Map_Info.rhomx);
	fprintf(drvui->flout,"Map Steps parallel to a, b, c: %5d %5d %5d\n",Map_Info.map_int[0],
		Map_Info.map_int[1], Map_Info.map_int[2]);
	fprintf(drvui->flout,"Map Calculated Region:            Axis     Min       Max\n"
		"                                    x: %8.3f %8.3f\n"
		"                                    y: %8.3f %8.3f\n"
		"                                    z: %8.3f %8.3f\n",
		Map_Info.xlim[0], Map_Info.xlim[1],
		Map_Info.ylim[0], Map_Info.ylim[1],
		Map_Info.zlim[0], Map_Info.zlim[1]);
}

void Progress_Window(int type, char *label, float max)
{

/* Routine to simulate a progress bar. The first argument defines the type of call
 *      type == -1, initialize. The second argument gives the label, 3rd the maximum.
 *      type == -2, close. 2nd and 3rd arguments ignored.
 *      type == 0, 3rd argument contains progress argument. 2nd ignored.
 */

    static char loc_label[1000];
    static float loc_max;
    
    switch (type) {
        
    case -1:
        strcpy(loc_label, label);
        loc_max = max;
        break;
    case -2:
        fprintf(stderr,"%s: 100 percent\n",loc_label);
        break;
    default:
        fprintf(stderr,"%s: %3d percent\r",loc_label,(int) (100.0f * max/loc_max + 0.5f));
    }
}

/* ************************************************************** */
/* ************************************************************** */

void label_cell(void)
{
// generate the labels corresponding to the unit-cell axes
  int i, j;
  int used[4] = {0,0,0,0};
  
  if (Labels) {

      if (fabs(offset[0]) + fabs(offset[1]) + fabs(offset[2]) < 0.01)
          Locate_Triple();

/* find any labels that are already in the list */

      if (drvui->nlabel > 1) {
          for (i=1; i<drvui->nlabel; i++) {
            if (!strcmp(drvui->label_label[i],"o")) {
                drvui->triple[0] = 0;
                used[0] = 1;
            }
            if (!strcmp(drvui->label_label[i],"triple_vect")) {
                drvui->triple[0] = i;
                used[0] = 1;
            }
            if (!strcmp(drvui->label_label[i],"a")) {
                drvui->triple[1] = i;
                used[1] = 1;
            }
            if (!strcmp(drvui->label_label[i],"b")) {
                drvui->triple[2] = i;
                used[2] = 1;
            }
            if (!strcmp(drvui->label_label[i],"c")) {
                drvui->triple[3] = i;
                used[3] = 1;
            }
          }
      }

/* following section adds labels to the list for those not already supplied */

      if (!Display_axes) {
          drvui->label_x[drvui->nlabel][0]=-0.5f/drvui->lat_con[0];
          drvui->label_x[drvui->nlabel][1]=-0.5f/drvui->lat_con[1];
          drvui->label_x[drvui->nlabel][2]=-0.5f/drvui->lat_con[2];
          drvui->label_x[drvui->nlabel][2]=1.0f+0.5f/drvui->lat_con[2];
          drvui->label_x[drvui->nlabel][1]=1.0f+0.5f/drvui->lat_con[1];
          drvui->label_x[drvui->nlabel][0]=1.0f+0.5f/drvui->lat_con[0];
      } else {
          drvui->label_x[drvui->nlabel][0]=1.4f/drvui->lat_con[0];
          drvui->label_x[drvui->nlabel][2]=1.4f/drvui->lat_con[2];
          drvui->label_x[drvui->nlabel][1]=1.4f/drvui->lat_con[1];
      }
      if (!used[0]) {
          drvui->label_fn[drvui->nlabel]=1;
          if (!Display_axes) {
              drvui->label_x[drvui->nlabel][0]=-0.5f/drvui->lat_con[0];
              drvui->label_x[drvui->nlabel][1]=-0.5f/drvui->lat_con[1];
              drvui->label_x[drvui->nlabel][2]=-0.5f/drvui->lat_con[2];
              drvui->triple[0] = 0;
              strcpy(drvui->label_label[drvui->nlabel++],"o");
          } else {
              drvui->triple[0] = drvui->nlabel;
              drvui->label_x[drvui->nlabel][0] = 0.0f;
              drvui->label_x[drvui->nlabel][1] = 0.0f;
              drvui->label_x[drvui->nlabel][2] = 0.0f;
              strcpy(drvui->label_label[drvui->nlabel++],"triple_vect");
          }
      }
      if (!used[1]) {
          drvui->label_fn[drvui->nlabel]=1;
          if (!Display_axes) {
              drvui->label_x[drvui->nlabel][0] = 1.0f+0.5f/drvui->lat_con[0];
              drvui->label_x[drvui->nlabel][1] = 0.0f;
              drvui->label_x[drvui->nlabel][2] = 0.0f;
          } else {
              drvui->label_x[drvui->nlabel][0] = 1.4f/drvui->lat_con[0];
              drvui->label_x[drvui->nlabel][1] = 0.0f;
              drvui->label_x[drvui->nlabel][2] = 0.0f;
          }
          drvui->triple[1] = drvui->nlabel;
          strcpy(drvui->label_label[drvui->nlabel++],"a");
      }
      if (!used[2]) {
          drvui->label_fn[drvui->nlabel]=1;
          if (!Display_axes) {
              drvui->label_x[drvui->nlabel][0] = 0.0f;
              drvui->label_x[drvui->nlabel][1] = 1.0f+0.5f/drvui->lat_con[1];
              drvui->label_x[drvui->nlabel][2] = 0.0f;
          } else {
              drvui->label_x[drvui->nlabel][0] = 0.0f;
              drvui->label_x[drvui->nlabel][1] = 1.4f/drvui->lat_con[1];
              drvui->label_x[drvui->nlabel][2] = 0.0f;
          }
          drvui->triple[2] = drvui->nlabel;
          strcpy(drvui->label_label[drvui->nlabel++],"b");
      }
      if (!used[3]) {
          drvui->label_fn[drvui->nlabel]=1;
          if (!Display_axes) {
              drvui->label_x[drvui->nlabel][0] = 0.0f;
              drvui->label_x[drvui->nlabel][1] = 0.0f;
              drvui->label_x[drvui->nlabel][2] = 1.0f+0.5f/drvui->lat_con[2];
          } else {
              drvui->label_x[drvui->nlabel][0] = 0.0f;
              drvui->label_x[drvui->nlabel][1] = 0.0f;
              drvui->label_x[drvui->nlabel][2] = 1.4f/drvui->lat_con[2];
          }
          drvui->triple[3] = drvui->nlabel;
          strcpy(drvui->label_label[drvui->nlabel++],"c");
      }
      for (j=0; j<4; j++)
          used[j] = 1;
  }
}

static void close_and_exit(void)
{
    FILE * ftmp;
    int i;
    char string[256];

    memset(string,0,256);
    fclose(drvui->fpoutp);
    drvui->fpoutp = fopen(drvui->Pov_File,"r");
    ftmp = fopen("pov.tmp","w");              /* open a temp file */
    while (!feof (drvui->fpoutp)) {
      if (!fgets (string, 256, drvui->fpoutp))        /* read a line */
          break;
      trim_string(string,256);
      if (!strncmp(string,"#declare xrot =",15))
          fprintf(ftmp,"#declare xrot =%7.1f;\n",xrot); /* replace xrot, etc. */
      else if (!strncmp(string,"#declare yrot =",15))
          fprintf(ftmp,"#declare yrot =%7.1f;\n",yrot);
      else if (!strncmp(string,"#declare zrot =",15))
          fprintf(ftmp,"#declare zrot =%7.1f;\n",zrot);
      else
          fprintf(ftmp,"%s\n",string);
    }
    fclose(ftmp);
    fclose(drvui->fpoutp);
    unlink(drvui->Pov_File);
    rename("pov.tmp",drvui->Pov_File);

    fprintf (drvui->flout,
	     "Updated values for xrot, yrot, zrot are %8.2f %8.2f %8.2f\n",
	     xrot, yrot, zrot);
    drvui->fpin = fopen(drvui->Str_File,"r");
    ftmp = fopen("pov.tmp","w");                      /* open a temp file */
    while (!feof (drvui->fpin)) {
      if (!fgets (string, 256, drvui->fpin))        /* read a line */
          break;
      trim_string(string,256);
      fprintf(ftmp,"%s\n",string);
      if (!strncmp(string,"end",3) || !strncmp(string,"END",3))
          break;                                      /* truncate input file after 'end' */
    }
    fclose(drvui->fpin);

    fprintf(ftmp,"rem - following lines indicate (possible) changes/additions to parameters\n");
    fprintf(ftmp,"view %.3f %.3f %.3f\n",xrot, yrot, zrot);
	if (nomits >0) {
		fprintf (drvui->flout, "Sites marked in the GUI: \n");
		fprintf (drvui->flout, "(add lines to input file to keep settings) \n");
        for (i=0;i<nomits;i++) {
			fprintf(drvui->flout,"special %d %d %d\n",omit1[i],omit2[i],omit3[i]);
			fprintf(ftmp,"special %d %d %d\n",omit1[i],omit2[i],omit3[i]);
        }
	}
	if (drvui->nlabel >1 && drvui->labels_moved) {
		fprintf(drvui->flout, "Labels: \n");
		fprintf(drvui->flout,"(these lines may need to be added to the input file)\n");
        for (i=1;i<drvui->nlabel;i++) {
            if (drvui->triple[0] == i) continue;       /* the "triple_vect" is not printed */
			fprintf(drvui->flout,"labeltext %6.3f %6.3f %6.3f %s\n",drvui->label_x[i][0],
                drvui->label_x[i][1],drvui->label_x[i][2],drvui->label_label[i]);
			fprintf(ftmp,"labeltext %6.3f %6.3f %6.3f %s\n",drvui->label_x[i][0],
                drvui->label_x[i][1],drvui->label_x[i][2],drvui->label_label[i]);
        }
        if (Display_axes) {
            fprintf(drvui->flout,"vectors %.3f %.3f %.3f\n",offset[0], offset[1], offset[2]);
            fprintf(ftmp,"vectors %.3f %.3f %.3f\n",offset[0], offset[1], offset[2]);
        }
	}			
    fclose(ftmp);
    unlink(drvui->Str_File);
    rename("pov.tmp",drvui->Str_File);
    (void) fclose (drvui->fpoutv);
    (void) fclose (drvui->flout);
    exit (0);
}
